
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;Uncomment the line below if you want the log messages in this executable to be compiled with the file.
;Thses messages will be logged if switch -l is used.
;Reasons for turing it off is preformance, turning on is for debugging!
CompileLogMessage


include 'include/function.inc'
include 'include/math.inc'
include 'include/savedatatodisk.inc'
include 'include/collection.inc'
include 'include/export_list.inc'
include 'include/flags.inc'
include 'include/init.inc'
include 'include/copy_structures.inc'
include 'include/internal_structures.inc'
include 'include/save_restore_data.inc'
include 'include/bind.inc'
include 'include/general_exports.inc'
include 'include/listen.inc'
include 'include/handle_connection.inc'
include 'include/control_panel_exports.inc'
include 'include/file.inc'
include 'include/timedate.inc'
include 'include/commandline.inc'
include 'include/stringFunctions.inc'

currentLoadingModule pointer 0
modules pointer 0 ;A linked-list of loaded modules
protocols pointer 0 ;A linked-list of loaded protocols
listenSessions pointer 0 ;A indexed-collection, indexed by the listensession
clientSessions pointer 0 ;A indexed-collection, indexed by the listensession
highestReservedListenSessionId pointer 0   ;the highest reserved, not yet released, ListensessionId

registerdProtocolIsLoadedFunctions pointer 0 ;A collection with pointers to ProtocolIsLoaded functions that need to be called when a new
registerdListenSessionIsCreatedFunctions pointer 0 ;A collection with pointers to ListenSessionIsCreated functions that need to be called when a new

log_options pointer 0

if used isLin
  commandLine pointer 0
end if

;start routine
Start:
if used is32 & used isWin
  ;do systemcheck (W 9x or NT)
  call doSystemCheck
end if

if used isLin
  mov [commandLine],pesp
end if

call initGetSmallestPortionPageAllocation

if used New_Memory_management & used MemoryMultiThreaded
  startLocalStorage
end if

if used isLin
  push 0
  call SetModuleDirectoryAsCurrentDirectory
end if

;call setExceptionHandler  ;=> only used during development

if used isWin
  ;process the commandline and get the amount of items in the command
  call processCommandline
  mov pecx,peax ;R/ECX => amount of items

  call getCommandLine
  mov pesi,peax  ;R/ESI => commandline
else
  mov pebx,pdword [commandLine] ;&cArg
  mov pecx,pdword [pebx]        ;R/ECX => amount of items
  add pebx,psize                ;first item
end if
xor edx,edx    ;flags 0x1 => -i, 0x2 => -u, 0x4 => -s, 0x8 => -b
xor edi,edi    ;pointer for next item

test pecx,pecx
jz noCommandlineArguments

loopThroughCommandline:
  if used isWin
    cmp word [pesi],0
  else
    mov pesi,pdword [pebx]
    cmp byte [pesi],0
  end if
  je nextLoopThroughCommandline

  test pedi,pedi
  jz itemIsNotValue

    mov [pedi],pesi
    xor edi,edi
    jmp nextLoopThroughCommandline

itemIsNotValue:
  if used isWin
    cmp word [pesi+2],0
  else
    cmp byte [pesi+1],0
  end if
  je nextLoopThroughCommandline
  if used isWin
    lodsd
  else
    lodsw
  end if

  mov dh,dl
  or dh,0x1 ;=> '-i'
  if used isWin
    cmp eax,0x0069002D ;'-i'
  else
    cmp ax,'-i'
  end if

  je nextLoopThroughCommandline

  mov dh,dl
  or dh,0x2 ;=> '-u'
  if used isWin
    cmp eax,0x0075002D ;'-u'
  else
    cmp ax,'-u'
  end if
  je nextLoopThroughCommandline

  mov dh,dl
  or dh,0x4 ;=> '-s'
  if used isWin
    cmp eax,0x0073002D ;'-s'
  else
    cmp ax,'-s'
  end if
  je nextLoopThroughCommandline

  mov dh,dl
  or dh,0x8 ;=> '-b'
  if used isWin
    cmp eax,0x0062002D ;'-b'
  else
    cmp ax,'-b'
  end if
  je nextLoopThroughCommandline

  mov dh,dl
  mov pedi,service_name_pointer
  if used isWin
    cmp eax,0x006E002D ;'-n'
  else
    cmp ax,'-n'
  end if
  je nextLoopThroughCommandline


  if used isWin
    mov pedi,keyPointerW
    cmp eax,0x0072002D ;'-r'
  else
    mov pedi,keyPointer
    cmp ax,'-r'
  end if
  je nextLoopThroughCommandline

  mov pedi,service_description
  if used isWin
    cmp eax,0x0064002D ;'-d'
  else
    cmp ax,'-d'
  end if
  je nextLoopThroughCommandline

  mov pedi,log_location
  if used isWin
    cmp eax,0x006C002D ;'-l
  else
    cmp ax,'-l'
  end if
  je nextLoopThroughCommandline

  mov pedi,log_options
  if used isWin
    cmp eax,0x006F002D ;'-o'
  else
    cmp ax,'-o'
  end if
  je nextLoopThroughCommandline

  ;unknown command ?
  xor edi,edi

nextLoopThroughCommandline:
  mov dl,dh

  if used isWin
    push pesi
    call strlenW
    add pesi,peax
    add pesi,2
  else
    add pebx,psize
  end if

dec pecx
jnz loopThroughCommandline
;loop loopThroughCommandline

noCommandlineArguments:
mov pesi,[log_options]
test pesi,pesi
jz noLogOptions

  xor ecx,ecx
  if used isWin
    cmp dword [pesi],0x000030 ;du '0',
  else
    cmp word [pesi],'0'
  end if
  je setLogOptions

  loopLogOptions:
    if used isWin
      lodsw
    else
      lodsb
    end if

    mov ch,cl
    or ch,LOG_MESSAGE
    if used isWin
      cmp ax,'m' ;message
    else
      cmp al,'m' ;message
    end if
    je addLogOption

    mov ch,cl
    or ch,LOG_WARNING
    if used isWin
      cmp ax,'w' ;warning
    else
      cmp al,'w' ;warning
    end if
    je addLogOption

    mov ch,cl
    or ch,LOG_ERROR
    if used isWin
      cmp ax,'e' ;error
    else
      cmp al,'e' ;error
    end if
    je addLogOption

    mov ch,cl

    addLogOption:
    mov cl,ch
    xor ch,ch

  if used isWin
    test ax,ax
  else
    test al,al
  end if
  jnz loopLogOptions

setLogOptions:
mov [allowedLog],ecx

noLogOptions:
cmp [log_location],0
je dontStartLogging

if used isWin
  push 0
  push [log_location]
  push 0x2 ;or 0x20000 ;UTF-16 and free pointer
  push 0 ;output string (CreateOneStringUtf8 creates new buffer)
  push 0 ;flags
  call CreateOneStringUtf8
  mov [log_location],peax
end if

startLog



dontStartLogging:

log LOG_WELCOME_MESSAGE,"" ;new line
log LOG_WELCOME_MESSAGE,"" ;new line
log LOG_WELCOME_MESSAGE,"-------------------------------------------"
log LOG_WELCOME_MESSAGE,"TV's server starting..."

;-if used LOG Always log the time
  sub pesp,0x60 ;make space for the 'Date' string

  ;get string with the date
  push pesp
  call GetGMTTime
  mov pecx,pesp
;-;end if

plog LOG_WELCOME_MESSAGE,"Time: ",0x11,pecx

;if used LOG
  add pesp,0x60
;end if

;Since in 32-bit mode Windows doesn't automaticly enable data execution prevention
if used is32 & used isWin
  call setDEBenabled
end if

push pedx ;save commandline flags

;check on cmov instructions support and cmpxchg instruction
mov eax,1
CPUID

pop pecx ;restore commandline flags

;check on cmpxchg instruction
test edx,0x100
jz missingCmpxchgInstruction

;check on cmov instruction
test edx,0x8000
jz missingCmovInstruction

push 0           ;parameter for exitProcess
ppush exitProcess ;return address

test cl,0x1 ;Install service
jnz InstallService

test cl,0x2 ;Uninstall service
jnz UninstallService

test cl,0x4 ;Stop service
jnz StopService

add pesp,psize*2 ;remove exitProcess and parameter from stack

test cl,0x8 ;boot as user  (no service)
jnz RunWithoutService

;-call getCommandLine

;-;find the end of the string
;-mov pedi,peax ;(R/E)DI => pointer to ASCII string
;-xor ecx,ecx   ; zero ECX
;-dec pecx      ; (R/E)CX => -1  (scasb will be repeated for 0xFFFFFFFF(FFFFFFFF) times, or until zero byte is found)
;-xor eax,eax   ; byte to find with scasb => '\0'

;-repnz scasb   ;loop string until zero byte is found

;-;check if size commadline is <2
;-add pecx,2
;-jc noCommandLineGiven

;-mov ax,word [pedi-3] ;get 2 last bytes of the commandline

;-push 0           ;parameter for exitProcess
;-push exitProcess ;return address

;-cmp ax,'-i' ;Install service
;-je InstallService

;-cmp ax,'-u' ;Uninstall service
;-je UninstallService

;-cmp ax,'-s' ;Stop service
;-je StopService

;-add pesp,psize*2 ;remove exitProcess and parameter from stack

;-cmp ax,'-b' ;boot as user  (no service)
;-je RunWithoutService

noCommandLineGiven:
;add pesp,psize*2 ;remove exitprocess and the 0 parameter
call CheckAndStartService
test eax,eax
jnz _endprocess

RunWithoutService:
log LOG_WELCOME_MESSAGE, "Starting TV's server..."
call StartTVServer ;Load the plugins and start listensessions (in a user process)
jmp quit

missingCmpxchgInstruction:
log LOG_ERROR,"ERROR: CPU doesn't support compare-and-swap instruction (cmpxchg), exiting..."
if used LOG
  jmp quit
end if

missingCmovInstruction:
log LOG_ERROR,"ERROR: CPU doesn't support conditional mov instruction (cmov), exiting..."
if used LOG
  jmp quit
end if

quit:
endThread

if used New_Memory_management & used MemoryMultiThreaded
  add pesp,SizeUsedByInitNewPointerList
end if

;if used New_Memory_management & used MemoryMultiThreaded
;  call uninitNewPointerListInTEB
;end if
;never release the main pointer list, dangerous!!!

; instead of uninitNewPointerListInTEB
ret



_endprocess:

if used New_Memory_management & used MemoryMultiThreaded
  add pesp,SizeUsedByInitNewPointerList
end if

push 0
call exitProcess

;------------------------------------------------------------------------------
;This function will load the plugins and start the ListenSessions
;
;
;input:
;       none
;output:none
;------------------------------------------------------------------------------
StartTVServer:
log LOG_MESSAGE, "Opening location with stored data..."
call PrepareSaveToDisk

log LOG_MESSAGE, "Allocate memory..."
call CreateCollection
mov [listenSessions],peax

call CreateCollection
mov [clientSessions],peax

call CreateCollection
mov [registerdProtocolIsLoadedFunctions],peax

call CreateCollection
mov [registerdListenSessionIsCreatedFunctions],peax

log LOG_MESSAGE, "Loading modules..."
log LOG_MESSAGE,""
;load modules
push 0                      ;no param
ppush CallbackLoadModules    ;callback function
push 0                      ;no subKey
push 0                      ;no subKey
push 0                      ;no subKey
ppush subkeyMudules          ;subKey
push pdword [hKey]          ;hkey
call EnumRegValue

;load lsitensessions
call LoadListenSessionsFromDisk
if used isLin
  call registerSignalHandlers
end if
ret

;------------------------------------------------------------------------------
;This function is a callback function of EnumRegValue and will fill in the
;item[] member of the structure Row
;
;input:
;       1:hKey
;       2:param  (pointer to Row structure)
;       3:valueName
;       4:value
;       5:valueSize
;       6:valueType
;output: EAX: true / false
;------------------------------------------------------------------------------
CallbackLoadModules:
call denter
MakeStack

;check if the type is a string
cmp pdword [pebp+psize*7],REG_SZ
jne endCallbackLoadModules

plog LOG_MESSAGE,"Loading ", 0x11, pdword [pebp+psize*5], "..."

;load the plugin
push 0x1 ;NOT_SAFE_ON_DISK
push pdword [pebp+psize*5]
call loadPlugin

if used LOG
test peax,peax
jz errLoadPlugin

log LOG_MESSAGE,"Sucess."
log LOG_MESSAGE,""
jmp endCallbackLoadModules

errLoadPlugin:
plog LOG_ERROR,"Failed loading ", 0x11, pdword [pebp+psize*5], "!"

end if

endCallbackLoadModules:
call dleave
ret psize*6

;------------------------------------------------------------------------------
;This function sets error code from eax/rax to the psess param
;------------------------------------------------------------------------------
setLastError:
push pecx
;set lastError member
mov pecx,pdword [pebp+(psize*2)]
mov dword [pecx+(psize*2)],eax

;if eax>0 then eax=0, if eax=0 then eax=1
xor ecx,ecx
STC
rcr ecx,1
dec eax
and eax,ecx
rol eax,1
pop pecx
ret