
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

FULL_LOCK equ 0x1                ;=>for functions outside this file
INTERNAL_COLLECTION_LOCK equ 0x2 ;=>for functions in this file that call other functions sepperately from this file and don't want those functions to unlock the list (so set this bit)

ITEM_IS_COLLECTION equ 0x4
REMOVE_INDEX_LIST equ 2

ONLY_ADD_IF_NOT_EXISTS equ 0x1
ADD_POINTER equ 0x2
USE_FREE_ITEM_FUNCTION equ 0x8

;------------------------------------------------------------------------------
;This function will create a list (not a linked one, but a collection)
;List     ->  |size    (or |Dword item0
;reserved     |item0       |Dword item1
;theadLock    |_____       ______)
;lockflags
;------------------------------------------------------------------------------
if used CreateCollection
CreateCollection:
push pebp
mov pebp,pesp
push pecx

_CreateCollection:
push 0x10
call GetMemLocation
push peax


push psize*5
call GetMemLocation

pop pecx
mov pdword [peax],pecx

pop pecx
mov pesp,pebp
pop pebp
ret
end if

;------------------------------------------------------------------------------
;This function will create a list (not a linked one, but a collection)
;This should only be called by fucntions in memfunctions.inc
;ONLY memfunctions.inc
;List     ->  |Dword item0
;reserved     |Dword item1
;theadLock    |______
;lockflags

;  1:First pointer in list
;------------------------------------------------------------------------------
if used __CreateCollectionVirtual
__CreateCollectionVirtual:
push pebp
mov pebp,pesp
push pecx

push 0x10
call GetMemLocationVirtual
push peax

;set first item
push pdword [pebp+psize*2]
pop pdword [peax+psize]

mov dword [peax],1 ;set count

push psize*5 + psize*2 ;THE LAST will be used for the structure in TEB!
call GetMemLocationVirtual

pop pecx
mov pdword [peax],pecx

pop pecx
mov pesp,pebp
pop pebp
ret psize
end if

;------------------------------------------------------------------------------
;This function will get a lock on the list so the function can be saftly executed 1:Collection to lock 2:flags
;Flags:
;FULL_LOCK 0x1
;--Example:
;push FULL_LOCK
;push pebx
;call LockListAcces
;--Unlock:
;push FULL_LOCK
;push pebx
;call ExternalUnlockListAccesEx
;------------------------------------------------------------------------------
if used LockListAcces
LockListAcces:
call denter

mov pecx,pdword [pebp+psize*2]
;test pecx,pecx
jpcxz errLockListAcces

call getCurThreadid
mov pedx,peax
cmp pdword [pecx+psize*2],pedx
je sucessLockListAcces

jmp firstTryGrantAcces

;Try to lock the list
grantAcces:
        push 1
        call sleep

        ;check if pointer still exist, if not, the list must be deleted
        ppush errLockListAcces
        push RIGHT_READ
        push pecx
        call saveCheckPointer

firstTryGrantAcces:
        xor eax,eax  ;Comparand

        ;Try to set a value on one (to lock the list)
        lock cmpxchg pdword [pecx+psize*2], pedx    ;Destination, Exchange,
jnz grantAcces

sucessLockListAcces:
;set optional flags
mov eax,dword [pebp+psize*3]
or dword [pecx+psize*3],eax

xor eax,eax
inc eax

endLockListAcces:
call dleave
ret psize*2

errLockListAcces:
xor eax,eax
jmp endLockListAcces
end if

;------------------------------------------------------------------------------
;This function will get a WRITE lock on the list so the function can be saftly executed 1:Collection to lock 2:flags
;This function should only be called by function who want to preform a read action.
;Flags:
;--Example:
;push pebx
;call LockListWriteAcces
;--Unlock:
;push pebx
;call ExternalUnlockListWriteAcces
;------------------------------------------------------------------------------
if used LockListWriteAcces
LockListWriteAcces:
call denter

mov pesi,pdword [pebp+psize*2]
test pesi,pesi
jz errLockListWriteAcces

;check if the list is locked for this thread
call getCurThreadid
cmp pdword [pesi+psize*2],peax
je sucessLockListWriteAcces

jmp firstTryGrantReadAcces

;Try to lock the list
grantReadAcces:
        push 1
        call sleep

        ;check if pointer still exist, if not, the list must be deleted
        ppush errLockListWriteAcces
        push RIGHT_READ
        push pesi
        call saveCheckPointer

firstTryGrantReadAcces:
        mov peax,pdword [pesi+psize*2] ;(R/E)AX => Comparand (low (D/Q)word)
        test peax,peax ;zero means the list is not locked
        je AccesIsFree

        cmp peax,-1     ;-1 means the list is locked for write actions
        jne grantReadAcces

        AccesIsFree:
        mov pedx,pdword [pesi+psize*3] ;(R/E)DX => Comparand (high (D/Q)word)

        xor pebx,pebx  ;(R/E)BX => Exchange (low (D/Q)word)
        dec pebx

        mov pecx,pedx  ;(R/E)CX => Exchange (high (D/Q)word)
        inc pecx

        ;Try to set a value on one (to lock the list)
        if used is64
          lock cmpxchg16b [pesi+psize*2]    ;Destination
        else
          lock cmpxchg8b [pesi+psize*2]     ;Destination
        end if
jnz grantReadAcces

sucessLockListWriteAcces:
xor eax,eax
inc eax

endLockListWriteAcces:
call dleave
ret psize

errLockListWriteAcces:
xor eax,eax
jmp endLockListWriteAcces
end if

;------------------------------------------------------------------------------
;This function reverse the LockListAcces effect, this can be called from outside functions
;------------------------------------------------------------------------------
if used ExternalUnlockListWriteAcces
ExternalUnlockListWriteAcces:
call denter

call UnlockListWriteAcces

call dleave
ret psize
end if

;------------------------------------------------------------------------------
;This function reverse the LockListWriteAcces effect
;------------------------------------------------------------------------------
if used UnlockListWriteAcces
UnlockListWriteAcces:
push pesi ;save (R/E)SI
push peax ;save (R/E)AX
push pedx ;save (R/E)DX
push pebx ;save (R/E)BX
push pecx ;save (R/E)CX

;get pointer to list
mov pesi,pdword [pebp+psize*2]
test pesi,pesi
jz endUnlockListWriteAcces

;check if the list is locked for this thread
call getCurThreadid
cmp pdword [pesi+psize*2],peax
je endUnlockListWriteAcces

lock dec pdword [pesi+psize*3] ;decrement the ammount of threads running read functions

xor edx,edx ;(R/E)DX => Comparand (high (D/Q)word)
xor eax,eax ;(R/E)DX => Comparand (low (D/Q)word)
dec peax

xor ebx,ebx  ;(R/E)BX => Exchange (low (D/Q)word)
xor ecx,ecx  ;(R/E)CX => Exchange (high (D/Q)word)

;free acces to the list if no thread locked the list for write actions
if used is64
  lock cmpxchg16b [pesi+psize*2]    ;Destination
else
  lock cmpxchg8b [pesi+psize*2]     ;Destination
end if

endUnlockListWriteAcces:
pop pecx
pop pebx
pop pedx
pop peax
pop pesi
ret
end if

;------------------------------------------------------------------------------
;This function reverse the LockListAcces effect, this can be called from outside functions
;------------------------------------------------------------------------------
if used ExternalUnlockListAccesEx
ExternalUnlockListAccesEx:
call denter

push pdword [pebp+psize*3]
call UnlockListAccesEx

call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function reverse the LockListAcces effect
;------------------------------------------------------------------------------
if used UnlockListAcces
UnlockListAcces:
push peax ;save the EAX value

xor eax,eax                     ;EAX=0
xchg peax,pdword [pesp+psize]   ;xchange return value with zero (flags)
xchg peax,pdword [pesp]         ;xchange saved EAX with return value
;jmp UnlockListAccesEx
end if

;------------------------------------------------------------------------------
;This function reverse the LockListAcces effect 1:flags
;------------------------------------------------------------------------------
if used UnlockListAccesEx | used UnlockListAcces
UnlockListAccesEx:
push pecx ;save (R/E)CX
push peax ;save (R/E)AX

;get pointer to list
mov pecx,pdword [pebp+psize*2]
jecxz endUnlockListAccesEx

call getCurThreadid
cmp pdword [pecx+psize*2],peax
jne endUnlockListAccesEx

;get the flags and test on zero
mov eax,dword [pecx+psize*3]
test eax,eax
jz releaseListLock

;test is the flags are set
and eax, dword [pesp+psize*3]
jz endUnlockListAccesEx

;zero flags
xor dword [pecx+psize*3],eax
jnz endUnlockListAccesEx

releaseListLock:
mov pdword [pecx+psize*2],0

endUnlockListAccesEx:
pop peax
pop pecx
ret psize
end if

;------------------------------------------------------------------------------
;This function will check pointer on zero and lock the list
;------------------------------------------------------------------------------
if used defaultListCheckup
defaultListCheckup:
;Get pointer to the list and check if not zero
xor eax,eax
mov pecx,pdword [pebp+psize*2]
test pecx,pecx
jz errDefaultCheckup

;Get a lock on the list
push 0 ;a normal lock
push pecx
call LockListAcces
test eax,eax
jz errDefaultCheckup

cmp pdword [pecx],0
jz errDefaultCheckup

endDefaultCheckup:
;just return
ret psize*1

errDefaultCheckup:
xor eax,eax
;jump to the error pointer
add pesp,psize*2
jmp pdword [pesp-psize]
end if

;------------------------------------------------------------------------------
;This function will check pointer on zero and lock the list
;------------------------------------------------------------------------------
if used defaultReadListCheckup
defaultReadListCheckup:
;Get pointer to the list and check if not zero
xor eax,eax
mov pecx,pdword [pebp+psize*2]
test pecx,pecx
jz errDefaultReadListCheckup

;Get a lock on write actions on the list
push pecx
call LockListWriteAcces
test eax,eax
jz errDefaultReadListCheckup

cmp pdword [pecx],0
jz errDefaultReadListCheckup

endDefaultReadListCheckup:
;just return
ret psize*1

errDefaultReadListCheckup:
xor eax,eax
;jump to the error pointer
add pesp,psize*2
jmp pdword [pesp-psize]
end if

;------------------------------------------------------------------------------
;This function will return the ammount of items in the list
;------------------------------------------------------------------------------
if used CountItems
CountItems:
call denter

ppush errCountItems ;error function
call defaultReadListCheckup

;Get the ammount of values in the list
mov pedx,pdword [pecx]
mov eax,dword [pedx]

l:
call UnlockListWriteAcces

errCountItems:
call dleave
ret psize*1
end if

;------------------------------------------------------------------------------
;This function copy's a collection of dwords
;input:
;  1:collection
;output: pointer to InternalClientSession
;------------------------------------------------------------------------------
if used copyDwordList
copyDwordList:
call denter

ppush errCopyDwordList ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ESI/RSI
mov pesi,pdword [pecx]

call CreateCollection
mov pebx,peax ;Collection
mov pedi,pdword [peax] ;list

push pesi ;source
push pedi ;destination
call cpyMemAndRealocPage

;set new pointer of list
mov pdword [pebx],peax

;return new list
mov peax,pebx

endCopyDwordList:
call UnlockListWriteAcces

errCopyDwordList:
call dleave
ret psize*1
end if

;------------------------------------------------------------------------------
;This function will transfer a high-case to a low case
;input:
;    AL => low or high case byte
;output:
;    AL => low case byte
;------------------------------------------------------------------------------

if used SetCaseLow
        SetCaseLow:
        cmp al,40h
        jna NextSetCaseLow
        
        cmp al,5Bh
        jb SetCaseLowjump
        
        NextSetCaseLow:
        cmp al,0BFh
        jna endSetCaseLow
        
        cmp al,0DDh
        jnb endSetCaseLow
        
        
        SetCaseLowjump:
        add al,20h
        endSetCaseLow:
        ret
end if


;------------------------------------------------------------------------------
;This function will generate a simple hash from the input string
;input:
;    (R/E)SI => input string
;output:
;    (R/E)DI => output hash of string
;    (R/E)SI => end of string
;Corupted: EAX, ESI
;------------------------------------------------------------------------------
if used GenerateHash
GenerateHash:
;generate hash
xor eax, eax
xor edi, edi

;calculate the has of the string and set it on EDI
next_character_hash:
lodsb
test al, al
jz end_of_name_hash

ror edi, 13
add edi, eax
jmp next_character_hash

end_of_name_hash:
ret
end if

if used GenerateHashCaseInsensitive | used GenerateHashCaseInsensitiveSafe
;------------------------------------------------------------------------------
;This function will generate a simple hash from the input string (case insensitive)
;input:
;    1: input string
;output:
;    (R/E)AX => output hash of string
;Corupted: none
;------------------------------------------------------------------------------
GenerateHashCaseInsensitiveSafe:
call denter

mov pesi,pdword [pebp+psize*2] ;input string
call GenerateHashCaseInsensitive

mov peax,pedi ;output hash

call dleave
ret psize

;------------------------------------------------------------------------------
;This function will generate a simple hash from the input string (case insensitive)
;input:
;    (R/E)SI => input string
;output:
;    (R/E)DI => output hash of string
;    (R/E)SI => end of string
;Corupted: EAX, ESI
;------------------------------------------------------------------------------
GenerateHashCaseInsensitive:
;generate hash
xor eax, eax

test pesi,pesi
jz end_of_name_hash_case_insensitive

xor edi, edi

;calculate the has of the string and set it on EDI
next_character_hash_case_insensitive:
lodsb
call SetCaseLow
test al, al
jz end_of_name_hash_case_insensitive

ror edi, 13
add edi, eax
jmp next_character_hash_case_insensitive

end_of_name_hash_case_insensitive:
ret
end if

;------------------------------------------------------------------------------
;This function will find a pointer to the desired item of the index name 1:collection 2:index name

;The way the collection will look in memory:
;hash of index name       --------------
;list-------------------->|index name
;hash of index name       |data pointer
;list----------------\    |index name
;                    |    |data pointer
;                    |    ---------------
;                    |          -------------
;                    \_________>|index name
;                               |data pointer
;                               --------------
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used GetIndexedStringItemWithHash_c
if used is64
    GetIndexedStringItemWithHash_c:
    StartProc 2

    mov rcx,retGetIndexedStringItemWithHash_c
    xchg qword [rsp],rcx
    jmp GetIndexedStringItemWithHash

    retGetIndexedStringItemWithHash_c:
    sub rsp,2*psize
    EndProc 2
    jmp rcx
else
    GetIndexedStringItemWithHash_c:
end if
end if
;---------
;asm implementation
;---------
if used GetIndexedStringItemWithHash | used GetIndexedStringItemWithHash_c
GetIndexedStringItemWithHash:
call denter

ppush errGetIndexedStringItemWithHash ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]

xor edx,edx ;clear (R/E)DX

;The list is now locked by defaultReadListCheckup, this will lock write actions until
;--;Lock the list so no other threads can use it
;--push INTERNAL_COLLECTION_LOCK
;--push pecx
;--call LockListAcces


;generate the hash of the index name EDI=HASH
mov pesi,pdword [pebp+psize*3]
call GenerateHashCaseInsensitive

;check if the hash already exists in the list
push pedi ;index
push pecx
call FindDwordByIndex
test peax,peax
jz endGetIndexedStringItemWithHash

;get the collection that belongs to this hash
push peax
push pecx
call GetDwordItem
mov pedx,peax ;move the pointer of the list to (R/E)DX

;Lock the list so no write actions can accor
;--push INTERNAL_COLLECTION_LOCK  ;This was a argument for LockListAcces, however since we now use LockListWriteAcces it's not needed annymore
push pedx
call LockListWriteAcces

;find out if the index string exists
push pdword [pebp+psize*3]
push pedx
call FindDwordByStringIndex
test peax,peax
jz endGetIndexedStringItemWithHash

push peax
push pedx
call GetItem

endGetIndexedStringItemWithHash:
  push peax

  test pedx,pedx
  jz dontReleaseRDXGetIndexedString
  ;Unlock the list so other threads can use it
  ;--push INTERNAL_COLLECTION_LOCK  ;This was a argument for ExternalUnlockListAccesEx, however since we now use ExternalUnlockListWriteAcces it's not needed annymore
  push pedx
  call ExternalUnlockListWriteAcces

dontReleaseRDXGetIndexedString:
  ;Unlock the list so other threads can use it
  ;--push INTERNAL_COLLECTION_LOCK  ;This was a argument for ExternalUnlockListAccesEx, however since we now use ExternalUnlockListWriteAcces it's not needed annymore
  push pecx
  call ExternalUnlockListWriteAcces

pop peax

errGetIndexedStringItemWithHash:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will add an indexed string to the collection 1:collection 2:index name 3:data 4:datalen

;The way the collection will look in memory:
;hash of index name       --------------
;list-------------------->|index name
;hash of index name       |data pointer
;list----------------\    |index name
;                    |    |data pointer
;                    |    ---------------
;                    |          -------------
;                    \_________>|index name
;                               |data pointer
;                               --------------
;------------------------------------------------------------------------------
if used AddIndexedStringItemWithHash | used AddIndexedStringItemWithHashNoCopy | used AddIndexedStringItemWithHashEx
AddIndexedStringItemWithHash:
xor eax,eax
xchg peax,pdword [pesp+psize*4] ;exchange datalen with zero (no flags)
xchg peax,pdword [pesp+psize*3] ;exchange data with datalen
xchg peax,pdword [pesp+psize*2] ;exchange index name with data
xchg peax,pdword [pesp+psize]   ;exchange collection with index name
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
jmp AddIndexedStringItemWithHashEx

;------------------------------------------------------------------------------
;This function will add an indexed string to the collection 1:collection 2:index name 3:data
;------------------------------------------------------------------------------
AddIndexedStringItemWithHashNoCopy:
push pdword [pesp+psize*1]      ;push collection
push pdword [pesp+psize*1]      ;push return valiue

xor eax,eax
xchg pdword [pesp+psize*4],peax  ;exchange the index name with data len (0)
mov pdword [pesp+psize*2],peax   ;set index name
mov eax,3
xchg pdword [pesp+psize*5],peax ;exchange data with new flags (DONT_COPY_INDEX | DONT_COPY_STRING | DONT_FREE_ORIGINAL_STRING)
mov pdword [pesp+psize*3],peax  ;set data
;jmp AddIndexedStringItemWithHashEx

;------------------------------------------------------------------------------
;This function will add an indexed string to the collection 1:collection 2:index name 3:data 4:datalen 5:flags
;flags:
;       0x0 COPY_STRING_AND_INDEX
;       0x1 DONT_COPY_INDEX  WaRnInG:: GetItemSize will not be able to be used !!!
;       0x2 DONT_COPY_STRING WaRnInG:: GetItemSize will not be able to be used !!!
;------------------------------------------------------------------------------
AddIndexedStringItemWithHashEx:
call denter

ppush errAddIndexedStringItem ;error function
call defaultListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]
xor edx,edx ;clear (R/E)DX

;Lock the list so no other threads can use it
push INTERNAL_COLLECTION_LOCK
push pecx
call LockListAcces

;generate the hash of the index name

mov pesi,pdword [pebp+psize*3]
call GenerateHashCaseInsensitive

;check if the hash already exists in the list
push pedi ;index
push pecx
call FindDwordByIndex
test eax,eax
jnz getSubList

;add a new colelction with the hash as index
call CreateCollection

;add the collection with his index to the list
push ADD_POINTER
push 0
push peax
push pedi
push pecx
call AddIndexedItemEx

getSubList:
;get the collection that belongs to this hash
push peax
push pecx
call GetDwordItem
mov pedx,peax ;move the pointer of the list to (R/E)DX

;Lock the list so no other threads can use it
push INTERNAL_COLLECTION_LOCK
push pedx
call LockListAcces

;find out if the index string already exists
push pdword [pebp+psize*3]
push pedx
call FindDwordByStringIndex
test eax,eax
jnz onlyModifyStringData

test pdword [pebp+psize*6],0x1 ;DONT_COPY_INDEX
jz CopyIndexString

;don't copy index string, just add pointer
sub pdword [pebp+psize*3],4 ;substract by 4, since it will be threaded as an item

push pdword [pebp+psize*3]
push pedx
call AddDword

add pdword [pebp+psize*3],4 ;restore index pointer
jmp IndexCopied

CopyIndexString:
;Add the index string
push pdword [pebp+psize*3]
push pedx
call AddString

IndexCopied:
test pdword [pebp+psize*6],0x2 ;DONT_COPY_STRING
jz CopyData

sub pdword [pebp+psize*4],4 ;substract by 4, since it will be threaded as an item, not dword

push pdword [pebp+psize*4] ;item
push pedx
call AddDword

add pdword [pebp+psize*3],4 ;restore index pointer
jmp unlockListAddIndexedStringItemWithHash

CopyData:
;Add the item
push pdword [pebp+psize*5] ;size
push pdword [pebp+psize*4] ;item
push pedx
call AddItem
jmp unlockListAddIndexedStringItemWithHash

onlyModifyStringData:
test pdword [pebp+psize*6],0x2 ;DONT_COPY_STRING
jz CopyDataModify

sub pdword [pebp+psize*4],4 ;substract by 4, since it will be threaded as an item, not dword

push pdword [pebp+psize*4] ;item
push peax ;itemId
push pedx
call ModifyDword

add pdword [pebp+psize*3],4 ;restore index pointer
jmp unlockListAddIndexedStringItemWithHash

CopyDataModify:
;modify item
push pdword [pebp+psize*5] ;size
push pdword [pebp+psize*4] ;item
push peax ;itemId
push pedx
call ModifyItem

unlockListAddIndexedStringItemWithHash:
test pedx,pedx
jz dontReleaseRDXAddIndexedString
;Unlock the list so other threads can use it
push INTERNAL_COLLECTION_LOCK
push pedx
call ExternalUnlockListAccesEx

dontReleaseRDXAddIndexedString:
;Unlock the list so other threads can use it
push INTERNAL_COLLECTION_LOCK
push pecx
call ExternalUnlockListAccesEx

errAddIndexedStringItem:
call dleave
ret psize*5
end if

;----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
;------------------------------------------------------------------------------
;This function will remove a pointer to the desired item of the index name 1:collection 2:index name

;The way the collection will look in memory:
;hash of index name       --------------
;list-------------------->|index name
;hash of index name       |data pointer
;list----------------\    |index name
;                    |    |data pointer
;                    |    ---------------
;                    |          -------------
;                    \_________>|index name
;                               |data pointer
;                               --------------
;------------------------------------------------------------------------------
if used RemoveIndexedStringItemWithHash | used RemoveIndexedStringItemWithHashEx
RemoveIndexedStringItemWithHash:
xor eax,eax
xchg pdword [pesp+psize*2],peax ;swap flags(zero) with index name
xchg pdword [pesp+psize*1],peax ;swap index name with collection
xchg pdword [pesp+psize*0],peax ;swap collection with returnpointer
push peax                       ;push return pointer
;jmp RemoveIndexedStringItemWithHashEx

;------------------------------------------------------------------------------
;This function will remove a pointer to the desired item of the index name 1:collection 2:index name 3:flags (reserved)

;The way the collection will look in memory:
;hash of index name       --------------
;list-------------------->|index name
;hash of index name       |data pointer
;list----------------\    |index name
;                    |    |data pointer
;                    |    ---------------
;                    |          -------------
;                    \_________>|index name
;                               |data pointer
;                               --------------
;------------------------------------------------------------------------------
RemoveIndexedStringItemWithHashEx:
call denter

ppush errRemoveIndexedStringItemWithHash ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]

xor edx,edx ;clear (R/E)DX

;The list is now locked by defaultReadListCheckup, this will lock write actions until
;--;Lock the list so no other threads can use it
;--push INTERNAL_COLLECTION_LOCK
;--push pecx
;--call LockListAcces


;generate the hash of the index name EDI=HASH
mov pesi,pdword [pebp+psize*3]
call GenerateHashCaseInsensitive

;check if the hash already exists in the list
push pedi ;index
push pecx
call FindDwordByIndex
test eax,eax
jz endRemoveIndexedStringItemWithHash

;get the collection that belongs to this hash
push peax
push pecx
call GetDwordItem
mov pedx,peax ;move the pointer of the list to (R/E)DX

;Lock the list so no write actions can accor
push INTERNAL_COLLECTION_LOCK
push pedx
call LockListAcces

;find out if the index string exists
push pdword [pebp+psize*3]
push pedx
call FindDwordByStringIndex
test eax,eax
jz endRemoveIndexedStringItemWithHash

dec peax ;decrement the id to get the index

push peax  ;push for RemoveItem

;remove the index
push peax
push pedx
call RemoveItem

;remove the item
;push peax already pushed
push pedx
call RemoveItem

endRemoveIndexedStringItemWithHash:
  push peax

  test pedx,pedx
  jz dontReleaseRDXGetIndexedString1
  ;Unlock the list so other threads can use it
  ;--Unlock the list
  push INTERNAL_COLLECTION_LOCK
  push pedx
  call ExternalUnlockListAccesEx

dontReleaseRDXGetIndexedString1:
  ;Unlock the list so other threads can use it
  ;--push INTERNAL_COLLECTION_LOCK  ;This was a argument for ExternalUnlockListAccesEx, however since we now use ExternalUnlockListWriteAcces it's not needed annymore
  push pecx
  call ExternalUnlockListWriteAcces

pop peax

errRemoveIndexedStringItemWithHash:
call dleave
ret psize*3
end if
;----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------


;------------------------------------------------------------------------------
;This function will get an indexed item of the collection 1:collection 2:index
;------------------------------------------------------------------------------
if used GetIndexedItem
GetIndexedItem:
call denter

ppush errGetIndexedItem ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]


push pdword [pebp+psize*3]; index
push pecx
call FindDwordByIndex
test eax,eax
jz endGetIndexedItem

push peax
push pecx
call GetItem

endGetIndexedItem:
call UnlockListWriteAcces

errGetIndexedItem:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will get an indexed dword item of the collection 1:collection 2:index
;------------------------------------------------------------------------------
if used GetIndexedDwordItem
GetIndexedDwordItem:
call denter

ppush errGetIndexedDwordItem ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]

push pdword [pebp+psize*3]; index
push pecx
call FindDwordByIndex
test eax,eax
jz endGetIndexedDwordItem

push peax
push pecx
call GetDwordItem

endGetIndexedDwordItem:
call UnlockListWriteAcces
;save eax

errGetIndexedDwordItem:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will add an indexed (D/Q)word item to the collection 1:collection 2:index 3:itemDword
;------------------------------------------------------------------------------
if used AddIndexedDwordItem
AddIndexedDwordItem:
;Rebuild stack for function AddIndexedItemEx
xor eax,eax
xchg peax,pdword [pesp+psize*2] ;exchange index with size
xchg peax,pdword [pesp]         ;xchange return pointer with index
push peax                       ;push return pointer

xor eax,eax
inc eax
inc eax
xchg peax,pdword [pesp+psize*4] ;exchange itemDword with 2 (flag: ADD_POINTER)
xchg peax,pdword [pesp+psize*2] ;exchange collection with itemDword
xchg peax,pdword [pesp]         ;xchange return pointer with collection
push peax                       ;push return pointer
jmp AddIndexedItemEx
end if

;------------------------------------------------------------------------------
;This function will add an indexed item to the collection 1:collection 2:index 3:item 4:itemsize
;------------------------------------------------------------------------------
if used AddIndexedItem
AddIndexedItem:
xor eax,eax
xchg peax,pdword [pesp+psize*4] ;exchange itemsize with zero (no flags)
xchg peax,pdword [pesp+psize*3] ;exchange item with itemsize
xchg peax,pdword [pesp+psize*2] ;exchange index with item
xchg peax,pdword [pesp+psize]   ;exchange collection with index
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
;jmp AddIndexedItemEx
end if

;------------------------------------------------------------------------------
;This function will add an indexed item to the collection 1:collection 2:index 3:item 4:itemsize 5:flags
;Flags
;MODIFY_IF_EXISTING 0x0       modify the item with the index if it already exists in the collection
;ONLY_ADD_IF_NOT_EXISTS 0x1   do not modify if the index already exists
;ADD_POINTER 0x2              the item should be threaded as a dword / qword, this means it will just move it to the list, not copy the content of the pointer (itemsize will be ignored)
;------------------------------------------------------------------------------
if used AddIndexedItemEx | used AddIndexedItem
AddIndexedItemEx:
call denter

ppush errAddIndexedItem ;error function
call defaultListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]


push INTERNAL_COLLECTION_LOCK
push pecx
call LockListAcces

push pdword [pebp+psize*3]; index
push pecx
call FindDwordByIndex
test eax,eax
jz AIIE_addItem

;check if we need to modify the existing item or don't touch it
test pdword [pebp+psize*6],ONLY_ADD_IF_NOT_EXISTS
jnz endAddIndexedItem

;modify item
push pdword [pebp+psize*5] ;size
push pdword [pebp+psize*4] ;item
push peax ;itemId
push pecx
call ModifyItem
jmp endAddIndexedItem

AIIE_addItem:
push pdword [pebp+psize*3] ;index
push pecx
call AddDword

;check if we need to copy a buffer or just copy a pointer
test pdword [pebp+psize*6],ADD_POINTER
jz copyItem

push pdword [pebp+psize*4] ;item
push pecx
call AddDword
jmp endAddIndexedItem

copyItem:
push pdword [pebp+psize*5] ;size
push pdword [pebp+psize*4] ;item
push pecx
call AddItem

endAddIndexedItem:
push peax

;--Unlock:
push INTERNAL_COLLECTION_LOCK
push pecx
call ExternalUnlockListAccesEx

pop peax

errAddIndexedItem:
call dleave
ret psize*5
end if

;------------------------------------------------------------------------------
;This function will add an string ASCII to the collection 1:collection 2:string
;------------------------------------------------------------------------------
if used AddString
AddString:
;calculate the size of the string
push pdword [pesp+psize*2]
call strlen
inc eax

xchg peax,pdword [pesp+psize*2] ;exchange string with the size of the string
xchg peax,pdword [pesp+psize]   ;exchange collection with the string
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
jmp AddItem
end if

;------------------------------------------------------------------------------
;This function will add a string UNICODE to the collection 1:collection 2:string
;------------------------------------------------------------------------------
if used AddStringW
AddStringW:
;calculate the size of the string
push pdword [pesp+psize*2]
call strlenW
add eax,2

xchg peax,pdword [pesp+psize*2] ;exchange string with the size of the string
xchg peax,pdword [pesp+psize]   ;exchange collection with the string
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
;jmp AddItem
end if

;------------------------------------------------------------------------------
;This function will add an item to the collection 1:collection 2:data 3:lenData
;------------------------------------------------------------------------------
if used AddItem | used AddStringW
AddItem:
call denter

ppush g ;error function
call defaultListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]

;set size on edx and check for zero
mov edx,dword [pebp+psize*4]
test edx,edx
jz g

;increment ammount of items
inc dword [pecx]
mov eax,dword [pecx]

;calculate the size of the list
if used is64
   shl rax,3 ;*8 (size for each item
else
   shl eax,2 ;*4  (size for each item
end if

add eax,psize ;<= size parameter

;reallocate the memory if necessary
push peax  ;the size of the new memory
push pecx  ;the memory to be resized
call ReAllocMemPage

;save the new pointer and set pointer on (r/e)cx
mov pecx,pdword [pebp+psize*2]
mov pdword [pecx],peax
mov pecx,peax

;add some spare memory to avoid memory acces failure and to make space for the size member
add edx,0x10

;accolate memory for the new ietm
push pedx
call GetMemLocation

;substract the space memory
sub edx,0x10

;set the size of the item
mov dword [peax],edx

;get items in the list
mov edi,dword [pecx]

;calculate the pointer of the new item in the list
if used is64
   shl rdi,3 ;*8 (size of a pointer)
else
   shl edi,2 ;*4  (size of a pointer)
end if

;set the pointer of the new item in the list
mov pdword [pecx+pedi],peax
add peax,4

;copy the item
push pedx
push pdword [pebp+psize*3]
push peax
call strcpyn

;return item Id
mov peax,pdword [pebp+psize*2]
mov peax,pdword [peax]
mov peax,pdword [peax]

g:
call UnlockListAcces

call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will add an dword /ineger to the collection 1:collection 2:int
;------------------------------------------------------------------------------
if used AddDword
AddDword:
call denter

ppush errAddDword ;error function
call defaultListCheckup

;Get pointer in list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]

mov eax,dword [pecx]
inc peax

;calculate the size of the list
if used is64
   shl rax,3 ;*8 (size for each item
else
   shl eax,2 ;*4  (size for each item
end if

add eax,4 ;<= size parameter

;reallocate the memory if necessary
push peax  ;the size of the new memory
push pecx  ;the memory to be resized
call ReAllocMemPage

;save the new pointer and set pointer on (r/e)cx
mov pecx,pdword [pebp+psize*2]
mov pdword [pecx],peax
mov pecx,peax

;get items in the list
mov edi,dword [pecx]
inc edi

;calculate the pointer of the new item in the list
if used is64
   shl rdi,3 ;*8 (size of a pointer)
else
   shl edi,2 ;*4  (size of a pointer)
end if

;set the integer of the new item in the list
mov peax,pdword [pebp+psize*3]
mov pdword [pecx+pedi],peax

mov peax,pdword [pebp+psize*2]
mov peax,pdword [peax]

;increment ammount of items
inc dword [peax]

;return item Id
mov peax,pdword [peax]

errAddDword:
call UnlockListAcces

call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will return the size of an item 1:collection 2:itemId
;------------------------------------------------------------------------------
if used GetItemSize
GetItemSize:
call denter

ppush errGetItemSize ;error function
call defaultReadListCheckup

;Get pointer to list and set it on ECX/RCX
mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]

;check if the the item exists
mov eax,dword [pecx]
mov edx,dword [pebp+psize*3]
cmp eax,edx
jae a

h:
xor eax,eax
jmp b

a:
;itemId starts at 1, not zero
test edx,edx
cmovz eax,edx
jz b

;calculate the pointer of the new item in the list
if used is64
   shl rdx,3 ;*8 (size of a pointer)
else
   shl edx,2 ;*4  (size of a pointer)
end if


mov pecx,pdword [pecx+pedx]
mov eax,dword [pecx]

b:
call UnlockListWriteAcces

errGetItemSize:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will return the size of an item 1:collection 2:itemId
;------------------------------------------------------------------------------
if used GetItem
GetItem:
call denter

ppush errGetItem ;error function
call defaultReadListCheckup

mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]

;check if the the item exists
mov eax,dword [pecx]
mov edx,dword [pebp+psize*3]
cmp eax,edx
jae c

i:
xor eax,eax
jmp d

c:
;itemId starts at 1, not zero
test edx,edx
cmovz eax,edx
jz d

;calculate the pointer of the new item in the list
if used is64
   shl rdx,3 ;*8 (size of a pointer)
else
   shl edx,2 ;*4  (size of a pointer)
end if

mov peax,pdword [pecx+pedx]
;mov edx,eax
add peax,4

d:
call UnlockListWriteAcces

errGetItem:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will return a integer from the collection 1:collection 2:itemId
;------------------------------------------------------------------------------
if used GetDwordItem
GetDwordItem:
;rebuild stack
pop peax ;return address
xchg peax,pdword [pesp+psize] ;xchange return address with itemId
xchg peax,pdword [pesp] ;xchange itemId with collection
push peax ;push collection
;call GetItem
call GetItem
test peax,peax
jz endGetDwordItem

sub peax,4

endGetDwordItem:
ret
end if

;------------------------------------------------------------------------------
;This function will Modify an string ASCII on a collection 1:collection 2:itemId 2:string
;------------------------------------------------------------------------------
if used ModifyString
ModifyString:
;calculate the size of the string
push pdword [pesp+psize*3]
call strlen
inc eax

xchg peax,pdword [pesp+psize*3] ;exchange string with the size of the string
xchg peax,pdword [pesp+psize*2] ;exchange itemId with the string
xchg peax,pdword [pesp+psize*1] ;exchange collection  with the itemId
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
jmp ModifyItem
end if

;------------------------------------------------------------------------------
;This function will Modify an string ASCII on a collection 1:collection 2:itemId 2:string
;------------------------------------------------------------------------------
if used ModifyStringW
ModifyStringW:
;calculate the size of the string
push pdword [pesp+psize*3]
call strlenW
inc eax

xchg peax,pdword [pesp+psize*3] ;exchange string with the size of the string
xchg peax,pdword [pesp+psize*2] ;exchange itemId with the string
xchg peax,pdword [pesp+psize*1] ;exchange collection  with the itemId
xchg peax,pdword [pesp]         ;exchange return address  with the collection
push peax                       ;push return address
jmp ModifyItem
end if

;------------------------------------------------------------------------------
;This function will modify an integer in the list 1:collection 2:itemId 3:newItem
;------------------------------------------------------------------------------
if used ModifyDword
ModifyDword:
xor eax,eax
xchg peax,pdword [pesp+psize*3] ;exchange newItem with zero (the size of an integer)
xchg peax,pdword [pesp+psize*2] ;exchange itemId with newItem
xchg peax,pdword [pesp+psize*1] ;exchange collection with itemId
xchg peax,pdword [pesp]         ;exchange return address with collection
push peax                       ;push return address
;jmp ModifyItem
end if

;------------------------------------------------------------------------------
;This function will modify an item in the list 1:collection 2:itemId 3:newItem 4:itemSize
;------------------------------------------------------------------------------
if used ModifyItem | used ModifyDword
ModifyItem:
call denter

ppush o ;error function
call defaultListCheckup

;get list pointer
mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]

;check if the the item exists
mov eax,dword [pecx]
mov edx,dword [pebp+psize*3]
cmp eax,edx
jae n

q:
xor eax,eax
jmp o

n:
;itemId starts at 1, not zero
test edx,edx
cmovz eax,edx
jz o

;calculate the pointer of the new item in the list
if used is64
   shl rdx,3 ;*8 (size of a pointer)
else
   shl edx,2 ;*4  (size of a pointer)
end if

;check if the size param is zero, is that the case, just move the integer
mov eax,dword [pebp+psize*5]
test eax,eax
cmovz pesi,pdword [pebp+psize*4]
jz dwordMove

add eax,14

;reallocate the memory if necessary
push peax                  ;the size of the new memory
push pdword [pecx+pedx]    ;the memory to be resized
call ReSizeMem
mov pesi,peax

mov edi,dword [pebp+psize*5]
mov dword [peax],edi
add peax,4

push pedi
push pdword [pebp+psize*4]
push peax
call strcpyn

dwordMove:
mov pdword [pecx+pedx],pesi

xor eax,eax
inc eax

o:
call UnlockListAcces
call dleave
ret 4*psize
end if

;------------------------------------------------------------------------------
;This function will remove an item from the list by it's index
;1:collection 2:ndex
;return: deleted value
;------------------------------------------------------------------------------
if used RemoveByIndex
RemoveByIndex:
call denter

ppush errRemoveByIndex ;error function
call defaultListCheckup

;Lock the list so no other threads can use it
push INTERNAL_COLLECTION_LOCK
push pecx
call LockListAcces

;find the item of the index
push pdword [pebp+psize*3] ;index
push pecx                  ;collection
call FindDwordByIndex
test eax,eax
jz endRemoveByIndex

;decrement (index of the indexItem)
dec eax
push peax ;push for later use with RemoveDwordItem

push peax                  ;index
push pecx                  ;collection
call RemoveDwordItem

;remove the item itself
;push pdword [pebp+psize*3] ;index already pushed
push pecx                  ;collection
call RemoveItem

endRemoveByIndex:
push peax ;save return value

;--Unlock the list
push INTERNAL_COLLECTION_LOCK
push pecx
call ExternalUnlockListAccesEx

pop peax

errRemoveByIndex:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will remove an integer from the list by it's index
;1:collection 2:ndex
;return: deleted value
;------------------------------------------------------------------------------
if used RemoveDwordByIndex
RemoveDwordByIndex:
call denter

ppush errRemoveDwordByIndex ;error function
call defaultListCheckup

;Lock the list so no other threads cannot use it
push INTERNAL_COLLECTION_LOCK
push pecx
call LockListAcces

;find the item of the index
push pdword [pebp+psize*3] ;index
push pecx                  ;collection
call FindDwordByIndex
test eax,eax
jz endRemoveDwordByIndex

;decrement (index of the indexItem)
dec eax
push peax ;push for later use with RemoveDwordItem

push peax                  ;index
push pecx                  ;collection
call RemoveDwordItem

;remove the item itself
;push pdword [pebp+psize*3] ;index already pushed
push pecx                  ;collection
call RemoveDwordItem

endRemoveDwordByIndex:
push peax ;save return value

;--Unlock the list
push INTERNAL_COLLECTION_LOCK
push pecx
call ExternalUnlockListAccesEx

pop peax

errRemoveDwordByIndex:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will remove an integer from the list 1:collection 2:itemId
;return: deleted value
;------------------------------------------------------------------------------
if used RemoveDwordItem
RemoveDwordItem:
xor eax,eax
inc eax
xchg peax,pdword [pesp+psize*2] ;exchange itemId with zero (the flag REMOVE_INTEGER)
xchg peax,pdword [pesp+psize*1] ;exchange collection with itemId
xchg peax,pdword [pesp]         ;exchange return address with collection
push peax                       ;push return address
jmp RemoveItemEx
end if

;------------------------------------------------------------------------------
;This function will remove an item from the list 1:collection 2:itemId
;return: deleted value
;------------------------------------------------------------------------------
if used RemoveItem
RemoveItem:
xor eax,eax
xchg peax,pdword [pesp+psize*2] ;exchange itemId with zero (the flag REMOVE_ITEM)
xchg peax,pdword [pesp+psize*1] ;exchange collection with itemId
xchg peax,pdword [pesp]         ;exchange return address with collection
push peax                       ;push return address
;jmp RemoveItemEx
end if

;------------------------------------------------------------------------------
;This function will remove an item or integer from the list 1:collection 2:itemId 3:flags
;flags:
;-REMOVE_ITEM 0x0
;-REMOVE_INTEGER 0x1
;return: deleted value
;------------------------------------------------------------------------------
if used RemoveItemEx | used RemoveItem
RemoveItemEx:
call denter

ppush f ;error function
call defaultListCheckup

;get list pointer
mov pecx,pdword [pebp+psize*2]
mov pecx,pdword [pecx]
mov pedx,pecx ;backup list

;check if the the item exists
mov eax,dword [pecx]
mov edi,dword [pebp+psize*3]
cmp eax,edi
jae e

j:
xor eax,eax
jmp f

e:
;itemId starts at 1, not zero
test edi,edi
cmovz eax,edi
jz f

;calculate the pointer of the new item in the list
if used is64
   shl rdi,3 ;*8 (size of a pointer)
else
   shl edi,2 ;*4  (size of a pointer)
end if

;get pointer to the item in the list
add pedi,pecx

;test if flag REMOVE_INTEGER is set
test pdword [pebp+psize*4],1
jnz dontReleaseIntger

;free the item
push pdword [pedi]
call FreeMemLocation

dontReleaseIntger:
push pdword [pedi]  ;save deleted value (to return)
mov pdword [pedi],0 ;zero item
;set (r/e)si to the next pointer in the list
mov pesi,pedi
add pesi,psize

;get the ammount of items that need to be moved
mov ecx,dword [pecx]
sub ecx,dword [pebp+psize*3]

;copy every item to its new place
rep movsp

;decrement the ammount of items in the list
dec dword [pedx]

;xor eax,eax
;inc eax
pop peax ;return the old value

f:
call UnlockListAcces
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will remove the lsit itself 1:collection
;------------------------------------------------------------------------------
if used FreeIndexedCollectionWithItemCollections
FreeIndexedCollectionWithItemCollections:
mov peax,pdword [pesp+psize*1] ;collection
push peax                      ;Push collection on it's new place

mov peax,pdword [pesp+psize*1] ;return address
push peax                      ;push return address on it's new place

;fill in the new created parameter with flags
mov pdword [pesp+psize*2],USE_FREE_ITEM_FUNCTION or REMOVE_INDEX_LIST
;fill in the second new created parameter with FreeCollection
pmov pdword [pesp+psize*3],FreeCollection

jmp FreeCollectionEx
end if

;------------------------------------------------------------------------------
;This function will remove the lsit itself 1:collection
;------------------------------------------------------------------------------
if used FreeDwordCollection
FreeDwordCollection:
xor eax,eax
inc eax
xchg peax,pdword [pesp+psize] ;exchange collection with 1 (flag REMOVE_INTEGER)
xchg peax,pdword [pesp] ;exchange return address with collection
push peax                       ;push return address
jmp FreeCollectionEx
end if

;------------------------------------------------------------------------------
;This function will remove the list and uneven items in it 1:collection
;------------------------------------------------------------------------------
if used FreeIndexedCollection
FreeIndexedCollection:
mov eax,2
xchg peax,pdword [pesp+psize] ;exchange collection with 2 (flag REMOVE_INDEX_LIST)
xchg peax,pdword [pesp] ;exchange return address with collection
push peax                       ;push return address
jmp FreeCollectionEx
end if

;------------------------------------------------------------------------------
;This function will remove every item in the list and the lsit itself 1:collection
;------------------------------------------------------------------------------
if used FreeCollection
FreeCollection:
xor eax,eax
xchg peax,pdword [pesp+psize] ;exchange collection with zero (flag REMOVE_ITEM)
xchg peax,pdword [pesp] ;exchange return address with collection
push peax                       ;push return address
;jmp FreeCollectionEx
end if

;------------------------------------------------------------------------------
;This function will remove every item in the list and the lsit itself 1:collection 2:flags 3:FreeItemFunction (optional)
;flags:
;-REMOVE_ITEM 0x0
;-REMOVE_INTEGER 0x1
;-REMOVE_INDEX_LIST 0x2
;-ITEM_IS_COLLECTION 0x4
;-USE_FREE_ITEM_FUNCTION 0x8
;------------------------------------------------------------------------------
if used FreeCollectionEx | used FreeCollection
FreeCollectionEx:
call denter

ppush k ;error function
call defaultListCheckup

;get pointer to list
mov pedx,pdword [pebp+psize*2]

;Lock the list so no other threads can use it
push INTERNAL_COLLECTION_LOCK
push pedx
call LockListAcces

mov pedx,pdword [pedx]

;test of flag REMOVE_INTEGER
test dword [pebp+psize*3],1
jnz NoItems

;get ammount of items in the list
mov ecx,dword [pedx]
jpcxz NoItems

;set flags on EBX
mov ebx,dword [pebp+psize*3]

;Set (r/e)SI on the first item on the list
mov pesi,pedx
add pesi,psize

UndoItems:
  lodsp

  test ebx,2 ;test on REMOVE_INDEX_LIST
  jz justRemoveItem

  ;(un)set the flag unused flag REMOVE_INTEGER to detect the current item is even(index) or uneven (item)
  xor ebx,1

  test ebx,1
  jnz nextItem

  justRemoveItem:
  ;--calculate the pointer of the new item in the list
  ;--if used is64
  ;-   shl rdi,3 ;*8 (size of a pointer)
  ;-else
  ;-   shl edi,2 ;*4  (size of a pointer)
  ;-end if

  ;test if flag USE_FREE_ITEM_FUNCTION is set
  test ebx,USE_FREE_ITEM_FUNCTION
  jz DoNotCallFreeItemFunction

  ;free the item
  push peax
  call pdword [pebp+psize*4]
  jmp nextItem

  DoNotCallFreeItemFunction:
  ;check if we need to free the memory or a collection
  test ebx,ITEM_IS_COLLECTION
  jz justFreeMemory

  ;free the collection
  push peax
  call FreeCollection
  jmp nextItem

  justFreeMemory:
  ;free memory
  push peax
  call FreeMemLocation

  nextItem:
loop UndoItems
NoItems:

push pedx
call FreeMemLocation
push pdword [pebp+psize*2]
call FreeMemLocation

EndFreeCollectionEx:
xor eax,eax
inc eax

k:
;test if flag USE_FREE_ITEM_FUNCTION is set
test dword [pebp+psize*3],USE_FREE_ITEM_FUNCTION
jz returnByTwoParams

;call UnlockListAcces  We deleted the list, unlocking is not possible :)
call dleave
ret psize*3

;if USE_FREE_ITEM_FUNCTION flag is not set, there are only 2 parameters to release
returnByTwoParams:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will find a integer in the list 1:collection 2:integer to find
;return:Id of found item
;------------------------------------------------------------------------------
if used FindDword
FindDword:
call denter

ppush errFindDword ;error function
call defaultReadListCheckup

;get pointer to list
mov pedi,pdword [pebp+psize*2]
mov pedi,pdword [pedi]

;check if there are items in the list
mov ecx,dword [pedi]
mov edx,ecx ;backup ammount of items
jpcxz Foundtt

;set (r/e)DI pointer on the first item
add pedi,psize

;loop every item and see if the integer is equal to the one we try to find
mov peax,pdword [pebp+psize*3]

repnz scasp

cmovnz ecx,edx ;move is NOT ZERO flag is set

Foundtt:
;calculate the itemId
mov eax,edx
sub eax,ecx

tt:
call UnlockListWriteAcces

errFindDword:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will find a integer in the list, looking at the even items(those are indexes) and return the index in the list of the belonging item
;!!! not true, it returns the index of the item+1
;1:collection 2:integer to find
;return:Id of found item+1 <<<<<<<<<<<<<<<<
;------------------------------------------------------------------------------
if used FindDwordByEvenIndex
FindDwordByEvenIndex:
mov peax,checkEven
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
jmp FindDwordByIndexEx
end if

;------------------------------------------------------------------------------
;This function will find a integer in the list, looking at the uneven items(those are indexes) and return the index in the list of the belonging item
;1:collection 2:integer to find
;return:Id of found item
;------------------------------------------------------------------------------
if used FindDwordByIndex
FindDwordByIndex:
mov peax,checkNotEven
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
;jmp
end if

;------------------------------------------------------------------------------
;This function will find a integer in the list, looking at the uneven items or even, depending on the function passed in the paramd return the index in the list of the belonging item
;1:collection 2:integer to find  3:decidefunction
;return:Id of found item
;------------------------------------------------------------------------------
if used FindDwordByIndexEx | used FindDwordByIndex
FindDwordByIndexEx:
call denter

ppush errFindDwordIndexCollection ;error function
call defaultReadListCheckup

;get pointer to list
mov pedi,pdword [pebp+psize*2]
mov pedi,pdword [pedi]
mov pesi,pdword [pebp+psize*4] ;even/uneven function

;check if there are items in the list
mov ecx,dword [pedi]
mov edx,ecx ;backup ammount of items
jpcxz endFindDwordIndexCollection

;set (r/e)DI pointer on the first item
add pedi,psize

;loop every item and see if the integer is equal to the one we try to find
mov peax,pdword [pebp+psize*3]

rcr ecx,1
unevenIndexFound:

rcl ecx,1
repnz scasp

jpcxz notFoundFindDwordByIndex

;check in the index of the list is an even number
rcr ecx,1
jmp pesi

notFoundFindDwordByIndex:
inc ecx

notFoundFindDwordByIndexjump:

cmovnz ecx,edx ;move is NOT ZERO flag is set

endFindDwordIndexCollection:
;calculate the itemId
mov eax,edx
sub eax,ecx

call UnlockListWriteAcces

errFindDwordIndexCollection:
call dleave
ret psize*3
end if

;------------
;--function not even check
;------------
if used checkNotEven
checkNotEven:
jnc unevenIndexFound ;jump if not even

;restore shifted bit by a zero (the same as retoring ecx and a operation decrement )
;shl can't be used becease this operation sets the zero flag and that one is needed by the cmovnz operation
clc
rcl ecx,1

jmp notFoundFindDwordByIndexjump
end if

;------------
;--function even check
;------------
if used checkEven
checkEven:
jc unevenIndexFound ;jump if even

;restore shifted bit by a zero (the same as retoring ecx and a operation decrement )
;shl can't be used becease this operation sets the zero flag and that one is needed by the cmovnz operation
clc
rcl ecx,1
jmp notFoundFindDwordByIndexjump
end if

;------------------------------------------------------------------------------
;This function will look for a index that is bigger or smaller than the one given in param 2 and return the index of the belonging item
;1:collection 2:integer to find return:Id of found item
;------------------------------------------------------------------------------
if used FindDwordBiggerOrEqual | used FindDwordBySize
FindDwordBiggerOrEqual:
FindDwordBySize:
mov peax,checkUnevenIndexBiggerEqual
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
jmp FindDwordByStringIndexEx
end if

;------------------------------------------------------------------------------
;This function will look for a EVEN index that is bigger or smaller than the one given in param 2 and return the index of the belonging item
;1:collection 2:integer to find return:Id of found item
;return: index of found index
;------------------------------------------------------------------------------
if used FindDwordEvenLowerOrEqual
FindDwordEvenLowerOrEqual:
mov peax,checkEvenIndexLowerOrEqual
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
jmp FindDwordByStringIndexEx
end if

;------------------------------------------------------------------------------
;This function will look for a EVEN index that is bigger or smaller than the one given in param 2 and return the index of the belonging item
;1:collection 2:integer to find return:Id of found item 3:maximum (this argument has to be deleted from stack by the caller)
;return: index of found index
;------------------------------------------------------------------------------
if used FindDwordEvenBetween
FindDwordEvenBetween:
mov peax,checkEvenIndexBetween
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
jmp FindDwordByStringIndexEx
end if

;------------------------------------------------------------------------------
;This function will look for a index that is a string and return the index in the list of the belonging item
;1:collection 2:integer to find return:Id of found item
;------------------------------------------------------------------------------
if used FindDwordByStringIndex
FindDwordByStringIndex:
mov peax,CheckString
xchg pdword [pesp+psize*2],peax  ;integer to find => decidefunction
xchg pdword [pesp+psize*1],peax  ;collection => integer to find
xchg pdword [pesp],peax  ;return address => collection
push peax                ;push return address
;jmp
end if

;------------------------------------------------------------------------------
;This function will look for a index that is a string and return the index in the list of the belonging item
;1:collection 2:integer to find return:Id of found item 3:decidefunction
;------------------------------------------------------------------------------
if used FindDwordByStringIndexEx | used FindDwordByStringIndex
FindDwordByStringIndexEx:
call denter

ppush errFindDwordByStringIndex ;error function
call defaultReadListCheckup

;get pointer to list
mov pesi,pdword [pebp+psize*2]
mov pesi,pdword [pesi]

;check if there are items in the list
mov ecx,dword [pesi]
mov edx,ecx ;backup ammount of items
xor eax,eax
jpcxz endFindDwordByStringIndex

shr ecx,1 ;ecx = ecx / 2

;set (r/e)SI pointer on the first item
add pesi,psize


mov pedi,pdword [pebp+psize*3]

if used is32
  ;set decidefunction on (R/E)BP and push original (R/E)BP
  push pdword [pebp+psize*4]
  xchg pebp,pdword [pesp]
else
  ;R12 => decidefunction
  mov r12,pdword [pebp+psize*4]
end if

;loop every item compair the string
loopIndexes:

  lodsp

  if used is32
    jmp ebp
  else
    jmp r12
  end if

nextLoopIndexes:
loop loopIndexes

if used is32
  pop pebp
end if

xor eax,eax ;clear found index
jmp endFindDwordByStringIndex

FoundIndex:
if used is32
  pop pebp
end if

;calculate the itemId
mov eax,edx
shr eax,1 ;eax = eax / 2
sub eax,ecx
inc eax
shl eax,1 ;eax = eax * 2

endFindDwordByStringIndex:
call UnlockListWriteAcces

errFindDwordByStringIndex:
call dleave
ret psize*2
end if

;-----
;check string
;-----
if used CheckString
CheckString:
add peax,4 ;skip the size member

push peax
push pedi
call strcmpUtf8CaseInsensitive
test eax,eax
jz FoundIndex

lodsp ;Add 4 (to skip the item and go to the next index )
jmp nextLoopIndexes
end if

;-----
;check value bigger or equal than
;-----
if used checkUnevenIndexBiggerEqual
checkUnevenIndexBiggerEqual:
cmp peax,pedi
jae FoundIndex

lodsp ;Add 4 (to skip the item and go to the next index )
jmp nextLoopIndexes
end if

;-----
;check value smaller or equal than
;-----
if used checkEvenIndexLowerOrEqual
checkEvenIndexLowerOrEqual:
lodsp ;Add 4 (to skip the item and go to the next index )

cmp peax,pedi
jbe FoundIndex

jmp nextLoopIndexes
end if

;-----
;check value smaller or equal than x and not higher than pebp-psize*4
;-----
if used checkEvenIndexBetween
checkEvenIndexBetween:
lodsp ;Add 4 (to skip the item and go to the next index )

cmp peax,pedi
ja _nextcheckEvenIndexBetween

cmp peax,pdword [pebp-psize*4]
ja FoundIndex

_nextcheckEvenIndexBetween:
jmp nextLoopIndexes
end if