
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;------------------------------------------------------------------------------
;The GetCopyOfInternalProtocolStructures function will create a copy of the internal protocol structures and returns the first structure in the linked list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetCopyOfInternalProtocolStructures_c:
    StartProc 1

    mov rcx,retGetCopyOfInternalProtocolStructures_c
    xchg qword [rsp],rcx
    jmp GetCopyOfInternalProtocolStructures_asm

    retGetCopyOfInternalProtocolStructures_c:
    sub rsp,psize
    EndProc 1
    jmp rcx
else
    GetCopyOfInternalProtocolStructures_c:
end if

;---------
;asm implementation
;---------
GetCopyOfInternalProtocolStructures_asm:
call denter

;get internal protocol structure
push pdword [pebp+psize*2] ;psess
call getInternalProtocolStructure
test peax,peax
jz endGetCopyOfInternalProtocolStructures

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endGetCopyOfInternalProtocolStructures ;jump if no control panel rights

mov pebx,[protocols]    ;the first internal protocol structure
xor pedx,pedx

copyInternalProtocolStructure:
  push SIZEOF_INTERNALPROTOCOLSTRUCTURE
  call GetMemLocation
  test pedx,pedx
  cmovz pedx,peax ;(R/E)DX => internal protocol structure
  cmovz pecx,peax ;(R/E)CX => first protocol structure
  jz firstInternalProtocolStructure

  mov pdword [pedx],peax ;'next' structure
  mov pedx,peax

  firstInternalProtocolStructure:
  mov peax,pdword [pebx+psize*1]  ;psess
  mov pdword [pedx+psize*1], peax ;set psess

  ;create copyPsess
  mov peax,pdword [pebx+psize*2]  ;copyPsess
  push pdword [peax+(psize*1)] ;Protocolname
  call CreateProtocolSession
  mov pdword [pedx+psize*2],peax ;set copyPsess

  ;copy main
  push 0x1 ;FLAG_USE_COPYINTERNALROWSCOLLECTION
  push pdword [pebx+(psize*5)] ;main
  add pdword [pesp],psize ;skip the next pointer
  call copyControlPanelOptiona
  mov pdword [pedx+psize*5],peax ;set main

  ;copy subOptions
  mov peax,pdword [pebx+(psize*6)] ;source subOptions
  test peax,peax
  jz noSubOptionsToCopy
  lea pedi,pdword [pedx+psize*6]   ;ouput subOptions

  copySubOptions:
    mov pesi,peax

    push 0x1 ;FLAG_USE_COPYINTERNALROWSCOLLECTION
    push peax ;subOptions[i]
    add pdword [pesp],psize ;skip the 'next' pointer
    call copyControlPanelOptiona
    stosp ;store the new pointer on ouput subOptions
    mov pedi,peax ;ouput = 'next' pointer

    lodsp  ;get 'next' structure
    test peax,peax
  jnz copySubOptions

  noSubOptionsToCopy:
  lea pesi,pdword [pebx+(psize*7)] ;source status
  lea pedi,pdword [pedx+psize*7]   ;ouput status

  if used is64
    movsq ;copy status, type
    movsq ;copy highestSupportedVersion and extraSize
  else
    movsd ;copy status
    movsd ;copy type
    movsd ;copy highestSupportedVersion
    movsd ;copy extraSize
  end if

  ;copy ProtocolSignature
  push pdword [pebx+((psize*7)+0x10+(psize*2))]
  call copyProtocolSignature
  mov pdword [pedx+((psize*7)+0x10+(psize*2))],peax

nextCopyInternalProtocolStructure:
  mov pebx,pdword [pebx] ;get next structure
  test pebx,pebx  ;jump if there is an next structure
jnz copyInternalProtocolStructure

mov peax,pecx ;return first structure

endGetCopyOfInternalProtocolStructures:
call dleave
ret psize

;------------------------------------------------------------------------------
;The ReleaseCopyOfInternalProtocolStructures function will release all recources allocated by function GetCopyOfInternalProtocolStructures.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    ReleaseCopyOfInternalProtocolStructures_c:
    StartProc 1

    mov rcx,retReleaseCopyOfInternalProtocolStructures_c
    xchg qword [rsp],rcx
    jmp ReleaseCopyOfInternalProtocolStructures_asm

    retReleaseCopyOfInternalProtocolStructures_c:
    sub rsp,psize*1
    EndProc 1
    jmp rcx
else
    ReleaseCopyOfInternalProtocolStructures_c:
end if

;---------
;asm implementation
;---------
ReleaseCopyOfInternalProtocolStructures_asm:
call denter

;since GetCopyOfInternalProtocolStructures does not make a copy, but return internal pointer, do nothing...
mov pesi,pdword [pebp+psize*2] ;InternalProtocolStructure
test pesi,pesi
jz endReleaseCopyOfInternalProtocolStructures

loopThoughInternalProtocolStructures:
  ;free copy of ProtocolSession
  push pdword [pesi+psize*2]
  call FreeProtocolSession

  ;free main
  push pdword [pesi+psize*5]
  call FreeLinkedlistControlPanelOptions

  ;free subOptions
  push pdword [pesi+psize*6]
  call FreeLinkedlistControlPanelOptions

  ;free ProtocolSignature
  push pdword [pesi+((psize*7)+0x10+(psize*2))]
  call FreeMemLocation

  push pesi ;push InternalProtocolStructure for FreeMemLocation
  mov pesi,pdword [pesi] ;get next InternalProtocolStructure
  call FreeMemLocation ;free InternalProtocolStructure

  test pesi,pesi
jnz loopThoughInternalProtocolStructures

endReleaseCopyOfInternalProtocolStructures:
call dleave
ret psize*1

;------------------------------------------------------------------------------
;The CountListenSessions function will return the ammount of active ListenSessions.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    CountListenSessions_c:
    StartProc 1

    mov rcx,retCountListenSessions_c
    xchg qword [rsp],rcx
    jmp CountListenSessions_asm

    retCountListenSessions_c:
    sub rsp,psize*1
    EndProc 1
    jmp rcx
else
    CountListenSessions_c:
end if

;---------
;asm implementation
;---------
CountListenSessions_asm:
;get internal protocol structure
push pdword [pesp+psize] ;psess
call getInternalProtocolStructure
test peax,peax
jz endCountListenSessions

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endCountListenSessions ;jump if no control panel rights


;get the ammount of items in the collection listenSessions
push [listenSessions]
call CountItems
shr eax,1 ;EAX = EAX / 2. Divide by two because the items are indexed by the Listensession (lsess) and we shouldn't count those.

endCountListenSessions:
ret psize

;------------------------------------------------------------------------------
;The GetCopyOfInternalListenStructure function will create a copy of the internal Listen structure and returns the ListenSession.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetCopyOfInternalListenStructure_c:
    StartProc 3

    mov rcx,retGetCopyOfInternalListenStructure_c
    xchg qword [rsp],rcx
    jmp GetCopyOfInternalListenStructure_asm

    retGetCopyOfInternalListenStructure_c:
    sub rsp,psize*3
    EndProc 3
    jmp rcx
else
    GetCopyOfInternalListenStructure_c:
end if

;---------
;asm implementation
;---------
GetCopyOfInternalListenStructure_asm:
call denter

;get internal protocol structure
push pdword [pebp+psize*2] ;psess
call getInternalProtocolStructure
test peax,peax
jz endGetCopyOfInternalListenStructure

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endGetCopyOfInternalListenStructure ;jump if no control panel rights

;lock the list for write actions
push [listenSessions]
call LockListWriteAcces

mov ecx,dword [pebp+psize*3] ;parameter id
shl ecx,1 ;ECX = ECX * 2 (since the list is indexed)
dec ecx   ;--ECX since we want the index (this is a pointer to the ListenSession)

;get the ListenSession
push pecx
push [listenSessions]
call GetDwordItem
test peax,peax
jz endAndUnlockGetCopyOfInternalListenStructure

mov pesi,pdword [pebp+psize*4] ;paramter *lsess
test pesi,pesi
jz dontReturnListenSession

  mov pdword [pesi],peax ;set *lsess parameter

dontReturnListenSession:
inc ecx ;++ECX; since we want the the InternalListenSessionStructure.

;get the InternalListenSession
push pecx
push [listenSessions]
call GetDwordItem
test peax,peax
jz endAndUnlockGetCopyOfInternalListenStructure

mov pesi,peax ;(R/E)SI => InternalListenSession Structure

;create and InternalListenSession Structure
push SIZEOF_INTERNALLISTENSESSION
call GetMemLocation
mov pebx,peax ;(R/E)BX => new created InternalListenSession Structure

if used is32
  mov eax,dword [pesi+psize*5] ;EAX = member status from original InternalListenSession Structure
  mov dword [pebx+psize*5],eax ;Set member status in the new created InternalListenSession Structure

  mov eax,dword [pesi+(psize*5)+4] ;EAX = member id from original InternalListenSession Structure
  mov dword [pebx+(psize*5)+4],eax ;Set member id in the new created InternalListenSession Structure
else
  mov peax,qword [pesi+psize*5] ;EAX = member status and id from original InternalListenSession Structure
  mov qword [pebx+psize*5],peax ;Set member status and id in the new created InternalListenSession Structure
end if

;copy the name
push pdword [pesi+(5*psize)+8+psize]
call copyStringA
mov pdword [pebx+(5*psize)+8+psize],peax ;fill in the ListenSession name

;Create an collection pointer without acces flags (so a direct pointer) and size as pointer (not integer) so it is aligned for C compatibility
mov pesi,pdword [pesi] ;get pointer to bindedToHandle

;make a copy of the collection
push pesi
call copyDwordList
mov pesi,pdword [peax] ;get pionter to the actual collection
mov pdword [pebx],pesi ;set member BindedToHandles in new created InternalListenSession

;free the 'header' of the collection
push peax
call FreeMemLocation

mov peax,pebx ;Set the new created InternalListenSession Structure as return value

endAndUnlockGetCopyOfInternalListenStructure:
push peax

;unlock the list for write actions
push [listenSessions]
call ExternalUnlockListWriteAcces

pop peax

endGetCopyOfInternalListenStructure:
call dleave
ret psize*3

;---------------------------------------------------------------------------------------------------------------------------------------------------------
;------------------------------------------------------------------------------
;The GetCopyOfInternalListenStructureByLsess function will create a copy of the internal Listen structure that belongs to the passed ListenSession.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetCopyOfInternalListenStructureByLsess_c:
    StartProc 2

    mov rcx,retGetCopyOfInternalListenStructureByLsess_c
    xchg qword [rsp],rcx
    jmp GetCopyOfInternalListenStructureByLsess_asm

    retGetCopyOfInternalListenStructureByLsess_c:
    sub rsp,psize*2
    EndProc 2
    jmp rcx

else
    GetCopyOfInternalListenStructureByLsess_c:
end if

;---------
;asm implementation
;---------
GetCopyOfInternalListenStructureByLsess_asm:
call denter

;lock the list for write actions
push [listenSessions]
call LockListWriteAcces

;get Id for the ListenSession
push pdword [pebp+psize*3]  ;lsess to look for
push [listenSessions]       ;collection
call FindDword
test eax,eax
jz endGetCopyOfInternalListenStructureByLsess

inc eax
shr eax,1 ;divide by 2 to the the Id

push 0    ;output ListenSession
push peax ;id
push pdword [pebp+psize*2]  ;psess
call GetCopyOfInternalListenStructure_asm

push peax

;unlock the list for write actions
push [listenSessions]
call ExternalUnlockListWriteAcces

pop peax

endGetCopyOfInternalListenStructureByLsess:
call dleave
ret psize*2

;------------------------------------------------------------------------------
;The ReleaseCopyOfInternalListenStructure function will release all recources allocated by function GetCopyOfInternalListenStructure.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    ReleaseCopyOfInternalListenStructure_c:
    StartProc 1

    mov rcx,retReleaseCopyOfInternalListenStructure_c
    xchg qword [rsp],rcx
    jmp ReleaseCopyOfInternalListenStructure_asm

    retReleaseCopyOfInternalListenStructure_c:
    sub rsp,psize*1
    EndProc 1
    jmp rcx
else
    ReleaseCopyOfInternalListenStructure_c:
end if

;---------
;asm implementation
;---------
ReleaseCopyOfInternalListenStructure_asm:

mov peax,pdword [pesp+psize] ;parameter cpyInternalListenStructure

push pdword [peax] ;push for FreeMemLocation (to free BindedToHandles)
push pdword [peax+(5*psize)+8+psize] ;push name
push peax          ;push for FreeMemLocation (to free InternalListenStructure)

call FreeMemLocation
call FreeMemLocation
call FreeMemLocation

endReleaseCopyOfInternalListenStructure:
ret psize*1

;------------------------------------------------------------------------------
;The RegisterProtocolLoadedCallbackFunction function will register a callback function that will be called every time a new protocol is loaded.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    RegisterProtocolLoadedCallbackFunction_c:
    StartProc 2

    mov rcx,retRegisterProtocolLoadedCallbackFunction_c
    xchg qword [rsp],rcx
    jmp RegisterProtocolLoadedCallbackFunction_asm

    retRegisterProtocolLoadedCallbackFunction_c:
    sub rsp,psize*2
    EndProc 2
    jmp rcx
else
    RegisterProtocolLoadedCallbackFunction_c:
end if

;---------
;asm implementation
;---------
RegisterProtocolLoadedCallbackFunction_asm:
call denter

;get internal protocol structure
push pdword [pebp+psize*2] ;psess
call getInternalProtocolStructure
test peax,peax
jz endRegisterProtocolLoadedCallbackFunction

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endRegisterProtocolLoadedCallbackFunction ;jump if no control panel rights

push pdword [pebp+psize*3] ;function to register
push [registerdProtocolIsLoadedFunctions]
call AddDword

xor eax,eax
inc eax ;return true

endRegisterProtocolLoadedCallbackFunction:
call dleave
ret psize*2

;------------------------------------------------------------------------------
;The RegisterProtocolLoadedCallbackFunction function will register a callback function that will be called every time a new protocol is loaded.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    RegisterListenSessionIsCreatedCallbackFunction_c:
    StartProc 2

    mov rcx,retRegisterListenSessionIsCreatedCallbackFunction_c
    xchg qword [rsp],rcx
    jmp RegisterListenSessionIsCreatedCallbackFunction_asm

    retRegisterListenSessionIsCreatedCallbackFunction_c:
    sub rsp,psize*2
    EndProc 2
    jmp rcx
else
    RegisterListenSessionIsCreatedCallbackFunction_c:
end if

;---------
;asm implementation
;---------
RegisterListenSessionIsCreatedCallbackFunction_asm:
call denter

;get internal protocol structure
push pdword [pebp+psize*2] ;psess
call getInternalProtocolStructure
test peax,peax
jz endRegisterListenSessionIsCreatedCallbackFunction

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endRegisterListenSessionIsCreatedCallbackFunction ;jump if no control panel rights

push pdword [pebp+psize*3] ;function to register
push [registerdListenSessionIsCreatedFunctions]
call AddDword

xor eax,eax
inc eax ;return true

endRegisterListenSessionIsCreatedCallbackFunction:
call dleave
ret psize*2

;------------------------------------------------------------------------------
;The ShutDownServer function will signal all the ListenSessions to stop
;------------------------------------------------------------------------------   \
;---------
;Internel implementation (skip ProtocolSession check and don't stop service)
ShutdownServer_internal:
xchg pdword [pesp],peax ;Set ProtocolSession to random number
push peax               ;Set returnvalue

call denter
jmp jmpShutdownServer_internal

;---------
;C implementation
;---------
if used is64
    ShutdownServer_c:
    StartProc 1

    mov rcx,retShutdownServer_c
    xchg qword [rsp],rcx
    jmp ShutdownServer_asm

    retShutdownServer_c:
    sub rsp,1*psize
    EndProc 1
    jmp rcx
else
    ShutdownServer_c:
end if
;---------
;asm implementation
;---------
ShutdownServer_asm:
call denter

;get internal protocol structure
push pdword [pebp+psize*2] ;psess
call getInternalProtocolStructure
test peax,peax
jz endShutdownServer

  mov eax,dword [peax+(psize*7)+4] ;get member type
  and eax,PT_CONTROLPANEL ;check on CONTROL_PANEL rights
  jz endShutdownServer ;jump if no control panel rights

if used is32 & used isWin
  ;are we 9x?
  cmp [is9x],0
  jne stopListenThreads
end if

;Are we running as service?
cmp [Handle],0
je stopListenThreads

;--;Signal service to shut down
;--call StopService ;StopCurrentService

;Create a new thread to tell the serice to stop (for non-blocking effect)
push 0   ;STOP_SERVICE ;flag
ppush StopDeleteService ;base
call createthread
jmp endShutdownServer

jmpShutdownServer_internal:
stopListenThreads:
push [listenSessions]
call CountItems
mov ecx,eax
test eax,eax
jz endShutdownServer

closeAllListenSessionsLoop:
  rcr ecx,1 ;check first bit for even/uneven
  jnc nextCloseListenSessionLoop  ;jump if even (even items are Internal ListenSessions)
  rcl ecx,1

  push pecx
  push [listenSessions]
  call GetDwordItem
  test peax,peax
  jz nextCloseListenSessionLooprcr

  push 0    ;flags
  push peax ;ListenSession

  ;find internal ListenSession
  push peax                  ;name (lsess)
  push [listenSessions]      ;collection
  call GetIndexedDwordItem
  mov peax,pdword [peax]     ;BindedTohandle

  ;get first psess of binded protocols
  push 1
  push peax
  call GetDwordItem

  push peax ;Push the Psess (the reveiver protocol)

  call CloseListenSessionEx_asm

nextCloseListenSessionLooprcr:
  rcr ecx,1

nextCloseListenSessionLoop:
  rcl ecx,1
loop closeAllListenSessionsLoop

endShutdownServer:
call dleave
ret psize