
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;-;I don't need UTF-8 High-low case conversion in this project and since the table used takes up ~2kB in size I exclude it.
;-virtual at 0
;-dontImplementCaseUTF8:
;-call dontImplementCaseUTF8
;-end virtual



;TODO:
;-see StrToIntWithBaseAsTerminator

;----------------------------------------------------
;                      strlen
;
;input:
;    1: Pointer to ASCII/ANSI/UTF-8 string
;
;output: (R/E)AX = size of string in bytes
;
;----------------------------------------------------
if used strlen
strlen:
  call denter

  xor ecx,ecx  ; zero ECX

  dec pecx     ; (R/E)CX => -1  (scasb will be repeated for 0xFFFFFFFF(FFFFFFFF) times, or until zero byte is found)
  xor eax,eax  ; byte to find with scasb => '\0'
  mov pedi,pdword [pebp+(psize*2)] ;(R/E)DI => pointer to ASCII string
  mov pedx,pedi ; (R/E)DX => copy of pointer to ASCII string

  test pedi,pedi
  jz _endStrlen    ;jump if pointer is zero

  repnz scasb   ;loop string until zero byte is found

endStrlen:
  sub pedi,pedx ; current pos of zero byte - begin of string = size of string
  mov peax,pedi ;(R/E)AX => size of string
  dec peax

_endStrlen:
  call dleave
  ret (psize*1)
end if

;----------------------------------------------------
;                      strlenW
;
;input:
;    1: Pointer to UTF-16 string
;
;output: (R/E)AX = size of string IN BYTES!!!
;
;----------------------------------------------------
if used strlenW
strlenW:
  call denter

  xor ecx,ecx  ; zero ECX

  dec pecx     ; (R/E)CX => -1  (scasb will be repeated for 0xFFFFFFFF(FFFFFFFF) times, or until zero byte is found)
  xor eax,eax  ; byte to find with scasb => '\0'
  mov pedi,pdword [pebp+(psize*2)] ;(R/E)DI => pointer to UNICODE string
  mov pedx,pedi ; (R/E)DX => copy of pointer to UNICODE string

  test pedi,pedi
  jz _endStrlenW    ;jump if pointer is zero

  repnz scasw   ;loop string until zero word is found

endStrlenW:
  sub pedi,pedx ; current pos of zero byte - begin of string = size of string
  mov peax,pedi ;(R/E)AX => size of string
  sub peax,2

_endStrlenW:
  call dleave
  ret (psize*1)
end if

;----------------------------------------------------
;                      strcpy
;This function copies source string to pointer destination.
;
;input:
;    1: Pointer to destination ASCII/ANSI/UTF-8 string
;    2: Pointer to source ASCII string
;
;output: none
;
;----------------------------------------------------
if used strcpy
strcpy:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  cld
        
  StartCpy:
    lodsb
    stosb
    test al,al
  jnz StartCpy

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      strcat
;This function copies source string to pointer destination.
;
;input:
;    1: Pointer to destination Asci/UTF-8 string
;    2: Pointer to source  Asci/UTF-8 string
;
;output: none
;
;----------------------------------------------------
if used strcat
strcat:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  cld

  push pedi
  call strlen
  add pedi,peax
        
  StartCatA:
    lodsb
    stosb
    test al,al
  jnz StartCatA

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      strcpyn
;This function copies source string to pointer destination with specific size.
;
;input:
;    1: Pointer to destination ASCII/ANSI/UTF-8 string
;    2: Pointer to source ASCII/ANSI/UTF-8 string
;    3: Size to copy source to destination
;
;output: none
;
;----------------------------------------------------
if used strcpyn
strcpyn:
  call denter

  mov pecx,pdword [pebp+(psize*4)]       ;ECX = size
  mov pedi,pdword [pebp+(psize*2)]       ;EDI= destination
  mov pesi,pdword [pebp+(psize*3)]       ;ESI = source
  rep movsb                              ;Copy [ESI] to [EDI] with size ECX

  call dleave
  ret (psize*3)
end if

;----------------------------------------------------
;                      strcpyW
;This function copies source UTF-16 string to pointer UTF-16 destination.
;
;input:
;    1: Pointer to destination UTF-16 string
;    2: Pointer to source UTF-16 string
;
;output: none
;
;----------------------------------------------------
if used strcpyW
strcpyW:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld
        
  call InternalStrcpyW

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalStrcpyW
;This function copies source UTF-16 string to pointer UTF-16 destination.
;
;input:
;    (R/E)DI: Pointer to destination UTF-16 string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source UTF-16 string + size(Pointer to source UTF-16 string)
;
;
;----------------------------------------------------
if used InternalStrcpyW
InternalStrcpyW:
  lodsw
  stosw
  test ax,ax
jnz InternalStrcpyW

  ret
end if

;----------------------------------------------------
;                      InternalStrcpyWCaseHigh
;This function copies source UTF-16 string to pointer UTF-16 destinationand higher low characters.
;
;input:
;    (R/E)DI: Pointer to destination UTF-16 string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source UTF-16 string + size(Pointer to source UTF-16 string)
;
;
;----------------------------------------------------
if used InternalStrcpyWCaseHigh
InternalStrcpyWCaseHigh:
  lodsw
  test ah,ah        ;don;t higher the case if the higher byte in the unicode character is set.
  jnz dontHigherCase
    call SetHighCase
  dontHigherCase:
  stosw
  test ax,ax
jnz InternalStrcpyWCaseHigh

  ret
end if

;----------------------------------------------------
;                      strcpyAW
;This function copies source ANSI string to pointer UTF-16 destination.
;
;input:
;    1: Pointer to destination UTF-16 string
;    2: Pointer to source ANSI string
;
;output: none
;
;----------------------------------------------------
if used strcpyAW
strcpyAW:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld
        
  call InternalStrcpyAW

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalStrcpyAW
;This function copies source ANSI string to pointer UTF-16 destination.
;
;input:
;    (R/E)DI: Pointer to destination UTF-16 string
;    (R/E)SI: Pointer to source ANSI string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)SI => Pointer to source ANSI string + size(Pointer to source ANSI string)
;
;
;----------------------------------------------------
if used InternalStrcpyAW
InternalStrcpyAW:

InternalStrcpyAW_loop:
  xor ax,ax
  lodsb

  call aToUTFCode

  stosw
  test ax,ax
jnz InternalStrcpyAW_loop

  ret
end if

;----------------------------------------------------
;                      strlenUtf8W
;
;input:
;    1: Pointer to UTF-8 string
;    2: flags
;      0x2:escape json characters
;
;output: (R/E)AX = the size of string in UTF-16 without zero byte
;
;----------------------------------------------------
if used strlenUtf8W
strlenUtf8W:
  call denter

  xor ecx,ecx  ; zero ECX

  mov pesi,pdword [pebp+(psize*2)] ;(R/E)DI => pointer to ASCII string

  test pesi,pesi
  jz _endStrlenUtf8W    ;jump if pointer is zero

  loopStrlenUtf8W:
    xor eax,eax
    lodsb

    call Utf8ToUnicode

    test eax,eax
    je _endStrlenUtf8W

    test pdword [pebp+psize*3],2
    jz dontEscapeJsonChar__strlenUtf8W

      call GetEscapeCharJson
      test dx,dx
      jz dontEscapeJsonChar__strlenUtf8W ;not an character to escape

      add pecx,2 ;add one extra escape word
      mov ax,dx

    dontEscapeJsonChar__strlenUtf8W:

    ;at least 1 word char in UTF-16
    add pecx,2

    cmp eax,0x010000
    jb loopStrlenUtf8W

    ;at least 2 words char in UTF-16
    add pecx,2
  jmp loopStrlenUtf8W

_endStrlenUtf8W:
  mov peax,pecx

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      strlenAUtf8
;
;input:
;    1: Pointer to ANSI string
;    2: flags
;      0x2:escape json characters
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9
;
;output: (R/E)AX = the size of string in UTF-8 without zero byte
;
;----------------------------------------------------
if used strlenAUtf8
strlenAUtf8:
  call denter

  xor ecx,ecx  ; zero ECX

  mov pesi,pdword [pebp+(psize*2)] ;(R/E)DI => pointer to ASCII string

  test pesi,pesi
  jz _endStrlenAutf8    ;jump if pointer is zero

  loopStrlenAutf8:
    xor ax,ax
    lodsb
    call aToUTFCode

    test ax,ax
    je _endStrlenAutf8

     test pdword [pebp+psize*3],0x40
    jz onlyAZ09__strlenAUtf8
      cmp ax,' '
      jne notSpace
 
        mov ax,'-'
        jmp dontEscapeJsonChar__strlenAUtf8

      notSpace:
      cmp ax,'0'
      jb loopStrlenAutf8
      cmp ax,' '
      jne notSpace5
 
        mov ax,'-'
        jmp dontEscapeJsonChar__strlenAUtf8

      notSpace5:
      cmp ax,'9'+1
      jc dontEscapeJsonChar__strlenAUtf8

      cmp ax,'A'
      jc loopStrlenAutf8

      cmp ax,'Z'+1
      jc dontEscapeJsonChar__strlenAUtf8

      cmp ax,'a'
      jc loopStrlenAutf8

      cmp ax,'z'
      ja loopStrlenAutf8

    onlyAZ09__strlenAUtf8:
     test pdword [pebp+psize*3],0x20
    jz dontEscapeTerminalChar__strlenAUtf8

      call GetEscapeCharTerminal
      test dx,dx
      jz dontEscapeJsonChar__strlenAUtf8 ;not an character to escape

      inc pecx ;add one extra escape byte
      mov ax,dx

    dontEscapeTerminalChar__strlenAUtf8:
    test pdword [pebp+psize*3],2
    jz dontEscapeJsonChar__strlenAUtf8

      call GetEscapeCharJson
      test dx,dx
      jz dontEscapeJsonChar__strlenAUtf8 ;not an character to escape

      inc pecx ;add one extra escape byte
      mov ax,dx

    dontEscapeJsonChar__strlenAUtf8:

    ;at least 1 bytes char in UTF-8
    inc pecx

    cmp ax,0x80
    jb loopStrlenAutf8

    ;at least 2 bytes char in UTF-8
    inc pecx

    cmp ax,0x800
    jb loopStrlenAutf8

    ;at least 3 bytes char in UTF-8
    inc pecx

    ;-cmp ax,0x10000
    ;-jb loopStrlenAutf8

    ;-;4 bytes char in UTF-8
    ;-inc pecx
  jmp loopStrlenAutf8

_endStrlenAutf8:
  mov peax,pecx

  call dleave
  ret (psize*2)
end if


;----------------------------------------------------
;                      strcpyUtf8ToW
;This function copies source UTF-8 string to pointer UTF-16 destination.
;
;input:
;    1: Pointer to destination UTF-16 string
;    2: Pointer to source UTF-8 string
;
;output: none
;
;----------------------------------------------------
if used strcpyUtf8ToW
strcpyUtf8ToW:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld

  push 0
  call InternalstrcpyUtf8ToW

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      strlenWUtf8
;
;input:
;    1: Pointer to UTF-16 string  / UTF-8 string
;    2: flags
;      0x2:escape json characters
;      0x4:is UTF-8 string instead of UTF-16 string
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9
;
;output: (R/E)AX = the size of string in UTF-8 without zero byte
;
;----------------------------------------------------
if used strlenWUtf8
strlenWUtf8:
  call denter

  xor ecx,ecx  ; zero ECX

  mov pesi,pdword [pebp+(psize*2)] ;(R/E)DI => pointer to ASCII string

  test pesi,pesi
  jz _endStrlenWutf8    ;jump if pointer is zero

  loopStrlenWutf8:
    xor eax,eax
    test pdword [pebp+psize*3],0x4
    jnz isUTF8StringInput

      lodsw
      call utf16ToUnicodeCode
      jmp isUTF16StringInput

    isUTF8StringInput:

      lodsb
      call Utf8ToUnicode

    isUTF16StringInput:
    test eax,eax
    je _endStrlenWutf8

    test pdword [pebp+psize*3],0x40
    jz onlyAZ09__strlenWutf8
      cmp eax,' '
      jne notSpace2
 
        mov eax,'-'
        jmp dontEscapeJsonChar__strlenWUtf8

      notSpace2:
      cmp eax,'0'
      jb loopStrlenWutf8

      cmp eax,'9'+1
      jc dontEscapeJsonChar__strlenWUtf8

      cmp eax,'A'
      jc loopStrlenWutf8

      cmp eax,'Z'+1
      jc dontEscapeJsonChar__strlenWUtf8

      cmp eax,'a'
      jc loopStrlenWutf8

      cmp eax,'z'
      ja loopStrlenWutf8

    onlyAZ09__strlenWutf8:

    test pdword [pebp+psize*3],0x20
    jz dontEscapeTerminalChar__strlenWUtf8

      call GetEscapeCharTerminal
      test dx,dx
      jz dontEscapeJsonChar__strlenWUtf8 ;not an character to escape

      inc pecx ;add one extra escape byte
      mov ax,dx

    dontEscapeTerminalChar__strlenWUtf8:
    test pdword [pebp+psize*3],2
    jz dontEscapeJsonChar__strlenWUtf8

      call GetEscapeCharJson
      test dx,dx
      jz dontEscapeJsonChar__strlenWUtf8 ;not an character to escape

      inc pecx ;add one extra escape byte
      mov ax,dx

    dontEscapeJsonChar__strlenWUtf8:

    ;at least 1 bytes char in UTF-8
    inc pecx

    cmp eax,0x80
    jb loopStrlenWutf8

    ;at least 2 bytes char in UTF-8
    inc pecx

    cmp eax,0x800
    jb loopStrlenWutf8

    ;at least 3 bytes char in UTF-8
    inc pecx

    cmp eax,0x10000
    jb loopStrlenWutf8

    ;4 bytes char in UTF-8
    inc pecx
  jmp loopStrlenWutf8

_endStrlenWutf8:
  mov peax,pecx

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalstrcpyUTF8
;This function copies source UTF-8 string to pointer UTF-8 destination.
;
;input:
;    1: flags
;      0x1:higher case
;      0x2:escape json characters
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9

;    (R/E)DI: Pointer to destination UTF-8 string
;    (R/E)SI: Pointer to source UTF-8 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source UTF-8 string + size(Pointer to source UTF-8 string)
;
;
;----------------------------------------------------
if used InternalstrcpyUTF8
InternalstrcpyUTF8:
push pedx

strcpyUft8Toutf8_loop:
  xor eax,eax

  lodsb
  call Utf8ToUnicode
  
  call InternalUnicodePointToUtf8

  test eax,eax
jnz strcpyUft8Toutf8_loop

pop pedx
  ret psize
end if

;----------------------------------------------------
;                      strcpyAToUtf8
;This function copies source ANSI string to pointer UTF-8 destination.
;
;input:
;    1: Pointer to destination UTF-8 string
;    2: Pointer to source ANSI string
;
;output: none
;
;----------------------------------------------------
if used strcpyAToUtf8
strcpyAToUtf8:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld

push 0
call InternalstrcpyAToUtf8

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalstrcpyAToUtf8
;This function copies source UTF-16 string to pointer UTF-8 destination.
;
;input:
;    1: flags
;      0x1:higher case
;      0x2: escape json characters
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9
;      0x80:lower case
;    (R/E)DI: Pointer to destination ANSI string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source ANSI string + size(Pointer to source ANSI string)
;
;
;----------------------------------------------------
if used InternalstrcpyAToUtf8
InternalstrcpyAToUtf8:
push pedx

strcpyAToutf8_loop:
  xor eax,eax
  lodsb

  call aToUTFCode

  test pdword [pesp+psize*2],0x80
  jz dontLowerCase__AToUtf8

    call SetLowerCaseUnicode

  dontLowerCase__AToUtf8:
  test pdword [pesp+psize*2],1
  jz dontHigherCase__AToUtf8

    call SetHighCaseUnicode

  dontHigherCase__AToUtf8:
    test pdword [pebp+psize*3],0x40
    jz onlyAZ09___AToUtf8
      test ax,ax
      jz dontEscapeJsonChar__AToUtf8
      cmp ax,' '
      jne notSpace3
 
        mov ax,'-'
        jmp dontEscapeJsonChar__AToUtf8

      notSpace3:
      cmp ax,'0'
      jb strcpyAToutf8_loop

      cmp ax,'9'+1
      jc dontEscapeJsonChar__AToUtf8

      cmp ax,'A'
      jc strcpyAToutf8_loop

      cmp ax,'Z'+1
      jc dontEscapeJsonChar__AToUtf8

      cmp ax,'a'
      jc strcpyAToutf8_loop

      cmp ax,'z'
      ja strcpyAToutf8_loop

    onlyAZ09___AToUtf8:
  test pdword [pebp+psize*3],0x20
  jz dontEscapeTerminalChar___AToUtf8

    call GetEscapeCharTerminal
    test dx,dx
    jz dontEscapeJsonChar__AToUtf8 ;not an character to escape

    inc pecx ;add one extra escape byte
    mov ax,dx

  dontEscapeTerminalChar___AToUtf8:
  test pdword [pesp+psize*2],2
  jz dontEscapeJsonChar__AToUtf8

    call GetEscapeCharJson
    test dx,dx
    jz dontEscapeJsonChar__AToUtf8 ;not an character to escape

    mov ax,'\'
    stosb
    mov ax,dx

  dontEscapeJsonChar__AToUtf8:

  cmp ax,0x80
  jb cpyByte

  cmp ax,0x800
  jb cpyWord

  ;-cmp ax,0x10000
  ;-jb cpy3Bytes


;cpy3Bytes:
  shl eax,4
  shr ax,2
  shr al,2

  or eax,0xE08080

  ror eax,16

  stosb
  rol eax,8
  stosb
  rol eax,8
  stosb
  jmp strcpyAToutf8_next

cpyWord:
  shl ax,2
  shr al,2
  or ax,0xC080


  ror eax,8
  stosb
  rol eax,8
  stosb
  jmp strcpyAToutf8_next

cpyByte:
  stosb

strcpyAToutf8_next:
  test ax,ax
jnz strcpyAToutf8_loop 

pop pedx
  ret psize
end if

;----------------------------------------------------
;                      InternalstrcpyWToUtf8
;This function copies source UTF-8 string to pointer UTF-16 destination.
;
;input:
;    1: flags
;      0x1:higher case
;      0x2:escape json characters
;      0x80:lower case

;    (R/E)DI: Pointer to destination UTF-16 string
;    (R/E)SI: Pointer to source UTF-8 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source UTF-8 string + size(Pointer to source UTF-8 string)
;
;
;----------------------------------------------------
if used InternalstrcpyUtf8ToW
InternalstrcpyUtf8ToW:
push pedx

strcpyUtf8ToW_loop:
  xor eax,eax
  lodsb

  call Utf8ToUnicode

  test pdword [pesp+psize*2],0x80
  jz dontLowerCase__Utf8ToW

    call SetLowerCaseUnicode

  dontLowerCase__Utf8ToW:
  test pdword [pesp+psize*2],1
  jz dontHigherCase__Utf8ToW

    call SetHighCaseUnicode

  dontHigherCase__Utf8ToW:
  test pdword [pesp+psize*2],2
  jz dontEscapeJsonChar__Utf8ToW

    call GetEscapeCharJson
    test dx,dx
    jz dontEscapeJsonChar__Utf8ToW ;not an character to escape

    mov ax,'\'
    stosw
    mov ax,dx

  dontEscapeJsonChar__Utf8ToW:

  cmp eax,0x010000
  jb cpyWord__Uft8

  ;00000000 000ccccc bbbbbbbb aaaaaaaa
  ror eax,8  ;
  ;aaaaaaaa 00000000 000ccccc bbbbbbbb
  dec ah     ;ccccc - 1 = dddd
  ;aaaaaaaa 00000000 0000dddd bbbbbbbb
  ror eax,2
  ;bbaaaaaa aa000000 000000dd ddbbbbbb
  shl ax,6
  ;bbaaaaaa aa000000 ddddbbbb bb000000
  rol eax,10
  ;000000dd ddbbbbbb 000000bb aaaaaaaa
  or eax,0xD800DC00
  ;110110dd ddbbbbbb 110111bb aaaaaaaa

  ror eax,16
  stosd
  jmp strcpyUtf8ToW_next

cpyWord__Uft8:
  stosw

strcpyUtf8ToW_next:
  test eax,eax
jnz strcpyUtf8ToW_loop

pop pedx
  ret psize
end if

;----------------------------------------------------
;                      strcpyWToUtf8
;This function copies source UTF-16 string to pointer UTF-8 destination.
;
;input:
;    1: Pointer to destination UTF-8 string
;    2: Pointer to source UTF-16 string
;
;output: none
;
;----------------------------------------------------
if used strcpyWToUtf8
strcpyWToUtf8:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld

  push 0
  call InternalstrcpyWToUtf8

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalstrcpyWToUtf8
;This function copies source UTF-16 string to pointer UTF-8 destination.
;
;input:
;    1: flags
;      0x1:higher case
;      0x2:escape json characters
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9
;      0x80:lower case

;    (R/E)DI: Pointer to destination UTF-8 string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)SI => pointer to last converted byte + 1
;        (R/E)DI => Pointer to source UTF-16 string + size(Pointer to source UTF-16 string)
;
;
;----------------------------------------------------
if used InternalstrcpyWToUtf8
InternalstrcpyWToUtf8:
push pedx

strcpyUft16Toutf8_loop:
  xor eax,eax
  lodsw
     call utf16ToUnicodeCode

  call InternalUnicodePointToUtf8

  test ax,ax
jnz strcpyUft16Toutf8_loop

pop pedx
  ret psize
end if

;----------------------------------------------------
;                      InternalUnicodePointToUtf8
;This function converts a UNICODE POINT to  UTF-8 and writes it to a string.
;
;input:
;    [pesp+psize*3]: flags
;      0x1:higher case
;      0x2:escape json characters
;      0x20:escape terminal characters
;      0x40:only a-z, A-Z and 0-9
;      0x80:lower case

;     EAX: UNICODE POINT
;    (R/E)DI: Pointer to destination UTF-8 string
;
;output: (R/E)DI => new end othe string with the unicode character in it
;
;
;----------------------------------------------------
InternalUnicodePointToUtf8:
  test pdword [pesp+psize*3],0x80
  jz dontLowerCase__WToUtf8

    call SetLowerCaseUnicode

  dontLowerCase__WToUtf8:
  test pdword [pesp+psize*3],1
  jz dontHigherCase__WToUtf8

    call SetHighCaseUnicode

  dontHigherCase__WToUtf8:
  test pdword [pesp+psize*3],0x40
  jz onlyAZ09___WToUtf8
    test eax,eax
    jz dontEscapeJsonChar__WToUtf8
    cmp eax,' '
    jne notSpace4
 
      mov eax,'-'
      jmp dontEscapeJsonChar__WToUtf8

    notSpace4:
    cmp eax,'0'
    jb strcpyUft16Toutf8_next

    cmp eax,'9'+1
    jc dontEscapeJsonChar__WToUtf8

    cmp eax,'A'
    jc strcpyUft16Toutf8_next

    cmp eax,'Z'+1
    jc dontEscapeJsonChar__WToUtf8

    cmp eax,'a'
    jc strcpyUft16Toutf8_next

    cmp eax,'z'
    ja strcpyUft16Toutf8_next

    onlyAZ09___WToUtf8:
  test pdword [pesp+psize*3],0x20
  jz dontEscapeTerminalChar__WToUtf8

    call GetEscapeCharTerminal
    test dx,dx
    jz dontEscapeJsonChar__WToUtf8 ;not an character to escape

    mov ax,'\'
    stosb
    mov ax,dx

  dontEscapeTerminalChar__WToUtf8:
  test pdword [pesp+psize*3],2
  jz dontEscapeJsonChar__WToUtf8

    call GetEscapeCharJson
    test dx,dx
    jz dontEscapeJsonChar__WToUtf8 ;not an character to escape

    mov ax,'\'
    stosb
    mov ax,dx

  dontEscapeJsonChar__WToUtf8:

  cmp eax,0x80
  jb cpyByte__Uft16

  cmp eax,0x800
  jb cpyWord__Uft16

  cmp eax,0x10000
  jb cpy3Bytes__Uft16

  ;00000000 000ccccc bbbbbbbb aaaaaaaa
  ror eax,10
  ;bbaaaaaa aa000000 00000ccc ccbbbbbb
  shl ah,4
  ;bbaaaaaa aa000000 0ccc0000 ccbbbbbb
  ror eax,2 ;3->2
  ;bbbbaaaa aaaa0000 000ccc00 00ccbbbb
  shl al,2
  ;bbbbaaaa aaaa0000 000ccc00 ccbbbb00
  ror eax,16
  ;000ccc00 ccbbbb00 bbbbaaaa aaaa0000
  shr ax,2
  ;000ccc00 ccbbbb00 00bbbbaa aaaaaa00
  shl ah,2
  ;000ccc00 ccbbbb00 bbbbaa00 aaaaaa00

  shr eax,2
  ;00000ccc 00ccbbbb 00bbbbaa 00aaaaaa
  or eax,0xF0808080
  ;11110ccc 10ccbbbb 10bbbbaa 10aaaaaa


  ror eax,24
  stosb
  rol eax,8
  stosb
  rol eax,8
  stosb
  rol eax,8
  stosb
  jmp strcpyUft16Toutf8_next


cpy3Bytes__Uft16:
  shl eax,4
  shr ax,2
  shr al,2

  or eax,0xE08080

  ror eax,16

  stosb
  rol eax,8
  stosb
  rol eax,8
  stosb
  jmp strcpyUft16Toutf8_next

cpyWord__Uft16:
  shl ax,2
  shr al,2
  or ax,0xC080


  ror eax,8
  stosb
  rol eax,8
  stosb
  jmp strcpyUft16Toutf8_next

cpyByte__Uft16:
  stosb

strcpyUft16Toutf8_next:
ret

;----------------------------------------------------
;                      UTF8ToUnicode
;This function changes an UTF8 character to UNICODE
;
;input:
;    EAX: byte from  string
;    R/ESI: pointer to current position in string
;
;output: EAX => UNICODE character
;        R/ESI => new pointer to current position in string
;
;Table source: https://nl.wikipedia.org/wiki/UTF-8#Beschrijving
;---------------------------------------------------- 
Utf8ToUnicode:
push pedx
push pecx
xor edx,edx

mov dl,0xF0 ;total of 4 bits
mov dh,4 ;4 bytes

mov ah,al
and ah,dl  ;& 0xF0
cmp ah,dl
je processBits

shl dl,1
dec dh ;3 bytes

mov ah,al
and ah,dl  ;& 0xE0
cmp ah,dl
je processBits

shl dl,1
dec dh ;2 bytes

mov ah,al
and ah,dl  ;& 0xC0
cmp ah,dl
je processBits

processOneByte:
xor ah,ah
;one byte
jmp endUtf8ToUnicode

processBits:
  xor ah,ah
  xor al,dl
  shl ax,8
  
  mov dl,dh
  dec dl
  loopThroughBytes:
    lodsb
    test al,al
    jz errLoopThroughBytes

    mov cl,al
    and cl,0x80
    cmp cl,0x80
    jne errLoopThroughBytes
    
    xor al,0x80
    shl al,2
    shl eax,6
    
    dec dl
  jnz loopThroughBytes
  shr eax,8
  
  jmp endUtf8ToUnicode
    
errLoopThroughBytes:
  sub dh,dl
  inc dh
  shr edx,8
  sub pesi,pedx
  lodsb


endUtf8ToUnicode:
pop pecx
pop pedx
ret

;----------------------------------------------------
;                      GetEscapeCharJson
;This function changes an ANSI character to UNICODE
;
;input:
;    EAX: UNICODE character
;
;output: EAX => UNICODE character
;        DX => escape character or zero if none
;
;----------------------------------------------------
if used GetEscapeCharJson
GetEscapeCharJson:
  xor dx,dx

  test eax,0xFFFFFF00
  jnz endGetEscapeCharJson

  mov dx,ax
  cmp al,'\'
  je endGetEscapeCharJson

  mov dx,ax
  cmp al,'"'
  je endGetEscapeCharJson

  ;cmp al,"'"
  ;je escapeCharA

  mov dx,'b'
  cmp al,0x0008 ;\b
  je endGetEscapeCharJson

  mov dx,'t'
  cmp al,0x0009 ;\t ???
  je endGetEscapeCharJson

  ;\v???

  mov dx,'n'
  cmp ax,0x000A ;\n
  je endGetEscapeCharJson

  mov dx,'f'
  cmp al,0x000C ;\f
  je endGetEscapeCharJson

  mov dx,'r'
  cmp al,0x000D ;\r
  je endGetEscapeCharJson

  xor dx,dx

endGetEscapeCharJson:
ret
end if

;----------------------------------------------------
;                      GetEscapeCharTerminal
;This function escapes characters: space, dubble quote and backslash
;
;input:
;    EAX: UNICODE character
;
;output: EAX => UNICODE character
;        DX => escape character or zero if none
;
;----------------------------------------------------
if used GetEscapeCharTerminal
GetEscapeCharTerminal:
  xor dx,dx

  test eax,0xFFFFFF00
  jnz endGetEscapeCharTerminal

  cmp al,'\'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'"'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,' '
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'!'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'?'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'$'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'%'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'#'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'&'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'*'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'('
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,')'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,0x09 ;tab
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'|'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,"'"
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,';'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'"'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'"'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'<'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'>'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'~'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'`'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'['
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,']'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'{'
  cmove dx,ax
  je endGetEscapeCharJson

  cmp al,'}'
  cmove dx,ax
  je endGetEscapeCharJson

endGetEscapeCharTerminal:
ret
end if

;----------------------------------------------------
;                      aToUTFCode
;This function changes an ANSI character to UNICODE
;
;input:
;    AX: original ANSI character
;
;output: AX => UNICODE character
;
;Table source: http://www.alanwood.net/demos/ansi.html
;----------------------------------------------------
if used aToUTFCode
aToUTFCode:
push pedx

  mov dx,0x20AC
  cmp ax,0x80
  cmovz ax,dx

  mov dx,0x201A
  cmp ax,0x82
  cmovz ax,dx

  mov dx,0x192
  cmp ax,0x83
  cmovz ax,dx

  mov dx,0x201E
  cmp ax,0x84
  cmovz ax,dx

  mov dx,0x2026
  cmp ax,0x85
  cmovz ax,dx

  mov dx,0x2020
  cmp ax,0x86
  cmovz ax,dx

  mov dx,0x2021
  cmp ax,0x87
  cmovz ax,dx

  mov dx,0x2c6
  cmp ax,0x88
  cmovz ax,dx

  mov dx,0x2030
  cmp ax,0x89
  cmovz ax,dx

  mov dx,0x160
  cmp ax,0x8A
  cmovz ax,dx

  mov dx,0x2039
  cmp ax,0x8B
  cmovz ax,dx

  mov dx,0x152
  cmp ax,0x8C
  cmovz ax,dx

  mov dx,0x17D
  cmp ax,0x8E
  cmovz ax,dx

  mov dx,0x2018
  cmp ax,0x91
  cmovz ax,dx

  mov dx,0x2019
  cmp ax,0x92
  cmovz ax,dx

  mov dx,0x201C
  cmp ax,0x93
  cmovz ax,dx

  mov dx,0x201D
  cmp ax,0x94
  cmovz ax,dx

  mov dx,0x2022
  cmp ax,0x95
  cmovz ax,dx

  mov dx,0x2013
  cmp ax,0x96
  cmovz ax,dx

  mov dx,0x2014
  cmp ax,0x97
  cmovz ax,dx

  mov dx,0x2DC
  cmp ax,0x98
  cmovz ax,dx

  mov dx,0x2122
  cmp ax,0x99
  cmovz ax,dx

  mov dx,0x161
  cmp ax,0x9A
  cmovz ax,dx

  mov dx,0x203A
  cmp ax,0x9B
  cmovz ax,dx

  mov dx,0x153
  cmp ax,0x9C
  cmovz ax,dx

  mov dx,0x17E
  cmp ax,0x9E
  cmovz ax,dx

  mov dx,0x178
  cmp ax,0x9F
  cmovz ax,dx

pop pedx
ret
end if

;----------------------------------------------------
;                      UTFToaCode
;This function changes an UNICODE to ANSI character
;
;input:
;    EAX: original UNICODE character
;
;output: AL => ANSI character
;
;Table source: http://www.alanwood.net/demos/ansi.html
;----------------------------------------------------
if used UTFToaCode
UTFToaCode:
push pedx

  mov dl,'?'
  test eax,0xFFFF0000
  jnz endAToUTFCode

  mov dl,0x80
  cmp ax,0x20AC
  jz endAToUTFCode

  mov dl,0x82
  cmp ax,0x201A
  jz endAToUTFCode

  mov dl,0x83
  cmp ax,0x192
  jz endAToUTFCode

  mov dl,0x84
  cmp ax,0x201E
  jz endAToUTFCode

  mov dl,0x85
  cmp ax,0x2026
  jz endAToUTFCode

  mov dl,0x86
  cmp ax,0x2020
  jz endAToUTFCode

  mov dl,0x87
  cmp ax,0x2021
  jz endAToUTFCode

  mov dl,0x88
  cmp ax,0x2c6
  jz endAToUTFCode

  mov dl,0x89
  cmp ax,0x2030
  jz endAToUTFCode

  mov dl,0x8A
  cmp ax,0x160
  jz endAToUTFCode

  mov dl,0x8B
  cmp ax,0x2039
  jz endAToUTFCode

  mov dl,0x8C
  cmp ax,0x152
  jz endAToUTFCode

  mov dl,0x8E
  cmp ax,0x17D
  jz endAToUTFCode

  mov dl,0x91
  cmp ax,0x2018
  jz endAToUTFCode

  mov dl,0x92
  cmp ax,0x2019
  jz endAToUTFCode

  mov dl,0x93
  cmp ax,0x201C
  jz endAToUTFCode

  mov dl,0x94
  cmp ax,0x201D
  jz endAToUTFCode

  mov dl,0x95
  cmp ax,0x2022
  jz endAToUTFCode

  mov dl,0x96
  cmp ax,0x2013
  jz endAToUTFCode

  mov dl,0x97
  cmp ax,0x2014
  jz endAToUTFCode

  mov dl,0x98
  cmp ax,0x2DC
  jz endAToUTFCode

  mov dl,0x99
  cmp ax,0x2122
  jz endAToUTFCode

  mov dl,0x9A
  cmp ax,0x161
  jz endAToUTFCode

  mov dl,0x9B
  cmp ax,0x203A
  jz endAToUTFCode

  mov dl,0x9C
  cmp ax,0x153
  jz endAToUTFCode

  mov dl,0x9E
  cmp ax,0x17E
  jz endAToUTFCode

  mov dl,0x9F
  cmp ax,0x178
  jz endAToUTFCode

  mov dl,'?'
  cmp ax,0xFF
  ja endAToUTFCode

  mov dl,al

endAToUTFCode:
xor eax,eax
mov al,dl

pop pedx
ret
end if

;----------------------------------------------------
;                      UTF16ToUnicode
;This function changes an UTF16 character to UNICODE
;
;input:
;    EAX: word from  string
;    R/ESI: pointer to current position in string
;
;output: EAX => UNICODE character
;        R/ESI => new pointer to current position in string
;
;Table source: http://www.alanwood.net/demos/ansi.html
;----------------------------------------------------
if used utf16ToUnicodeCode
utf16ToUnicodeCode:
push pedx

  mov dx,ax
  and dx,0xD800
  cmp dx,0xD800
  jne notTwoWords

  shl eax,16
  lodsw
  mov dx,ax

  and dx,0xDC00
  cmp dx,0xDC00
  jne isNotTwoWords

  ;110110dd ddbbbbbb 110111bb aaaaaaaa
  xor eax,0xD800DC00
  ;000000dd ddbbbbbb 000000bb aaaaaaaa
  ror eax,22
  ;bbbbbb00 0000bbaa aaaaaa00 0000dddd
  inc al
  ;bbbbbb00 0000bbaa aaaaaa00 000ccccc
  ror eax,20
  ;bbaaaaaa aa00000c ccccbbbb bb000000
  shr ax,6
  ;bbaaaaaa aa00000c 000000cc ccbbbbbb
  ror eax,2
  ;bbbbaaaa aaaa0000 0c000000 ccccbbbb
  shr ah,6
  ;bbbbaaaa aaaa0000 0000000c ccccbbbb
  ror eax,20
  ;00000000 000ccccc bbbbbbbb aaaaaaaa
  jmp isTwoWords

  isNotTwoWords:
  shr eax,16
  sub pesi,2


  isTwoWords:
  notTwoWords:

pop pedx
ret
end if

;----------------------------------------------------
;                      InternalStrcpyAWCaseHigh
;This function copies source ASCII string to pointer unicode destination.
;
;input:
;    (R/E)DI: Pointer to destination UTF-16 string
;    (R/E)SI: Pointer to source ANSI string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)SI => Pointer to source ANSI string + size(Pointer to source ASCII string)
;
;
;----------------------------------------------------
if used InternalStrcpyAWCaseHigh
InternalStrcpyAWCaseHigh:
  xor ax,ax

  lodsb

  call aToUTFCode
  call SetHighCase

  stosw
  test ax,ax
jnz InternalStrcpyAWCaseHigh

  ret
end if

;----------------------------------------------------
;                      strcpyWA
;This function copies source unicode string to pointer ASCII destination.
;
;input:
;    1: Pointer to destination ANSI string
;    2: Pointer to source UTF-16 string
;
;output: none
;
;----------------------------------------------------
if used strcpyWA
strcpyWA:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  cld
        
  call InternalStrcpyWA

  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      InternalStrcpyWA
;This function copies source unicode string to pointer ASCII destination.
;
;input:
;    (R/E)DI: Pointer to destination ANSI string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)SI => Pointer to source UTF-16 string + size(Pointer to source UTF-16 string)
;
;
;----------------------------------------------------
if used InternalStrcpyWA
InternalStrcpyWA:
  lodsw

  call utf16ToUnicodeCode

  call UTFToaCode

  stosb
  test al,al
jnz InternalStrcpyWA
  ret
end if

;----------------------------------------------------
;                      InternalStrcpyWACaseHigh
;This function copies source UTF-16 string to pointer ANSI destination and higher low characters.
;
;input:
;    (R/E)DI: Pointer to destination ANSI string
;    (R/E)SI: Pointer to source UTF-16 string
;
;output: (R/E)DI => pointer to last converted byte + 1
;        (R/E)SI => Pointer to source UTF-16 string + size(Pointer to source UTF-16 string)
;
;
;----------------------------------------------------
if used InternalStrcpyWACaseHigh
InternalStrcpyWACaseHigh:
  lodsw

  call utf16ToUnicodeCode
  call SetHighCase
  call UTFToaCode

  stosb
  test al,al
jnz InternalStrcpyWACaseHigh
  ret
end if

;----------------------------------------------------
;                      strcmp
;This function compairs string 1 to string 2 (case-insentitive)
;
;input:
;    1: Pointer to ASCII string 1
;    2: Pointer to ASCII string 2
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal
;----------------------------------------------------
if used strcmp
strcmp:
  call denter

  ;(R/E)DI =>string 1; (R/E)DI =>string 2
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ecx,ecx ; ecx=0 (return value)
  mov al,1 ;set byte to one
  
  StartCmpLoop:
    ;check for terminating zero byte
    test al,al
    jz endStrCmp

    ;AL=>byte [(R/E)SI]; (R/E)SI++;
    lodsb
  
    inc pedi ;string1++;

    ;check if byte string 1 matches string 2
    cmp al,byte [pedi-0x1]
    je StartCmpLoop

    ;Do a case sensitive match
    ppush StartCmpLoop ;jump addres if match
    call SetCase

  errStrCmp: ;The strings do not match
  inc ecx

  endStrCmp: ;The strings do match
  mov eax,ecx

  call dleave
  test eax,eax ;set return flags
  ret (psize*2)
end if

;----------------------------------------------------
;                      strcmpW
;This function compairs string 1 to string 2 (case-insentitive)
;
;input:
;    1: Pointer to UNICODE string 1
;    2: Pointer to UNICODE string 2
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal
;----------------------------------------------------
if used strcmpW
strcmpW:
  call denter

  ;(R/E)DI =>string 1; (R/E)DI =>string 2
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ecx,ecx ; ecx=0 (return value)
  mov ax,1 ;set byte to one
  
  StartCmpWLoop:
    ;check for terminating zero byte
    test ax,ax
    jz endStrCmpW

    ;AL=>byte [(R/E)SI]; (R/E)SI++;
    lodsw
  
    add pedi,2 ;string1++;

    ;check if byte string 1 matches string 2
    cmp ax,word [pedi-0x2]
    je StartCmpWLoop

    ;setcase does not handle words > sizeof(char)
    test ah,ah
    jnz errStrCmpW

    ;Do a case sensitive match
    ppush StartCmpWLoop ;jump addres if match
    call SetCase

  errStrCmpW: ;The strings do not match
  inc ecx

  endStrCmpW: ;The strings do match
  mov eax,ecx

  call dleave
  test eax,eax ;set return flags
  ret (psize*2)
end if


;----------------------------------------------------
;                      strcmpCaseSensitive
;This function compairs string 1 to string 2 (case-sentitive)
;
;input:
;    1: Pointer to ASCII string 1
;    2: Pointer to ASCII string 2
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal
;----------------------------------------------------
if used strcmpCaseSensitive
strcmpCaseSensitive:
  call denter

  ;(R/E)DI =>string 1; (R/E)DI =>string 2
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ecx,ecx ; ecx=0 (return value)
  
  StartCmpCaseLoop:
    lodsb
    scasb
    jz endStartCmpCaseLoop

    test al,al
  jnz StartCmpCaseLoop

  errStrcmpCaseSensitive: ;The strings do not match
  inc ecx

  endStartCmpCaseLoop: ;The strings do match
  mov eax,ecx

  call dleave
  test eax,eax ;set return flags
  ret (psize*2)
end if

;----------------------------------------------------
;                      strcmpUtf8CaseInsensitive
;This function compares source UTF-8 string with another UTF-8 string. Case insensitive compare.
;
;input:
;    1: Pointer to destination UTF-16 string
;    2: Pointer to source UTF-8 string
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal  
;
;----------------------------------------------------
if used strcmpUtf8CaseInsensitive
strcmpUtf8CaseInsensitive:
  call denter
        
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ax,ax
  cld

strcmpUtf8CaseInsensitive_loop:
  xor eax,eax
  lodsb

  call Utf8ToUnicode
  call SetHighCaseUnicode

  mov edx,eax ;save char

  xchg pedi,pesi

  xor eax,eax
  lodsb
  call Utf8ToUnicode
  call SetHighCaseUnicode
  
  xchg pedi,pesi
  
  cmp eax,edx
  jne notEqualUtf8USC

  test eax,eax
  jnz strcmpUtf8CaseInsensitive_loop

EqualUtf8USC:
xor eax,eax
jmp endStrcmpUtf8CaseInsensitive

notEqualUtf8USC:
  xor eax,eax
  inc eax

endStrcmpUtf8CaseInsensitive:
  call dleave
  test eax,eax
  ret (psize*2)
end if

;----------------------------------------------------
;                      strcmpCaseSensitive
;This function compairs string 1 to string 2 (case-sentitive)
;
;input:
;    1: Pointer to ASCII string 1
;    2: Pointer to ASCII string 2
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal
;----------------------------------------------------
if used strcmpCaseSensitiveW
strcmpCaseSensitiveW:
  call denter

  ;(R/E)DI =>string 1; (R/E)DI =>string 2
  mov pedi,pdword [pebp+(psize*2)]
  mov pesi,pdword [pebp+(psize*3)]
  xor ecx,ecx ; ecx=0 (return value)
  
  StartCmpCaseLoopW:
    lodsw
    scasw
    jz endStrcmpCaseSensitiveW

    test ax,ax
  jnz StartCmpCaseLoopW

  errStrcmpCaseSensitiveW: ;The strings do not match
  inc ecx

  endStrcmpCaseSensitiveW: ;The strings do match
  mov eax,ecx

  call dleave
  test eax,eax ;set return flags
  ret (psize*2)
end if


;----------------------------------------------------
;                comapairMemCaseSen
;This function compairs memory 1 to memory 2 (case-insentitive) with specific size.
;
;input:
;    1: Pointer to ASCII memory 1
;    2: Pointer to ASCII memory 2
;    3: Size to compaire
;
;output: (R/E)AX: 0 = strings match
;                 1 = strings do not match
;
;output: Zero flag:  set = strings are equal
;                    not set = strings are not equal
;----------------------------------------------------
if used comapairMemCaseSen ;| used bufcmp
comapairMemCaseSen:
;bufcmp:
  call denter

  ;(R/E)DI =>buf 1; (R/E)DI =>buf 2; (R/E)CX = size
  mov pedi,pdword [pebp+psize*2]
  mov pesi,pdword [pebp+psize*3]
  mov pecx,pdword [pebp+psize*4]
  xor edx,edx ; ECX = 0 (return value)

  startComapairMemCaseSenLoop:
    jecxz endComapairMemCaseSen
    ;compaire [edi] to [esi] with size ECX
    repz cmpsb
    jz endComapairMemCaseSen  ;jump if buffers match

    ;check on lower-higher case
    mov al,byte [pesi-1]

    ;Do a case sensitive match
    ppush startComapairMemCaseSenLoop ;jump addres if match
    call SetCase

errComapairMemCaseSen:
  inc edx     ;return value = 1  (and unset zero flag)

endComapairMemCaseSen:
  mov eax,edx ;return the return value

  call dleave

  test eax,eax
  ret (psize*3)
end if

;----------------------------------------------------
;                comapairMem
;This function compairs memory 1 to memory 2 with specific size.
;
;input:
;    1: Pointer to memory 1
;    2: Pointer to memory 2
;    3: Size to compaire
;
;output: Zero flags: set = strings are equal
;                    not set = strings are not equal
;
;----------------------------------------------------
if used comapairMem
comapairMem:
  push pebp
  mov pebp,pesp

  push pedi
  push pesi
  push pecx

  ;(R/E)DI =>buf 1; (R/E)DI =>buf 2; (R/E)CX = size
  mov pedi,pdword [pebp+psize*2]
  mov pesi,pdword [pebp+psize*3]
  mov pecx,pdword [pebp+psize*4]

    ;test pecx,pecx
    jpcxz endComapairMem

    ;compaire [edi] to [esi] with size ECX
    repz cmpsb
    ;jz endBufcmp  ;jump if buffers match

  endComapairMem:
  pop pecx
  pop pesi
  pop pedi

  mov pesp,pebp
  pop pebp
  ret (psize*3)
end if

;----------------------------------------------------
;                      SetCase
;This function lowers or highers a byte it's case
;
;input:
;    1:Pointer to jump to when the higher/lower byte match
;    AL:ASCII byte
;    byte [pedi-0x1]: byte to compair the higher or lower byte with
;
;output: AL: Higher, lowered or same byte as input
;
;corrupts: AH
;
;----------------------------------------------------
if used SetCase
SetCase:
      ;add the difference between high and low case  (0x20)
      mov ah,al
      add ah,0x20

      ;is char between A and Z?
      cmp al,0x40
      jna isnotHighCase

      cmp al,0x5B
      jnae caseChanged

    isnotHighCase:
      ;is char between  and ?
      cmp al,0xBF
      jna isnotHighSpecialCase

      cmp al,0xDF
      jnae caseChanged

    isnotHighSpecialCase:
      sub ah,0x40 ;set case to high

      ;is char between a and z?
      cmp al,0x60
      jna isNotLowCase

      cmp al,0x7B
      jnae caseChanged

    isNotLowCase:
      ;is char between  and ?
      cmp al,0x0DF
      jnae notCaseProblem

      ;-cmp al,0xFF
      ;-jnb notCaseProblem  ;the char is not high or low case

    caseChanged:
      ;check again, but now with the low or high case
      cmp ah,byte [pedi-0x1]
      jne notCaseProblem

      add pesp,psize*2
      jmp pdword [pesp-psize]


    notCaseProblem:

    ret psize
end if

;----------------------------------------------------
;                   SetHighCaseUnicode
;This function highers a unicode char case
;
;input:
;    eax:unicode char
;
;output: eax: Higher or same byte as input
;
;
;----------------------------------------------------
if used SetHighCaseUnicode
SetHighCaseUnicode:
test eax,0xFFFFFF00
jz SetHighCase

if ~ used dontImplementCaseUTF8
test eax,0xFFFF0000
jnz noHighCase

push peax
call denter

mov peax,pdword [pebp+psize]

mov pecx,peax ;Char
mov pesi,UnicodeLowHighTable
jmp startLoopThroughUpperTable

loopThroughUpperTable:
add pesi,2
startLoopThroughUpperTable:
  test ax,ax
  jz lastCharInTableUpper

  lodsw ;get lower case

  cmp ax,cx
  jne loopThroughUpperTable

  lodsw ;get higher case
  mov cx,ax

lastCharInTableUpper:
mov ax,cx
call dleave
add pesp,psize
end if

;FF = 178
noHighCase:
ret
end if

;----------------------------------------------------
;                      SetHighCase
;This function highers a byte it's case
;
;input:
;    AL:ASCII/ ANSI byte
;
;output: AL: Higher or same byte as input
;
;
;----------------------------------------------------
if used SetHighCase
SetHighCase:
      test al,al
      jz endSetHighCase

      ;is char between A and Z?
      cmp al,0x40
      jna isnotHighCase1

      cmp al,0x5B
      jnae endSetHighCase

    isnotHighCase1:
      ;is char between  and ?
      cmp al,0xBF
      jna isnotHighSpecialCase1

      cmp al,0xDD
      jnae endSetHighCase

    isnotHighSpecialCase1:
      ;is char between a and z?
      cmp al,0x60
      jna isNotLowCase1

      cmp al,0x7B
      jnae setCaseHigher

    isNotLowCase1:
      ;is char between  and ?
      cmp al,0x0DF
      jnae endSetHighCase;setCaseHigher

      ;-cmp al,0xFF
      ;-jnb endSetHighCase  ;the char is not high or low case

    setCaseHigher:
      sub al,0x20 ;set case to high (add the difference between high and low case  (0x20))

    endSetHighCase:
    ret
end if

;----------------------------------------------------
;                   SetLowerCaseUnicode
;This function lower a unicode char case
;
;input:
;    eax:unicode char
;
;output: eax: lower or same byte as input
;
;
;----------------------------------------------------
if used SetLowerCaseUnicode
SetLowerCaseUnicode:
test eax,0xFFFFFF00
jz SetLowCase

if ~ used dontImplementCaseUTF8
test eax,0xFFFF0000
jnz noLowCase

push peax
call denter

mov peax,pdword [pebp+psize]

mov pecx,peax ;Char
mov pesi,UnicodeLowHighTable

loopThroughLowerTable:
  test ax,ax
  jz lastCharInTableLower

  lodsw ;get lower case
  mov dx,ax ;save lower case
  lodsw ;get higher case

  cmp ax,cx
  jne loopThroughLowerTable

  mov cx,dx  ;set lower case

lastCharInTableLower:
mov ax,cx
call dleave
add pesp,psize
end if

;FF = 178
noLowCase:
ret
end if

;----------------------------------------------------
;                      SetLowCase
;This function lowers or highers a byte it's case
;
;input:
;    AL:ASCII/ ANSI byte
;
;output: AL: Higher or same byte as input
;
;
;----------------------------------------------------
if used SetLowCase
SetLowCase:
      test al,al
      jz endSetLowCase

      ;is char between a and z?
      cmp al,0x60
      jna isnotLowCase1

      cmp al,0x7B
      jnae endSetLowCase

    isnotLowCase1:
      ;is char between  and ?
      cmp al,0xBF
      jna isnotLowSpecialCase1

      cmp al,0xDD
      jnae endSetLowCase

    isnotLowSpecialCase1:
      ;is char between a and z?
      cmp al,0x40
      jna isNotHighCase2

      cmp al,0x5B
      jnae setCaseLower

    isNotHighCase2:
      ;is char between  and ?
      cmp al,0x0BF
      jnae endSetLowCase

      ;-cmp al,0xFF
      ;-jnb endSetHighCase  ;the char is not high or low case

    setCaseLower:
      add al,0x20 ;set case to high (add the difference between high and low case  (0x20))

    endSetLowCase:
    ret
end if

;----------------------------------------------------
;                      copyStringA
;This function copies a string to a new allocated memory
;
;input:
;    1:Pointer to string to copy
;
;output: (R/E)AX: pointer to memory with copied string
;
;----------------------------------------------------
if used copyStringA
copyStringA:
  call denter

  ;(R/E)SI => String to copy
  mov pesi,pdword [pebp+(psize*2)]

  ;Calculate string to copy
  push pesi
  call strlen
  mov pecx,peax

  ;make sure you request at least 1 byte (if the string is empty eax will be zero and GetMemLocation will fail)
  inc peax

  ;Allocate new memory pointer
  push peax
  call GetMemLocation

  ;return a empty pointer if the input string is zero
  test pesi,pesi
  jz endCopyStringA

  push peax  ;save new memory pointer

  ;copy the string to the new memory pointer
  mov pedi,peax
  rep movsb

  pop peax   ;restore new memory pointer

endCopyStringA:
  call dleave

  ret (psize*1)
end if

;----------------------------------------------------
;                      copyStringW
;This function copies a Wide Char string to a new allocated memory
;
;input:
;    1:Pointer to Wide Char string to copy
;
;output: (R/E)AX: pointer to memory with copied string
;
;----------------------------------------------------
if used copyStringW
copyStringW:
  call denter

  ;(R/E)SI => String to copy
  mov pesi,pdword [pebp+(psize*2)]

  ;Calculate string to copy
  push pesi
  call strlenW
  mov ecx,eax
  shr ecx,1 ;get size in words, instead of bytes (ECX=ECX/2)

  ;make sure you request at least 2 bytes (if the string is empty eax will be zero and GetMemLocation will fail)
  add peax,2

  ;Allocate new memory pointer
  push peax
  call GetMemLocation

  ;return a empty pointer if the input string is zero
  test pesi,pesi
  jz endCopyStringW

  push peax  ;save new memory pointer

  ;copy the string to the new memory pointer
  mov pedi,peax
  rep movsw

  pop peax   ;restore new memory pointer

endCopyStringW:
  call dleave

  ret (psize*1)
end if

;----------------------------------------------------
;                      copyBuffer
;This function copies a buffer to a new allocated memory
;
;input:
;    1:Pointer to buffer to copy
;    2:size of buffer to copy
;
;output: (R/E)AX: pointer to memory with copied buffer
;
;----------------------------------------------------
if used copyBuffer
copyBuffer:
  call denter

  ;(R/E)SI => Buffer to copy; (R/E)CX => Size
  mov pesi,pdword [pebp+(psize*2)]
  mov pecx,pdword [pebp+(psize*3)]

  test pecx,pecx
  cmovz peax,pesi
  jz endCopyBuffer

  ;Allocate new memory pointer
  push pecx
  call GetMemLocation

  push peax  ;save new memory pointer

  ;copy buffer to the new memory pointer
  mov pedi,peax
  rep movsb

  pop peax   ;restore new memory pointer

endCopyBuffer:
  call dleave

  ret (psize*2)
end if

;----------------------------------------------------
;                      StrToInt
;Converts an ASCII format number to an integer
;
;input:
;    1:source ASCII string
;    2:base for conversion
;
;output: (R/E)AX => integer
;
;----------------------------------------------------
if used StrToInt
StrToInt:
  call denter
  mov pedi,pdword [pebp+(psize*2)] ;(R/E)DI => source ASCII string
  mov pebx,pdword [pebp+(psize*3)] ;(R/E)BX => base for conversion

  xor peax,peax  ;Zero output value
  xor pedx,pedx  ;Zero (R/E)DX
  .loop:
    ;Get digit from string and test on terminating zero byte
    mov dl,byte [pedi]
    test dl,dl
    jz .endx

    ;multiply result by base
    imul peax,pebx

    cmp dl,'a'
    jc .notHighAlpha

      sub dl,0x20

    .notHighAlpha:  
    ;substract ASCII char '0'
    sub dl,'0'

    ;substract the distance between the char '9' and ALPHABETIC char 'A' if nececery
    cmp dl,9
    jbe .ok
    sub dl,7
  .ok:

    ;add the converted digit to the result
    add peax,pedx
    inc pedi ;seek to next char in string
    jmp .loop
  .endx:
  call dleave
  ret (psize*2)
end if

;----------------------------------------------------
;                      StrToIntQ
;Converts an ASCII format number to an 64-bit integer in 32-bit mode
;
;input:
;    1:source ASCII string
;    2:base for conversion
;    3:pointer to 64-bit integer
;
;output: none
;
;----------------------------------------------------
if used StrToIntQ
StrToIntQ:
  call denter
  mov pedi,pdword [pebp+(psize*2)] ;(R/E)DI => source ASCII string
  mov pecx,pdword [pebp+(psize*3)] ;(R/E)BX => base for conversion
  mov pesi,pdword [pebp+(psize*4)] ;(R/E)SI => pointer to 64-bit integer

  xor pedx,pedx  ;Zero (R/E)DX
  ;zero 64-bit integer
  mov dword [pesi],edx
  mov dword [pesi+4],edx

  .loop:
    ;Get digit from string and test on terminating zero byte
    mov dl,byte [pedi]
    test dl,dl
    jz .endx

    ;multiply result by base
    push pesi
    call MulQword

    ;substract ASCII char '0'
    sub dl,'0'

    ;substract the distance between the char '9' and ALPHABETIC char 'A' if nececery
    cmp dl,9
    jbe .ok
    sub dl,7
  .ok:
    ;add the converted digit to the result
    push pesi
    call AddEDXToQword

    inc pedi ;seek to next char in string
    jmp .loop
  .endx:
  call dleave
  ret (psize*3)
end if

;----------------------------------------------------
;                      StrToIntWithBaseAsTerminator
;Converts an ASCII format number to an integer, ending coversion by a zero-char of if char > base
;
;input:
;    1:source ASCII string
;    2:base for conversion
;
;output: (R/E)AX => integer
;        1:position of input source ASCII string + size of converted integer
;
;
;TODO:fix if base >10  part between cmp dl,9+7 and .ok (the chars 3A t/m 41 are read as 0 t/m 6)
;----------------------------------------------------
if used StrToIntWithBaseAsTerminator
StrToIntWithBaseAsTerminator:
  call denter
  mov pedi,pdword [pebp+(psize*2)] ;(R/E)DI => source ASCII string
  mov pebx,pdword [pebp+(psize*3)] ;(R/E)BX => base for conversion

  xor peax,peax  ;Zero output value
  xor pedx,pedx  ;Zero (R/E)DX
  .loop:
    ;Get digit from string and test on terminating zero byte
    mov dl,byte [pedi]
    test dl,dl
    jz .endx

    inc pedi ;seek to next char in string

    ;substract ASCII char '0'
    sub dl,'0'

    ;substract the distance between the char '9' and ALPHABETIC char 'A' if nececery
    cmp dl,9+7
    jbe .ok
    sub dl,7
  .ok:

    cmp dl,bl   ;if current digit => base  then end
    jae .endx

    ;multiply result by base
    imul peax,pebx

    ;add the converted digit to the result
    add peax,pedx
    jmp .loop

  .endx:
    mov pdword [pebp+(psize*2)],pedi ;replace source ASCII string with the end of the source string
    call dleave
    ret (psize*2)
end if

;----------------------------------------------------
;                      IntToStr
;Converts an integer to ASCII format
;
;input:
;    1:integer to convert
;    2:target ASCII string
;    3:base for conversion
;
;output: pointer to last converted byte + 1
;
;----------------------------------------------------
if used IntToStr
IntToStr:
  call denter

  mov peax,pdword [pebp+(psize*2)]
  mov pedi,pdword [pebp+(psize*3)]
  mov pebx,pdword [pebp+(psize*4)]

  call internalIntToStr

  xor al,al
  stosb

  mov peax,pedi

  call dleave
  ret (psize*3)
end if

;----------------------------------------------------
;                  internalIntToStr
;Converts an integer to ASCII format
;
;input:
;    (R/E)AX => integer
;    (R/E)DI => string
;    ebx     => base
;
;output: (R/E)DI => pointer to last converted byte + 1
;
;Corupts:
;  edx, ecx, eax
;----------------------------------------------------
if used internalIntToStr
internalIntToStr:
  cld
  xor pecx,pecx  ;zero ammount of rest from divide instructions pushed on stack

  .new: ;divide integer by base until result is zero
    ;divide by base
    xor pedx,pedx
    div pebx

    ;push rest on stack and increment counter
    push pedx
    inc pecx

    ;is integer zero?
    test peax,peax
    jnz .new

  .loop: ;convert rest from divide function to ASCII-chars
    ;get rest from stack and add '0' as base
    pop peax
    add al,30h

    ;if char>'9' add 7 to make is an alphabetic character ('9'+7='A')
    cmp al,'9'
    jng .ok
    add al,7

  .ok:
    stosb ;store ASCII char to string
    loop .loop
ret
end if

;----------------------------------------------------
;                      IntToStrW
;Converts an integer to Unicode format
;
;input:
;    1:integer to convert
;    2:target ASCII string
;    3:base for conversion
;
;output: none
;
;----------------------------------------------------
if used IntToStrW
IntToStrW:
  call denter

  mov peax,pdword [pebp+(psize*2)]
  mov pedi,pdword [pebp+(psize*3)]
  mov pebx,pdword [pebp+(psize*4)]

  call internalIntToStrW

  xor ax,ax
  stosw

  call dleave
  ret (psize*3)
end if

;----------------------------------------------------
;                  internalIntToStrW
;Converts an integer to UNICODE format
;
;input:
;    (R/E)AX => integer
;    (R/E)DI => string
;    ebx     => base
;
;output: (R/E)DI => pointer to last converted byte + 2
;
;Corupts:
;  edx, ecx
;----------------------------------------------------
if used internalIntToStrW
internalIntToStrW:
  cld
  xor pecx,pecx  ;zero ammount of rest from divide instructions pushed on stack

  .new: ;divide integer by base until result is zero
    ;divide by base
    xor pedx,pedx
    div pebx

    ;push rest on stack and increment counter
    push pedx
    inc pecx

    ;is integer zero?
    test peax,peax
    jnz .new

  .loop: ;convert rest from divide function to ASCII-chars
    ;get rest from stack and add '0' as base
    pop peax
    add ax,30h

    ;if char>'9' add 7 to make is an alphabetic character ('9'+7='A')
    cmp ax,'9'
    jng .ok
    add ax,7

  .ok:
    stosw ;store UNICODE char to string
    loop .loop
ret
end if

;----------------------------------------------------
;                      IntQToStr
;Converts an 64-bit integer to ASCII format in 32-bit mode
;
;input:
;    1:pointer to 64-bit integer to convert
;    2:target ASCII string
;    3:base for conversion
;
;output: (R/E)AX => pointer to last converted byte + 1
;
;----------------------------------------------------
if used IntQToStr & used is32
IntQToStr:
  call denter
        
  mov peax,pdword [pebp+(psize*2)]  ;pointer to Qword
  mov pedi,pdword [pebp+(psize*3)]  ;pointer to output buffer
  mov pebx,pdword [pebp+(psize*4)]  ;base

  call internalIntQToStr

  xor     al,al
  stosb
  
  mov peax,pedi

  call dleave
  ret (psize*3)
end if

;----------------------------------------------------
;                  internalIntQToStr
;Converts an 64-bit integer to ASCII format in 32-bit code
;
;input:
;    (R/E)AX => pointer to 64-bit integer
;    (R/E)DI => string
;    ebx     => base
;
;output: (R/E)DI => pointer to last converted byte + 1
;
;Corupts:
;  edx, ecx
;----------------------------------------------------
if used internalIntQToStr & used is32
internalIntQToStr:
  push dword [peax+0x4]
  push dword [peax]
  mov peax,pesp

  cld
  xor pecx,pecx  ;zero ammount of rest from divide instructions pushed on stack

  .new: ;divide integer by base until result is zero
    ;divide by base
    xor pedx,pedx
    push peax
    call DivQword

    ;push rest on stack and increment counter
    push pedx
    inc pecx

    ;is integer zero?
    cmp    dword [peax],0
    jnz     .new

    cmp    dword [peax+4],0
    jnz     .new

  .loop: ;convert rest from divide function to ASCII-chars
    ;get rest from stack and add '0' as base
    pop peax
    add al,30h

    ;if char>'9' add 7 to make is an alphabetic character ('9'+7='A')
    cmp al,'9'
    jng .ok
    add al,7

  .ok:
    stosb ;store ASCII char to string
    loop .loop

  add pesp,0x8
ret
end if

;----------------------------------------------------
;                      CreateOneString
;Creates a string on a new pointer out of different strings and integers.
;
;input:
;    1..*:List of strings and integers
;        {Type: 0x0=last item; 0x1=ANSI string; 0x11=UTF-8 string; 0x12=Multi UTF-8 string; 0x2=Unicode string; 0x3=Decimal number; 0x31=Hexidecimal number; 0x10000=Copy caracters upper case; 0x20000=Free pointer;,
;         Pointer or integer}
;
;output: (R/E)AX: pointer to memory with new UTF-16 string
;
;----------------------------------------------------
if used CreateOneString
CreateOneString:
xor eax,eax
push peax                       ;push flags
xchg peax,pdword [pesp+psize]   ;R/EAX => return value, pdword [pesp] => optional output string
push peax                       ;push return value
end if

;----------------------------------------------------
;                      CreateOneStringEx
;Creates a string on a new pointer out of different strings and integers.
;
;input:
;    1:flags
;       0x1 The return value should be the size neccecery for this string
;    2:optional output string. If not set, the function will create it's own buffer, as long as flag 0x1 is not set
;    3..*:List of strings and integers
;        {Type: 0x0=last item; 0x1=ANSI string; 0x11=UTF-8 string; 0x12=Multi UTF-8 string; 0x2=Unicode string; 0x3=Decimal number; 0x31=Hexidecimal number; 0x10000=Copy caracters upper case; 0x20000=Free pointer;,
;         Pointer or integer}
;
;output: (R/E)AX: pointer to memory with new UTF-16 string
;
;----------------------------------------------------
if used CreateOneStringEx | used CreateOneString
CreateOneStringEx:
  call denter

  lea pesi,pdword [pebp+psize*4]  ;get pointer to first item in list                ;
  xor pecx,pecx                   ;zero (R/E)CX, the size of the new string to create

  test pdword [pebp+psize*2],0x1  ;is flag 0x1 set and should we return the neccecery size
  jnz lenItemsLoop

    mov peax,pdword [pebp+psize*3] ;R/EAX => buffer
    test peax,peax ;if there a return pointer we don't need to calculate the size
    jnz bufferIsSet_Ex

  ;--
  lenItemsLoop:
    lodsp             ;Get type of current item in list
    mov edx,eax       ;EDX => type of item with flags
    and eax,0xFFFF    ;only get lower bytes
    mov edi,eax       ;EDI => type of item

    ;zero is last item in list
    test edi,edi
    jz endLenItemsLoop

    lodsp             ;Get pointer of current item in list

    ;1 is type ASCII string
    dec edi
    jz LenASCIString

    ;2 is type UNICODE string
    dec edi
    jz LenUNICODEString

    cmp edi,0x11-2
    jz LenUTF8String

    cmp edi,0x12-2
    jz LenMultiUTF8String

    ;3 or 31 is type integer
    if used is32
      mov eax,(10+1)*2 ;maximum size of a integer in ASCII format in x86 mode
    else
      mov eax,(20+1)*2 ;maximum size of a integer in ASCII format in x64 mode
    end if
    jmp StrCounted

  LenUNICODEString:
    ;calculate size of Unicode string
    push peax
    call strlenW
    jmp StrCounted

  LenASCIString:
    ;calculate size of ASCII string
    push peax
    call strlen
    shl peax,1  ;double size (since the output will be an Unicode string)
    jmp StrCounted

  LenUTF8String:
    ;calculate size of UTF-8 string
    and edx,0x200000  ;escape json?
    shr edx,20        ;make 0x2 from 0x200000
    ;calculate size of Unicode string
    push pedx
    push peax
    call strlenUtf8W
    jmp StrCounted

  LenMultiUTF8String:
    push pesi
    push pecx

    mov pesi,peax
    xor ecx,ecx

    and edx,0x200000  ;escape json?
    shr edx,20        ;make 0x2 from 0x200000

    cntMultiStr:
      ;calculate size of Multi UTF-8 string
      push pedx
      push pesi
      call strlenUtf8W
      add pecx,peax
      add pecx,2

      push pesi
      call strlen
      add pesi,peax
      inc pesi
      cmp byte[pesi],0
    jne cntMultiStr

    mov peax,pecx
    pop pecx
    pop pesi


  StrCounted:
    ;add size of item to total size of the new string.
    add pecx,peax
    add pecx,2 ;add space (needed if flag 0x40000 is set)
    jmp lenItemsLoop


endLenItemsLoop:
  inc pecx ;make sure the size of the new string is >0

  test pdword [pebp+psize*2],0x1  ;is flag 0x1 set and should we return the neccecery size
  jz createBufferEx

    sub pesi,psize
    if used is64
      mov pedx,pesi ;R/EDX => end of the list
    else
      push pesi ;save end of the list
    end if
    mov pebx,pecx ;size needed for the buffer
    jmp EndStrCopyLoop_Return_Size


  createBufferEx:

  ;allocate the new string
  push pecx
  call GetMemLocation

  bufferIsSet_Ex:

  mov pebx,peax ;base address of the new string
  mov pedi,peax ;current position in string
  lea pedx,pdword [pebp+psize*2] ;pointer to stack (where list is located)

  if used is32
    push edx ;save EDX
  end if

  ;---
  DoStrCopyLoop:
    if used is32
      pop edx ;restore EDX
    end if

    add pedx,(psize*2)           ;seek to next item in the list
    mov peax,pdword [pedx]       ;Get type of current item in list

    if used is32
      push edx ;save EDX
    end if

    ;is this the last item in the list?
    test ax,ax
    jz EndStrCopyLoop

    mov pesi,pdword [pedx+psize] ;Get pointer of current item in list

    if used is32
      mov edx,eax  ;EDX => type
    else
      mov r12d,eax ;R12D => type
    end if

    ;1 is type ASCII string
    dec ax
    jz CopyASCIString

    ;2 is type UNICODE string
    dec ax
    jz CopyUNICODEString

    cmp ax,0x11-2 ;Size format
    jz CopyUTF8String

    cmp ax,0x12-2
    jz CopyMultiUTF8String

    ;0x31 = hexidecimal, check for the '1'
    shr ax,4

    push pedx     ;save (R/E)DX
    push pebx     ;save (R/E)BX

    jz Decimal

    ;add hexidecimal sign '0x' to output string
    mov eax,0x00780030 ;L'0x'
    stosd

    mov peax,pesi ;(R/E)AX => integer to convert
    mov ebx,0x10           ;base = 16 (hexidecimal)
    call internalIntToStrW ;convert integer to Unicode string

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX
    jmp nextDoStrCopyLoop

  ;--
  Decimal:
    mov peax,pesi ;(R/E)AX => integer to convert
    mov ebx,10             ;base = 10 (decimal)
    call internalIntToStrW ;convert integer to Unicode string

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX
    jmp nextDoStrCopyLoop

  ;--
  CopyUNICODEString:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationUnicode
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationUnicode:

    ppush endCopyUNICODEString ;push return address

    test eax,0x10000 ;should the characters be transfered to high case?
    jnz InternalStrcpyWCaseHigh ;copy the Unicode string and set every character to high case
    ;copy the Unicode string
    jmp InternalStrcpyW

    endCopyUNICODEString:
    sub pedi,2 ;substract the zero word
    jmp IsBitCallFreeMemLocatioSet

CopyUTF8String:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationUtf8
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationUtf8:

    and peax,0x10000 or 0x200000 ;should the characters be transfered to high case?
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2
    shr peax,9
    shr ah,4
    shr peax,7

    push peax ;High or low case
    call InternalstrcpyUtf8ToW

    endCopyUTF8String:
    sub pedi,2 ;substract the zero word
    jmp IsBitCallFreeMemLocatioSet

  ;--
CopyMultiUTF8String:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationMultiUtf8
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationMultiUtf8:

    and peax,0x10000 or 0x200000 ;should the characters be transfered to high case?
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2
    shr peax,9
    shr ah,4
    shr peax,7
    push pecx

    mov pecx,peax
    CopyMultiUTF8ToW:
      push pecx ;High or low case
      call InternalstrcpyUtf8ToW

      cmp byte [pesi],0
    jne CopyMultiUTF8ToW

    xor ax,ax
    stosw

    pop pecx
    sub pedi,2 ;substract the zero word
    jmp IsBitCallFreeMemLocatioSet

  ;--
  CopyASCIString:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocation
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocation:
    ppush endCopyASCIString ;push return address

    test eax,0x10000 ;should the characters be transfered to high case?
    jnz InternalStrcpyAWCaseHigh ;copy the Unicode string and set every character to high case
    ;copy the ASCII string
    jmp InternalStrcpyAW

    endCopyASCIString:
    sub pedi,2 ;substract the zero byte

    ;call FreeMemLocation is flag 0x20000 is set
    IsBitCallFreeMemLocatioSet:
      if used is32
        test edx,0x20000
      else
        test r12d,0x20000
      end if
      jz nextDoStrCopyLoop
        call FreeMemLocation

nextDoStrCopyLoop:
    ;Add extra zero-character at the end of this current item?
    if used is32
      test edx,0x40000
    else
      test r12d,0x40000
    end if
    jz dontAddZeroChar

      xor ax,ax
      stosw ;store zero char

dontAddZeroChar:
    jmp DoStrCopyLoop

  ;--
EndStrCopyLoop:
  cmp pdword [pebp+psize*3],0
  je dontReturnEndOfBuffer_2

    mov pebx,pedi ;R/EBX => end of the buffer

  dontReturnEndOfBuffer_2:
  EndStrCopyLoop_Return_Size:

  if used is32
    pop edx ;restore EDX
  end if

  mov pdword [pedx],pebx  ;Save pointer of the new created string in the end of the list
  mov peax,pedx           ;(R/E)AX => pointer to the end of the list

  call dleave

  mov pesp,pdword [pesp]  ;(R/E)SP => return address
  xchg pdword [peax],pesp ;(R/E)SP => pointer to the new string, (d/q)word [(R/E)AX] => return address
  xchg pesp,peax          ;(R/E)SP => pointer to return address on stack, (R/E)AX => pointer to new created string
  ret
end if

;----------------------------------------------------
;                      CreateOneStringA
;Creates an ASCII string on a new pointer out of different strings and integers.
;
;input:
;    1...*:List of strings and integers
;        {Type: 0x0=last item; 0x1=ASCII/ANSI string; 0x2=Unicode string; 0x3=Decimal number; 0x31=Hexidecimal number; 0x12=Escape UTF-8 Url ;0x10000=Copy caracters upper case; 0x20000=Free pointer; 0x40000=Add zero-char behind;,
;         Pointer or integer}
;
;output: (R/E)AX: pointer to memory with new ANSI string
;
;----------------------------------------------------
if used CreateOneStringA
CreateOneStringA:
  call denter

  lea pesi,pdword [pebp+psize*2]  ;get pointer to first item in list                ;
  xor pecx,pecx                   ;zero (R/E)CX, the size of the new string to create

  ;--
  lenItemsLoopA:
    lodsp             ;Get type of current item in list
    and eax,0xFFFF    ;get only the lower bytes
    mov edi,eax       ;(R/E)DI => type of item

    ;zero is last item in list
    test pedi,pedi
    jz endLenItemsLoopA

    lodsp             ;Get pointer of current item in list

    ;1 is type ASCII string
    dec edi
    jz LenASCIStringA

    ;2 is type UNICODE string
    dec edi
    jz LenUNICODEStringA

    cmp edi,0x12-2
    if ~used EscapeValue
      jz LenASCIStringA
    else
      jz LenEscapedUrlA
    end if

    ;3 or 31 is type integer
    if used is32
      mov eax,(10+1)*2 ;maximum size of a integer in ASCII format in x86 mode
    else
      mov eax,(20+1)*2 ;maximum size of a integer in ASCII format in x64 mode
    end if
    jmp StrCountedA

  LenUNICODEStringA:
    ;calculate size of Unicode string
    push peax
    call strlenW
    shr peax,1  ;half size (since the output will be an ASCII string)
    jmp StrCountedA

  LenASCIStringA:
    ;calculate size of ASCII string
    push peax
    call strlen

  if used EscapeValue
    jmp StrCountedA

    LenEscapedUrlA:
      if ~ used EscapeValue_is_not_pointer
        cmp pdword [EscapeValue],0
        je LenASCIStringA


        push DONT_REPLACE_PATH_CHARACTERS
        push 0
        push peax
        call [EscapeValue]
      else
        push DONT_REPLACE_PATH_CHARACTERS
        push 0
        push peax
        call EscapeValue
      end if

  end if

  StrCountedA:
    ;add size of item to total size of the new string.
    add pecx,peax
    inc pecx ;add one byte (this for if flag 0x40000 is set)
    jmp lenItemsLoopA


endLenItemsLoopA:
  inc pecx ;make sure the size of the new string is >0

  ;allocate the new string
  push pecx
  call GetMemLocation

  mov pebx,peax ;base address of the new string
  mov pedi,peax ;current position in string
  mov pedx,pebp ;pointer to stack (where list is located)

  if used is32
    push edx ;save EDX
  end if

  ;---
  DoStrCopyLoopA:
    if used is32
      pop edx ;restore EDX
    end if

    add pedx,(psize*2)           ;seek to next item in the list
    mov peax,pdword [pedx]       ;Get type of current item in list

    if used is32
      push edx ;save EDX
    end if

    ;is this the last item in the list?
    test peax,peax
    jz EndStrCopyLoopA

    mov pesi,pdword [pedx+psize] ;Get pointer of current item in list

    if used is32
      mov edx,eax  ;EDX => type
    else
      mov r12d,eax ;R12D => type
    end if

    ;1 is type ASCII string
    dec ax
    jz CopyASCIStringA

    ;2 is type UNICODE string
    dec ax
    jz CopyUNICODEStringA

    cmp ax,0x12-2
    if ~used EscapeValue
      jz CopyASCIStringA
    else
      jz CopyEscapedUrlA
    end if

    ;0x31 = hexidecimal, check for the '1'
    shr ax,4

    push pedx     ;save (R/E)DX
    push pebx     ;save (R/E)BX

    jz DecimalA

    ;add hexidecimal sign '0x' to output string
    mov ax,'0x'
    stosw

    mov peax,pesi ;(R/E)AX => integer to convert
    mov ebx,0x10           ;base = 16 (hexidecimal)
    call internalIntToStr  ;convert integer to Unicode string

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX
    jmp nextDoStrCopyLoopA

  ;--
  DecimalA:
    mov peax,pesi ;(R/E)AX => integer to convert
    mov ebx,10             ;base = 10 (decimal)
    call internalIntToStr  ;convert integer to Unicode string

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX
    jmp nextDoStrCopyLoopA

  ;--
  CopyUNICODEStringA:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationUnicodeA
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationUnicodeA:

    ppush endInternalStrCpyWA ;push return address

    test eax,0x10000 ;should the characters be transfered to high case?
    jnz InternalStrcpyWACaseHigh ;copy the Unicode string and set every character to high case
    ;copy the Unicode string
    jmp InternalStrcpyWA

    endInternalStrCpyWA:
    dec pedi ;substract the zero byte
    jmp IsBitCallFreeMemLocatioSetA

  ;--
  CopyASCIStringA:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationA
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationA:

    ;copy the ASCII string
    StartCpyA:
      lodsb

      ;check if we need to replace lower case with upper case
      if used is32
        test edx,0x10000
      else
        test r12d,0x10000
      end if
      jz donSetCaseToHigh
        call SetHighCase

      donSetCaseToHigh:

      stosb
      test al,al
    jnz StartCpyA
    dec pedi ;substract the zero byte

  ;--
    if used EscapeValue
    jmp IsBitCallFreeMemLocatioSetA

      CopyEscapedUrlA:
      if ~ used EscapeValue_is_not_pointer
        cmp [EscapeValue],0
        je CopyASCIStringA
      end if

      test eax,0x20000
      jz dontPushPointerForFreeMemLocationCopyEscapedUrlA
        push pesi ;push pointer for FreeMemLocation
      dontPushPointerForFreeMemLocationCopyEscapedUrlA:

      push DONT_REPLACE_PATH_CHARACTERS
      push pedi
      push pesi

      if ~ used EscapeValue_is_not_pointer
        call [EscapeValue]
      else
        call EscapeValue
      end if
      add pedi,peax
    end if

    ;call FreeMemLocation is flag 0x20000 is set
    IsBitCallFreeMemLocatioSetA:
      if used is32
        test edx,0x20000
      else
        test r12d,0x20000
      end if
      jz nextDoStrCopyLoopA
        call FreeMemLocation

nextDoStrCopyLoopA:
    ;Add extra zero-character at the end of this current item?
    if used is32
      test edx,0x40000
    else
      test r12d,0x40000
    end if
    jz dontAddZeroCharA

      xor al,al
      stosb ;store zero char

dontAddZeroCharA:
  jmp DoStrCopyLoopA

  ;--
EndStrCopyLoopA:
  if used is32
    pop edx ;restore EDX
  end if

  mov pdword [pedx],pebx  ;Save pointer of the new created string in the end of the list
  mov peax,pedx           ;(R/E)AX => pointer to the end of the list

  call dleave

  mov pesp,pdword [pesp]  ;(R/E)SP => return address
  xchg pdword [peax],pesp ;(R/E)SP => pointer to the new string, (d/q)word [(R/E)AX] => return address
  xchg pesp,peax          ;(R/E)SP => pointer to return address on stack, (R/E)AX => pointer to new created string
  ret
end if

;----------------------------------------------------
;                      CreateOneStringUtf8
;Creates a string on a new pointer out of different strings and integers.
;
;input:
;    1:flags
;       0x1 The return value should be the size neccecery for this string
;    2:optional output string. If not set, the function will create it's own buffer, as long as flag 0x1 is not set
;    1..*:List of strings and integers
;        {Type: 0x0=last item; 0x1=ANSI string; 0x11=UTF-8 string 0x2=UTF-16 string;; 0x22=Multi UTF-16 string
;               0x3=Decimal number; 0x31=Hexidecimal number;
;               0x4=decimal 64 bit number (pointer in x86, same as 3 in x64);
;               0x41=Hexidecimal number 64 bit number (pointer in x86, same as 31 in x64);
;               0x5=Pointer to long long that needs to formated to "5GB (5.000.000.000 bytes)"
;               0xA=Begin JSON object, 0xB=end JSON object, 0xE=Boolean, 0xF=begin JSON array
;               0x10=end JSON array

;NOT SUPPORTED: 0xC then=File attributes, 0xD=FILETIME that needs to transformed to a string

;
;               0x10000=Copy caracters upper case; 0x20000=Free pointer; 0x80000 format number as size;
;               0x100000=format number with dots; 0x200000=escape string as JSON string
;               0x400000=is item name in JSON object; 0x800000=is value in JSON object
;               0x1000000=don't add coma in front of JSON object name if 0x400000 is set, else is adds a coma
;               0x2000000=escape string as Terminal string, 0x4000000=only allow a-z A-Z 0-9, 0x8000000=lower case
;         Pointer or integer}
;
;output: (R/E)AX:
;                 if flag 0x1 is set ,the size if returned pointer to memory with new UTF-8 string
;                 else if output string is set, a pointer to the end of the buffer (the zero byte) is returned
;                 else a new created utf-8 string is returned
;
;----------------------------------------------------
if used CreateOneStringUtf8
CreateOneStringUtf8:
  call denter

  lea pesi,pdword [pebp+psize*4]  ;get pointer to first item in list                ;
  xor pecx,pecx                   ;zero (R/E)CX, the size of the new string to create

  test pdword [pebp+psize*2],0x1  ;is flag 0x1 set and should we return the neccecery size
  jnz lenItemsLoop_Utf8

  mov peax,pdword [pebp+psize*3] ;R/EAX => buffer
  test peax,peax ;if there a return pointer we don't need to calculate the size
  jnz bufferIsSet

  ;--
  lenItemsLoop_Utf8:
    lodsp             ;Get type of current item in list
    mov pedx,peax     ;R/EDX => type with flags
    and eax,0xFFFF    ;only get lower bytes
    mov edi,eax       ;EDI => type of item

    ;zero is last item in list
    test edi,edi
    jz endlenItemsLoop_Utf8

    lodsp             ;Get pointer of current item in list

    ;1 is type ASCII string
    dec edi
    jz LenASCIString_Utf8

    ;2 is type UNICODE string
    dec edi
    jz LenUNICODEString_Utf8

    cmp edi,0x22-2
    jz LenMultiUTF16String_Utf8

    ;11 is type UNICODE string
    cmp edi,0x11-2
    jz LenUTF8String_Utf8

    cmp edi,5-2
    jz isFullSizeFormat_Utf8

    cmp edi,0xA-2
    jz isJSONObject

    cmp edi,0xB-2
    jz isJSONObject

    cmp edi,0xF-2
    jz isJSONObject

    cmp edi,0x10-2
    jz isJSONObject

    ;-cmp edi,0xC-2
    ;-jz isisFileAttributes

    ;-cmp edi,0xD-2
    ;-jz isFiletime_Utf8

    cmp edi,0xE-2
    jz isBoolean_Utf8

    ;3, 31, 4 or 41  is type integer
    mov eax,20+1 ;maximum size of a integer in UNICODE format in x64 mode

    mov edx,20 + 6 ;MAX_LONG_LONG + dots
    test pdword [pesi-psize*2],0x100000 ;format number with dots
    cmovnz eax,edx

    add edx,6 + 1 ;" bytes" + arrows + space + rounded size
    test pdword [pesi-psize*2],0x80000 ;format number as size
    cmovnz eax,edx
    jmp StrCounted_Utf8

  isFullSizeFormat_Utf8:
    mov eax,20 + 6 + 6 + 2 + 1 + 7 ;MAX_LONG_LONG + dots + " bytes" + arrows + space + rounded size
    jmp StrCounted_Utf8

  LenUNICODEString_Utf8:
    and edx,0x200000 or 0x2000000 or 0x4000000  ;escape json or terminal?
    shr edx,20        ;make 0x2 from 0x200000 and 0x20 from 0x2000000 and 0x40 from 0x4000000
    ;calculate size of Unicode string
    push pedx
    push peax
    call strlenWUtf8
    jmp StrCounted_Utf8

  LenASCIString_Utf8:
    and edx,0x200000 or 0x2000000 or 0x4000000  ;escape json or terminal?
    shr edx,20       ;make 0x2 from 0x200000 and 0x20 from 0x2000000 and 0x40 from  0x4000000
    ;calculate size of ASCII string
    push pedx
    push peax
    call strlenAUtf8
    jmp StrCounted_Utf8

  LenUTF8String_Utf8:
    and edx,0x200000 or 0x2000000 or 0x4000000 ;escape json or terminal?
    shr edx,20       ;make 0x2 from 0x200000 and 0x20 from 0x2000000 and 0x40 from  0x4000000
    or edx,0x4        ;input string is UTF-8
    ;calculate size of UTF-8 string
    push pedx
    push peax
    call strlenWUtf8
    jmp StrCounted_Utf8

  LenMultiUTF16String_Utf8:
    push pesi
    push pecx

    mov pesi,peax
    xor ecx,ecx
        
    and edx,0x200000 or 0x2000000 or 0x4000000  ;escape json or terminal?
    shr edx,20       ;make 0x2 from 0x200000 and 0x20 from 0x2000000  and 0x40 from 0x4000000

    cntMultUTF16Str:
      ;calculate size of Multi UTF-8 string
      push pedx
      push pesi
      call strlenWUtf8
      add pecx,peax
      inc pecx

      push pesi
      call strlenW
      add pesi,peax
      add pesi,2
      cmp word [pesi],0
    jne cntMultUTF16Str

    mov peax,pecx
    pop pecx
    pop pesi
    jmp StrCounted_Utf8

  isBoolean_Utf8:
    mov eax,4 ;'true'
    test peax,peax
    jnz StrCounted_Utf8

    inc eax ;'false'
    jmp StrCounted_Utf8

  ;isisFileAttributes:
  ;  push 0               ;pointer
  ;  push 0x1            ;flags
  ;  push peax           ;push attributes
  ;  push FileAttributes ;push names for the flags
  ;  call getFlagsByNameUtf8 ;get Unicode string with attributes
  ;  jmp StrCounted_Utf8

  isJSONObject:
    mov eax,1

  ;-isFiletime_Utf8:
  ;-  mov eax,4 + 1 + 2 + 1 + 4 + 5 + 1 + 2 + 1 + 2 + 1 + 2 + 3 + 1 ;day + space + day number + space + monthname + year + space + hour + ':' + minutes + ':' + seconds + GMT + zero byte

  StrCounted_Utf8:

    test pdword [pesi-psize*2],0x40000
    jz no4FlagSet

      inc peax ;add space (needed if flag 0x40000 is set)

  no4FlagSet:

;--don't add coma in front of JSON object name if 0x400000 is set, else is adds a coma

     test pdword [pesi-psize*2],0x400000   ;name in JSON object
     jz isNotJsonName

       add peax,4 ;',"":'

       test pdword [pesi-psize*2],0x1000000  ;don't add coma in front of JSON object name
       jz comaWillBeSet


       dec peax ; -1 ','

     isNotJsonName:
     test pdword [pesi-psize*2],0x800000   ;value in JSON opject
     jz isNotJsonValue

       add peax,2 ;'""'

     isNotJsonValue:
     test pdword [pesi-psize*2],0x1000000  ; add coma in front of item
     jz isNoExtraComa

       inc peax ;+1 ','

     isNoExtraComa:
    comaWillBeSet:
    ;add size of item to total size of the new string.
    add pecx,peax
    jmp lenItemsLoop_Utf8


endlenItemsLoop_Utf8:
  inc pecx ;make sure the size of the new string is >0

  test pdword [pebp+psize*2],0x1  ;is flag 0x1 set and should we return the neccecery size
  jz createBuffer

    sub pesi,psize
    if used is64
      mov pedx,pesi ;R/EDX => end of the list
    else
      push pesi ;save end of the list
    end if
    mov pebx,pecx ;size needed for the buffer
    jmp EndStrCopyLoop_Utf8_Return_Size


  createBuffer:

  ;allocate the new string
  push pecx
  call GetMemLocation

bufferIsSet:
  mov pebx,peax ;base address of the new string
  mov pedi,peax ;current position in string
  lea pedx,pdword [pebp+psize*2] ;pointer to stack (where list is located)

  if used is32
    push edx ;save EDX
  end if

  ;---
  DoStrCopyLoop_Utf8:
    if used is32
      pop edx ;restore EDX
    end if

    add pedx,(psize*2)           ;seek to next item in the list
    mov peax,pdword [pedx]       ;Get type of current item in list

    if used is32
      push edx ;save EDX
    end if

    ;is this the last item in the list?
    test ax,ax
    jz EndStrCopyLoop_Utf8

    mov pesi,pdword [pedx+psize] ;Get pointer of current item in list

    if used is32
      mov edx,eax  ;EDX => type
    else
      mov r12d,eax ;R12D => type
    end if

    if used is64
      test r12d,0x400000   ;name in JSON object
    else
      test edx,0x400000   ;name in JSON object
    end if
   jz dontSetJsonName

     if used is64
      test r12d,0x1000000  ;don't add coma in front of JSON object name
     else
       test edx,0x1000000  ;don't add coma in front of JSON object name
     end if
     jnz dontSetComa

       mov ax,','
       stosw
       dec pedi

     dontSetComa:
     mov ax,'"'
     stosw
     dec pedi
     jmp beginJSONObjectNameIsSet

   dontSetJsonName:
   if used is64
     test r12d,0x800000   ;value in JSON opject
   else
     test edx,0x800000  ;value in JSON opject
   end if
   jz dontSetJsonValue


     mov ax,'"'
     stosw
     dec pedi

   dontSetJsonValue:
   if used is64
     test r12d,0x1000000    ; add coma in front of item
   else
     test edx,0x1000000   ; add coma in front of item
   end if
   jz dontSetExtraComa

     mov ax,','
     stosw
     dec pedi

   dontSetExtraComa:
   beginJSONObjectNameIsSet:
    if used is32
      mov eax,edx  ;EAX => type
    else
      mov eax,r12d ;EAX => type
    end if

    ;1 is type ASCII string
    dec ax
    jz CopyASCIString_Utf8

    ;2 is type UNICODE string
    dec ax
    jz CopyUNICODEString_Utf8

    ;0x11 is type UTF-8 string
    cmp ax,0x11-2 ;Size format
    jz CopyUTF8String_Utf8

    cmp ax,0x22-2
    jz CopyMultiUTF16String_Utf8

    cmp ax,5-2 ;Size format
    jz isSizeFormat_Utf8

    cmp ax,0xA-2
    jz copyBeginJSONObject

    cmp ax,0xB-2
    jz copyEndJSONObject

    cmp ax,0xF-2
    jz copyBeginJSONArray

    cmp ax,0x10-2
    jz copyEndJSONArray

    ;-cmp ax,0xC-2
    ;-jz copyFileAttributesFlags_Utf8

    ;-cmp ax,0xD-2
    ;-jz copyFiletime_Utf8

    cmp ax,0xE-2
    jz copyBoolean_Utf8

    ;this is for intToStr
    push pedi ;Save current position to be able to format it later if flag 0x80000 or 0x100000  is set

    push pedx     ;save (R/E)DX
    push pebx     ;save (R/E)BX


    if used is32
      mov ebx,10             ;base = 10 (decimal)
      cmp ax,0x3-2
      je Decimal_Utf8

      cmp ax,0x31-2
      je HexDecimal_Utf8

      ;0x31 = hexidecimal, check for the '1'
      shr ax,4

      jz Decimal64_Utf8
    else
      ;0x31 = hexidecimal, check for the '1'
      shr ax,4

      mov ebx,10             ;base = 10 (decimal)
      jz Decimal_Utf8
    end if

  if used is32
  HexDecimal64_Utf8:
    ;add hexidecimal sign '0x' to output string
    mov ax,'0x' ;L'0x'
    stosw
    mov ebx,0x10             ;base = 10 (decimal)

  Decimal64_Utf8:
    mov peax,pesi ;(R/E)AX => integer to convert
    ;-mov ebx,10             ;base = 10 (decimal)
    call internalIntQToStr ;convert integer to UTF-8 string

    xor al,al
    stosb
    dec pedi

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX
    jmp _doFormatSize_Utf8
  end if

  HexDecimal_Utf8:
    ;add hexidecimal sign '0x' to output string
    mov ax,'0x' ;L'0x'
    stosw
    mov ebx,0x10             ;base = 0x10 (hexdecimal)

  ;--
  Decimal_Utf8:
    mov peax,pesi ;(R/E)AX => integer to convert
    ;--mov ebx,10             ;base = 10 (decimal)
    call internalIntToStr ;convert integer to UTF-8 string

    xor al,al
    stosb
    dec pedi

    pop pebx ;restore (R/E)BX
    pop pedx ;restore (R/E)DX

  _doFormatSize_Utf8:
    if used is32
      test edx,0x80000
    else
      test r12d,0x80000
    end if
    jz _doFormatNumber_Utf8

    call FormatBytesA
    jmp findEndOfString_Utf8

  _doFormatNumber_Utf8:
    lea peax,pdword [pesp+psize]   ;pop the push for FormatBytes from stack in EAX
    if used is32
      test edx,0x100000
    else
      test r12d,0x100000
    end if
    cmovz pesp,peax                ;pop is from Rsp
    jz nextDoStrCopyLoop_Utf8

    call FormatNumberA

  findEndOfString_Utf8:
    cmp byte [pedi],0
    je endOfStringFound_Utf8

      inc pedi

    jmp findEndOfString_Utf8

  endOfStringFound_Utf8:
    jmp nextDoStrCopyLoop_Utf8

  isSizeFormat_Utf8:
    ;format as "5GB (5.000.000.000 bytes)"
    push pedi  ;output string
    push pesi  ;(x64) long lnog (x86) pointer to long long
    call neatSizeFormatA
    dec peax ;substract the zero char

    mov pedi,peax
    jmp nextDoStrCopyLoop_Utf8

  ;-copyFiletime_Utf8:
  ;-  push pedi           ;bufer
  ;-  push pesi           ;push TimeDate
  ;-  call FileTimeToStringBuf
  ;-  mov pedi,peax
  ;-  dec pedi ;substract the zero char
  ;-  jmp nextDoStrCopyLoop_Utf8

  ;--
  CopyUNICODEString_Utf8:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationUnicode_Utf8
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationUnicode_Utf8:

    and peax,0x10000 or 0x200000 or 0x2000000 or 0x4000000 or 0x8000000 ;should the characters be transfered to high case/escaped as json/terinal
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2 and 0x2000000 0x20 and 0x4000000 0x40
    shr eax,9
    shr ah,4
    shr eax,3
    shr ah,4
    shr eax,4

    push peax ;High or low case
    call InternalstrcpyWToUtf8
    
    endCopyUNICODEString_Utf8:
    dec pedi ;substract the zero char
    jmp IsBitCallFreeMemLocatioSet_Utf8

  ;--
  CopyUTF8String_Utf8:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationUTF8_Utf8
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationUTF8_Utf8:

    and eax,0x10000 or 0x200000 or 0x2000000 or 0x4000000 or 0x8000000 ;should the characters be transfered to high case/escaped as json/terinal
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2 and 0x2000000 0x20 and 0x4000000 0x40
    ;plog LOG_MESSAGE,"The flags are:",0x31,peax
    shr eax,9
    shr ah,4
    shr eax,3
    shr ah,4
    shr eax,4

    push peax ;High or low case
    call InternalstrcpyUTF8

    dec pedi ;substract the zero char
    jmp IsBitCallFreeMemLocatioSet_Utf8

  ;--
  CopyMultiUTF16String_Utf8:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocationMultiUtf16
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocationMultiUtf16:

    and eax,0x10000 or 0x200000 or 0x2000000 or 0x4000000 or 0x8000000 ;should the characters be transfered to high case/escaped as json/terinal
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2 and 0x2000000 0x20 and 0x4000000 0x40
    shr eax,9
    shr ah,4
    shr eax,3
    shr ah,4
    shr eax,4
    push pecx

    mov pecx,peax
    CopyMultiUTF16To8:
      push pecx ;High or low case
      call InternalstrcpyWToUtf8

      cmp word [pesi],0
    jne CopyMultiUTF16To8

    xor al,al
    stosb

    pop pecx
    dec pedi ;substract the zero byte
    jmp IsBitCallFreeMemLocatioSet_Utf8

  ;--
    copyBeginJSONObject:
    mov ax,'{'
    stosw
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8

    copyEndJSONObject:
    mov ax,'}'
    stosw
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8

    copyBeginJSONArray:
    mov ax,'['
    stosw
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8

    copyEndJSONArray:
    mov ax,']'
    stosw
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8

  ;--
  copyBoolean_Utf8:
  test pesi,pesi
  jz copyFalse

    mov eax,'true'
    stosd
    xor al,al
    stosb
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8


  copyFalse:
    mov eax,'fals'
    stosd
    mov  ax,'e'
    stosw
    dec pedi ;set point not on zero byte
    jmp nextDoStrCopyLoop_Utf8

  ;-copyFileAttributesFlags_Utf8:
  ;-  push pedi               ;pointer
  ;-  push 0              ;flags
  ;-  push pesi           ;push attributes
  ;-  push FileAttributes ;push names for the flags
  ;-  call getFlagsByNameUtf8 ;get Unicode string with attributes
  ;-  mov pedi,peax
  ;-  jmp nextDoStrCopyLoop_Utf8


  ;--
  CopyASCIString_Utf8:
    test eax,0x20000
    jz dontPushPointerForFreeMemLocation_Utf8
      push pesi ;push pointer for FreeMemLocation
    dontPushPointerForFreeMemLocation_Utf8:
    and eax,0x10000 or 0x200000 or 0x2000000 or 0x4000000 or 0x8000000 ;should the characters be transfered to high case/escaped as json/terinal
    ;make 0x10000 flag 0x1 and 0x200000 flag 0x2 and 0x2000000 0x20 and 0x4000000 from 0x40
    shr eax,9
    shr ah,4
    shr eax,3
    shr ah,4
    shr eax,4

    push peax ;High or low case
    call InternalstrcpyAToUtf8

    endCopyASCIString_Utf8:
    dec pedi ;substract the zero char

    ;call FreeMemLocation is flag 0x20000 is set
    IsBitCallFreeMemLocatioSet_Utf8:

      if used is32
        test edx,0x20000
      else
        test r12d,0x20000
      end if
      jz nextDoStrCopyLoop_Utf8
        call FreeMemLocation



nextDoStrCopyLoop_Utf8:
     if used is64
       test r12d,0x400000  ;name in JSON object
     else
       test edx,0x400000   ;name in JSON object
     end if
     jz dontSetEndJsonObjectName

       mov ax,'":'
       stosw       ;store ':"'
       xor ax,ax
       stosb       ;store last zeor byte
       dec pedi    ;set pointer on the zero byte
       jmp EndJsonObjectNameIsSet

    dontSetEndJsonObjectName:
    if used is64
      test r12d,0x800000  ;value in JSON opject
    else
      test edx,0x800000   ;value in JSON opject
    end if
    jz dontSetEndJsonValue

      mov ax,'"'
      stosw    ;store '"',0
      dec pedi ;set pointer on zero

    dontSetEndJsonValue:
    EndJsonObjectNameIsSet:
    ;Add extra zero-character at the end of this current item?
    if used is32
      test edx,0x40000
    else
      test r12d,0x40000
    end if
    jz dontAddZeroChar_Utf8

      xor al,al
      stosb ;store zero char

dontAddZeroChar_Utf8:
    jmp DoStrCopyLoop_Utf8

  ;--
EndStrCopyLoop_Utf8:
  cmp pdword [pebp+psize*3],0
  je dontReturnEndOfBuffer

    mov pebx,pedi ;R/EBX => end of the buffer

  dontReturnEndOfBuffer:
  EndStrCopyLoop_Utf8_Return_Size:
  if used is32
    pop edx ;restore EDX
  end if

  mov pdword [pedx],pebx  ;Save pointer of the new created string in the end of the list
  mov peax,pedx           ;(R/E)AX => pointer to the end of the list

  call dleave

  mov pesp,pdword [pesp]  ;(R/E)SP => return address
  xchg pdword [peax],pesp ;(R/E)SP => pointer to the new string, (d/q)word [(R/E)AX] => return address
  xchg pesp,peax          ;(R/E)SP => pointer to return address on stack, (R/E)AX => pointer to new created string
  ret
end if

;source:  https://www.ibm.com/docs/en/i/7.2?topic=tables-unicode-lowercase-uppercase-conversion-mapping-table
if used UnicodeLowHighTable
UnicodeLowHighTable:
;lower   , upper
dw 0x0101, 0x0100
dw 0x0103, 0x0102
dw 0x0105, 0x0104
dw 0x0107, 0x0106
dw 0x0109, 0x0108
dw 0x010B, 0x010A
dw 0x010D, 0x010C
dw 0x010F, 0x010E
dw 0x0111, 0x0110
dw 0x0113, 0x0112
dw 0x0115, 0x0114
dw 0x0117, 0x0116
dw 0x0119, 0x0118
dw 0x011B, 0x011A
dw 0x011D, 0x011C
dw 0x011F, 0x011E
dw 0x0121, 0x0120
dw 0x0123, 0x0122
dw 0x0125, 0x0124
dw 0x0127, 0x0126
dw 0x0129, 0x0128
dw 0x012B, 0x012A
dw 0x012D, 0x012C
dw 0x012F, 0x012E
dw 0x0131, 0x0049
dw 0x0133, 0x0132
dw 0x0135, 0x0134
dw 0x0137, 0x0136
dw 0x013A, 0x0139
dw 0x013C, 0x013B
dw 0x013E, 0x013D
dw 0x0140, 0x013F
dw 0x0142, 0x0141
dw 0x0144, 0x0143
dw 0x0146, 0x0145
dw 0x0148, 0x0147
dw 0x014B, 0x014A
dw 0x014D, 0x014C
dw 0x014F, 0x014E
dw 0x0151, 0x0150
dw 0x0153, 0x0152
dw 0x0155, 0x0154
dw 0x0157, 0x0156
dw 0x0159, 0x0158
dw 0x015B, 0x015A
dw 0x015D, 0x015C
dw 0x015F, 0x015E
dw 0x0161, 0x0160
dw 0x0163, 0x0162
dw 0x0165, 0x0164
dw 0x0167, 0x0166
dw 0x0169, 0x0168
dw 0x016B, 0x016A
dw 0x016D, 0x016C
dw 0x016F, 0x016E
dw 0x0171, 0x0170
dw 0x0173, 0x0172
dw 0x0175, 0x0174
dw 0x0177, 0x0176
dw 0x017A, 0x0179
dw 0x017C, 0x017B
dw 0x017E, 0x017D
dw 0x0183, 0x0182
dw 0x0185, 0x0184
dw 0x0188, 0x0187
dw 0x018C, 0x018B
dw 0x0192, 0x0191
dw 0x0199, 0x0198
dw 0x01A1, 0x01A0
dw 0x01A3, 0x01A2
dw 0x01A5, 0x01A4
dw 0x01A8, 0x01A7
dw 0x01AD, 0x01AC
dw 0x01B0, 0x01AF
dw 0x01B4, 0x01B3
dw 0x01B6, 0x01B5
dw 0x01B9, 0x01B8
dw 0x01BD, 0x01BC
dw 0x01C6, 0x01C4
dw 0x01C9, 0x01C7
dw 0x01CC, 0x01CA
dw 0x01CE, 0x01CD
dw 0x01D0, 0x01CF
dw 0x01D2, 0x01D1
dw 0x01D4, 0x01D3
dw 0x01D6, 0x01D5
dw 0x01D8, 0x01D7
dw 0x01DA, 0x01D9
dw 0x01DC, 0x01DB
dw 0x01DF, 0x01DE
dw 0x01E1, 0x01E0
dw 0x01E3, 0x01E2
dw 0x01E5, 0x01E4
dw 0x01E7, 0x01E6
dw 0x01E9, 0x01E8
dw 0x01EB, 0x01EA
dw 0x01ED, 0x01EC
dw 0x01EF, 0x01EE
dw 0x01F3, 0x01F1
dw 0x01F5, 0x01F4
dw 0x01FB, 0x01FA
dw 0x01FD, 0x01FC
dw 0x01FF, 0x01FE
dw 0x0201, 0x0200
dw 0x0203, 0x0202
dw 0x0205, 0x0204
dw 0x0207, 0x0206
dw 0x0209, 0x0208
dw 0x020B, 0x020A
dw 0x020D, 0x020C
dw 0x020F, 0x020E
dw 0x0211, 0x0210
dw 0x0213, 0x0212
dw 0x0215, 0x0214
dw 0x0217, 0x0216
dw 0x0253, 0x0181
dw 0x0254, 0x0186
dw 0x0257, 0x018A
dw 0x0258, 0x018E
dw 0x0259, 0x018F
dw 0x025B, 0x0190
dw 0x0260, 0x0193
dw 0x0263, 0x0194
dw 0x0268, 0x0197
dw 0x0269, 0x0196
dw 0x026F, 0x019C
dw 0x0272, 0x019D
dw 0x0275, 0x019F
dw 0x0283, 0x01A9
dw 0x0288, 0x01AE
dw 0x028A, 0x01B1
dw 0x028B, 0x01B2
dw 0x0292, 0x01B7
dw 0x03AC, 0x0386
dw 0x03AD, 0x0388
dw 0x03AE, 0x0389
dw 0x03AF, 0x038A
dw 0x03B1, 0x0391
dw 0x03B2, 0x0392
dw 0x03B3, 0x0393
dw 0x03B4, 0x0394
dw 0x03B5, 0x0395
dw 0x03B6, 0x0396
dw 0x03B7, 0x0397
dw 0x03B8, 0x0398
dw 0x03B9, 0x0399
dw 0x03BA, 0x039A
dw 0x03BB, 0x039B
dw 0x03BC, 0x039C
dw 0x03BD, 0x039D
dw 0x03BE, 0x039E
dw 0x03BF, 0x039F
dw 0x03C0, 0x03A0
dw 0x03C1, 0x03A1
dw 0x03C3, 0x03A3
dw 0x03C4, 0x03A4
dw 0x03C5, 0x03A5
dw 0x03C6, 0x03A6
dw 0x03C7, 0x03A7
dw 0x03C8, 0x03A8
dw 0x03C9, 0x03A9
dw 0x03CA, 0x03AA
dw 0x03CB, 0x03AB
dw 0x03CC, 0x038C
dw 0x03CD, 0x038E
dw 0x03CE, 0x038F
dw 0x03E3, 0x03E2
dw 0x03E5, 0x03E4
dw 0x03E7, 0x03E6
dw 0x03E9, 0x03E8
dw 0x03EB, 0x03EA
dw 0x03ED, 0x03EC
dw 0x03EF, 0x03EE
dw 0x0430, 0x0410
dw 0x0431, 0x0411
dw 0x0432, 0x0412
dw 0x0433, 0x0413
dw 0x0434, 0x0414
dw 0x0435, 0x0415
dw 0x0436, 0x0416
dw 0x0437, 0x0417
dw 0x0438, 0x0418
dw 0x0439, 0x0419
dw 0x043A, 0x041A
dw 0x043B, 0x041B
dw 0x043C, 0x041C
dw 0x043D, 0x041D
dw 0x043E, 0x041E
dw 0x043F, 0x041F
dw 0x0440, 0x0420
dw 0x0441, 0x0421
dw 0x0442, 0x0422
dw 0x0443, 0x0423
dw 0x0444, 0x0424
dw 0x0445, 0x0425
dw 0x0446, 0x0426
dw 0x0447, 0x0427
dw 0x0448, 0x0428
dw 0x0449, 0x0429
dw 0x044A, 0x042A
dw 0x044B, 0x042B
dw 0x044C, 0x042C
dw 0x044D, 0x042D
dw 0x044E, 0x042E
dw 0x044F, 0x042F
dw 0x0451, 0x0401
dw 0x0452, 0x0402
dw 0x0453, 0x0403
dw 0x0454, 0x0404
dw 0x0455, 0x0405
dw 0x0456, 0x0406
dw 0x0457, 0x0407
dw 0x0458, 0x0408
dw 0x0459, 0x0409
dw 0x045A, 0x040A
dw 0x045B, 0x040B
dw 0x045C, 0x040C
dw 0x045E, 0x040E
dw 0x045F, 0x040F
dw 0x0461, 0x0460
dw 0x0463, 0x0462
dw 0x0465, 0x0464
dw 0x0467, 0x0466
dw 0x0469, 0x0468
dw 0x046B, 0x046A
dw 0x046D, 0x046C
dw 0x046F, 0x046E
dw 0x0471, 0x0470
dw 0x0473, 0x0472
dw 0x0475, 0x0474
dw 0x0477, 0x0476
dw 0x0479, 0x0478
dw 0x047B, 0x047A
dw 0x047D, 0x047C
dw 0x047F, 0x047E
dw 0x0481, 0x0480
dw 0x0491, 0x0490
dw 0x0493, 0x0492
dw 0x0495, 0x0494
dw 0x0497, 0x0496
dw 0x0499, 0x0498
dw 0x049B, 0x049A
dw 0x049D, 0x049C
dw 0x049F, 0x049E
dw 0x04A1, 0x04A0
dw 0x04A3, 0x04A2
dw 0x04A5, 0x04A4
dw 0x04A7, 0x04A6
dw 0x04A9, 0x04A8
dw 0x04AB, 0x04AA
dw 0x04AD, 0x04AC
dw 0x04AF, 0x04AE
dw 0x04B1, 0x04B0
dw 0x04B3, 0x04B2
dw 0x04B5, 0x04B4
dw 0x04B7, 0x04B6
dw 0x04B9, 0x04B8
dw 0x04BB, 0x04BA
dw 0x04BD, 0x04BC
dw 0x04BF, 0x04BE
dw 0x04C2, 0x04C1
dw 0x04C4, 0x04C3
dw 0x04C8, 0x04C7
dw 0x04CC, 0x04CB
dw 0x04D1, 0x04D0
dw 0x04D3, 0x04D2
dw 0x04D5, 0x04D4
dw 0x04D7, 0x04D6
dw 0x04D9, 0x04D8
dw 0x04DB, 0x04DA
dw 0x04DD, 0x04DC
dw 0x04DF, 0x04DE
dw 0x04E1, 0x04E0
dw 0x04E3, 0x04E2
dw 0x04E5, 0x04E4
dw 0x04E7, 0x04E6
dw 0x04E9, 0x04E8
dw 0x04EB, 0x04EA
dw 0x04EF, 0x04EE
dw 0x04F1, 0x04F0
dw 0x04F3, 0x04F2
dw 0x04F5, 0x04F4
dw 0x04F9, 0x04F8
dw 0x0561, 0x0531
dw 0x0562, 0x0532
dw 0x0563, 0x0533
dw 0x0564, 0x0534
dw 0x0565, 0x0535
dw 0x0566, 0x0536
dw 0x0567, 0x0537
dw 0x0568, 0x0538
dw 0x0569, 0x0539
dw 0x056A, 0x053A
dw 0x056B, 0x053B
dw 0x056C, 0x053C
dw 0x056D, 0x053D
dw 0x056E, 0x053E
dw 0x056F, 0x053F
dw 0x0570, 0x0540
dw 0x0571, 0x0541
dw 0x0572, 0x0542
dw 0x0573, 0x0543
dw 0x0574, 0x0544
dw 0x0575, 0x0545
dw 0x0576, 0x0546
dw 0x0577, 0x0547
dw 0x0578, 0x0548
dw 0x0579, 0x0549
dw 0x057A, 0x054A
dw 0x057B, 0x054B
dw 0x057C, 0x054C
dw 0x057D, 0x054D
dw 0x057E, 0x054E
dw 0x057F, 0x054F
dw 0x0580, 0x0550
dw 0x0581, 0x0551
dw 0x0582, 0x0552
dw 0x0583, 0x0553
dw 0x0584, 0x0554
dw 0x0585, 0x0555
dw 0x0586, 0x0556
dw 0x10D0, 0x10A0
dw 0x10D1, 0x10A1
dw 0x10D2, 0x10A2
dw 0x10D3, 0x10A3
dw 0x10D4, 0x10A4
dw 0x10D5, 0x10A5
dw 0x10D6, 0x10A6
dw 0x10D7, 0x10A7
dw 0x10D8, 0x10A8
dw 0x10D9, 0x10A9
dw 0x10DA, 0x10AA
dw 0x10DB, 0x10AB
dw 0x10DC, 0x10AC
dw 0x10DD, 0x10AD
dw 0x10DE, 0x10AE
dw 0x10DF, 0x10AF
dw 0x10E0, 0x10B0
dw 0x10E1, 0x10B1
dw 0x10E2, 0x10B2
dw 0x10E3, 0x10B3
dw 0x10E4, 0x10B4
dw 0x10E5, 0x10B5
dw 0x10E6, 0x10B6
dw 0x10E7, 0x10B7
dw 0x10E8, 0x10B8
dw 0x10E9, 0x10B9
dw 0x10EA, 0x10BA
dw 0x10EB, 0x10BB
dw 0x10EC, 0x10BC
dw 0x10ED, 0x10BD
dw 0x10EE, 0x10BE
dw 0x10EF, 0x10BF
dw 0x10F0, 0x10C0
dw 0x10F1, 0x10C1
dw 0x10F2, 0x10C2
dw 0x10F3, 0x10C3
dw 0x10F4, 0x10C4
dw 0x10F5, 0x10C5
dw 0x1E01, 0x1E00
dw 0x1E03, 0x1E02
dw 0x1E05, 0x1E04
dw 0x1E07, 0x1E06
dw 0x1E09, 0x1E08
dw 0x1E0B, 0x1E0A
dw 0x1E0D, 0x1E0C
dw 0x1E0F, 0x1E0E
dw 0x1E11, 0x1E10
dw 0x1E13, 0x1E12
dw 0x1E15, 0x1E14
dw 0x1E17, 0x1E16
dw 0x1E19, 0x1E18
dw 0x1E1B, 0x1E1A
dw 0x1E1D, 0x1E1C
dw 0x1E1F, 0x1E1E
dw 0x1E21, 0x1E20
dw 0x1E23, 0x1E22
dw 0x1E25, 0x1E24
dw 0x1E27, 0x1E26
dw 0x1E29, 0x1E28
dw 0x1E2B, 0x1E2A
dw 0x1E2D, 0x1E2C
dw 0x1E2F, 0x1E2E
dw 0x1E31, 0x1E30
dw 0x1E33, 0x1E32
dw 0x1E35, 0x1E34
dw 0x1E37, 0x1E36
dw 0x1E39, 0x1E38
dw 0x1E3B, 0x1E3A
dw 0x1E3D, 0x1E3C
dw 0x1E3F, 0x1E3E
dw 0x1E41, 0x1E40
dw 0x1E43, 0x1E42
dw 0x1E45, 0x1E44
dw 0x1E47, 0x1E46
dw 0x1E49, 0x1E48
dw 0x1E4B, 0x1E4A
dw 0x1E4D, 0x1E4C
dw 0x1E4F, 0x1E4E
dw 0x1E51, 0x1E50
dw 0x1E53, 0x1E52
dw 0x1E55, 0x1E54
dw 0x1E57, 0x1E56
dw 0x1E59, 0x1E58
dw 0x1E5B, 0x1E5A
dw 0x1E5D, 0x1E5C
dw 0x1E5F, 0x1E5E
dw 0x1E61, 0x1E60
dw 0x1E63, 0x1E62
dw 0x1E65, 0x1E64
dw 0x1E67, 0x1E66
dw 0x1E69, 0x1E68
dw 0x1E6B, 0x1E6A
dw 0x1E6D, 0x1E6C
dw 0x1E6F, 0x1E6E
dw 0x1E71, 0x1E70
dw 0x1E73, 0x1E72
dw 0x1E75, 0x1E74
dw 0x1E77, 0x1E76
dw 0x1E79, 0x1E78
dw 0x1E7B, 0x1E7A
dw 0x1E7D, 0x1E7C
dw 0x1E7F, 0x1E7E
dw 0x1E81, 0x1E80
dw 0x1E83, 0x1E82
dw 0x1E85, 0x1E84
dw 0x1E87, 0x1E86
dw 0x1E89, 0x1E88
dw 0x1E8B, 0x1E8A
dw 0x1E8D, 0x1E8C
dw 0x1E8F, 0x1E8E
dw 0x1E91, 0x1E90
dw 0x1E93, 0x1E92
dw 0x1E95, 0x1E94
dw 0x1EA1, 0x1EA0
dw 0x1EA3, 0x1EA2
dw 0x1EA5, 0x1EA4
dw 0x1EA7, 0x1EA6
dw 0x1EA9, 0x1EA8
dw 0x1EAB, 0x1EAA
dw 0x1EAD, 0x1EAC
dw 0x1EAF, 0x1EAE
dw 0x1EB1, 0x1EB0
dw 0x1EB3, 0x1EB2
dw 0x1EB5, 0x1EB4
dw 0x1EB7, 0x1EB6
dw 0x1EB9, 0x1EB8
dw 0x1EBB, 0x1EBA
dw 0x1EBD, 0x1EBC
dw 0x1EBF, 0x1EBE
dw 0x1EC1, 0x1EC0
dw 0x1EC3, 0x1EC2
dw 0x1EC5, 0x1EC4
dw 0x1EC7, 0x1EC6
dw 0x1EC9, 0x1EC8
dw 0x1ECB, 0x1ECA
dw 0x1ECD, 0x1ECC
dw 0x1ECF, 0x1ECE
dw 0x1ED1, 0x1ED0
dw 0x1ED3, 0x1ED2
dw 0x1ED5, 0x1ED4
dw 0x1ED7, 0x1ED6
dw 0x1ED9, 0x1ED8
dw 0x1EDB, 0x1EDA
dw 0x1EDD, 0x1EDC
dw 0x1EDF, 0x1EDE
dw 0x1EE1, 0x1EE0
dw 0x1EE3, 0x1EE2
dw 0x1EE5, 0x1EE4
dw 0x1EE7, 0x1EE6
dw 0x1EE9, 0x1EE8
dw 0x1EEB, 0x1EEA
dw 0x1EED, 0x1EEC
dw 0x1EEF, 0x1EEE
dw 0x1EF1, 0x1EF0
dw 0x1EF3, 0x1EF2
dw 0x1EF5, 0x1EF4
dw 0x1EF7, 0x1EF6
dw 0x1EF9, 0x1EF8
dw 0x1F00, 0x1F08
dw 0x1F01, 0x1F09
dw 0x1F02, 0x1F0A
dw 0x1F03, 0x1F0B
dw 0x1F04, 0x1F0C
dw 0x1F05, 0x1F0D
dw 0x1F06, 0x1F0E
dw 0x1F07, 0x1F0F
dw 0x1F10, 0x1F18
dw 0x1F11, 0x1F19
dw 0x1F12, 0x1F1A
dw 0x1F13, 0x1F1B
dw 0x1F14, 0x1F1C
dw 0x1F15, 0x1F1D
dw 0x1F20, 0x1F28
dw 0x1F21, 0x1F29
dw 0x1F22, 0x1F2A
dw 0x1F23, 0x1F2B
dw 0x1F24, 0x1F2C
dw 0x1F25, 0x1F2D
dw 0x1F26, 0x1F2E
dw 0x1F27, 0x1F2F
dw 0x1F30, 0x1F38
dw 0x1F31, 0x1F39
dw 0x1F32, 0x1F3A
dw 0x1F33, 0x1F3B
dw 0x1F34, 0x1F3C
dw 0x1F35, 0x1F3D
dw 0x1F36, 0x1F3E
dw 0x1F37, 0x1F3F
dw 0x1F40, 0x1F48
dw 0x1F41, 0x1F49
dw 0x1F42, 0x1F4A
dw 0x1F43, 0x1F4B
dw 0x1F44, 0x1F4C
dw 0x1F45, 0x1F4D
dw 0x1F51, 0x1F59
dw 0x1F53, 0x1F5B
dw 0x1F55, 0x1F5D
dw 0x1F57, 0x1F5F
dw 0x1F60, 0x1F68
dw 0x1F61, 0x1F69
dw 0x1F62, 0x1F6A
dw 0x1F63, 0x1F6B
dw 0x1F64, 0x1F6C
dw 0x1F65, 0x1F6D
dw 0x1F66, 0x1F6E
dw 0x1F67, 0x1F6F
dw 0x1F80, 0x1F88
dw 0x1F81, 0x1F89
dw 0x1F82, 0x1F8A
dw 0x1F83, 0x1F8B
dw 0x1F84, 0x1F8C
dw 0x1F85, 0x1F8D
dw 0x1F86, 0x1F8E
dw 0x1F87, 0x1F8F
dw 0x1F90, 0x1F98
dw 0x1F91, 0x1F99
dw 0x1F92, 0x1F9A
dw 0x1F93, 0x1F9B
dw 0x1F94, 0x1F9C
dw 0x1F95, 0x1F9D
dw 0x1F96, 0x1F9E
dw 0x1F97, 0x1F9F
dw 0x1FA0, 0x1FA8
dw 0x1FA1, 0x1FA9
dw 0x1FA2, 0x1FAA
dw 0x1FA3, 0x1FAB
dw 0x1FA4, 0x1FAC
dw 0x1FA5, 0x1FAD
dw 0x1FA6, 0x1FAE
dw 0x1FA7, 0x1FAF
dw 0x1FB0, 0x1FB8
dw 0x1FB1, 0x1FB9
dw 0x1FD0, 0x1FD8
dw 0x1FD1, 0x1FD9
dw 0x1FE0, 0x1FE8
dw 0x1FE1, 0x1FE9
dw 0x24D0, 0x24B6
dw 0x24D1, 0x24B7
dw 0x24D2, 0x24B8
dw 0x24D3, 0x24B9
dw 0x24D4, 0x24BA
dw 0x24D5, 0x24BB
dw 0x24D6, 0x24BC
dw 0x24D7, 0x24BD
dw 0x24D8, 0x24BE
dw 0x24D9, 0x24BF
dw 0x24DA, 0x24C0
dw 0x24DB, 0x24C1
dw 0x24DC, 0x24C2
dw 0x24DD, 0x24C3
dw 0x24DE, 0x24C4
dw 0x24DF, 0x24C5
dw 0x24E0, 0x24C6
dw 0x24E1, 0x24C7
dw 0x24E2, 0x24C8
dw 0x24E3, 0x24C9
dw 0x24E4, 0x24CA
dw 0x24E5, 0x24CB
dw 0x24E6, 0x24CC
dw 0x24E7, 0x24CD
dw 0x24E8, 0x24CE
dw 0x24E9, 0x24CF
dw 0xFF41, 0xFF21
dw 0xFF42, 0xFF22
dw 0xFF43, 0xFF23
dw 0xFF44, 0xFF24
dw 0xFF45, 0xFF25
dw 0xFF46, 0xFF26
dw 0xFF47, 0xFF27
dw 0xFF48, 0xFF28
dw 0xFF49, 0xFF29
dw 0xFF4A, 0xFF2A
dw 0xFF4B, 0xFF2B
dw 0xFF4C, 0xFF2C
dw 0xFF4D, 0xFF2D
dw 0xFF4E, 0xFF2E
dw 0xFF4F, 0xFF2F
dw 0xFF50, 0xFF30
dw 0xFF51, 0xFF31
dw 0xFF52, 0xFF32
dw 0xFF53, 0xFF33
dw 0xFF54, 0xFF34
dw 0xFF55, 0xFF35
dw 0xFF56, 0xFF36
dw 0xFF57, 0xFF37
dw 0xFF58, 0xFF38
dw 0xFF59, 0xFF39
dw 0xFF5A, 0xFF3A
dw 0, 0
end if