DONT_COPY_OLD_DATA equ 0x1
SizeUsedByInitNewPointerList equ psize*3

AllocatedMemorys pointer 0

if used MemoryMultiThreaded
  basePointerList pointer 0  ;used for the first thread and the service thread :)
  unusedPointerListCollection pointer 0
end if

;Default (no macro used) is Virtual memory mode
;To use Virtual_Divider_Memory, comment the line ;Virtual_Thread_Divider_Memory
;and uncomment Virtual_Divider_Memory

;Virtual_Divider_Memory
Virtual_Thread_Divider_Memory

;-------------------------------------------
;----There are three memory modes:
;-------------------------------------------
;!!WARNING: changing from OR to Virtual Thread Divider
;will require a fully recompile all files in the project
;-------------------------------------------
;--Virtual: quick, but memory consuming. All memory is requested
;by calling VirtualAlloc. The smallest amount of memory it returns
;is ~4K. This means that you require 10 bytes and get 4K.

;--Virtual Divider: slower than Virtual, but AL LOT less memory
;consuming. This will allocate memory though VirtualAlloc
;and divides it over multple requests. The overhead for every
;peace of memory is 0x10 bytes and all memory is rounded up by 0x10.

;--Virtual Thread Divider: A litle bit more memory consuming
;than Virtual divider, but a lot quicker. It's basicly the same as
;Virtual Divider, with the difference that every thread divides
;there own peace of memory, instead of centralizing it. This means
;that different threads don't have to wait on each other, which is
;a hugh win in time. The pointer to the AllocatedMemory structure (see
;below) is placed in the TEB.
;-----------------------------------------------

;Structures used by Virtual Divider or Virtual Thread Divider:
;-----------------------------------------------

;AllocatedMemory structure
;  - This is a linked list with pointers to "pointer spots" (Allocated
;  by VirtualAlloc) and a lock byte for protection against multiple
;  threads usage at the same time. Use "lock cmpxchg" to set
;  this value. When pointer spots are not used anymore they are dismissed by VirtualFree

  ;[lock 1 byte]
  ;[total pointers 4 bytes]
  ;[next list] (this is a pointer, 4 bytes on 32 bit and 8 on 64bit)
  ;[First pointer spot]
  ;[...]
  ;[Last pointer spot]

;Pointer spots structure
;  - this memory holds memory that is divided on multiple requests. This structure
;  can have multiple headers and is initiated with one. Every header is 16 bytes.

;  "Biggest free pointer" is a dword and indicates the size of the biggest memory spot that can
;  be allocated in this "Pointer spots" structure. (on init this is 4K - 16b). This value
;  is only valid on top of the "Pointer spots" structure and is 'reserved' in the following
;  headers.

;  "used mem" is a boolean, '1' if the "mem location" is allocated and '0' if it's freed (or not been
;  used in the first place.

;  "size" is a dword that holds the size of the "mem location"

;  "nextpointer" is a boolean that indicates if there is a next header after the "mem location"

;  "mem location" is the memory that can allocated or is already in use (see "used mem") with size "size".

  ;[Biggest free pointer 4 bytes]
  ;[used mem, 1 byte]
  ;[size 4 bytes]
  ;[nextpointer byte]
  ;[reserved 6 bytes]
  ;[mem location]
  ;[next Pointer spots structure] (used to be mempointer structure)
  ;[...]
  ;[last Pointer spots structure]

;Stack Structure:
;  -This structure is placed on stack (which is placed in the TEB)

;  "AllocatedMemory" is a pointer the AllocatedMemory structure for the current thread itself

;  "AllocatedMemorys" is a pointer to a collection with all AllocatedMemory structures

;  "unusedPointerListCollection" is a pointer with all the AllocatedMemory structures
;   that are not assigned to any thread, but still had located memory in it.

;-----------------------------------------------

;------------------------------------------------------------------------------
;Removes all Pointer spots structure and then the AllocatedMemory structure itself
;input:
;  (R/E)DX => Pointer to a AllocatedMemory structure
;output: unknown
;------------------------------------------------------------------------------
if used New_Memory_management & used FreePointerList
  FreePointerList:
  call denter

  mov pesi,[pebp+psize*2]
  xor ebx,ebx

  test pesi,pesi
  jz endFree

  ;lock
  lodsb

  ;total pointers
  lodsd
  mov ecx,eax

  lodsp ;next list

  test peax,peax
  jz freePointers

  push peax
  call FreePointerList
  mov pebx,peax ;set 'next pointer' to (R/E)BX
  mov pdword [pesi-psize],peax ;set next list

  ;Copy pointers [ESI] to [EDI] with size ECX
  freePointers:
    lodsp ;get pointer
    test peax,peax
    jz noPointerToFree

      mov pebx,pdword [pebp+psize*2] ;next pointer will be this structure itself
    ;push peax
    ;call FreeMemLocationVirtual

    noPointerToFree:
  loop freePointers

  endFree:
  cmp pebx,pdword [pebp+psize*2]
  je listStillHasPointers

    push pdword [pebp+psize*2]
    call FreeMemLocationVirtual

  listStillHasPointers:

  mov peax,pebx
  call dleave
  ret psize
end if


;------------------------------------------------------------------------------
;Removes Pointer spots structure from AllocatedMemory structure
;INPUT:
;    1: Local Storage (only when MemoryMultiThreaded is set)
;    (R/E)DX => Pointer to PointerStruct
;    (R/E)BX => BaseAddess
;------------------------------------------------------------------------------
if used removeMemPointerFromPointerList
  removeMemPointerFromPointerList:
  call denter

  mov pesi,pedx           ;(R/E)SI => List
  xor edi,edi             ;EDI => counter of active pointers in Pointer list

  ;lock
  lodsb

  ;total pointers
  lodsd
  mov ecx,eax

  lodsp ;next list

  ;Copy pointers [ESI] to [EDI] with size ECX
  readPointers:
    lodsp ;get pointer
    cmp peax,pebx
    jne notPointerToRemove

      push peax
      call FreeMemLocationVirtual
      mov pdword [pesi-psize],0
      jmp nextReadPointers

    notPointerToRemove:
      test peax,peax
      jz nextReadPointers

        inc pedi ;we found an active pointer in the list

  nextReadPointers:
  loop readPointers

  endReadPointers:
    test pedi,pedi
    jnz endRemoveMemPointerFromPointerList

    if used MemoryMultiThreaded
       mov pesi,[pebp+psize*2] ;AllocatedList
    else
      mov pesi,[AllocatedMemorys]
    end if

    cmp pedx,pesi
    je endRemoveMemPointerFromPointerList ;dont remove first list

    ;remove list since zero pointers in it
    push pdword [pedx+5] ;next list

    removePointerListFromLinkedList:
      cmp pdword [pesi+5],pedx
      jne nextRemovePointerListFromLinkedList

        pop pdword [pesi+5]
        push pedx
        call FreeMemLocationVirtual
        jmp endRemoveMemPointerFromPointerList

    nextRemovePointerListFromLinkedList:
      mov pesi,pdword [pesi+5]
      test pesi,pesi
    jnz removePointerListFromLinkedList

  endRemoveMemPointerFromPointerList:
  call dleave
  if used MemoryMultiThreaded
    ret psize
  else
    ret
  end if
end if

;------------------------------------------------------------------------------
;This sould be called when a new thread starts, a AllocatedMemory structure will
;be init and be placed on Stack, and pointer to stack will be placed on TEB.
;WARNING: Dont'forget to call uninitNewPointerListInTEB before exit the thread
;to clean up this structure from the stack!!

;INPUT:
;    NONE
;------------------------------------------------------------------------------
if used New_Memory_management & used MemoryMultiThreaded
  macro startLocalStorage
  { common
    call initNewPointerListInTEB
    push peax
    call __CreateCollectionVirtual

    mov [AllocatedMemorys],peax
    mov pdword [pesp+psize],peax ;set AllocatedMemory in struct in TEB

    call CreateCollection
    mov [unusedPointerListCollection],peax
    mov pdword [pesp+psize*2],peax ;set unusedPointerListCollection in struct in TEB
  }

  ;------------------------------------------------------------------------------
  ;Same as initNewPointerListInTEB, with difference it does not corrupt the stack!
  ;INPUT:
  ;    NONE
  ;------------------------------------------------------------------------------
  if used initNewPointerListInTEBForService
    initNewPointerListInTEBForService:
    call denter

    ;call initNewPointerListInTEB
    push [unusedPointerListCollection]
    push [AllocatedMemorys]
    push [basePointerList]

    push SizeUsedByInitNewPointerList
    call GetMemLocationVirtual

    pop pdword [peax]
    pop pdword [peax+psize]
    pop pdword [peax+psize*2]

    call SetUserReservedValueTeb

    call dleave

    ret
  end if

  ;------------------------------------------------------------------------------
  ;This will init a Stack Structure and places it on TEB:UserReservedValue
  ;  Warning, this will move the stack pointer when returning, use uninitNewPointerListInTEB
  ;           to fix this
  ;
  ; OUTPUT:
  ;   when AllocatedMemory is zero: pointer to local PointerList
  ;   when AllocatedMemory is not zero: index in AllocatedMemory for pointer to local PointerList
  ;------------------------------------------------------------------------------
  if used initNewPointerListInTEB
    initNewPointerListInTEB:
    pop peax ;return pointer

    sub pesp,SizeUsedByInitNewPointerList ;psize*3 ;make room on stack

    push peax
    push pebx ;save EBX to be used as a boolean

    ;make THREADMEM structure with local storage and a pointer to a collection with all local storages
    mov peax,[AllocatedMemorys]
    mov pdword [pesp+psize*3], peax ;set AllocatedMemory

    mov peax,[unusedPointerListCollection]
    mov pdword [pesp+psize*4], peax ;set unusedPointerListCollection

    ;Get unused PointerList if avaidable
    push 1
    push pdword [unusedPointerListCollection] ;collection with all the unused Pointerstructs
    call RemoveDwordItem

    mov pebx,peax ;save if the pointer list is from the unused list
    test peax,peax
    jnz foundUnusedList
      call initNewPointerList

    foundUnusedList:
    mov pdword [pesp+psize*2],peax ;set Local PointerList

    mov peax,pesp
    add peax,psize*2
    ;use WOW32Reserved in TEB to save the local memory storage
    call SetUserReservedValueTeb

    ;list is already added to the AllocatedMemory list if this list is from the unusedPointerCollection
    test pebx,pebx
    jnz setFirstPointerList

    cmp [AllocatedMemorys],0
    je setFirstPointerList

      push 0x1 ;FULL_LOCK
      push [AllocatedMemorys]
      call LockListAcces

      mov peax,pdword [pesp+psize*2] ;return local list

      push peax
      push [AllocatedMemorys]
      call AddDword

      ;unlock the list
      push 0x1 ;FULL_LOCK
      push [AllocatedMemorys]
      call ExternalUnlockListAccesEx
      jmp endInitNewPointerListInTEB

    setFirstPointerList:
      mov peax,pdword [pesp+psize*2] ;return local list
      mov [basePointerList],peax

    endInitNewPointerListInTEB:
      mov peax,pdword [pesp+psize*2] ;return local list

      pop pebx ;restore (R/E)BX
    ret
  end if

  ;------------------------------------------------------------------------------
  ;Frees a AllocatedMemory structure and restores stack
  ;
  ;INPUT: NONE
  ;OUTPUT: NONE
  ;------------------------------------------------------------------------------
  if used uninitNewPointerListInTEB
    uninitNewPointerListInTEB:
    call denter

    mov peax,pdword [pebp+psize*3] ;collection with all local storages

    push 0x1 ;FULL_LOCK
    push peax
    call LockListAcces

    mov peax,pdword [pebp+psize*3] ;collection with all local storages

    push pdword [pebp+psize*2] ;pointer list
    push peax
    call FindDword

    push peax
    mov peax,pdword [pebp+psize*3] ;collection with all local storages
    push peax
    call RemoveDwordItem

    mov peax,pdword [pebp+psize*2] ;get Pointer list

    push peax
    call FreePointerList
    mov pdword [pebp+psize*2],peax ;set Pointer list
    test peax,peax
    jz endUninitNewPointerListInTEB
      push peax ;push for next AddDword

      ;Since there are still pointers in use, add it to the local storages again
      push peax
      push pdword [pebp+psize*3] ;collection with all local storages
      call AddDword

      ;Add to the unusedPointerListCollection
      ;push peax push already done
      push pdword [pebp+psize*4] ;collection with all the unused Pointerstructs
      call AddDword

  endUninitNewPointerListInTEB:
    mov peax,pdword [pebp+psize*3] ;collection with all local storages
    ;unlock the list
    push 0x1 ;FULL_LOCK
    push peax
    call ExternalUnlockListAccesEx

    call dleave
    ret SizeUsedByInitNewPointerList ;psize*3 ;make room on stack
  end if

  ;------------------------------------------------------------------------------
  ; Retrieves collection of all AllLocalMemory structures from TEB to (R/E)DI
    ;OUTPUT:
  ;     (R/E)DI: pointer to collection with all the AllLocalMemory structures
  ;------------------------------------------------------------------------------
  if used getListAllLocalMemoryStorageEDI
    getListAllLocalMemoryStorageEDI:
    push peax
    call GetUserReservedValueTeb
    mov pedi,peax
    pop peax

    mov pedi,pdword [pedi+psize]
    ret
  end if

  ;------------------------------------------------------------------------------
  ; Retrieves AllLocalMemory structure from TEB to (R/E)SI
  ;OUTPUT:
  ;    (R/E)SI: pointer to AllLocalMemory structure
  ;------------------------------------------------------------------------------
  if used getLocalMemoryStorageESI
    getLocalMemoryStorageESI:
    push peax
    call GetUserReservedValueTeb
    mov pesi,peax
    pop peax

    mov pesi,pdword [pesi]
    ret
  end if

  ;------------------------------------------------------------------------------
  ; Retrieves AllLocalMemory structure from TEB to (R/E)AX
  ;OUTPUT:
  ;    (R/E)AX: pointer to AllLocalMemory structure
  ;------------------------------------------------------------------------------
  if used getLocalMemoryStorage
    getLocalMemoryStorage:
    call GetUserReservedValueTeb

    mov peax,pdword [peax]
    ret
  end if

  ;------------------------------------------------------------------------------
  ; Retrieves AllLocalMemory structure from TEB to (R/E)DI
  ;OUTPUT:
  ;   (R/E)DI: pointer to AllLocalMemory structure
  ;------------------------------------------------------------------------------
  if used getLocalMemoryStorageEDI
    getLocalMemoryStorageEDI:
    push peax
    call GetUserReservedValueTeb
    mov pedi,peax
    pop peax

    mov pedi,pdword [pedi]
    ret
  end if
end if

;------------------------------------------------------------------------------
;This will init a Pointer Spot Structure
;OUTPUT:
;    (R/E)AX: pointer to a Pointer Spot Structure
;------------------------------------------------------------------------------
if used New_Memory_management
  initNewPointerList:
  push pecx

    mov ecx, [smallestPortionOfPageSize]
    push pecx
    call GetMemLocationVirtual

    sub ecx,(5+psize) ;sub 5 bytes used for lock byte,  free pointers dword and next list
    if used is32
      shr ecx,2      ;divide by 4 (to get the ammount of pointers)
    else
      shr ecx,3      ;divide by 8 (to get the ammount of pointers)
    end if

    mov dword [peax+1],ecx ;ammount of total pointer spots
    mov byte [peax],0      ;set lock
   pop pecx
  ret


  ;------------------------------------------------------------------------------
  ;This will init a Pointer Spot Structure
  ;INPUT:
  ;  (only when Virtual Thread Divider is selected:)
  ;    1: pointer to a Pointer Spot Structur
  ;--------------------------------------
  if used getLockPointerStructure
    getLockPointerStructure:
      call denter
      if used MemoryMultiThreaded
        mov pesi,pdword [pebp+psize*2]
      else
        mov pesi,pdword [AllocatedMemorys]
      end if

      jmp noSleepGetLockPointerStructure
      ;get lock on AllocatedMemory
      lockAllocatedMemory:
        push 1
        call sleep

      noSleepGetLockPointerStructure:
        xor al,al
        mov cl,1
        lock cmpxchg byte [pesi], cl
      jnz lockAllocatedMemory
      call dleave
    if used MemoryMultiThreaded
      ret psize
    else
      ret
    end if
  end if

  ;--------------------------------------
  ; looks through all AllocatedMemory collection to find the Base Address of Pointer spot structure.
  ; If not found in the local AllocatedMemory, this function will look in other local storages
  ; WARNING it will lock the returned Pointer spot structure, unlock it when you are done!
  ;INPUT
  ;  1: pointer to memory location to find base address from
  ;OUTPUT:
  ;  When fails:
  ;    (R/E)AX => Zero, other registers untouched
  ;  When succeeds:
  ;    (R/E)AX => pointer to Pointer Spots structure where the memory location is located in
  ;    (R/E)DX => pointer to AllocatedMemory structure where the Pointer Spots structure is located in
  ;--------------------------------------
  if used getBaseaddressPointerstructureAndBelongingLocalStorage & used MemoryMultiThreaded
    getBaseaddressPointerstructureAndBelongingLocalStorage:
    call denter

    call getLocalMemoryStorageESI

    push pesi
    call getLockPointerStructure

    push pesi
    push pdword [pebp+psize*2]
    call findBaseAddressAndPointerStructure

    test peax,peax
    jnz endGetBaseaddressPointerstructureAndPointerlist

    push pesi
    call releaseLockPointerStructure

    call getListAllLocalMemoryStorageEDI
    ;lock the list for write actions
    push pedi
    call LockListWriteAcces

    push pedi
    call CountItems
    mov pecx,peax

    loopThoughLocalStorages:
      push pecx
      push pedi
      call GetDwordItem
      test peax,peax
      jz nextLoopThoughLocalStorages

      mov pesi,peax

      push pesi
      call getLockPointerStructure

      push pesi
      push pdword [pebp+psize*2]
      call findBaseAddressAndPointerStructure
      test peax,peax
      jnz endLoopThoughLocalStorages

      push pesi
      call releaseLockPointerStructure
    nextLoopThoughLocalStorages:
      loop loopThoughLocalStorages

    endLoopThoughLocalStorages:
      push peax

      push pedi
      call ExternalUnlockListWriteAcces

      pop peax

    endGetBaseaddressPointerstructureAndPointerlist:
      mov pdword [pebp+psize*2],pesi

      push 0 ;push dummy (will be overwritten by dleave method
      push peax
      push pedx
      mov peax,pesp ;AX is only register that is saved by dleave
      call dleave

      cmp pdword [peax+psize],0
      je dontOverwriteEDX

      mov pedx,pdword [peax]

      dontOverwriteEDX:
        mov peax,pdword [peax+psize]

      ret psize
  end if

  ;--------------------------------------
  ; This releases the lock on a AllocatedMemory structure
  ;INPUT ONLY when Virtual * Divided memory is used:
  ;  1: pointer to Allocated memory structure
  ;--------------------------------------
  if used releaseLockPointerStructure
    releaseLockPointerStructure:
      push peax

      if used MemoryMultiThreaded
        mov peax,pdword [pesp+psize*2]
      else
        mov peax,pdword [AllocatedMemorys]
      end if
      mov byte [peax],0

      pop peax

    if used MemoryMultiThreaded
      ret psize
    else
      ret
    end if
  end if

  ;------------------------------------------------------------------------------
  ;This function places "pointer spots structure" in AllocatedMemory structure and
  ;allocates a new AllocatedMemory structure if neccecery
  ;INPUT
  ;  1: a pointer to a pointer spots structure
  ;------------------------------------------------------------------------------
  if used setMemPointerStructureInPointerList
    setMemPointerStructureInPointerList:
    call denter

      if used MemoryMultiThreaded
        call getLocalMemoryStorageESI
      else
        mov pesi,pdword [AllocatedMemorys]
      end if

      push pesi ;save current Pointerlist

      inc pesi
      lodsd
      mov ecx,eax ;total pointer

      lodsp ;get 'next'

      findEmptyPointerSpot:
        lodsp

        push pecx ;save ECX

        test peax,peax
        jnz nextfindEmptyPointerSpot

          ;Save the memPointerStructure
          mov peax,pdword [pebp+psize*2]
          mov pdword [pesi-psize],peax
          add pesp,psize ;remove current Pointerlis from stack
          jmp endSetMemPointerStructureInPointerList

      nextfindEmptyPointerSpot:
        pop pecx ;restore ECX
        loop findEmptyPointerSpot

      pop pesi ;restore current Pointerlist
      inc pesi
      lodsd
      lodsp    ;get next pointer
      test peax,peax
      jz noEmptyPointerSpot

      mov pesi,peax
      push pesi ;save current Pointerlist

      inc pesi ;skip lock byte
      lodsd
      mov ecx,eax ;total pointer

      lodsp ;get 'next'
      jmp findEmptyPointerSpot

    noEmptyPointerSpot:
      call initNewPointerList
      ;--pop pesi ;restore current PointerList
      mov pdword [pesi-psize],peax ;set next list 

      ;set Pointerlist in memPointerStructure
      mov pesi,pdword [pebp+psize*2]
      mov pdword [peax+5+psize],pesi
      xchg pesi,peax


    endSetMemPointerStructureInPointerList:
    call dleave
    ret psize
  end if


  ;------------------------------------------------------------------------------
  ;Finds the Pointer Spots structure where the momeory is is allocated and the
  ;AllocatedMemory structure where this pointer spots structure is placed.
  ;  1:memAddress
  ;  (Only when Virtual Thread Divider is selected:)
  ;    2:AllocatedMemory structure
  ; Return:
  ;  (R/E)AX => Pointer Spots structure (or zero)
  ;  (R/E)DX => AllocatedMemory structure (or unknown if AX = zero)
  ;------------------------------------------------------------------------------
  if used findBaseAddressAndPointerStructure
    findBaseAddressAndPointerStructure:
    call denter

      if used MemoryMultiThreaded
        mov pesi,pdword [pebp+psize*3]
     else
        mov pesi,pdword [AllocatedMemorys]
      end if

      push pesi ;save current Pointerlist
      inc pesi
      lodsd
      mov ecx,eax ;total pointer

      lodsp ;get 'next list'

      findBaseAddress:
        lodsp

        push pecx ;save ECX

        test peax,peax
        jz nextFindBaseAddress

        mov pedx,pdword [pebp+psize*2]
        ;address < BaseAddress, not the base address we are looking for
        cmp pedx,peax
        jb nextFindBaseAddress

        add peax,pdword [smallestPortionOfPageSize]
        ;address > BaseAddress+smallest size, not the base address we are looking for
        cmp pedx,peax
        jae nextFindBaseAddress

          ;SET (R/E)AX = Base Address
          ;    (R/E)DX = PointerStructure
          sub peax,pdword [smallestPortionOfPageSize]
          add pesp,psize ;skip ECX
          pop pedx

          push 0 ;push dummy (will be overwritten by dleave method
          push peax
          push pedx
          mov peax,pesp ;AX is only register that is saved by dleave
          call dleave
          mov pedx,pdword [peax]
          mov peax,pdword [peax+psize]
          jmp _endFindBaseAddress

      nextFindBaseAddress:
        pop pecx ;restore ECX
        loop findBaseAddress

      pop pesi ;restore current Pointerlist
      inc pesi ;skip lock
      lodsd    ;skip total pointers
      lodsp    ;get 'next'
      test peax,peax
      jz noBaseAddressFound

      mov pesi,peax
      push pesi ;save current Pointerlist

      inc pesi ;skip lock byte
      lodsd
      mov ecx,eax ;total pointer

      lodsp ;get 'next'
      jmp findBaseAddress

      noBaseAddressFound:
      xor peax,peax

    endFindBaseAddress:
    call dleave
    _endFindBaseAddress:
    if used MemoryMultiThreaded
      ret psize*2
    else
      ret psize
    end if
  end if

  ;------------------------------------------------------------------------------
  ;groups different free mem locations that are next to each other in a
  ;pointer spots structure.
  ;INPUT:
  ;  1:pointer spots structure
  ;------------------------------------------------------------------------------
  if used resizeMemoryPointers
    resizeMemoryPointers:
    call denter

    mov pesi,pdword [pebp+psize*2]

    findFreeMemPointer:
      cmp byte [pesi+4],1 ;is memory used?
      je nextFindFreeMemPointer

        CombineStructures:
          cmp byte [pesi+9],1 ;is there a next pointer?
          jne endCombineStructures

          ;(R/E)AX => next mem pointer in structure (head)
          mov eax,dword [pesi+5] ;EAX = size
          add peax,pesi ;firstUnusedMemoryPointer + size
          add peax,0x10 ;add header

          cmp byte [peax+4],1 ;is memory used?
          je endCombineStructures

          ;Add size to previous header + 0x10 (for old header)
          mov edx,dword [peax+5] ;EDX = size
          add dword [pesi+5],edx
          add dword [pesi+5],0x10

          ;is there a next pointer?
          mov dl,byte [peax+9] ;next pointer?
          mov byte [pesi+9],dl

          ;Zero header
          mov pdword [peax],0
          mov pdword [peax+psize],0
          if used is32
            mov pdword [peax+psize*2],0
            mov pdword [peax+psize*3],0
          end if

        jmp CombineStructures


      endCombineStructures:

    nextFindFreeMemPointer:
      cmp byte [pesi+9],1
      jne endResizeMemoryPointers

      ;get next mempointer structure
      mov eax,dword [pesi+5]
      add pesi,peax
      add pesi,0x10
    jmp findFreeMemPointer

    endResizeMemoryPointers:
    call dleave
    ret psize
  end if

  ;------------------------------------------------------------------------------
  ;Recalculate the "BiggestFreePointer" in the pointer spots structure
  ;INPUT:
  ;  1:pointer spots structure
  ;------------------------------------------------------------------------------
  if used recalculateBiggestFreePointer
    recalculateBiggestFreePointer:
    call denter

    mov pesi,pdword [pebp+psize*2]
    xor ecx,ecx ;ECX = biggest size

    findBigestAvaidableSize:
      mov edi,dword [pesi+5]
      cmp ecx,dword [pesi+5]
      jae notHighestSize

      cmp byte [pesi+4],1 ;is memory used?
      je notHighestSize

        mov ecx,dword [pesi+5]

    notHighestSize:
      cmp byte [pesi+9],1
      jne endFindBigestAvaidableSize

      ;get next mempointer structure
      mov eax,dword [pesi+5]
      add pesi,peax
      add pesi,0x10
    jmp findBigestAvaidableSize

    endFindBigestAvaidableSize:
    ;set biggest size
      mov dword [pesi],ecx

    call dleave
    ret psize
  end if

  ;------------------------------------------------------------------------------
  ;Allocate the smallest possible size of memory and init with pointer spots structure
  ;------------------------------------------------------------------------------
  if used initiateNewMemPointerStructure
    initiateNewMemPointerStructure:
    call denter

      call GetSmallestPortionPageAllocation
      mov ecx,eax

      push pecx
      call GetMemLocationVirtual

      ;set Biggest free pointer
      sub ecx,0x10
      mov dword [peax],ecx

      ;set used mem
      mov byte [peax+4],0

      ;set size
      mov dword [peax+5],ecx

      ;set next pointer
      mov byte [peax+9],0


    call dleave
     ret
  end if
end if

;------------------------------------------------------------------------------
;Receive the smallest size that the VirtualAlloc function can allocate
;OUTPUT:
;    EAX: the smallest size VirtualAlloc can allocate
;------------------------------------------------------------------------------
if used GetSmallestPortionPageAllocation
  GetSmallestPortionPageAllocation:
  mov eax,dword [smallestPortionOfPageSize]
  ret
end if

;------------------------------------------------------------------------------
;This function check pointer 1 with rights 2 and jumps to 3 if invalid
;input:
;   1:Pointer to check
;   2:rights to check on
;   3:Pionter to jump to if pointer does not exists
;output:
;    INVALID_POINTER (4) or 1 (VALID_POINTER)
;------------------------------------------------------------------------------
if used saveCheckPointer
  saveCheckPointer:
  call denter
   mov pedx,invalidResultSaveCheckPointer
  mov pecx,pdword [pebp+psize*3]
  mov peax,pdword [pebp+psize*2]
  call checkPointer

  call dleave
  ret psize*3

  invalidResultSaveCheckPointer:
  call dleave

  add pesp,psize*4
  jmp pdword [pesp-(psize*1)]
end if

;------------------------------------------------------------------------------
;This function check pointer on eax/rax with rights ecx\rcx and jumps to edx\rdx if invalid
;input:
;   (R/E)AX: Pointer to check
;   (R/E)CX: Rights the memory should have
;   (R/E)DX: Pointer to jump to if the pointer of (R/E)AX does not exists
;corupted:
;       (R/E)DI
;output:
;    INVALID_POINTER (4) or 1 (VALID_POINTER)
;------------------------------------------------------------------------------
if used checkPointer
  checkPointer:
  pop pedi

  ;check pointer to ControlPanelOptions structure
  push pecx ;RIGHT_READ
  push peax ;pdword [pebp+(psize*3)]
  call doesVMLocationExists
  test eax,eax
  jnz continueCheckPointer

  mov eax,4;INVALID_POINTER
  jmp pedx

  continueCheckPointer:
  jmp pedi
end if

;------------------------------------------------------------------------------
;This function will make a new memory location with the required size 1:mem location
;------------------------------------------------------------------------------
if used GetMemLocation | used GetMemLocationVirtual
  GetMemLocation:
  if used New_Memory_management
    call denter

    ;round up size by 0x10 (for x64 systems very important
    mov eax,dword [pebp+psize*2]  ;requeste mem size
    test eax,eax
    jz quitGetMemLocation

    mov ecx,eax  ;to be rounded up

    and eax,0xF
    jz numberAlreadyRoundedUp

      sub ecx,eax
      add ecx,0x10
      mov dword [pebp+psize*2],ecx

    numberAlreadyRoundedUp:
    mov eax,[smallestPortionOfPageSize]
    sub eax,0x10
    cmp ecx,eax
    ja noDenter_GetMemLocation2

    if used MemoryMultiThreaded
      call getLocalMemoryStorageESI
    else
      mov pesi,pdword [AllocatedMemorys]
    end if
    ;get lock on AllocatedMemory
    if used MemoryMultiThreaded
      push pesi
    end if
    call getLockPointerStructure

    inc pesi
    lodsd
    mov ecx,eax ;total pointer

    lodsp ;get 'next'
    push peax ;save 'next'

      loopThroughPointers:
        lodsp

        push pecx ;save ECX

        test peax,peax
        jz nextloopThroughPointers

          ;check if bigest pointer reaches the requested size
          mov ecx,dword [pebp+psize*2] ;ECX = requested size

          cmp dword [peax],ecx
          jb nextloopThroughPointers ;not enough free space if the pointer

          push peax ;save top of mempointerstructure

          loopThroughMempointers:
            cmp byte [peax+4],1 ;check if memory already in use ?
            je nextloopThroughMempointers

            mov edx,dword [peax+5] ;EDX = size of this mempointer
            cmp edx,ecx ;check if size fits
            jb nextloopThroughMempointers

            mov byte [peax+4],1   ;used mem boolean = true
            sub edx,ecx
            cmp edx,0x20 ;Check if there is some leftover memory  (0x10 to alloc and 0x10 for the net header)
            jb notEnoughMemoryForNextMemPointer

              mov dword [peax+5],ecx ;set size
              ;watch out CL = part of ECX!!!
              mov cl, byte [peax+9]  ;get original boolean
              mov byte [peax+9],1    ;nextpointer boolean = true

              mov ebx,edx ;save size for nextpointer
              sub pebx,0x10 ;substract with new header strcuture size
              ;get next mem pointer
              mov edx,dword [pebp+psize*2] ;ECX = requested size
              add pedx,peax
              add pedx,0x10

              ;set size and original boolean for next mem pointer
              mov byte [pedx+9],cl    ;set nextpointer boolean

              mov dword [pedx+5],ebx ;set size
              mov byte [pedx+4],0    ;not used

            notEnoughMemoryForNextMemPointer:
              xchg peax,pdword [pesp] ;change pointer to return with top of mempointer

              push peax
              call recalculateBiggestFreePointer

              pop peax

              add peax,0x10
              add pesp,psize ;pop the 'next' pointer
              jmp endGetMemLocation ;found a memory location that fits the specifications :)

          nextloopThroughMempointers:
            cmp byte [peax+9],0
            je _nextloopThroughPointers

            ;get next pointer header
            mov edx,dword [peax+5]
            add peax,pedx
            add peax,0x10
          jmp loopThroughMempointers

      _nextloopThroughPointers:
        add pesp,psize ;pop top of mempointer

      nextloopThroughPointers:
      pop pecx ;restore ECX
      loop loopThroughPointers

    pop pecx ;get 'next'
    jpcxz noFreePointers

    mov pesi,pecx
    inc pesi
    lodsd
    mov ecx,eax ;total pointer

    lodsp ;get 'next'
    push peax ;save 'next'
    jmp loopThroughPointers

    noFreePointers:
    call initiateNewMemPointerStructure

    push 0 ;'next'
    push 1 ;ammount of pointers
    push peax ;save top of mempointerstructure

    push peax ;save top of mempointerstructure
    call setMemPointerStructureInPointerList

    pop peax
    push peax

    mov ecx,dword [pebp+psize*2] ;ECX = requested size
    jmp loopThroughMempointers

    endGetMemLocation:
    push peax

    if used MemoryMultiThreaded
      call getLocalMemoryStorageESI
      push pesi
    end if

    call releaseLockPointerStructure

    pop peax

    quitGetMemLocation:
      call dleave
      ret (psize*1)

  end if
  jmp GetMemLocationVirtual
end if

;------------------------------------------------------------------------------
;This function will free a memory location 1:mem location
;------------------------------------------------------------------------------
if used FreeMemLocation | used FreeMemLocationVirtual
  FreeMemLocation:
  if used New_Memory_management
    call denter

    mov pesi,pdword [pebp+psize*2] ;mem
    test pesi,pesi
    jz quitFreeMemLocation

    if used MemoryMultiThreaded
      push pesi
      call getBaseaddressPointerstructureAndBelongingLocalStorage
      mov pedi,pdword [pesp-psize] ;(R/E)DI => save Local storage
    else
      mov pedi,pdword [AllocatedMemorys]

      call getLockPointerStructure

      push pesi
      call findBaseAddressAndPointerStructure
    end if

    ;AX => pointer to BaseAddess
    ;DX => pointer to PointerStruct
    test peax,peax
    jz gotoFreeMemLocationVirtual

    mov pebx,peax ;(R/E)BX => pointer to BaseAddess

    push pedi ;save (R/E)DI

    ;(R/E)DI = pointer to memlocation
    mov pedi,pesi
    sub pesi,0x10
    mov ecx,dword [pesi+5] ;ECX = size

    ;since memory if rounded up by 0x10 we can use stosd and stosq ^-^
    if used is32
      shr ecx,2      ;divide by 4
    else
      shr ecx,3      ;divide by 8
    end if

    xor peax,peax
    rep stosp

    ;'Free' memory
    mov byte [pesi+4],0 ;unused

    pop pedi ;restore (R/E)DI

    push pebx
    call resizeMemoryPointers

    push pebx
    call recalculateBiggestFreePointer

    mov eax,[smallestPortionOfPageSize]
    sub eax,0x10
    cmp dword [pebx+5],eax
    jne endFreeMemLocation

      ;(R/E)DX => Pointer to PointerStruct
      ;(R/E)BX => BaseAddess
      ;on stack: 1: Local storage (will not be removed from stack)
      if used MemoryMultiThreaded
        push pedi ;Local storage (will not be removed from stack)
      end if
      call removeMemPointerFromPointerList

    endFreeMemLocation:
      if used MemoryMultiThreaded
        push pedi ;on stack: 1: Local storage
      end if
      call releaseLockPointerStructure

    quitFreeMemLocation:
      call dleave
      ret (psize*1)

    gotoFreeMemLocationVirtual:
    if ~ used MemoryMultiThreaded
      call releaseLockPointerStructure
    end if
    call dleave
  end if
  jmp FreeMemLocationVirtual
end if

;------------------------------------------------------------------------------
;This function will resize the memory if necessary, but will not restore the current data 1:original mem location 2:required space
;------------------------------------------------------------------------------
if used ReSizeMem
ReSizeMem:
        xor eax,eax
        inc eax
        xchg peax,pdword [pesp+psize*2] ;xchange required space with one (flag DONT_COPY_OLD_DATA)
        xchg peax,pdword [pesp+psize]   ;xchange original mem location with required space
        xchg peax,pdword [pesp]         ;xchange return address  with original mem location
        push peax                       ;push return address
        jmp ReAllocMemEx
end if

;------------------------------------------------------------------------------
;This function will resize the memory if necessary 1:original mem location 2:required space
;------------------------------------------------------------------------------
if used ReAllocMem | used ReAllocMemPage
ReAllocMemPage:
ReAllocMem:
        xor eax,eax
        xchg peax,pdword [pesp+psize*2] ;xchange required space with zero (flag COPY_DATA)
        xchg peax,pdword [pesp+psize]   ;xchange original mem location with required space
        xchg peax,pdword [pesp]         ;xchange return address  with original mem location
        push peax                       ;push return address
        jmp ReAllocMemEx
end if

;------------------------------------------------------------------------------
;This function will resize the memory without checking if necessary 1:original mem location 2:required space
;------------------------------------------------------------------------------
if used ReAllocMemNoCheck
ReAllocMemNoCheck:
        mov eax,2
        xchg peax,pdword [pesp+psize*2] ;xchange required space with zero (flag COPY_DATA ;| DONT_CHECK_ORIGINAL_SIZE)
        xchg peax,pdword [pesp+psize]   ;xchange original mem location with required space
        xchg peax,pdword [pesp]         ;xchange return address  with original mem location
        push peax                       ;push return address
        jmp ReAllocMemEx
end if

;------------------------------------------------------------------------------
;This function will resize the memory if necessary
;  1:original mem location
;  2:required space
;  3:flags
;Flags:
;COPY_DATA 0x0
;DONT_COPY_OLD_DATA 0x1
;DONT_CHECK_ORIGINAL_SIZE 0x2  => danarous (wrong size it will copy), this flag is disabled
;------------------------------------------------------------------------------
if used ReAllocMemEx
        ReAllocMemEx:
        call denter

        ;load original adress of (r/e)si and the required size on (r/e)dx
        mov pesi,pdword [pebp+psize*2]


      if used New_Memory_management
        if used MemoryMultiThreaded
          push pesi
          call getBaseaddressPointerstructureAndBelongingLocalStorage
          mov pedi,pdword [pesp-psize] ;(R/E)DI => save Local storage

          test peax,peax
          jz noPointerStructureFound
        ;----
        else
        ;----
          mov pedi,pdword [AllocatedMemorys]
          push pedi
          call getLockPointerStructure

          push pesi
          call findBaseAddressAndPointerStructure
        ;----
        end if
        ;----

        mov pedx,pdword [pebp+psize*3]
        push peax

        if used MemoryMultiThreaded
          push pedi
        ;----
        end if
        ;----
        call releaseLockPointerStructure

        pop peax

        noPointerStructureFound:
        test peax,peax
        jz memoryIsVirtualPointer

        mov ecx,dword [pesi-11] ;get from memPointer header original size
        jmp gotoSizeCheck

        memoryIsVirtualPointer:
      ;----
      end if
      ;----
        ;--DISABLEDtest pdword [pebp+psize*4],0x2 ;DONT_CHECK_ORIGINAL_SIZE
        ;--DISABLEDjnz skipSizeCheck

        ;Get the size of the current memory
        push 0

        push pesp
        push RIGHT_READ
        push pesi
        call doesVMLocationExistsEx

        pop pecx ;the original size
        mov pedx,pdword [pebp+psize*3]  ;the requested size

      gotoSizeCheck:
        ;check if there is already sufficient space, return
        cmp pecx,pedx
        cmovae peax,pesi
        jae endReAllocMem

     ;skipSizeCheck:
        push pedi ;save (R/E)DI
        ;get a new (bigger) memory location
        push pedx
        call GetMemLocation
        mov pedi,peax
        mov pebx,peax

        test dword [pebp+psize*4],1
     jnz dontCopyOldData

        ;copy the old data to the new pointer
        rep movsb

        dontCopyOldData:
        push peax

        ;release the original memory
        push pdword [pebp+psize*2]
        call FreeMemLocation

        pop peax

        endReAllocMem:
        call dleave
        ret psize*3
end if

;------------------------------------------------------------------------------
;This function will resize the destination to the size of the source memory and copy the data
;1:destination mem location
;2:source mem location
;------------------------------------------------------------------------------
if used cpyMemAndRealoc | used cpyMemAndRealocPage
        cpyMemAndRealocPage:
        cpyMemAndRealoc:
        call denter

        ;load original adress of (r/e)si and the required size on (r/e)dx

        mov pesi,pdword [pebp+psize*3] ;source
        mov pedi,pdword [pebp+psize*2] ;destination
        mov pebx,pedi                  ;EBX = destination

      if used New_Memory_management
        if used MemoryMultiThreaded
          push pesi
          call getBaseaddressPointerstructureAndBelongingLocalStorage

          test peax,peax
          jz dontReleasePointerStructure
            push pdword [pesp-psize] ;local storage
            call releaseLockPointerStructure
          dontReleasePointerStructure:
        else
          call getLockPointerStructure

          push pesi
          call findBaseAddressAndPointerStructure
        end if

        test peax,peax
        jz memoryDestinationIsVirtualPointer

          mov ecx,dword [pesi-11] ;get source size
          jmp gotSizeDestinationFromMemPointerHeader

        memoryDestinationIsVirtualPointer:

      end if
        ;Get the size of the source
        push 0

        push pesp
        push RIGHT_READ
        push pesi
        call doesVMLocationExistsEx

        pop pecx ;the original size
      if used New_Memory_management
        gotSizeDestinationFromMemPointerHeader:
        if used MemoryMultiThreaded
          push pedi
          call getBaseaddressPointerstructureAndBelongingLocalStorage

          test peax,peax
          jz dontReleasePointerStructure2
            push pdword [pesp-psize] ;local storage
            call releaseLockPointerStructure
          dontReleasePointerStructure2:
        else
            push pedi
            call findBaseAddressAndPointerStructure
        end if

        test peax,peax
        jz memorySourceIsVirtualPointer

        mov edx,dword [pedi-11] ;get destination size
        jmp gotSizeSourceFromMemPointerHeader

      memorySourceIsVirtualPointer:
      end if

        ;Get the size of destination
        push 0

        push pesp
        push RIGHT_READ
        push pedi
        call doesVMLocationExistsEx

        pop pedx ;the original size

      gotSizeSourceFromMemPointerHeader:
      if used New_Memory_management & ~ used MemoryMultiThreaded
          call releaseLockPointerStructure
      end if

        ;check if there is already sufficient space, return
        cmp pecx,pedx
        cmovbe peax,pesi
        jbe copyTheSourceToDestination

        ;resize the destination
        push pecx
        push pedi
        call ReSizeMem
        mov pebx,peax   ;EBX = destination
        mov pedi,peax   ;EDI = destination

        copyTheSourceToDestination:
        ;copy the old data to the new pointer
        rep movsb

        mov peax,pebx

        endcpyMemAndRealoc:
        call dleave
        ret psize*2
end if