
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;TODO:   EDIT 6/8/2014

;------------------------------------------------------------------------------
;This function will restore data of a protocol/listen or client session
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetPluginData_c:
    StartProc 6

    mov rcx,retGetPluginData_c
    xchg qword [rsp],rcx
    jmp GetPluginData_asm

    retGetPluginData_c:
    sub rsp,7*psize
    EndProc 6
    jmp rcx
else
    GetPluginData_c:
end if

;---------
;asm implementation
;---------

GetPluginData_asm:
call denter

internal_GetPluginData_asm_Call:
test dword [pebp+psize*8],GDPD_USED_PSESS
jz flag_GDPD_USED_PSESS_NotSet

  xor dword [pebp+psize*8],GDPD_USED_PSESS    ;delete flag

flag_GDPD_USED_PSESS_NotSet:

test dword [pebp+psize*8],GDPD_USED_LSESS
jz flag_GDPD_USED_LSESS_NotSet

  xor dword [pebp+psize*8],GDPD_USED_LSESS    ;delete flag

flag_GDPD_USED_LSESS_NotSet:

mov pecx,endGetPluginData
call checkProtocolSession
;move internal protocol structure to edi/rdi
mov pedi,peax

mov ebx,POS_DATABYNAME_INTERNALPROTOCOL      ;name is a string
mov edx,POS_DATABYINTEGER_INTERNALPROTOCOL   ;name is an integer

;on 32-bit push a zero value, on 64-bit clear R12
;This is used to store a pointer to a listensession (if ClientsSession failed to have the required value)
if used is32
   push 0
else
   xor r12d,r12d
end if

;-----if Client session
mov peax,pdword [pebp+psize*4]
test peax,peax
jz noSess

;get internal ClientSession
push pdword [pebp+psize*4] ;name (sess)
push [clientSessions]      ;collection
call GetIndexedDwordItem
test peax,peax
jz noSess


;Get the ListenSession this ClientSession belongs to and store it on stack or R12(x64)
if used is32
  mov pecx,pdword [peax+psize]
  mov dword [pesp],ecx
else
  mov r12,qword [rax+psize]
end if

mov ebx,POS_DATABYNAME_INTERNALCLIENTSESSION      ;name is a string
mov edx,POS_DATABYINTEGER_INTERNALCLIENTSESSION   ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx

;get the correct collection that belongs to this listensession
push pdword [pebp+psize*2]      ;name (psess)
push pdword [peax+pebx]         ;collection
call GetIndexedDwordItem
test peax,peax
jz endGetPluginData

mov pedi,peax                                ;put the collection on (R/E)DI
mov pecx,GetIndexedStringItemWithHash         ;name is a string
mov peax,GetIndexedItem                       ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz pecx,peax

jmp endIfLsess

;-----if listen session
noSess:
mov peax,pdword [pebp+psize*3]
test peax,peax
jz notLsess

;get internal ListenSession
mov pecx,endGetPluginData
call checkListenSession

or dword [pebp+psize*8],GDPD_USED_LSESS    ;ListenSession is used

mov ebx,POS_DATABYNAME_INTERNALLISTENSESSION      ;name is a string
mov edx,POS_DATABYINTEGER_INTERNALLISTENSESSION   ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx

;get the correct collection that belongs to this listensession
push pdword [pebp+psize*2]      ;name (psess)
push pdword [peax+pebx]         ;collection
call GetIndexedDwordItem
test peax,peax
jz endGetPluginData


mov pedi,peax                                ;put the collection on (R/E)DI
mov pecx,GetIndexedStringItemWithHash         ;name is a string
mov peax,GetIndexedItem                       ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz pecx,peax

jmp endIfLsess

notLsess:
;-----end if lsess
mov pecx,GetIndexedStringItemWithHash         ;name is a string
mov peax,GetIndexedItem                       ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx
cmovnz pecx,peax

or dword [pebp+psize*8],GDPD_USED_PSESS    ;ProtocolSession is used

mov pedi,pdword [pedi+pebx]

endIfLsess:
;Lock the list so no other threads can write on it
push pedi               ;collection
call LockListWriteAcces

push pdword [pebp+psize*5] ;name
push pedi                  ;collection
call pecx
test peax,peax
jnz NoListenSessionToTry

test dword [pebp+psize*8],GDPD_IS_LIST
jz isNotListFirstTry
  if used is32
    cmp dword [pesp],0
  else
    test r12,r12
  end if
  jne ListenIsOutOfInternalClientSession

  cmp pdword [pebp+psize*3],0
  je NoListenSessionToTry

ListenIsOutOfInternalClientSession:
  or dword [pebp+psize*8],GDPD_USED_LSESS    ;ListenSession is used

  if used is32
    mov peax,pdword [pesp]
    test peax,peax
  else
    test r12,r12
  end if
  jnz ClientSessionIsNotUsed

  xor dword [pebp+psize*8],GDPD_USED_LSESS    ;ListenSession is used
  or dword [pebp+psize*8],GDPD_USED_PSESS     ;ListenSession is used

ClientSessionIsNotUsed:
  push peax

   ;Unlock the list so other threads can use it
   push pedi
   call ExternalUnlockListWriteAcces
   xor pedi,pedi

   pop peax

  push pdword [pebp+psize*8] ;flags
  or pdword [pesp],GDPD_NO_COPY_LIST
  push pdword [pebp+psize*7] ;len
  push pdword [pebp+psize*6] ;data
  push pdword [pebp+psize*5] ;name
  push 0                     ;ClientSession
  if used is32
    push eax                  ;ListenSession
  else
    push r12                  ;ListenSession
  end if
  push pdword [pebp+psize*2] ;ProtocolSession
  call GetPluginData_asm

  test peax,peax
  jz NoListenSessionToTry

  push pedi ;save (r/e)di

  mov pedi,pdword [pebp+psize*6] ;destination
  mov pesi,pedi                ;source
  mov pecx,psize
  jmp sufficientSpace


isNotListFirstTry:
;if R12 of dword [pesp] is set, there is a listensession that might have the required value
if used is32
  cmp dword [pesp],0
  je NoListenSessionToTry

  pop dword [pebp+psize*3]      ;set ListenSession
  mov dword [pebp+psize*4],eax  ;ClientSession = 0
else
  test r12,r12
  jz NoListenSessionToTry

  mov qword [pebp+psize*3],r12  ;set ListenSession
  mov qword [pebp+psize*4],rax  ;ClientSession = 0
end if

;Unlock the list so other threads can use it
push pedi
call ExternalUnlockListWriteAcces
xor pedi,pedi
jmp internal_GetPluginData_asm_Call


NoListenSessionToTry:
push pedi ;save (r/e)di

test peax,peax
jz _getPluginDatagetDefaultValue

;get pointer to size
mov pedx,pdword [pebp+psize*7]

;read out the needed space
mov ecx,dword [peax-4]

mov pesi,peax  ;pointer with the data to be copied
mov pedi,pdword [pebp+psize*6] ;buf

;read out the pointer and compair
cmp dword [pedx],ecx
jae sufficientSpace

  ;set the required space
  mov dword [pedx],ecx
  xor eax,eax ;error
  jmp exitGetPluginData ;return

sufficientSpace:
test dword [pebp+psize*8],GDPD_IS_LIST
jz isNotList

  mov peax,pdword [pebp+psize*4]                ;(R/E)AX = ClientSession

  test peax,peax                                ;is ClientSession set ?
  jz ClientSessionIsNotSet

  test dword [pebp+psize*8],GDPD_USED_LSESS    ;is ProtocolSession or ListenSession used ?
  jnz doCopyInternalRowsCollection
  jz isNotList

  ClientSessionIsNotSet:
  mov peax,pdword [pebp+psize*3]                ;(R/E)AX = listensession

  test peax,peax                                ;is ListenSession set ?
  jz isNotList

  test dword [pebp+psize*8],GDPD_USED_PSESS
  jz isNotList

doCopyInternalRowsCollection:
   push peax  ;push ListenSession or ClientSession for addToGarbageFromListenSession call

   mov pedi,pdword [pesp+psize] ;restore list to unlock

   ;Unlock the list so we can lock write actions
   push pedi
   call ExternalUnlockListWriteAcces
   mov pdword [pesp+psize],0

   pop peax ;restore ListenSession or ClientSession for addToGarbageFromListenSession call

   ;quit if flag GDPD_NO_COPY_LIST is set
   test dword [pebp+psize*8],GDPD_NO_COPY_LIST
   jnz successExitGetPluginData

   push peax  ;push ListenSession or ClientSession for addToGarbageFromListenSession call

   ;-push FULL_LOCK
   ;-push pedi
   ;-call ExternalUnlockListAccesEx
   ;-xor pedi,pedi

  ;copy the row list
  push pdword [pesi]
  call copyInternalRowsCollection
  mov pesi,peax ;set copy as new source

  push peax
  mov peax,pesp ;*data

  push pdword [pebp+psize*8] ;flags
  push psize                 ;len
  push peax                  ;data
  push pdword [pebp+psize*5] ;name
  push pdword [pebp+psize*4] ;ClientSession
  push pdword [pebp+psize*3] ;ListenSession
  push pdword [pebp+psize*2] ;ProtocolSession
  call SavePluginData_asm

  add pesp,psize

  pop peax  ;pop ListenSession or ClientSession
  push pesi ;push new pointer
  push peax ;push ListenSession or ClientSession
  push TD_LIST ;type of the pointer
  call addToGarbageFromListenSession

  ;set pointer to data param
  mov peax,pdword [pebp+psize*6]
  mov pdword [peax],pesi

  jmp successExitGetPluginData

isNotList:
rep movsb ;copy bytes [esi] to [edi] with size ecx

successExitGetPluginData:
xor eax,eax
inc eax  ;return value = 1

exitGetPluginData:
pop pedi ;restore (r/e)di
push peax  ;save eax

;Unlock the list so other threads can use it
push pedi
call ExternalUnlockListWriteAcces

pop peax

endGetPluginData:
call dleave
ret 7*psize

;--
_getPluginDatagetDefaultValue:
mov pecx,endGetPluginData
call checkProtocolSession
;move internal protocol structure to edi/rdi
mov pedi,peax

;Get the first ControlPaneloption from linked list
mov pesi, pdword [pedi+psize*6]

mov pecx, pdword [pebp+psize*8] ;Get flags
and pecx,GDPD_NAME_IS_INTEGER  ;The only flag also used in the controlpanelOption list

mov ebx,CPO_PRIV_LISTENSESSION_OPTION or CPO_GEN_LISTENSESSION_OPTION

mov pedx,pdword [pebp+psize*3]
test pedx,pedx  ;check if listensession is set
jnz _getPluginDataLoopControlPanelOptions

mov pedx,pdword [pebp+psize*4]
test pedx,pedx  ;check if clientsession is set
jnz _getPluginDataLoopControlPanelOptions

mov ebx,CPO_PRIV_PROTOCOL_OPTION or CPO_GEN_PROTOCOL_OPTION

_getPluginDataLoopControlPanelOptions:
  xor eax,eax
  test pesi,pesi
  jz exitGetPluginData ;Also unlock list

  ;check if the correct flag is set *LISTENSESSION_OPTION or *_PLUGIN_OPTION
  test dword [pesi+psize+4],ebx
  jz _getPluginDataLoopnext

  ;Test if the flag GDPD_NAME_IS_INTEGER is set
  test dword [pesi+psize+4],GDPD_NAME_IS_INTEGER

  jecxz _getPluginDataCompairString ;ECX = 0 then flag GDPD_NAME_IS_INTEGER must NOT been set, ECX = 0 the flag MUST been set
  jz _getPluginDataLoopnext ;(result test) jump if the flag GDPD_NAME_IS_INTEGER is SET

  mov peax,pdword [pebp+psize*5]
  cmp pdword [pesi+psize+8],peax
  je _getPluginDataFoundControlPanelOption
  jmp _getPluginDataLoopnext

_getPluginDataCompairString:
  jnz _getPluginDataLoopnext ;(result test) jump if the flag GDPD_NAME_IS_INTEGER is NOT set

  push pdword [pebp+psize*5]
  push pdword [pesi+psize+8]
  call strcmpUtf8CaseInsensitive
  test eax,eax
  jz _getPluginDataFoundControlPanelOption

_getPluginDataLoopnext:
  ;get next ControlPaneloption
  mov pesi,pdword [pesi]
jmp _getPluginDataLoopControlPanelOptions

_getPluginDataFoundControlPanelOption:
;Get userValue
mov peax,pdword [pesi+psize+8+(psize*7)]
test peax,peax
jz exitGetPluginData ;Also unlock list

;get size member
mov ecx,dword [peax+4]
jecxz sizeIsNotSet

 ;default value
  xor pedi,pedi

  ;load buffer pointer or (large)integer
  mov pesi,pdword [peax+8]
  test pesi,pesi
  cmovz peax,pedi
  jz exitGetPluginData
 jmp buffIsChecked

sizeIsNotSet:
 lea pesi,pdword [peax+8]
 mov ecx,psize

buffIsChecked:
;dest
mov pedi,pdword [pebp+psize*6]

mov pedx,pdword [pebp+psize*7]
cmp dword [pedx],ecx
jae _getPluginDatasufficientSpace

  ;set the required space
  mov dword [pedx],ecx
  xor eax,eax ;error
  jmp exitGetPluginData ;return

_getPluginDatasufficientSpace:
rep movsb

xor eax,eax
inc eax
jmp exitGetPluginData



;------------------------------------------------------------------------------
;This function will save data of a protocol/listen or client session
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    SavePluginData_c:
    StartProc 6

    mov rcx,retSavePluginData_c
    xchg qword [rsp],rcx
    jmp SavePluginData_asm

    retSavePluginData_c:
    sub rsp,7*psize
    EndProc 6
    jmp rcx
else
    SavePluginData_c:
end if


;---------
;asm implementation
;---------
SavePluginData_asm:
call denter

mov pecx,endSavePluginData
call checkProtocolSession
;move internal protocol structure to edi/rdi
mov pedi,peax

;-----if ClientSession
mov pecx,pdword [pebp+psize*4]
jecxz SavePluginDataNotClientsess

;get internal ClientSession
push pdword [pebp+psize*4] ;name (sess)
push [clientSessions]      ;collectiom
call GetIndexedDwordItem
test peax,peax
jz endSavePluginData

mov ebx,POS_DATABYNAME_INTERNALLISTENSESSION      ;name is a string
mov edx,POS_DATABYINTEGER_INTERNALLISTENSESSION   ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx

;get the correct collection that belongs to this plugin
push pdword [pebp+psize*2]      ;name (psess)
push pdword [peax+pebx]      ;collection
call GetIndexedDwordItem
test peax,peax
jz endSavePluginData

mov pedi,peax                                ;put the collection on (R/E)DI
mov pecx,AddIndexedStringItemWithHash         ;name is a string
mov peax,AddIndexedItem                       ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz pecx,peax

xor edx,edx
xor esi,esi

jmp SavePluginDataEndIfLsess

SavePluginDataNotClientsess:
;-----end if ClientSession

;-----if ListenSession
;get internal ListenSession
;jmp SavePluginDataNotLsess
mov pecx,SavePluginDataNotLsess
call checkListenSession

;------------------------------------------------------------------------------
;-----For internal calls, without a  'checkProtocolSession'
;-----(R/E)AX = Internal ListenSession
;-----denter should have been called
;-----checkProtocolSession should be called and (R/E)AX should be set to (R/E)DI
;-----(R/E)DI => Internal ProtocolSession
;------------------------------------------------------------------------------
SavePluginData_internal_lsess_jump:
;EDIT 6/8/2014
;test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
;cmovnz ebx,edx

mov ebx,POS_DATABYNAME_INTERNALLISTENSESSION      ;name is a string
mov edx,POS_DATABYINTEGER_INTERNALLISTENSESSION   ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx

mov esi,dword [peax+(5*psize)+4] ;(R/E)SI = unique number of the listensession

;get the correct collection that belongs to this plugin
push pdword [pebp+psize*2]      ;name (psess)
push pdword [peax+pebx]      ;collectiom
call GetIndexedDwordItem
test peax,peax
jz endSavePluginData

mov pedx,pdword [pedi+psize*2] ;(R/E)DX = copy of psess
mov pedx,pdword [pedx+psize]   ;(R/E)DX = name of protocol

mov pedi,peax                                ;put the collection on (R/E)DI
mov pecx,AddIndexedStringItemWithHash         ;name is a string
mov peax,AddIndexedItem                       ;name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz pecx,peax

jmp SavePluginDataEndIfLsess
;-----end if ListenSession

;------------------------------------------------------------------------------
;-----For internal calls, without a  'checkProtocolSession'
;-----denter should have been called
;-----checkProtocolSession should be called and (R/E)AX should be set to (R/E)DI
;-----(R/E)DI => Internal ProtocolSession
;------------------------------------------------------------------------------
SavePluginDataNotLsess:
;-----if ProtocolSession
mov ebx,POS_DATABYNAME_INTERNALPROTOCOL    ;Poisition of the list by string in structure InternalProtocolStructure
mov pecx,AddIndexedStringItemWithHash       ;function if the name is an string

mov edx,POS_DATABYINTEGER_INTERNALPROTOCOL ;Poisition of the list by integer in structure InternalProtocolStructure
mov peax,AddIndexedItem                     ;function if the name is an integer

test dword [pebp+psize*8],GDPD_NAME_IS_INTEGER
cmovnz ebx,edx                             ;switch the position of the list if flag GDPD_NAME_IS_INTEGER is set
cmovnz pecx,peax                             ;switch the function to call if flag GDPD_NAME_IS_INTEGER is set

mov pedx,pdword [pedi+psize*2] ;(R/E)DX = copy of psess
mov pedx,pdword [pedx+psize]   ;(R/E)DX = name of protocol
xor esi,esi                    ;ESI = 0
mov pedi,pdword [pedi+pebx]
;-----end if ProtocolSession

SavePluginDataEndIfLsess:
test dword [pebp+psize*8],GDPD_RESET_DATA
jnz removeData

;Make a call to AddIndexedStringItemWithHash OR AddIndexedItem
push pdword [pebp+psize*7] ;len
push pdword [pebp+psize*6] ;data
push pdword [pebp+psize*5] ;name
push pedi                  ;collection
call pecx

test pedx,pedx
jz endSavePluginData ;Do not safe data from a ClientSession on disk (only ListenSession and ProtocolSession

test dword [pebp+psize*8],GDPD_MEMORY_ONLY
jnz endSavePluginData

push pdword [pebp+psize*7] ;len
push pdword [pebp+psize*6] ;data
push pdword [pebp+psize*5] ;name
push pdword [pebp+psize*8] ;flags
push pesi                  ;Unique Id of the Listensession
push pedx                  ;Protocol name
call SavePluginDataToDisk
jmp endSavePluginData

removeData:
mov peax,RemoveIndexedStringItemWithHash
cmp pecx,AddIndexedStringItemWithHash
cmove pecx,peax

mov peax,RemoveByIndex
cmp pecx,AddIndexedItem
cmove pecx,peax

;remove the item from the list
push pdword [pebp+psize*5] ;name
push pedi                  ;collection
call pecx

test dword [pebp+psize*8],GDPD_MEMORY_ONLY
jnz endSavePluginData

;remove the item from disk
push pdword [pebp+psize*5] ;name
push pdword [pebp+psize*8] ;flags
push pesi                  ;Unique Id of the Listensession
push pedx                  ;Protocol name
call RemovePluginDataFromDisk

endSavePluginData:
call dleave
ret 7*psize

;------------------------------------------------------------------------------
;The CountListRows function will return the amount of rows the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    CountListRows_c:
    StartProc 1

    mov rcx,retCountListRows_c
    xchg qword [rsp],rcx
    jmp CountListRows_asm

    retCountListRows_c:
    sub rsp,1*psize
    EndProc 1
    jmp rcx
else
    CountListRows_c:
end if

;---------
;asm implementation
;---------

CountListRows_asm:
;get pointer to MultiColumnList
push pdword [pesp+psize*1] ;list
call CountItems
sub peax,2 ;the first two items are used for pointer MultiColumnList and ammount of items per column
ret psize*1

;------------------------------------------------------------------------------
;The AddRow function will add one or more new rows to the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    AddListRow_c:
    StartProc 2

    mov rcx,retAddListRow_c
    xchg qword [rsp],rcx
    jmp AddListRow_asm

    retAddListRow_c:
    sub rsp,2*psize
    EndProc 2
    jmp rcx
else
    AddListRow_c:
end if

;---------
;asm implementation
;---------

AddListRow_asm:
call denter

mov pecx,pdword [pebp+psize*2] ;list

;lock the list for other threads
push FULL_LOCK
push pecx
call LockListAcces

;get pointer to MultiColumnList
push 1
push pecx
call GetDwordItem

;copy the new rows to the list
push 0                      ;flags
push pecx                   ;pointer to list
push pdword [pebp+psize*3]  ;pointer to rows to add
push peax                   ;pointer to MultiColumnList
call copyRowsa

push peax

;lock the list
push FULL_LOCK
push pecx
call ExternalUnlockListAccesEx

pop peax

call dleave
ret psize*2

;------------------------------------------------------------------------------
;The GetListItem function will restore data saved in the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetListItem_c:
    StartProc 5

    mov rcx,retGetListItem_c
    xchg qword [rsp],rcx
    jmp GetListItem_asm

    retGetListItem_c:
    sub rsp,5*psize
    EndProc 5
    jmp rcx
else
    GetListItem_c:
end if

;---------
;asm implementation
;---------

GetListItem_asm:
call denter

mov pebx,pdword [pebp+psize*2] ;list
mov pedx,pdword [pebp+psize*3] ;rowId
add edx,2                      ;the first Id is taken bij the MultiColumnList and second by the ammount of items per row
mov pecx,pdword [pebp+psize*4] ;itemId
xor edi,edi                    ;EDI = 0

;lock the list for write actions
push pebx
call LockListWriteAcces

;get ammount of items per row
push 2
push pebx
call GetDwordItem
test peax,peax
jz endGetListItem

;check if the requested itemId exists
cmp peax,pecx
  cmovb peax,pedi   ;eax = 0
jb endGetListItem

;get the row and check on existance
push pedx
push pebx
call GetDwordItem
test peax,peax
jz endGetListItem

;calculate the pointer of the item in the list  (every item has two pointers, one with the size and one is the item)
if used is64
  shl ecx,4 ;multiply by 16
else
  shl ecx,3 ;multiply by 8
end if

mov pesi,pdword [peax+pecx+psize] ;get the pointer to item
mov pecx,pdword [peax+pecx]   ;get the size of the item

mov peax,pdword [pebp+psize*6] ;&len
cmp dword [peax],ecx
jae enoughSpaceForListItem

  mov dword [peax],ecx
  xor eax,eax
  jmp endGetListItem

enoughSpaceForListItem:
mov pedi,pdword [pebp+psize*5] ;buf

test pecx,pecx
jz copyDQWORD

;copy the item
rep movsb
jmp endGetListItem

copyDQWORD:
mov pdword [pedi],pesi

endGetListItem:
push peax

push pebx
call ExternalUnlockListWriteAcces

pop peax

call dleave
ret psize*5

;------------------------------------------------------------------------------
;The FindListItem function will search for data in a certain column and return the row identifier if found.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    FindListItem_c:
    StartProc 6

    mov rcx,retFindListItem_c
    xchg qword [rsp],rcx
    jmp FindListItem_asm

    retFindListItem_c:
    sub rsp,6*psize
    EndProc 6
    jmp rcx
else
    FindListItem_c:
end if

;---------
;asm implementation
;---------
if used FindListItem_asm

FindListItem_asm:
call denter

mov pebx,pdword [pebp+psize*2] ;list

;lock the list for write actions
push pebx
call LockListWriteAcces 

;get pointer to MultiColumnList
push 1
push pebx
call GetDwordItem
mov pesi,peax ;(R/E)SI => pointer to MultiColumnList

lodsd ;EAX => size member MultiColumnList
add pesi,4 ;skip flags
sub eax,SIZEOF_MULTICOLUNMLIST

;Do a divide by the size of COLUMN strcuture to get the ammount of COLUMN structures
if used is32
  shr eax,4 ;divide by 16 (the size of the Column structure)
else
  mov ecx,0x18
  xor edx,edx
  div ecx ;divide by 0x18 (the size of the Column structure)
end if

mov edx,dword [pebp+psize*3] ;EDX => columnId
xor ecx,ecx ;Zero ECX
cmp eax,edx ;compaire columnId to ammount of columns
 cmovb eax,ecx ;Zero return value if columnId>ammount of columns
jb endFindListItem

;;Do a multiply by the size of COLUMN strcuture to get the position of the COLUMN structure in the MultiColumnList structure
if used is32
  shl edx,4   ;multiply by 0x10 (the size of the Column structure)
else
  mov rax,rdx
  shl rax,4   ;multiply by 0x10
  lea rdx,qword [rdx*8+rax] ;multiply by 0x8 (result of *0x10 + this result  =>  *0x18) (the size of the Column structure)
end if

mov edi,dword [pesi+pedx+psize] ;EDI => type of selected column
mov pesi,pdword [pebp+psize*4] ;(R/E)SI => item
mov pecx,pesi                  ;(R/E)CX => item

mov edx,dword [pebp+psize*3] ;EDX => columnId
;;Do a multiply by the size of two pointers (one is the size, the other is the item for later use
if used is32
  shl edx,3 ;multiply by 0x8 (the size of two pointers)
else
  shl rdx,4 ;multiply by 0x10 (the size of two pointers)
end if


;ASCII String
mov peax,compairAsciiString
test edi,edi
jz typeSelected

;Unicode string
sub edi,2
mov peax,compairUnicodeString
jb typeSelected

;Boolean
mov ecx,dword [pesi] ;get dword as boolean
mov peax,compairBoolean
jz typeSelected

;char
mov cl,byte [pesi] ;get byte as item
sub edi,2
mov peax,compairChar
jb typeSelected

;short int
mov cx,word [pesi] ;get word as item
mov peax,compairShortInt
jz typeSelected

;integer
mov ecx,dword [pesi] ;get dword as item
sub edi,2
mov peax,compairInt
jb typeSelected

;large integer
if used is32
  mov ecx,esi ;get pointer as item
else
  mov rcx,qword [rsi] ;get qword as item
end if
mov peax,compairQword
jz typeSelected

;date & time
mov ecx,dword [pesi] ;get dword as item
sub edi,3
mov peax,compairInt
jb typeSelected

;dateTime
if used is32
  mov ecx,esi ;get pointer as item
else
  mov rcx,qword [rsi] ;get qword as item
end if
mov peax,compairQword
jz typeSelected

;progress
mov ecx,dword [pesi] ;get dword as item
dec edi
mov peax,compairInt
jz typeSelected


  jmp endFindListItem

;Register status
;(R/E)CX => pointer to compaire function
;(R/E)BX => pointer to list

typeSelected:
mov pesi,peax ;(R/E)SI => pointer to compair function

;get ammount of rows
push pebx
call CountItems
xchg pecx,peax ;ECX => ammount of rows  (R/E)AX => item to compair

;the first two items are used for pointer MultiColumnList and ammount of items per column
cmp pecx,2
jz endFindListItem

mov pedi,pdword [pebp+psize*7] ;startFromRow
test pedi,pedi ;rowId starts at 1, not zero
jz dontUseStartFromRow
inc pedi

;check if ammount of rows if lower then the row where to start searching
cmp pecx,pedi
jb dontUseStartFromRow

  mov pecx,pedi ;ECX => startFromRow

dontUseStartFromRow:

if used is64
  mov r12,rax ;R12 => item
  mov r13d,dword [rbp+psize*6] ;R13d => flags
end if

inc pecx ;increment the rowId, sinc it will be lowerd in the loop

jmp loopThrougRowsStart

loopThrougRows:
  mov edi,2

  if used is32
    test dword [pebp+psize*6],FLI_COMPARE_ONLY_ONE_ROW
  else
    test r13d,FLI_COMPARE_ONLY_ONE_ROW
  end if
  cmovnz ecx,edi ;if(flags&FLI_COMPARE_ONLY_ONE_ROW) { rowId = 2 }
  jnz endFindListItem


loopThrougRowsStart:
  if used is32
    push eax ;save item
  end if

  dec pecx ;lower the id
  cmp pecx,2
  je endFindListItem

  ;get row
  push pecx
  push pebx
  call GetDwordItem  

  lea pedi,pdword [peax+pedx+psize] ;(R/E)DI => the pointer to item
  ;--mov peax,pdword [peax+pedx]   ;get the size of the item

  if used is32
    pop eax ;restore item
  else
    mov rax,r12 ;RAX => item
  end if

  ppush compaired  ;return address
  jmp pesi

compairBoolean:
  test eax,eax ;check on zero compairand

  if used is32
    push peax ;save item
  end if

  mov eax,dword [pedi]

  jz compairandZero

  ;check if [edi] => not zero
  test edi,edi ;cehck on zero
  ;complement zero flag
  sete al
  test al,al

  if used is32
    pop peax ;restore item
  end if
  ret

compairandZero:
  ;check if [edi] => zero to
  test eax,eax

  if used is32
    pop peax ;restore item
  end if
  ret

compairChar:
  scasb
  ret

compairShortInt:
  scasw
  ret

compairInt:
  scasd
  ret

compairQword:
  if used is64
    scasq
    ret
  else
    if used is32
      push eax ;save item
    end if

    push 8
    push pdword [pedi]
    push peax
    call comapairMem

    if used is32
      pop eax ;restore item
    end if
    ret
  end if

compairAsciiString:
compairUnicodeString:
    if used is32
      push peax  ;save item
    end if

    push pdword [pedi]         ;String 2
    push peax                  ;String 1
    ppush endCompairAsciiString ;return address
    ;test if high/low case matters
    if used is32
      test dword [pebp+psize*6],FLI_CASE_SENSITIVE
    else
      test r13d,FLI_CASE_SENSITIVE
    end if
    jz strcmpUtf8CaseInsensitive ;check case insensitive
    jmp strcmpCaseSensitive ;check case sensitive

    endCompairAsciiString:
    if used is32
      pop peax  ;restore item
    end if
    ret

compaired:
jnz loopThrougRows

endFindListItem:
sub pecx,2 ;substract the MultiColumnList structure and ammount of items per column

;Unlock the list
push pebx
call ExternalUnlockListWriteAcces

mov peax,pecx

call dleave
ret 6*psize
end if


;------------------------------------------------------------------------------
;The RemoveListRow function will remove the specific row from the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    RemoveListRow_c:
    StartProc 2

    mov rcx,retRemoveListRow_c
    xchg qword [rsp],rcx
    jmp RemoveListRow_asm

    retRemoveListRow_c:
    sub rsp,2*psize
    EndProc 2
    jmp rcx
else
    RemoveListRow_c:
end if

;---------
;asm implementation
;---------

RemoveListRow_asm:
call denter

mov pebx,pdword [pebp+psize*2] ;list
mov pedx,pdword [pebp+psize*3] ;rowId
add edx,2                      ;the first Id is taken bij the MultiColumnList and second by the ammount of items per row

;lock the list for other threads
push FULL_LOCK
push pebx
call LockListAcces

;get ammount of items per row
push 2
push pebx
call GetDwordItem
test peax,peax
jz endRemoveListRow

if used is32
  mov edi,eax  ;EDI = ammount of items per row
else
  mov r12,peax ;r12 = ammount of items per row
end if

;remove the row
push pedx
push pebx
call RemoveDwordItem
test peax,peax
jz endRemoveListRow
mov pesi,peax   ;(R/E)SI = pointer to row to be deleted

;unlock the list
push FULL_LOCK
push pebx
call ExternalUnlockListAccesEx

mov pebx,pesi  ;(R/E)BX = save pointer to row to be deleted

call loopInternalRowCollectionItems

;delete pointer to row
push pebx
call FreeMemLocation
xor pebx,pebx ;zero (R/E)BX

xor pebx,pebx

endRemoveListRow:
test pebx,pebx
jnz quitRemoveListRow

push peax

  ;unlock the list
  push FULL_LOCK
  push pebx
  call ExternalUnlockListAccesEx

pop peax

quitRemoveListRow:
call dleave
ret psize*2

;------------------------------------------------------------------------------
;The GetListItemSize function will return the size, in bytes, of an item in the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    GetListItemSize_c:
    StartProc 3

    mov rcx,retGetListItemSize_c
    xchg qword [rsp],rcx
    jmp GetListItemSize_asm

    retGetListItemSize_c:
    sub rsp,3*psize
    EndProc 3
    jmp rcx
else
    GetListItemSize_c:
end if

;---------
;asm implementation
;---------

GetListItemSize_asm:
call denter

mov pebx,pdword [pebp+psize*2] ;list
mov pedx,pdword [pebp+psize*3] ;rowId
add edx,2                      ;the first Id is taken bij the MultiColumnList and second by the ammount of items per row
mov pecx,pdword [pebp+psize*4] ;itemId
xor edi,edi                    ;EDI = 0

;lock the list for write actions
push pebx
call LockListWriteAcces

;get ammount of items per row
push 2
push pebx
call GetDwordItem
test peax,peax
jz endGetListItemSize

;check if the requested itemId exists
cmp peax,pecx
  cmovb peax,pedi   ;eax = 0
jb endGetListItemSize

;get the row and check on existance
push pedx
push pebx
call GetDwordItem
test peax,peax
jz endGetListItemSize

;calculate the pointer of the item in the list  (every item has two pointers, one with the size and one is the item)
if used is64
  shl ecx,4 ;multiply by 16
else
  shl ecx,3 ;multiply by 8
end if

mov peax,pdword [peax+pecx]   ;get the size of the item

endGetListItemSize:
push peax

push pebx
call ExternalUnlockListWriteAcces

pop peax

call dleave
ret psize*3

;------------------------------------------------------------------------------
;The SetListItem function will replace data saved in the list.
;------------------------------------------------------------------------------
;---------
;C implementation
;---------
if used is64
    SetListItem_c:
    StartProc 5

    mov rcx,retSetListItem_c
    xchg qword [rsp],rcx
    jmp SetListItem_asm

    retSetListItem_c:
    sub rsp,5*psize
    EndProc 5
    jmp rcx
else
    SetListItem_c:
end if

;---------
;asm implementation
;---------

SetListItem_asm:
call denter

mov pebx,pdword [pebp+psize*2] ;list
mov pedx,pdword [pebp+psize*3] ;rowId
add edx,2                      ;the first Id is taken bij the MultiColumnList and second by the ammount of items per row
mov pecx,pdword [pebp+psize*4] ;itemId
xor edi,edi                    ;EDI = 0

;lock the list for other threads
push FULL_LOCK
push pebx
call LockListAcces

;get ammount of items per row
push 2
push pebx
call GetDwordItem
test peax,peax
jz endSetListItem

;check if the requested itemId exists
cmp peax,pecx
  cmovb peax,pedi   ;eax = 0
jb endSetListItem

;get the row and check on existance
push pedx
push pebx
call GetDwordItem
test peax,peax
jz endSetListItem
mov pedi,peax ;(R/E)DI = pointer to row belongs to

;calculate the pointer of the item in the list  (every item has two pointers, one with the size and one is the item)
if used is64
  shl ecx,4 ;multiply by 16
else
  shl ecx,3 ;multiply by 8
end if

push pecx   ;save location item

mov pesi,pdword [peax+pecx+psize] ;get the pointer to item
mov pecx,pdword [peax+pecx]   ;get the size of the item

jecxz doNotRealePointer

push pesi
call FreeMemLocation

doNotRealePointer:
pop pecx ;restore location item

mov pedx,pdword [pebp+psize*6] ;len

test edx,edx
jz copyDQWORDSetListItem

;create new memory location to copy the new item to
push pedx
call GetMemLocation

mov pdword [pedi+pecx+psize],peax ;set the pointer to item
mov pdword [pedi+pecx],pedx   ;set the size of the item

mov pedi,peax ;(R/E)SI = the pointer to the new item
mov pecx,pedx
mov pesi,pdword [pebp+psize*5] ;buf

;copy the item
rep movsb
jmp endSetListItem

copyDQWORDSetListItem:
mov peax,pdword [pebp+psize*5] ;buf
mov pdword [pedi+pecx+psize],peax ;set the pointer to item
mov pdword [pedi+pecx],0      ;set the size of the item

endSetListItem:
push peax

;unlock the list
push FULL_LOCK
push pebx
call ExternalUnlockListAccesEx

pop peax

call dleave
ret psize*5