
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;Uncomment the line below if you want the log messages in this DLL to be compiled with the file. Thses messages will be logged if switch -l is used.
CompileLogMessage

PLUGIN_VERSION equ ProductVersionPrefix, ".", FileVersion

;comment the below 4 lines if you want to disable Chuncking (HTTP protocol)
virtual at 0
call EnableChunking
EnableChunking:
end virtual

CPO_NAME_PATH equ 01         ;the name of the path configured in the ControlPanelOptions
CPO_NAME_REDIRECTIONS equ 02 ;the name of the redirections configured in the ControlPanelOptions
CPO_NAME_DIRLISTEN equ 03    ;the name of the directory listening configured in the ControlPanelOptions
CPO_NAME_MAX_SIZE equ 04
CPO_NAME_MAX_TIMEOUT equ 05
CPO_NAME_MAX_SIZE_SEND equ 06
CPO_NAME_CGI equ 07
CPO_NAME_MAX_CGI_RUNTIME equ 08
CPO_NAME_KEEPALIVE equ 09
CPO_NAME_INDEX equ 10

CPO_NAME_AUTHOR equ -1
CPO_NAME_VERSION equ -2

COLUMN_GENERAL_DOMAIN equ 0
COLUMN_GENERAL_ITEM equ 1

COLUMN_REDIR_DOMAIN equ COLUMN_GENERAL_DOMAIN
COLUMN_REDIR_PATH equ COLUMN_GENERAL_ITEM
COLUMN_REDIR_TO equ 2

COLUMN_CGI_DOMAIN equ COLUMN_GENERAL_DOMAIN
COLUMN_CGI_EXTENSION equ COLUMN_GENERAL_ITEM
COLUMN_CGI_APP equ 2


include 'include/function.inc'
include 'include/flags.inc'
include 'include/timedate.inc'       ;HTTP protocol
include 'include/math.inc'
include 'include_plugin_general/short_functions.inc'
include 'include_tvs_server_plugin_http/plugin_handle_connection.inc'
include 'include_tvs_server_plugin_http/directory_listening.inc'
include 'include/stringFunctions.inc'  

EXTRASIZE_LISTEN_AND_CLIENT_STRUCT equ psize

sizeSubportedProtocols equ endSubportedProtocols - subportedProtocols

include 'include_tvs_server_plugin_http/plugin_bind.inc'
include 'include_plugin_general/list_functions.inc'
include 'include/collection.inc'
include 'include/commandline.inc'
virtual at 0
call initHTTPProtocol
initHTTPProtocol:
end virtual
include 'include_plugin_general/__asm_exchange_tables_GetProtocolNames.inc'

if used InitGetFuncPointers
  InitGetFuncPointers db 'InitGetFuncPointers',0
end if

cpyInternalProtocolstructure pointer 0
cpyInternalListenSessionstructureCollection pointer 0

generatedMenuProtocols pointer 0
generatedMenuListenSession pointer 0

;flags
;   WRITING 0x1
;reading
;    counter of how many reading operations there are currently. The flag WRITING may not be set if this is not zero
align 0x10 ;this is used by instructure cmpxchg8b and this instruction requiers alignment
cpyInternalProtocolstructureSignal:
reading dd 0
flags dd 0
mymod pointer ?

main:
_init:
if used isWin
StartProc 3
end if
call denter

  if used isWin
    cmp dword [pebp+psize*3],1 ;DLL_PROCESS_ATTACH
    jne dontInitDLL
  end if

    call initGetSmallestPortionPageAllocation

    if used is32 & used isWin
      call doSystemCheckDLL
    end if

    if used isWin
      mov peax,pdword [pebp+psize*2] ;get HMODULE
      mov [mymod],peax
    else
      pmov [mymod],_init
    end if


    call CreateCollection
    mov [cpyInternalListenSessionstructureCollection],peax

    ;receive the size of one page
    call GetSmallestPortionPageAllocation
    mov pdword [defaultSendBuffer],peax

    call pushTVS_HTTP_PROTOCOL

      if used isWin
        if used is32
          db 'tvs_server_http_protocol_x86.dll',0
        else
          db 'tvs_server_http_protocol_x64.dll',0
        end if
      else
        if used is32
          db './tvs_server_http_protocol_lin_x86.so',0
        else
          db './tvs_server_http_protocol_lin_x64.so',0
        end if
      end if

    pushTVS_HTTP_PROTOCOL:
    call loadmodule ;load HTTP protocol
    test peax,peax
    jz endMain
    mov pedx,peax

    if used isWin
      xor ecx,ecx
      ;get pointers to functions
      mov esi,dword [pedx+0x3C] ;get offset of PE header
      cmp word [pedx+pesi+0x4+0x10],0x18  ;if SizeOfOptionalHeader<0x18 then BaseOfCode is not defined.
      cmovb eax,ecx
      jb endMain

      mov esi,dword [pedx+pesi+0x4+0x14+0x14] ;[Base + offset PE header + signature + fileheader + offset of BaseOfCode in optional header] = BaseOfCode
      test esi,esi
      cmovz eax,ecx
      jz endMain

      add pesi,pedx  ;pointer to first section with code
    else
      ppush InitGetFuncPointers
      push pedx
      call getfunc
      test peax,peax
      jz endMain
      mov pesi,peax
    end if

    ;check version dword
    lodsd ;get version
    cmp ax,1
    jz copyHTTPFunctions

    test ax,ax
    cmovnz eax,ecx
    jnz endMain

    copyHTTPFunctions:
    mov ecx,(endImportListHttp-beginImportListHttp)/psize

    shr eax,16
    test ax,ax
    jnz copyAll

      sub ecx,psize ; - function EscapeValue, which is not available

    copyAll:
    pmov  pedi,beginImportListHttp
    rep movsp ;copy the pointers

dontInitDLL:
  xor eax,eax
  inc eax

endMain:
call dleave
if used isWin
  EndProc 3
end if
ret (FunctionReturnFactor*3)

;------------------------------------------------------------------------------
;The StartProtocol function need to initiate the protocol named in the ProtocolSession structure.
;------------------------------------------------------------------------------
StartProtocol:
StartProc 1
call denter

;Get the protocolName member from ProtocolSession
mov peax,pdword [pebp+(psize*2)]
mov peax,pdword[peax+(psize*1)]

;Do we init http protocol
ppush http_name
push peax
call strcmp
test eax,eax
jne endStartProtocol

push V_TV_SERVER_102 ;version
push EXTRASIZE_LISTEN_AND_CLIENT_STRUCT ;extraSize
ppush HTTP_psign ;pSign
push PT_LABORER or PT_SUPPORT_PSIGN   ;type
push pdword [pebp+(psize*2)] ;psess
call [SetProtocolInformation]
test peax,peax
jz endStartProtocol

ppush httpOptions             ;cpo
push pdword [pebp+(psize*2)] ;psess
call [SetControlPanelOptions]

push pdword [pebp+(psize*2)] ;psess
call [AcceptProtocol]

;set protocolsession
mov pecx,pdword [pebp+(psize*2)] ;psess
mov [psess_http],pecx


;return
endStartProtocol:
call dleave
EndProc 1
ret (FunctionReturnFactor*1)