
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

S_IRWXU equ 0x1C0 ;user (file owner) has read, write, and execute permission
S_IRUSR equ 0x100 ;user has read permission
S_IWUSR equ 0x80  ;user has write permission
S_IXUSR equ 0x40  ;user has execute permission

S_IRWXG equ 0x38 ;group has read, write, and execute permission
S_IRGRP equ 0x20  ;group has read permission
S_IWGRP equ 0x10  ;group has write permission
S_IXGRP equ 0x8   ;group has execute permission

S_IRWXO equ 0x7   ;others have read, write, and execute permission
S_IROTH equ 0x4   ;others have read permission
S_IWOTH equ 0x2   ;others have write permission
S_IXOTH equ 0x1   ;others have execute permission

S_ISUID equ 0x800   ;set-user-ID bit
S_ISGID equ 0x400   ;set-group-ID bit (see inode(7)).
S_ISVTX equ 0x200   ;sticky bit (see inode(7)).

DT_UNKNOWN equ 0
DT_FIFO equ 1
DT_CHR equ 2
DT_DIR equ 4
DT_BLK equ 6
DT_REG equ 8
DT_LNK equ 10
DT_SOCK equ 12
DT_WHT equ 14

FIONREAD equ 21531

FILE_ATTRIBUTE_DIRECTORY equ 0x10

S_IFMT equ 0xF000

S_IFSOCK equ 0xC000
S_IFLNK equ 0xA000
S_IFREG equ 0x8000
S_IFBLK equ 0x6000
S_IFDIR equ 0x4000
S_IFCHR equ 0x2000
S_IFIFO equ 0x1000


;------------------------------------------------------------------------------
;This function will check if there was an error running system function
;if not, it will return to the pointer on the stack
;Else it will save error number on errno and set EAX/RAX to -1
;input:
;    1:pointer with filename
setErrNo:
if used is32
  test eax,0x80000000
else
  push pecx
  mov pecx,0x8000000000000000
  test rax,pecx
  pop pecx
end if
jz noError

  neg peax

  ;--;EAX = -1
  ;--mov peax,-1
  ret psize

noError:
add pesp,psize
ret

O_RDONLY equ 0x0
O_WRONLY equ 0x1
O_RDWR equ 0x2
O_TRUNC equ 0x200
O_LARGEFILE equ 0x8000
O_CREAT equ 0x40
O_EXCL equ 0x400
O_DIRECTORY equ 0x10000
O_PATH equ 0x200000


;------------------------------------------------------------------------------
;This function will open a file
;input:
;    1:pointer with filename
;    2:flags:
;       0x0     READ
;       0x1     WRITE&READ
;       0x2     CREATE_FILE
;       0x4     SHARE_READ_WHILE_WRITE
;       0x400   OPEN_DIRECTORY_ONLY
;       0x10000 filename is UTF-8
;       0x20000 don't make long path
;       0x40000 truncate file

;output:
;    (R/E)AX=> filehandle or -1 on failure
;------------------------------------------------------------------------------
if used openFileHandle
openFileHandle:
call denter

dontConvertString:

if used is64
  mov ecx,O_RDONLY  ;O_RDONLY
else
  mov ecx,O_RDONLY or O_LARGEFILE  ;O_RDONLY
end if
test pdword [pebp+psize*3],0x1 ;WRITE&READ
jz noWriteRightsRequired

  or ecx,O_RDWR ;O_RDWR
  ;test pdword [pebp+psize*3],0x4 ;SHARE_READ_WHILE_WRITE
  ;jnz noWriteRightsRequired

  ;xor edi,edi ;dwShareMode

noWriteRightsRequired:
;or edx,OPEN_EXISTING ;ESI => parameter CreationDisposition
test pdword [pebp+psize*3],0x2 ;CREATE_FILE
jz noFileCreate

  or ecx,O_CREAT

noFileCreate:
test pdword [pebp+psize*3],0x400 ;OPEN_DIRECTORY_ONLY
jz noOpenDirectoryOnly

  or ecx,O_DIRECTORY ;or O_PATH

noOpenDirectoryOnly:
test pdword [pebp+psize*3],0x40000
jz noTruncate

  or ecx,O_TRUNC

noTruncate:
argpushsys 3,S_IRUSR or S_IWUSR or S_IRGRP or S_IWGRP or S_IROTH ;mode
argpushsys 2,pecx                  ;flags
argpushsys 1,pdword [pebp+psize*2] ;filename
linuxCall 2,5  ;open

;;argpushsys 2,S_IRUSR or S_IWUSR or S_IRGRP or S_IWGRP or S_IROTH ;mode
;argpushsys 1,pdword [pebp+psize*2] ;filename
;linuxCall 85,8  ;create

ppush usedACreateFile
call setErrNo

mov peax,-1

usedACreateFile:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will return the x64 value of the stat64 in the pos set in the parameters
;input x86:
;    1:file handle
;    2:return pos in stat64
;    3:pointer to long long
;input x64:
;    1:file handle
;    2:return pos in stat64

;output x86:
;    EAX => low integer with size
;    [3] => low and high integer with value of stat

;output x64:
;    RAX => value of stat (qword)
;------------------------------------------------------------------------------
if used getFstatInfo
getFstatInfo:
call denter

sub pesp,0xA0 ;Clear memory for stat with off64_t instead of off_t

mov ecx,0xA0/psize
xor eax,eax
mov pedi,pesp
rep stosp

argpushsys 2,pesp                   ;stat pointer
argpushsys 1,pdword [pebp+psize*2] ;file descriptor
linuxCall 5,197  ;fstat / fstat64

ppush noErrorGetfileStatValue
call setErrNo

xor eax,eax
if used is32
  mov pedi,dword [pebp+psize*4]  ;long long
  stosd
  stosd
end if

xor eax,eax

jmp errGetfileStatValue

noErrorGetfileStatValue:
;mov ecx,0xA0/4
;mov pesi,pesp
;run:
;lodsd
;mov [errno],peax
;call printErrorNoNewLine
;call printSpace
;loop run

;call WriteNewLine

mov pecx,pdword [pebp+psize*3] ;return pos in stat64

if used is32
  mov pedi,dword [ebp+psize*4]  ;long long
  lea pesi,pdword [pesp+pecx]
  movsd ;low
  movsd ;high

  mov eax,dword [pesi-8]
else
  mov rax,qword [rsp+pecx]
end if

errGetfileStatValue:
call dleave
if used is64
  ret psize*2
else
  ret psize*3
end if
end if

;------------------------------------------------------------------------------
;This function will return the size of a file (with limit of 4GiB on x86)
;input:
;    1:file handle

;output:
;    (R/E)AX => low integer with size

if used getfilesize | (used getfilesizeLongLong & used is64)
getfilesize:
if used getfilesizeLongLong & used is64
  getfilesizeLongLong:
end if
call denter

if used is64
  push 0x30 ;the pos in Fstat64 we want returned
else
  ;create long long
  push 0
  push 0

  push esp  ;push return value for function getFstatInfo
  push 0x2C ;the pos in Fstat64 we want returned
end if

push pdword [pebp+psize*2]
call getFstatInfo

if used is32
  pop eax
  add esp,0x4
end if

errGetfilesize:
call dleave
ret psize
end if

;------------------------------------------------------------------------------
;This function will return the size of the file
;input x86:
;    1:file handle
;    2:pointer to long long
;input x64:
;    1:file handle

;output x86:
;    EAX => low integer with size
;    [2] => high integer with size

;output x64:
;    RAX => size of file
;------------------------------------------------------------------------------
if used getfilesizeLongLong & used is32
getfilesizeLongLong:
call denter

if used is64
  push 0x30 ;the pos in Fstat64 we want returned
else
  push pdword [pebp+psize*3]  ;push return value for function getFstatInfo
  push 0x2C                   ;the pos in Fstat64 we want returned
end if

push pdword [pebp+psize*2]
call getFstatInfo

errGetfilesizeLongLong:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will return the size that can be read from the pipe
;input:
;    1:pipe handle

;output:
;    EAX => integer with size
;------------------------------------------------------------------------------
getPipeSize:
call denter

push 0

argpushsys 3,pesp     ;buffer that receives the size available in the pipe
argpushsys 2,FIONREAD
argpushsys 1,pdword [pebp+psize*2] ;file descriptor
linuxCall 16,54  ;ioctl

pop peax
call dleave
ret psize


;------------------------------------------------------------------------------
;This function will return the last modified date of the file
;input x86:
;    1:file handle
;    2:pointer to long long
;input x64:
;    1:file handle

;output x86:
;    EAX => low integer with last modified date
;    [2] => high and low integer with last modified date

;output x64:
;    RAX => size of file
;------------------------------------------------------------------------------
if used getfileLastModifiedDate
getfileLastModifiedDate:
call denter

if used is64
  push 0x58 ;the pos in Fstat64 we want returned
else
  push pdword [pebp+psize*3]  ;push return value for function getFstatInfo
  push 0x48                   ;the pos in Fstat64 we want returned
end if

push pdword [pebp+psize*2]
call getFstatInfo

if used is32
  mov pesi,pdword [pebp+psize*3]
  mov dword [pesi+4],0
end if

errgetfileLastModifiedDate:
call dleave
if used is64
  ret psize*1
else
  ret psize*2
end if
end if

;------------------------------------------------------------------------------
;This function will read out a file
;input:
;    1:file handle
;    2:pointer to buffer
;    3:size to read

;output:
;    EAX => size read
;------------------------------------------------------------------------------
if used readfile
readfile:
call denter

argpushsys 3,pdword [pebp+psize*4] ;size to read
argpushsys 2,pdword [pebp+psize*3] ;buffer
argpushsys 1,pdword [pebp+psize*2] ;file descriptor
linuxCall 0,3  ;read

ppush noErrReadingFile
call setErrNo

xor eax,eax

noErrReadingFile:
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will write to a file
;input:
;    1:file handle
;    2:pointer to buffer
;    3:size to write

;output:
;    EAX => size written
;------------------------------------------------------------------------------
if used writefile
writefile:
call denter

argpushsys 3,pdword [pebp+psize*4]  ;size to write
argpushsys 2,pdword [pebp+psize*3] ;buffer
argpushsys 1,pdword [pebp+psize*2] ;file handle
linuxCall 1,4  ;write

ppush noErrWritingFile
call setErrNo

xor eax,eax

noErrWritingFile:
call dleave
ret psize*3
end if

SEEK_SET equ 0
SEEK_END equ 2

;------------------------------------------------------------------------------
;This function will set the file pointer to desired position
;input:
;    1:file handle
;    2:(x64)filepointer to seek to (x86)pointer to a long long with a filepointer to seek to

;output:
;    EAX => return value SetFilePointer (the low-order dword of the current file pointer
;------------------------------------------------------------------------------
if used seekToPointerFile
seekToPointerFile:
call denter

mov peax,pdword [pebp+psize*3] ;(x64)pointer to seek to (x86)pointer to a long long to seek to

if used is64
  argpushsys 3,SEEK_SET              ;seek from begining
  argpushsys 2,peax                  ;long long
else
  sub esp,0x8
  mov edx,esp ;=result
  argpushsys 5,SEEK_SET              ;seek from begining
  ;argpushsys 4,0                    ;result already set
  argpushsys 3,dword [peax]          ;low offset
  argpushsys 2,dword [peax+0x4]      ;high offset
end if
argpushsys 1,pdword [pebp+psize*2] ;file handle
linuxCall 8,140  ;lseek / _llseek

ppush noErrSeekingFile
call setErrNo

xor eax,eax

noErrSeekingFile:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will set the file pointer to the end of the file
;input:
;    1:file handle

;output:
;    EAX => return value SetFilePointer (the low-order dword of the current file pointer
;------------------------------------------------------------------------------
if used seekToEndPointerFile
seekToEndPointerFile:
call denter

if used is64
  argpushsys 3,SEEK_END              ;seek from begining
  argpushsys 2,0                     ;long long
else
  sub esp,0x8
  mov edx,esp ;=result
  argpushsys 5,SEEK_END              ;seek from begining
  ;argpushsys 4,0                    ;result already set
  argpushsys 3,0                     ;low offset
  argpushsys 2,0                     ;high offset
end if
argpushsys 1,pdword [pebp+psize*2] ;file handle
linuxCall 8,140  ;lseek / _llseek

ppush noErrSeekingFile2
call setErrNo

xor eax,eax

noErrSeekingFile2:

call dleave
ret psize*1
end if

;------------------------------------------------------------------------------
;This function will close a file handle
;input:
;    1:file handle

;output:
;    EAX => file handle
;------------------------------------------------------------------------------
if used closefile
closefile:
call denter

argpushsys 1,pdword [pebp+psize*2]
linuxCall 3,6  ;close

call dleave

mov peax,pdword [pesp+psize]

ret psize
end if

;------------------------------------------------------------------------------
;This function will check if a directorie exist
;input:
;    1:path (UTF-8)

;output:
;       eax:0 (no directoty) 0x10:directory
;flags: Z: No directory  z: directory
;------------------------------------------------------------------------------
if used isDirectory
isDirectory:
call denter

push 0x400 ;OPEN_DIRECTORY_ONLY
push pdword [pebp+psize*2] ;path
call openFileHandle
xor ecx,ecx
cmp peax,-1
cmovz peax,pecx
jz endIsDirectory

push peax
call closefile

mov eax,FILE_ATTRIBUTE_DIRECTORY

endIsDirectory:
call dleave

test peax,peax
ret psize
end if

;------------------------------------------------------------------------------
;This function will check if the open file handle is a directorie
;input:
;    1:file handle

;output:
;       eax:0 (no directoty) 0x10:directory
;flags: Z: No directory  z: directory
;------------------------------------------------------------------------------
if used isDirectoryByHandle
isDirectoryByHandle:
call denter

if used is64
  push 0x18 ;the pos in Fstat64 we want returned
else
  push 0
  push 0

  push pesp                   ;push return value for function getFstatInfo
  push 0x10                   ;the pos in Fstat64 we want returned
end if

push pdword [pebp+psize*2]
call getFstatInfo

if used is32
  pop peax
  add pesp,psize
end if

mov ecx,0x10

and eax,S_IFMT
cmp eax,S_IFDIR
cmove eax,ecx
je errIsDirectoryByHandle

xor eax,eax

errIsDirectoryByHandle:
call dleave

test eax,eax
ret psize
end if

;------------

if used __calltdents64
__calltdents64:
call denter

argpushsys 3,pdword [pebp+psize*4]  ;count
argpushsys 2,pdword [pebp+psize*3]  ;linux_dirent64 structure
argpushsys 1,pdword [pebp+psize*2]  ;file descriptor of a directory
linuxCall 217,220  ;getdents64

ppush end__calltdents64
call setErrNo

mov peax,-1

end__calltdents64:
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function start (and return an handle) to enumerate file(s) and directories in a directory
;input:
;    1:path (UTF-8)
;    2: return structure (total: 4+(7*2)+4+8+(260*4)=1070, rounded by 8=1072):
;       Atributes (dword):
;           0x10 Directory
;           0x10000 SYM LINK
;           0x20000 File
;       TIMEDATE structure (7*2 + 4)
;       Filesize (long long, no pointer)
;       Filename[260] (UTF-8)



;output:
;       eax:HANDLE or zero
;------------------------------------------------------------------------------
if used startDirectoryEnumeration
startDirectoryEnumeration:
xor eax,eax
xchg peax,pdword [pesp+psize*2] ;swap return structure with flags
xchg peax,pdword [pesp+psize]   ;swap path (UTF-8) with return structure
xchg peax,pdword [pesp]         ;swap return address with path (UTF-8)
push peax                       ;push return address
end if


;------------------------------------------------------------------------------
;This function start (and return an handle) to enumerate file(s) and directories in a directory
;input:
;    1:path (UTF-8) or file handle (see flags)
;    2: return structure (total: 4+(7*2)+4+8+(260*4)=1070, rounded by 8=1072):
;       Atributes (dword):
;           0x10 Directory
;           0x10000 SYM LINK
;           0x20000 File
;       TIMEDATE structure (7*2 + 4)
;       Filesize (long long, no pointer)
;       Filename[260] (UTF-8)
;    3:flags
;       0x1 PATH_IS_HANDLE path is a file handle to a directory, not a file



;output:
;       eax:HANDLE or zero
;------------------------------------------------------------------------------
if used startDirectoryEnumerationEx | used startDirectoryEnumeration
startDirectoryEnumerationEx:
call denter

mov peax,pdword [pebp+psize*2]  ;path (UTF-8) or file handle

test dword [pebp+psize*4],0x1 ;PATH_IS_HANDLE
jnz hasHandle

  push 0x400 ;OPEN_DIRECTORY_ONLY
  push pdword [pebp+psize*2] ;path
  call openFileHandle
  jmp opendFileHandle

hasHandle:
call parentDir

  db './',0

parentDir:
pop pesi

argpushsys 1,peax           ;hKey
argpushsys 3,O_RDONLY or O_DIRECTORY
argpushsys 2,pesi           ;subKey
linuxCall 257,295           ;openat

opendFileHandle:
xor ecx,ecx
cmp peax,-1
cmovz peax,pecx
jz endStartDirectoryEnumeration
mov pebx,peax ;R/EBX => file descriptor directory

;max ~3 times the linux_dirent64 structure + 18 extra free bytes
;18 free bytes:
; - open directory (file descriptor)
; - getdents return (dword) (amount of size writen by getdents)
; - total d_reclen (word)
; - 4/8(x64/x86) bytes free
; - first linux_dirent64 structure
push (0x13 + 255 + 6) * 3
call GetMemLocation
mov pedx,peax ;R/EDX -> linux_dirent64 + extra

push ((0x13 + 255 + 6) * 3) - 18 ;total receivable size
add peax,18
push peax    ;first linux_dirent64 structure
push pebx    ;file descriptor directory
call __calltdents64

cmp peax,-1
je errStartDirectoryEnumeration

test peax,peax
jz errStartDirectoryEnumeration

mov pdword [pedx],pebx     ;set file descriptor for directory
mov dword [pedx+psize],eax ;set returnvalue of getdents
mov ax,word [pedx+18+16]
mov word [pedx+psize+4],ax ;set total last d_reclen

add pedx,18 ;first linux_dirent64 structure

call fillInReturnStructureDirectoryEnumeration
;test eax,eax
;jz errStartDirectoryEnumeration

mov peax,pedx
sub peax,18

endStartDirectoryEnumeration:
call dleave
ret psize*3

;--
errStartDirectoryEnumeration:

;test dword [pebp+psize*4],0x1 ;PATH_IS_HANDLE
;jnz hasHandle2

  push pebx
  call closefile

hasHandle2:

push pedx
call FreeMemLocation

xor eax,eax
jmp endStartDirectoryEnumeration
end if

;------------------------------------------------------------------------------
;This function return next next file in the directory
;input:
;    1:handle
;    2: return structure (total: 4+(7*2)+4+8+(260*4)=1070, rounded by 8=1072):
;       Atributes (dword):
;           0x10 Directory
;           0x10000 SYM LINK
;           0x20000 FILE
;       TIMEDATE structure (7*2 + 4)
;       Filesize (long long, no pointer)
;       Filename[260] (UTF-8)



;output:
;       eax:true of false
;------------------------------------------------------------------------------
if used nextDirectoryEnumeration
nextDirectoryEnumeration:
call denter


mov pesi,pdword [pebp+psize*2] ;handle
mov pebx,pdword [pesi] ;R/EBX => open directory (file descriptor)

xor eax,eax
mov ax,word [pesi+psize+4] ;AX -> total d_reclen (word)
mov ecx,dword [pesi+psize] ;CX ->getdents return (dword) (amount of size writen by getdents)

cmp ax,cx
jb nextDirent64 ;next linux_dirent64 structure is laready loaded in memory

  push ((0x13 + 255 + 6) * 3) - 18 ;total receivable size
  mov peax,pesi
  add peax,18
  push peax    ;first linux_dirent64 structure
  push pebx    ;file descriptor directory
  call __calltdents64

  xor ecx,ecx
  cmp peax,-1
  cmove eax,ecx
  je errNextDirectoryEnumeration

  test peax,peax
  jz errNextDirectoryEnumeration

  mov dword [pesi+psize],eax ;set returnvalue of getdents
  mov word [pesi+psize+4],0
  xor eax,eax

nextDirent64:
lea pedx,pdword [pesi+18+peax] ;next linux_dirent64 structure

mov ax,word [pedx+16]
add word [pesi+psize+4],ax ;set total last d_reclen

call fillInReturnStructureDirectoryEnumeration
xor eax,eax
inc ax

errNextDirectoryEnumeration:
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function close the handle of the enumeration
;input:
;    1:handle


;output:
;       none
;------------------------------------------------------------------------------
if used closeDirectoryEnumeration
closeDirectoryEnumeration:
xor eax,eax
xchg pdword [pesp+psize],peax ;swap handle with flags
xchg pdword [pesp],peax       ;swap return address with handle
push peax                     ;push return address
end if

;------------------------------------------------------------------------------
;This function close the handle of the enumeration
;input:
;    1:handle
;    2:flags
;       0x1 PATH_IS_HANDLE path is a file handle to a directory, not a file



;output:
;       none
;------------------------------------------------------------------------------
if used closeDirectoryEnumerationEx | used closeDirectoryEnumeration
closeDirectoryEnumerationEx:
call denter

mov pesi,pdword [pebp+psize*2] ;handle  
mov pebx,pdword [pesi] ;R/EBX => open directory (file descriptor) 

;test dword [pebp+psize*3],0x1 ;PATH_IS_HANDLE
;jnz hasHandle3

  push pebx
  call closefile

hasHandle3:

push pesi
call FreeMemLocation

call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;R/EDX => linux_dirent64 structure
;[pebp+psize*3] => return structure
;R/EBX => File description of the working folder
;
;return: 0: error, 1:success
;------------------------------------------------------------------------------
if used fillInReturnStructureDirectoryEnumeration
fillInReturnStructureDirectoryEnumeration:
mov ecx,1072/psize
xor eax,eax
mov pedi,pdword [pebp+psize*3] ;return structure
rep stosp

lea pesi,pdword [pedx+0x13] ;R/ESI => d_name
mov pedi,pdword [pebp+psize*3] ;return structure
add pedi,4 + (7*2) + 4 + 8

;copy file name to struct
push pesi
push pedi
call strcpy


mov pedi,pdword [pebp+psize*3] ;return structure
mov dword [pedi],0   ;set attributes to zero

mov al,byte [pedx+0x12] ;AL => d_type
cmp al,DT_LNK
jne isNotLink

  mov dword [pedi],0x10000 ;SYM LINK
  jmp isNotDirTypeIsSet

isNotLink:
cmp al,DT_DIR
jne isNotDir

  mov dword [pedi],0x10 ;set attributes as directory

isNotDir:

call denter

argpushsys 2,pesi                  ;filename (set it here for x86 support)

argpushsys 4,S_IRUSR or S_IWUSR or S_IRGRP or S_IWGRP or S_IROTH ;mode
if used is64
  argpushsys 3,O_RDONLY                 ;flags
else
  argpushsys 3,O_RDONLY or O_LARGEFILE  ;flags
end if
argpushsys 1,pebx                  ;dfd
linuxCall 257,295  ;openat

call dleave

ppush openAtWasSuccessfull
call setErrNo

xor eax,eax
jmp errfillInReturnStructureDirectoryEnumeration

openAtWasSuccessfull:
mov pesi,peax ;R/ESI => file pointer

mov pedi,pdword [pebp+psize*3] ;return structure

mov al,byte [pedx+0x12] ;AL => d_type
cmp al,DT_DIR
jne isNotDir2

  setFileSizeToZero:
  ;set FileSize to zero
  mov pdword [pedi+0x16],0
  if used is32
    mov dword [pedi+0x1A],0
  end if

  jmp dirValuesAreSet

isNotDir2:
  mov dword [pedi],0x20000 ;FILE
isNotDirTypeIsSet:
  cmp al,DT_REG  ;is file
  jne setFileSizeToZero

    if used is32
      lea eax,dword [edi+0x16]
      push eax
    end if
    push pesi ;file handle current file
    call getfilesizeLongLong
    if used is64
      mov qword [rdi+0x16],rax
    end if

dirValuesAreSet:
    if used is32
      push 0
      push 0
      push esp
    end if

  push pesi ;file handle current file
  call getfileLastModifiedDate


  lea pecx,pdword [pedi+0x4]
  push pecx ;TIMEDATE
  if used is64
    push peax ;time in seconds
  else
    lea eax,dword [esp+0x4]
    push peax ;time in seconds
  end if
  if used RecoverDateTime_asm
    call RecoverDateTime_asm
  else
    cmp [RecoverDateTime],0
    jne callRecoverDateTime

       add pesp,psize*2
       jmp dontCallRecoverDateTime

    callRecoverDateTime:
    call [RecoverDateTime]
    dontCallRecoverDateTime:
  end if

  if used is32
    add esp,0x8
  end if

  push pesi
  call closefile

  xor eax,eax
  inc eax

errfillInReturnStructureDirectoryEnumeration:
ret
end if

;------------------------------------------------------------------------------
;Runs a program and wait for it to be terminated
;1: Args
;
;return: 0: error, 1:success
;------------------------------------------------------------------------------
if used ExecAndWait
ExecAndWait:
  call denter

  mov pebx,pdword [pebp+psize*2]

  linuxCall 57,2    ;fork
  xor ecx,ecx
  test peax,peax
  cmovc peax,pecx
  jl endExecAndWait
  jne processIsForked

    argpushsys 3,0
    argpushsys 2,pebx          ;args
    argpushsys 1,pdword [pebx] ;program
    linuxCall 59,11            ;execve

    push 127
    call exitProcess

  processIsForked:
  argpushsys 4,0       ;*ru
  argpushsys 3,0       ;options
  argpushsys 2,0       ;*stat_addr
  argpushsys 1,peax    ;pid
  linuxCall 61,114     ;wait4

  xor eax,eax
  inc eax

endExecAndWait:
call dleave
ret psize
end if

macro spush str {
local ..pushStr
local ..startStr
  call ..pushStr

..startStr:
    db str,0

  ..pushStr:
  pop peax
  mov ecx,..pushStr-..startStr
}

if used getMimeTypeFromExtension
;------------------------------------------------------------------------------
;This function will get a MIM
;input:
;    (R/E)AX:pointer with extension
;output:
;    (R/E)AX=> pointer with mime-type
;    (R/E)CX=> size of pointer
;------------------------------------------------------------------------------
getMimeTypeFromExtension:
test peax,peax
jz noExtension

mov eax,dword [peax+1]

cmp eax,'txt'
je isPlainText

cmp eax,'htm'
je isHtml

cmp eax,'html'
je isHtml

cmp eax,'html'
je isHtml

cmp eax,'png'
je isPng

cmp eax,'gif'
je isJif

cmp eax,'jpg'
je isjpg

cmp eax,'jpeg'
je isjpg

cmp eax,'css'
je isCss

cmp eax,'js'
je isJs

noExtension:
xor eax,eax
xor ecx,ecx
ret

isjpg:
spush "image/jpeg"
ret

isJif:
spush "image/gif"
ret

isPng:
spush "image/png"
ret

isHtml:
spush "text/html"
ret

isPlainText:
spush "text/plain"
ret

isCss:
spush "text/css"
ret

isJs:
spush "text/javascript"
ret

mov byte [peax+pecx-1],0
ret
end if

if used __FreeMimeTypeFromExtension_internal
__FreeMimeTypeFromExtension_internal:
ret psize
end if