
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;----------------------------------------------------
;                      LaunchCGI
;Launches CGI executable, send POST-command to CGI-application and send the header and result to the client
;
;input:
;    1:ProtocolSession
;    2:ClientSession
;    3:Pointer to a string with the CGI_path
;    4:Pointer to a string with the Enviroment
;    5:buffer size
;    6:Maximum timeout CGI-application
;    7:Pointer to a HTTPSession structure
;    8:Pointer to a ResponseHeader structure
;    9:flags:
;      The following flag is only available in Windows, linux is always UTF-8:
;          0x1 use UTF-8 in environment instead of TF1-6

;
;output: (R/E)AX: True/False
;
;----------------------------------------------------

LaunchCGI:
call denter
;Create space on stack:
;--psize*2: Return value for pipe() "CGI return Pipe"
;--psize*2: Return value for pipe() "Pipe For Post Command"
;--psize for keeping track of the timeout value
;--psize for saving pointer that is been replaced
;--psize for read bytes
sub pesp,psize*2 + psize*2 + psize + psize + psize
mov pebx,pesp ;save stack pointer

;--Create CGI return Pipe--
argpushsys 1,pebx  ;pipes
linuxCall 22,42    ;pipe
xor ecx,ecx
test peax,peax
cmovc peax,pecx
jl endLaunchCGI

;--Creating Pipe For Post Command--
lea pedi,pdword [pebx+psize*2]
argpushsys 1,pedi  ;pipes
linuxCall 22,42    ;pipe
xor ecx,ecx
test peax,peax
cmovc peax,pecx
jl endLaunchCGIDeleteCGIPipe

mov pebx,pesp ;save stack pointer

;stdin pdword [pebx+psize*2]
;stdout
;strerr pdword [pebx+psize*1]

;---------------
;Split commandline
;---------------
push pdword [pebp+psize*4]      ;CGI-APP
call splitCommandLine
mov pecx,peax ;R/ECX => amount of items in CGI-APP

;count space on stack
inc peax ; add one for the terminating zero
if used is64
  shl peax,3  ;*8
else
  shl peax,2    ;*4
end if
;free space for items
sub pesp,peax

mov pesi,pdword [pebp+psize*4]  ;R/ESI => Splited CGI-APP
sub peax,psize
xor pedx,pedx ;real position on stack

loopThroughAppPath:
  cmp byte [pesi],0
  je nextLoopThroughAppPath

mov peax,pesp
  mov pdword [pesp+pedx],pesi ;add item to stack
  add pedx,psize

nextLoopThroughAppPath:
  push pesi
  call strlen
  add pesi,peax
  inc pesi

dec pecx
jnz loopThroughAppPath
mov peax,pesp

mov pdword [pesp+pedx],0 ;zero last item
if used is64
  mov r13,pesp ;R13  => pointer to argv[]
else
  mov pedi,pesp ;R/EDI => pointer to argv[]
end if

;---------------
;Split Enviroment
;---------------
xor pedx,pedx ;R/EDX => envp[] = zero
mov pesi,pdword [pebp+psize*5]
test pesi,pesi
jz noEnviroment

push 0 ;last pointer in enviroment
pushAllEnviromentsItems:
  push pesi

  push pesi
  call strlen
  add pesi,peax
  inc pesi

cmp byte [pesi],0
jne pushAllEnviromentsItems

mov pedx,pesp ;R/EDX => envp[]

noEnviroment:

linuxCall 57,2    ;fork
xor ecx,ecx
test peax,peax
cmovc peax,pecx
jl endLaunchCGIDeletePostPipe
jnz isCurrentProcessNotforked

  ;
  ;Forked process to start CGI-application
  ;

  if used is64
    mov r12,pebx ;R/12 => now pointer to stack what was on R/EBX before
  else
    mov pesi,pebx  ;R/ESI => now pointer to stack what was on R/EBX before
  end if

  ;
  ;Set write pipe on stdout
  ;
  argpushsys 2,1              ;std out
  if used is64
    mov eax,dword [r12+4]     ;R/EAX =>write pipe
  else
    mov eax,pdword [pesi+4]   ;R/EAX =>write pipe
  end if
  argpushsys 1,peax           ;write pipe
  linuxCall 33,63             ;dup2
  test peax,peax
  jl exitForkedProcess

  ;
  ;Set write pipe on stderr
  ;
  argpushsys 2,2               ;std err
  if used is64
    mov eax,dword [r12+4]    ;R/EAX =>write pipe
  else
    mov eax,dword [pesi+4]   ;R/EAX =>write pipe
  end if
  argpushsys 1,peax           ;write pipe
  linuxCall 33,63             ;dup2
  test peax,peax
  jl exitForkedProcess

  ;
  ;Set read pipe on stdin
  ;
  argpushsys 2,0                   ;stdin
  if used is64
    mov eax,dword [r12+psize*2]   ;R/EAX =>read pipe
  else
    mov eax,dword [pesi+psize*2]  ;R/EAX =>read pipe
  end if
  argpushsys 1,peax               ;read pipe
  linuxCall 33,63                 ;dup2
  test peax,peax
  jl exitForkedProcess

  ;
  ;Close all pipes
  ;
  if used is64
    mov eax,dword [r12]       ;read pipe
    push rax                  ;read pipe
  else
    push dword [esi]          ;read pipe
  end if
  call closefile

  if used is64
    mov eax,dword [r12+4]   ;write pipe
    push rax                ;write pipe
  else
    push dword [esi+4]   ;write pipe
  end if
  call closefile

  if used is64
    mov eax,dword [r12+psize*2] ;read pipe
    push rax
  else
    push dword [esi+psize*2]    ;read pipe
  end if
  call closefile

  if used is64
    mov eax,dword [r12+(psize*2)+4] ;write pipe
    push rax
  else
    push dword [esi+(psize*2)+4]    ;write pipe
  end if
  call closefile

  ;
  ;Start CGI-application
  ;
  argpushsys 3,pedx             ;envp[]
  if used is64
    argpushsys 2,r13            ;argv[]
    argpushsys 1,pdword [r13]   ;pathname
  else
    argpushsys 2,pedi           ;argv[]
    argpushsys 1,pdword [pedi]  ;pathname
  end if
  linuxCall 59,11               ;execve

  exitForkedProcess:
  call exitThread

isCurrentProcessNotforked:
;
;the original process
;

mov pesi,peax ;R/ESI => PID of CGI process

mov eax,dword [pebx+4]    ;write pipe
push peax                 ;write pipe
call closefile
mov dword [pebx+4],0

mov eax,dword [pebx+psize*2]  ;read pipe
push peax                     ;read pipe
call closefile
mov dword [pebx+psize*2],0

mov pedx,pdword [pebp+psize*9] ;(R/E)DX => ResponeHeader
if used EnableChunking
  or dword [pedx],FRH_DONT_END_HEADER or FRH_DONT_SEND_CONTENT_TYPE or FRH_TRANSFER_ENCODING_CHUNKED ;add flags
else
  or dword [pedx],FRH_DONT_END_HEADER or FRH_DONT_SEND_CONTENT_TYPE  ;add flags
end if

if ~ used EnableChunking
;-;delete FRH_CONNECTION_KEEP_ALIVE flags
test dword [pedx],FRH_CONNECTION_KEEP_ALIVE
jz dontKeepConnectionAlive

  xor dword [pedx],FRH_CONNECTION_KEEP_ALIVE

dontKeepConnectionAlive:
end if
;check if a part of the POST command is in the buffer existing in HTTPSession
mov pedx,pdword [pebp+psize*8] ;(R/E)DX => HTTPSession
mov ecx,dword [pedx+psize*7+4] ;ECX => headerRestSize

test pecx,pecx
jz noPostCommandInBuffer
;jecxz noPostCommandInBuffer

  mov peax,pdword [pedx+psize*4] ;buffer
  if used is32
    add eax,dword [pedx+psize*7]  ;add headerSize to get pointer to headerRest
  else
    mov edx,dword [pedx+psize*7]  ;EDX => headerSize
    add rax,rdx                   ;add headerSize to get pointer to headerRest
  end if

  ;Write this POST command
  push pecx                         ;nNumberOfBytesToWrite
  push peax                         ;lpBuffer
  mov eax,dword [pebx+(psize*2)+4]  ;hWritePipe
  push peax                         ;hWritePipe
  call writefile

noPostCommandInBuffer:
;--;receive the size of one page
;--call GetSmallestPortionPageAllocation
mov eax,dword [pebp+psize*6] ;buffersize

if used is32
  mov ecx,eax ;ECX = size of buffer
else
  mov r12,rax ;r12 => size of buffer
end if

;allocate buffer
push peax
call GetMemLocation
mov pedi,peax ;(R/E)DI => pointer to buffer

if used is32
  sub ecx,4            ;substract the four bytes used by size of buffer integer
  mov dword [pedi],ecx ;dword [edi] = size of buffer
  add pedi,4           ;dword [edi-4] = size of buffer
end if

call getTickCount
if used is32
  mov pdword [pebx+psize*4],eax ;save TickCount on stack
else
  mov r13d,eax                  ;r13 => tickCount
end if

waitForInput:
  push 10
  call sleep ;Wait for 10 miliseconds

  mov edx,dword [pebp+psize*7] ;EDX => maximum timeout
  test edx,edx
  jz unlimitedExecutionTime

    call getTickCount
    mov edx,dword [pebp+psize*7] ;EDX => maximum timeout
    if used is32
      sub eax,pdword [pebx+psize*4] ;Tickcount - privious timeout => time CGI-application is running
    else
      sub eax,r13d                  ;Tickcount - privious timeout => time CGI-application is running
    end if
    cmp eax,edx
    ja timeoutReached

  unlimitedExecutionTime:
  ;check if connection is still alive
  push 1   ;buf (set to 'true' for if plugin doesn't support flag CSI_CONNECTION_ALIVE)
  mov peax,pesp

  push 4   ;len

  push pesp                                   ;*len
  push peax                                   ;*buf
  push CSI_CONNECTION_ALIVE                   ;type
  push pdword [pebp+psize*3]                  ;ClientSession
  push pdword [pebp+psize*2]                  ;ProtocolSession
  call [GetClientSessionInfo]
  test eax,eax
  jz connectionStillAlive

  pop peax ;len
  pop peax ;buf
  test peax,peax
  jz timeoutReached ;connection is not alive anymore

  connectionStillAlive:
  mov pedx,pdword [pebp+psize*8] ;(R/E)DX => HTTPSession
  mov pedx,pdword [pedx]         ;(R/E)DX => defaultVariables
  mov pedx,pdword [pedx+psize]   ;EDX => method
  cmp dword [pedx],'POST'        ;check if method is POST
  jne noPostCommand
    if used is32
      push dword [pedi-4]       ;len
    else
      push r12                 ;len
    end if

    push pesp                                   ;&len
    push pedi                                   ;buf
    push 100                                    ;timeout
    push RSF_RETURN_ON_FIRST_SUCCESSFUL_ATTEMPT ;flags
    push pdword [pebp+psize*3]                  ;ClientSession
    push pdword [pebp+psize*2]                  ;ProtocolSession
    call [RecvData]
    test eax,eax
    jz noRecvError

    cmp eax,3
    jbe timeoutReached

  noRecvError:
    pop peax ;Received data (len)

    test peax,peax
    jz noPostCommand
    ;jmp noPostCommand

    push peax                         ;nNumberOfBytesToWrite
    push pedi                         ;lpBuffer
    mov eax,dword [pebx+(psize*2)+4]  ;hWritePipe
    push peax                         ;hWritePipe
    call writefile

  noPostCommand:
  mov eax,dword [pebx]
  push peax            ;hReadPipe
  call getPipeSize
  test eax,eax
  jnz ReadBuffer

  push pesi
  call isProcessStillRunning

  test eax,eax
jnz waitForInput
jmp endLaunchCGIDeleteBufferEndProcessHandles

ReadBuffer:
  ;if(lenBuf<FileSize) FileSize=lenBuf;
  if used is32
    cmp dword [pedi-4],eax
    jae dontChangeReadSize

      mov eax,dword [pedi-4]

    dontChangeReadSize:
  else
    cmp r12,rax
    cmovb rax,r12
  end if

  push peax                  ;nNumberOfBytesToRead
  push pedi                  ;lpBuffer
  mov eax,dword [pebx]       ;hReadPipe
  push peax                  ;hReadPipe
  call readfile
  mov pdword [pebx+psize*6],peax ;size read

  cmp pdword [pebp+psize*9],0 ;ResponseHeader is set to zero is the header is send
  je HeaderIsAlreadySend

  cmp pdword [pebp+psize*9],1 ;ResponseHeader is set to one is the header is send
  je HeaderIsAlreadySend
    ;mov eax,0x0A0D0A0D ;two new lines
    mov pdword [pebx+psize*5],0 ;for the pointer that is set to zero for the Status
    mov pecx,pdword [pebx+psize*6]   ;lpNumberOfBytesRead
    mov pedx,pedi ;(R/E)DX => pointer to buffer
    mov peax,pedi ;(R/E)AX => pointer to buffer

    loopBuffer:
      cmp pecx,4 ;check if size of bufer > 4
      jb endHeaderReached

      cmp dword [pedx],0x0A0D0A0D ;check on new lines
      je endHeaderReached

      cmp byte [pedx],':'
      jne nextLoopBuffer

        mov byte [pedx],0x00 ;set zero byte on ':' location

        ;generate hash from the name
        push peax ;push for function GenerateHashCaseInsensitiveSafe

        ;find next next new line
        mov peax,pedx ;(R/E)AX => location of ':'

        findNewLine:
          cmp ecx,2
          jb endFindNewLine

          cmp word [pedx],0x0A0D
          je FoundNewLine

          inc pedx
        loop findNewLine

      endFindNewLine:
        add pesp,psize ;remove the push for GenerateHashCaseInsensitiveSafe
        mov byte [peax],':' ;set ':' back on it's location
        jmp endHeaderReached

      FoundNewLine:
        ;swap pointer to ':' and the pointer for function GenerateHashCaseInsensitiveSafe
        xchg peax,pdword [pesp]
        push peax    ;repush for function GenerateHashCaseInsensitiveSafe
        call GenerateHashCaseInsensitiveSafe
        cmp eax,0x45AF5DAD   ;hash of 'status'

        pop peax
        mov byte [peax],':' ;set ':' back on it's location

        jne findNextNewLine

          inc peax ;skip ':'
          skipSpaces:
            cmp byte [peax],' '
            jne endSkipSpaces

            inc peax
          endSkipSpaces:

          mov byte [pedx],0
          mov pdword [pebx+psize*5],pedx  ;pointer to replaced 0x0D byte

          mov pedx,pdword [pebp+psize*9] ;(R/E)DX => response header
          mov pdword [pedx+8],peax       ;set status code
          jmp endHeaderReached


        findNextNewLine:
        mov peax,pedx
        add peax,2 ;skip new line and send pointer to fname of CGI-variable

    nextLoopBuffer:
      inc pedx
    loop loopBuffer

  endHeaderReached:
    push DEFAULT_SENDTIMEOUT   ;timeout
    push pdword [pebp+psize*9] ;ResponseHeader
    push pdword [pebp+psize*3] ;ClientSession
    push pdword [pebp+psize*2] ;ProtocolSession
    call [SendHTTPHeader]


    mov peax,pdword [pebx+psize*5] ;(R/E)AX => pointer to replaced 0x0D byte
    test peax,peax
    jz statusWasNotSet

      mov byte [peax],0x0D           ;place the byte back
    statusWasNotSet:

    if used EnableChunking
      mov pdword [pebp+psize*9],1    ;set responseHeader to 1 as a signle to find the new line
    else
      mov pdword [pebp+psize*9],0    ;set responseHeader to 0
    end if

  HeaderIsAlreadySend:
  if used EnableChunking
  cmp pdword [pebp+psize*9],0
  je sendDataWithChunk

  mov pecx,pdword [pebx+psize*6]   ;lpNumberOfBytesRead
  mov pedx,pedi ;(R/E)DX => pointer to buffer
  HeaderCheck:
    cmp pecx,4 ;check if size of bufer > 4
    jb sendDataWithoutChunk

    cmp dword [pedx],0x0A0D0A0D ;check on new lines
    jne nextHeaderCheck

      mov pdword [pebp+psize*9],0
      sub pedx,pedi ;get size of header
      add pedx,2    ;add one 0A0D (the second will be send by the first chunk)
      mov pecx,pdword [pebx+psize*6] ;size
      sub pecx,pedx ;size of the content + 0A0D
      sub pecx,2    ;size of the content

      push pedx                   ;len

      push pesp                   ;&len
      push pedi                   ;buffer
      push DEFAULT_SENDTIMEOUT    ;timeout
      push RSF_RETURN_ON_COMPLETE ;flags
      push pdword [pebp+psize*3]  ;ClientSession
      push pdword [pebp+psize*2]  ;ProtocolSession
      call [SendData]

      pop peax ;remove len

      test pecx,pecx
      jz waitForInput ;continue reading the stream if there is no content
      call SendSizeChunk

      push pecx

      push pesp                   ;&len
      mov peax,pedi               ;(R/E)AX => buffer
      add peax,pedx               ;(R/E)AX => content - 2 (OAOD)
      add peax,2                  ;(R/E)AX => content
      push peax                   ;buffer
      push DEFAULT_SENDTIMEOUT    ;timeout
      push RSF_RETURN_ON_COMPLETE ;flags
      push pdword [pebp+psize*3]  ;ClientSession
      push pdword [pebp+psize*2]  ;ProtocolSession
      call [SendData]

      pop pecx  ;remove len
      test eax,eax
      jnz timeoutReached
      jmp waitForInput

  nextHeaderCheck:
    inc pedx
  dec pecx
  jnz HeaderCheck
  ;loop HeaderCheck

  sendDataWithChunk:
  mov pecx,pdword [pebx+psize*6] ;size
  call SendSizeChunk

  sendDataWithoutChunk:
  end if
  lea peax,pdword [pebx+psize*6]
  push peax                   ;&len
  push pedi                   ;buffer
  push DEFAULT_SENDTIMEOUT    ;timeout
  push RSF_RETURN_ON_COMPLETE ;flags
  push pdword [pebp+psize*3]  ;ClientSession
  push pdword [pebp+psize*2]  ;ProtocolSession
  call [SendData]
  test eax,eax
  jnz timeoutReached

jmp waitForInput

timeoutReached:
  ;terminate the CGI-application
  push pesi  ;PID
  call terminateProcess

endLaunchCGIDeleteBufferEndProcessHandles:
if used EnableChunking
call sendLastChunk
end if

;Free buffer
if used is32
  lea ecx,dword [pedi-4]
  push ecx
else
  push rdi
end if
call FreeMemLocation

;
;Close all pipes
;
endLaunchCGIDeletePostPipe:
mov eax,dword [pebx+psize*2] ;read pipe
push peax                    ;read pipe
call closefile
test eax,eax
jz ReadPipeAlreadyClosed

mov eax,dword [pebx+(psize*2)+4] ;write pipe
push peax                        ;write pipe
call closefile

ReadPipeAlreadyClosed:
endLaunchCGIDeleteCGIPipe:
mov eax,dword [pebx]         ;read pipe
push peax                    ;read pipe
call closefile

mov eax,dword [pebx+4]       ;write pipe
test eax,eax
jz WritePipeAlreadyClosed
push peax                    ;write pipe
call closefile

WritePipeAlreadyClosed:
endLaunchCGI:
call dleave
ret psize*9

