
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

REG_SZ equ 1
REG_BINARY equ 3
REG_MULTI_SZ equ 7
REG_DWORD equ 4
HKEY_LOCAL_MACHINE equ 0

;----------------------------------------------------
;                      RegMaxKey
;0x4  = hKey
;
;output: (R/E)AX = return value of RegQueryInfoKey
;        (R/E)DX = ammount of sub keys
;        (R/E)CX = length of the largest sub-key
;
;----------------------------------------------------

if used RegMaxKey
RegMaxKey:
push pebp
mov pebp,pesp
push pebx
push pesi

argpushsys 1,pdword [pebp+psize*2]   ;file descriptor of a directory
linuxCall 32,41  ;dup

sub pesp,1072
mov pesi,pesp ;return structure

push 0x1                    ;PATH_IS_HANDLE
push pesi                   ;return structure
push pdword [pebp+psize*2]  ;hFile
call startDirectoryEnumerationEx
mov pebx,peax ;R/EBX => directory handle

xor edx,edx  ;ammount of sub keys
xor ecx,ecx  ;length of the largest sub-key in WCHAR
inc ecx
test peax,peax
cmovz peax,pecx
cmovz pecx,pedx
jz ErrRegMaxKey

dec ecx

loopThroughDirectoryKeys:
  test dword [pesp],0x10
  jz isNoKey

   cmp word [pesp+4+(7*2)+4+8],'.'
   je isNoKey

   cmp word [pesp+4+(7*2)+4+8],'..'
   jne isNotPriviousDir

   cmp byte [pesp+4+(7*2)+4+8+2],0
   je isNoKey

   isNotPriviousDir:
   lea peax,pdword [pesp+4+(7*2)+4+8]
   push peax
   call strlen
   cmp ecx,eax
   jnc notBiggerFilename

     mov ecx,eax ;set highest filename

   notBiggerFilename:
   inc edx ;one more key

isNoKey:
  push pesp ;return structure
  push pebx ;directory handle
  call nextDirectoryEnumeration
  test peax,peax
jnz loopThroughDirectoryKeys

push 0x1                    ;PATH_IS_HANDLE
push pebx ;directory handle
call closeDirectoryEnumerationEx

xor eax,eax

ErrRegMaxKey:
add pesp,1072

pop pesi
pop pebx
mov pesp,pebp
pop pebp
ret psize
end if

;----------------------------------------------------
;                      RegAmmountOfKeys
;1: hKey
;2: SubKey
;3: SubSubKey
;4: SubSubSubKey
;5: SubSubSubSubKey
;
;output: (R/E)AX = ammount of sub keys
;
;----------------------------------------------------
if used RegAmmountOfKeys
RegAmmountOfKeys:
call denter

;open the register key
push pdword [pebp+psize*6]
push pdword [pebp+psize*5]
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
push 0
call OpenRegKey

xor edx,edx
test ecx,ecx   ;test return value
cmovnz peax,pedx ;EAX = 0 if OpenRegKey failed
jnz endRegAmmountOfKeys

push peax ;hKey

push peax ;hKey
call RegMaxKey
if used is32
  mov ebx,edx
end if

pop peax

cmp peax,pdword [pebp+psize*2]
je dontCloseKeyRegAmmountOfKeys

push peax
call RegCloseKeycall

dontCloseKeyRegAmmountOfKeys:
if used is32
  mov eax,ebx
else
  mov peax,pedx
end if

endRegAmmountOfKeys:
call dleave
ret psize*5
end if

;----------------------------------------------------
;                      RegAmmountOfValues
;1: hKey
;2: SubKey
;3: SubSubKey
;4: SubSubSubKey
;5: SubSubSubSubKey
;
;output: (R/E)AX = ammount of values
;
;----------------------------------------------------
if used RegAmmountOfValues
RegAmmountOfValues:
call denter

;open the register key
push pdword [pebp+psize*6]
push pdword [pebp+psize*5]
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
push 0
call OpenRegKey

xor edx,edx
test ecx,ecx   ;test return value
cmovnz eax,edx ;EAX = 0 if OpenRegKey failed
jnz endRegAmmountOfValues

push peax ;hKey

push peax ;hKey
call RegMaxValue
if used is32
  mov ebx,edx
end if

pop peax

cmp peax,pdword [pebp+psize*2]
je dontCloseRegAmmountOfValues

  push peax
  call RegCloseKeycall

dontCloseRegAmmountOfValues:
if used is32
  mov eax,ebx
else
  mov peax,pedx
end if

endRegAmmountOfValues:
call dleave
ret psize*5
end if

;----------------------------------------------------
;                      RegMaxValue
;0x4  = hKey
;
;output: (R/E)AX = return value of RegQueryInfoKey
;        (R/E)DX = ammount of values in this key
;        (R/E)CX = length of the largest name of all the values
;        (R/E)SI = length of the largest value
;
;---------------------------------------------------
if used RegMaxValue
RegMaxValue:
push pebp
mov pebp,pesp
push pebx

sub pesp,1072
mov pesi,pesp ;return structure

push 0x1                    ;PATH_IS_HANDLE
push pesi                   ;return structure
push pdword [pebp+psize*2]  ;hFile
call startDirectoryEnumerationEx
mov pebx,peax ;R/EBX => directory handle

xor edx,edx  ;ammount of sub keys
xor esi,esi  ;ammount of sub value
xor ecx,ecx  ;length of the largest sub-key in WCHAR
inc ecx
test peax,peax
cmovz peax,pecx
cmovz pecx,pedx
jz ErrRegMaxValue

dec ecx

loopThroughDirectoryValues:
  test dword [pesp],0x20000 ;file
  jz isNoValue

   lea peax,pdword [pesp+4+(7*2)+4+8] ;filename
   push peax
   call strlen
   cmp ecx,eax
   jnc notBiggerFilename2

     mov ecx,eax ;set highest filename

   notBiggerFilename2:
   mov peax,pdword [pesp+4+(7*2)+4] ;filesize
   cmp esi,eax
   jnc notBiggerSize

     mov esi,eax ;set highest filesize

   notBiggerSize:
   inc edx ;one more value

isNoValue:
  push pesp ;return structure
  push pebx ;directory handle
  call nextDirectoryEnumeration
  test peax,peax
jnz loopThroughDirectoryValues

push 0x1                    ;PATH_IS_HANDLE
push pebx ;directory handle
call closeDirectoryEnumerationEx

xor eax,eax

ErrRegMaxValue:
add pesp,1072

pop pebx
mov pesp,pebp
pop pebp
ret psize
end if

;----------------------------------------------------
;                      ReadRegDwordWithErrorJump
;1: hKey
;2: SubKey
;3: ValueName
;4: errorPointer
;
;----------------------------------------------------

if used ReadRegDwordWithErrorJump
ReadRegDwordWithErrorJump:
call denter

push pdword [pebp+psize*4]
push 0
push 0
push 0
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
call ReadRegEx

;  (r/e)cx = size of pointer (r/e)ax
;  (r/e)dx = return value of RegQueryValueEx
;  (r/e)ax = Memlocation with the read data
test edx,edx
jnz errReadRegDwordWithErrorJump

mov pecx,pdword [peax]

push peax
call FreeMemLocation

mov peax,pecx

call dleave
ret psize*4
end if

errReadRegDwordWithErrorJump:
call dleave
add pesp,psize*5
jmp pdword [pesp-psize]

;----------------------------------------------------
;                      ReadReg
;0x4  = hKey
;0x8  = SubKey
;0xC  = ValueName
;
;----------------------------------------------------

if used ReadReg
ReadReg:
push pebp
mov pebp,pesp

push pdword [pebp+psize*4]
push 0
push 0
push 0
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
call ReadRegEx

mov pesp,pebp
pop pebp
ret psize*3
end if

;------------

;redirect a call to RegCloseKey in 32bit, create a stack with arguments in 64 first.
if used RegCloseKeycall | used SafeRegCloseKeycall
  RegCloseKeycall:
  SafeRegCloseKeycall:
  jmp closefile
end if

;----------------------------------------------------
;                      OpenRegKey
;0x4  = CreateKey
;0x8  = hKey
;0xC  = SubKey
;0x10 = SubSubKey
;0x14 = SubSubSubKey
;0x18 = SubSubSubSubKey
;
;output
;  (r/e)cx = return value of RegOpenKeyEx
;  (r/e)ax = hKey
;----------------------------------------------------

if used OpenRegKey
OpenRegKey:
push pebp
mov pebp,pesp
push pedx
push pesi
push pedi

mov pedx,pdword [pebp+psize*3]
mov pesi,pebp
add pesi,psize*4

xor ecx,ecx

cmp pdword [pesi],pecx
cmove peax,pedx
je endOpenRegKey

FindFinalKey:
mov pecx,pdword [pebp+psize*2]
jpcxz DoNotCreateKey
push pesi ;save the position of the name of the current subKey
push pedx ;save hFile

argpushsys 1,pedx
argpushsys 3,S_IRWXU or S_IRWXG or S_IROTH or S_IXOTH
argpushsys 2,pdword [pesi]
linuxCall 258,296  ;mkdirat

pop pedx  ;restore hFile
pop pesi  ;restore the position of the name of the current subKey
DoNotCreateKey:

push pesi ;save the position of the name of the current subKey
push pedx ;save 'old' hKey


argpushsys 1,pedx           ;hKey
argpushsys 3,O_RDONLY or O_DIRECTORY
argpushsys 2,pdword [pesi]  ;subKey
linuxCall 257,295           ;openat

pop pedx  ;pop the old key
cmp pdword [pebp+psize*3],pedx   ;is the key a predefined hkey?
je DontCloseThisOne
push peax ;save return value of openat

;free the key if not predefined
push pedx
call RegCloseKeycall

pop peax  ;restore return value

DontCloseThisOne:
pop pesi   ;restore the position of the name of the current subKey
;--pop pedx   ;the new hKey
mov pedx,peax ;new file handle
xor ecx,ecx
xor eax,eax
inc eax;simutale error return as RegOpenKeyExA

test pedx,pedx
cmovl pedx,pecx
jl HasFinalKey    ;error


xor eax,eax
add pesi,psize

mov pedi,pebp
add pedi,8*psize
cmp pesi,pedi
je HasFinalKey


cmp pdword [pesi],0
jne FindFinalKey
;--

HasFinalKey:
xchg pedx,peax
mov pecx,pedx

endOpenRegKey:
pop pedi
pop pesi
pop pedx
mov pesp,pebp
pop pebp
ret 6*psize
end if


;----------------------------------------------------
;                      ReadRegEx
;0x4  = hKey
;0x8  = SubKey
;0xC  = SubSubKey
;0x10 = SubSubSubKey
;0x14 = SubSubSubSubKey
;0x18 = ValueName
;
;output
;  (r/e)cx = size of pointer (r/e)ax
;  (r/e)dx = return value of RegQueryValueEx
;  (r/e)ax = Memlocation with the read data
;----------------------------------------------------

if used ReadRegEx
ReadRegEx:
push pebp
mov pebp,pesp
push pesi  ;save ESI
push pedi  ;save EDI
push pebx  ;save EBX


;open the register key
push pdword [pebp+psize*6]
push pdword [pebp+psize*5]
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
push 0
call OpenRegKey
xor edx,edx
test pecx,pecx
cmovnz pedx,pecx
cmovnz pecx,peax
jnz endReadRegEx

mov pecx,peax  ;R/ECX => hKey

mov pesi,pdword [pebp+psize*7]   ;valueName

call denter

argpushsys 1,pecx                    ;dir handle
argpushsys 2,pesi                    ;valueName
argpushsys 3,O_RDONLY                ;flags
linuxCall 257,295                    ;openat

call dleave
mov pedx,peax ;hFile

cmp pecx,pdword [pebp+psize*2]
je dontCloseKeyReadRegEx

  ;close the directory
  push pecx
  call RegCloseKeycall

dontCloseKeyReadRegEx:
xor ecx,ecx ;size
xor eax,eax
test pedx,pedx
jl endReadRegEx

push pedx
call getfilesize
mov pesi,peax

add pesi,0x10 ;add 10 bytes to be sure

;allocate memory to read the item out
push pesi
call GetMemLocation
push peax  ;push memory location

sub pesi,0x10 ;add 10 bytes to be sure

push pesi                   ;size
push peax                   ;buffer
push pedx                   ;hKey
call readfile

;close the file
push pedx
call RegCloseKeycall

mov ecx,esi  ;size
xor edx,edx ;return value
pop peax ;buffer location


endReadRegEx:
pop pebx ;restore R/EBX
pop pedi ;restore R/EDI
pop pesi ;restore R/ESI
mov pesp,pebp
pop pebp
ret psize*6
end if


;----------------------------------------------------
;                      WriteRegString
;0x4  = SubKey
;0x8  = ValueName
;0xC  = ValueData
;0x10 = ValueDataLen
;
;----------------------------------------------------

if used WriteRegString
WriteRegString:
call denter

push REG_SZ
push pdword [pebp+psize*5]
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push 0
push 0
push 0
push pdword [pebp+psize*2]
if used is64
  mov peax,HKEY_LOCAL_MACHINE
  push peax
else
  push HKEY_LOCAL_MACHINE ;use key Local Machine
end if
call WriteRegEx

call dleave
ret psize*4
end if

;----------------------------------------------------
;                      WriteRegDword
;0x4  = SubKey
;0x8  = ValueName
;0xC  = ValueData
;
;----------------------------------------------------

if used WriteRegDword
WriteRegDword:
call denter

push REG_DWORD
push 4
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push 0
push 0
push 0
push pdword [pebp+psize*2]
if used is64
  mov peax,HKEY_LOCAL_MACHINE
  push peax
else
  push HKEY_LOCAL_MACHINE ;use key Local Machine
end if
call WriteRegEx

call dleave
ret psize*3
end if

;----------------------------------------------------
;                      WriteRegBinary
;0x4  = SubKey
;0x8  = ValueName
;0xC  = ValueData
;0x10 = ValueDataLen
;
;----------------------------------------------------

if used WriteRegBinary
WriteRegBinary:
call denter

push REG_BINARY
push pdword [pebp+psize*5]
push pdword [pebp+psize*4]
push pdword [pebp+psize*3]
push 0
push 0
push 0
push pdword [pebp+psize*2]
if used is64
  mov peax,HKEY_LOCAL_MACHINE
  push peax
else
  push HKEY_LOCAL_MACHINE ;use key Local Machine
end if
call WriteRegEx

call dleave
ret psize*4
end if

;----------------------------------------------------
;                      WriteRegEx
;0x4  = hKey
;0x8  = SubKey
;0xC  = SubSubKey
;0x10 = SubSubSubKey
;0x14 = SubSubSubSubKey
;0x18 = ValueName
;0x1C = ValueData
;0x20 = ValueDataLen
;0x24 = ValueDataType
;
;----------------------------------------------------

if used WriteRegEx
WriteRegEx:
call denter

push pdword [pebp+psize*6]  ;SubSubSubSubKey
push pdword [pebp+psize*5]  ;SubSubSubKey
push pdword [pebp+psize*4]  ;SubSubKey
push pdword [pebp+psize*3]  ;SubKey
push pdword [pebp+psize*2]  ;hKey
push 1                      ;create the key
call OpenRegKey

xchg peax,pecx

test peax,peax
jnz endWriteRegEx

push pecx ;save open directory

argpushsys 1,pecx                    ;valueName
argpushsys 2,pdword [pebp+psize*7]   ;valueName
argpushsys 3,O_RDONLY or O_TRUNC or O_RDWR or O_CREAT
argpushsys 4,S_IRUSR or S_IWUSR or S_IRGRP or S_IWGRP or S_IROTH ;mode
linuxCall 257,295                    ;openat
mov pedx,peax ;hFile

pop pecx ;restore open directory
cmp pecx,pdword [pebp+psize*2]
je dontFreeopenDir

  push pecx
  call RegCloseKeycall

dontFreeopenDir:
test pedx,pedx
jl endWriteRegEx

push pdword [pebp+psize*9]   ;ValueLen
push pdword [pebp+psize*8]   ;ValueData
push pedx
call writefile

push pedx
call RegCloseKeycall

endWriteRegEx:
call dleave
ret 9*psize
end if

;----------------------------------------------------
;                      EnumRegKey
;1: hKey
;2: SubKey
;3: SubSubKey
;4: SubSubSubKey
;5: SubSubSubSubKey
;6: CallBackFunction:
;               arg: 1:hKey
;                    2:param
;                    3:Keyname
;                    4:parent key (is popped of from the stack by the caller
;7: Param for callbackfunction
;
;----------------------------------------------------
if used EnumRegKey
EnumRegKey:
call denter

push pdword [pebp+psize*6]  ;SubSubSubSubKey
push pdword [pebp+psize*5]  ;SubSubSubKey
push pdword [pebp+psize*4]  ;SubSubKey
push pdword [pebp+psize*3]  ;SubKey
push pdword [pebp+psize*2]  ;hKey
push 0                      ;create the key
call OpenRegKey
test pecx,pecx
jnz endEnumRegKey

mov pedi,peax ;R/EDI => handle to directory

sub pesp,1072
mov pesi,pesp ;return structure

push 0x1                    ;PATH_IS_HANDLE
push pesi                   ;return structure
push pedi                   ;hFile
call startDirectoryEnumerationEx
mov pebx,peax ;R/EBX => directory handle

test peax,peax
jz NoSubKeysEnumRegKey


loopThroughDirectoryKeys3:
  test dword [pesp],0x10
  jz isNoKey3

   cmp word [pesp+4+(7*2)+4+8],'.'
   je isNoKey3

   cmp word [pesp+4+(7*2)+4+8],'..'
   jne isNotPriviousDir3

   cmp byte [pesp+4+(7*2)+4+8+2],0
   je isNoKey3

   isNotPriviousDir3:

   lea pecx,pdword [pesp+4+(7*2)+4+8]
   call denter

   ;open 'subkey'
   argpushsys 1,pedi           ;hKey
   argpushsys 3,O_RDONLY or O_DIRECTORY
   argpushsys 2,pecx           ;subKey
   linuxCall 257,295           ;openat

   call dleave

   push peax ;push for RegCloseKeycall

   push pedi                  ; parent key
   push pecx                  ;name of the key
   push pdword [pebp+psize*8] ;param
   push peax                  ;hKey
   call pdword [pebp+psize*7] ;callBack function

   add pesp,psize ;remove parent key from stack

   call RegCloseKeycall


isNoKey3:
  push pesp ;return structure
  push pebx ;directory handle
  call nextDirectoryEnumeration
  test peax,peax
jnz loopThroughDirectoryKeys3

push 0x1                    ;PATH_IS_HANDLE
push pebx ;directory handle
call closeDirectoryEnumerationEx

NoSubKeysEnumRegKey:
cmp pedi,pdword [pebp+psize*2]  ;hKey
je dontFreeThisKey

  push pedi
  call RegCloseKeycall

dontFreeThisKey:

endEnumRegKey:
call dleave
ret psize*7
end if

;----------------------------------------------------
;                      EnumRegValue
;1: hKey
;2: SubKey
;3: SubSubKey
;4: SubSubSubKey
;5: SubSubSubSubKey
;6: CallBackFunction:
;               arg: 1:hKey
;                    2:param
;                    3:valueName
;                    4:value
;                    5:valueSize
;                    6:valueType
;7: Param for callbackfunction
;
;----------------------------------------------------
if used EnumRegValue
EnumRegValue:
xor eax,eax
xchg pdword [pesp+psize*7],peax ;swap param with flags
xchg pdword [pesp+psize*6],peax ;swap flags with CallBackFunction
xchg pdword [pesp+psize*5],peax ;swap CallBackFunction with SubSubSubSubKey
xchg pdword [pesp+psize*4],peax ;swap SubSubSubSubKey with SubSubSubKey
xchg pdword [pesp+psize*3],peax ;swap SubSubSubKey with SubSubKey
xchg pdword [pesp+psize*2],peax ;swap SubSubKey with SubKey
xchg pdword [pesp+psize*1],peax ;swap SubKey with hKey
xchg pdword [pesp],peax ;swap hKey with return value
push peax               ;push return value
end if


;----------------------------------------------------
;                      EnumRegValueEx
;1: hKey
;2: SubKey
;3: SubSubKey
;4: SubSubSubKey
;5: SubSubSubSubKey
;6: CallBackFunction:
;               arg: 1:hKey
;                    2:param
;                    3:valueName
;                    4:value
;                    5:valueSize
;                    6:valueType
;7: Param for callbackfunction
;8: flags:
;   0x1 DONT_READ_VALUE
;
;----------------------------------------------------
if used EnumRegValueEx | used EnumRegValue
EnumRegValueEx:
call denter

push pdword [pebp+psize*6]  ;SubSubSubSubKey
push pdword [pebp+psize*5]  ;SubSubSubKey
push pdword [pebp+psize*4]  ;SubSubKey
push pdword [pebp+psize*3]  ;SubKey
push pdword [pebp+psize*2]  ;hKey
push 0                      ;create the key
call OpenRegKey
test pecx,pecx
jnz endEnumRegValue
mov pedi,peax ;R/EDI => Hey, but with a k

sub pesp,1072
mov pesi,pesp ;return structure

push 0x1                    ;PATH_IS_HANDLE
push pesi                   ;return structure
push pedi                   ;hFile
call startDirectoryEnumerationEx
mov pebx,peax ;R/EBX => directory handle

test peax,peax
jz endEnumRegValueCloseDir

loopThroughDirectoryValues2:
  test dword [pesp],0x20000 ;file
  jz isNoValue4

   ;output
   ;  (r/e)cx = size of pointer (r/e)ax
   ;  (r/e)dx = return value of RegQueryValueEx
   ;  (r/e)ax = Memlocation with the read data
   xor eax,eax
   xor ecx,ecx
   lea pedx,pdword [pesp+4+(7*2)+4+8] ;filename

   test pdword [pebp+psize*9],0x1 ;dont read value
   jnz  dontReadValue


   push pedx
   push 0
   push pedi  ;hFile
   call ReadReg
   test pedx,pedx
   jnz isNoValue4

   push peax ;for FreeMemLocation

   lea pedx,pdword [pesp+4+(7*2)+4+8+psize] ;filename

   dontReadValue:

   push REG_SZ                ;value type
   push pecx                  ;value size
   push peax                  ;value
   push pedx                  ;name of the value
   push pdword [pebp+psize*8] ;param
   push pedi                  ;hKey
   call pdword [pebp+psize*7] ;callBack function

   test pdword [pebp+psize*9],0x1 ;dont read value
   jnz  dontFreeValue

     call FreeMemLocation
   dontFreeValue:

isNoValue4:
  push pesp ;return structure
  push pebx ;directory handle
  call nextDirectoryEnumeration
  test peax,peax
jnz loopThroughDirectoryValues2

push 0x1                    ;PATH_IS_HANDLE
push pebx ;directory handle
call closeDirectoryEnumerationEx

xor eax,eax

endEnumRegValueCloseDir:
cmp pedi,pdword [pebp+psize*2]
je dontFreeHeyWithAKey

  push pedi
  call RegCloseKeycall

dontFreeHeyWithAKey:

endEnumRegValue:
call dleave
ret psize*8
end if

;----------------------------------------------------
;This function deletes the a value from the registry
;input:
;  1: hKey
;  2: SubKey
;  3: SubSubKey
;  4: SubSubSubKey
;  5: SubSubSubSubKey
;  6: value name
;
;----------------------------------------------------

if used DeleteRegValue
DeleteRegValue:
call denter

;open registry key
push pdword [pebp+psize*6]  ;SubSubSubSubKey
push pdword [pebp+psize*5]  ;SubSubSubKey
push pdword [pebp+psize*4]  ;SubSubKey
push pdword [pebp+psize*3]  ;SubKey
push pdword [pebp+psize*2]  ;hKey
push 0                      ;create the key
call OpenRegKey
test pecx,pecx
jnz endDeleteRegValue
mov pebx,peax

mov pedi,pdword [pebp+psize*7]

call denter

argpushsys 3,0                     ;flags
argpushsys 2,pedi                  ;valueName
argpushsys 1,pebx                  ;dir handle
linuxCall 263,301                  ;unlinkat

call dleave

cmp pebx,pdword [pebp+psize*2]
je dontFreeKeyForDeleteRegValue

  push pebx
  call RegCloseKeycall

dontFreeKeyForDeleteRegValue:
endDeleteRegValue:
call dleave
ret psize*6
end if

;----------------------------------------------------
;This function deletes the sub keys of the key and the key itself
;input
;  1: hKey
;  2: SubKey
;
;----------------------------------------------------

if used DeleteRegKey
DeleteRegKey:
call denter

push 0x1 ;dont readvalue
push 0 ;param
ppush deleteValues
push 0
push 0
push 0
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
call EnumRegValueEx

push 0 ;param (original key)
ppush deleteKeyRecursive
push 0
push 0
push 0
push pdword [pebp+psize*3]
push pdword [pebp+psize*2]
call EnumRegKey

argpushsys 3,512                   ;flags   (AT_REMOVEDIR )(0x8?)
argpushsys 2,pdword [pebp+psize*3] ;valueName
argpushsys 1,pdword [pebp+psize*2] ;dir handle
linuxCall 263,301                  ;unlinkat


endDeleteRegKey:
call dleave
ret psize*2

;----------------------------------------------------
;     arg: 1:hKey
;          2:param
;          3:Keyname
;          4:Current key (is popped of from the stack by the caller
;----------------------------------------------------
deleteKeyRecursive:
call denter


push 0x1 ;dont readvalue
push 0 ;param
ppush deleteValues
push 0
push 0
push 0
push 0;pdword [pebp+psize*4]
push pdword [pebp+psize*2]
call EnumRegValueEx

push 0 ;param
ppush deleteKeyRecursive
push 0
push 0
push 0
push 0;pdword [pebp+psize*4]
push pdword [pebp+psize*2]
call EnumRegKey

argpushsys 3,512                   ;flags   (AT_REMOVEDIR )(0x8?)
argpushsys 2,pdword [pebp+psize*4] ;valueName
argpushsys 1,pdword [pebp+psize*5] ;dir handle
linuxCall 263,301                  ;unlinkat

call dleave
ret psize*3


;----------------------------------------------------
;     arg: 1:hKey
;          2:param
;          3:valueName
;          4:value
;          5:valueSize
;          6:valueType
;----------------------------------------------------
deleteValues:
call denter

argpushsys 3,0                     ;flags
argpushsys 2,pdword [pebp+psize*4] ;valueName
argpushsys 1,pdword [pebp+psize*2] ;dir handle
linuxCall 263,301                  ;unlinkat

call dleave
ret psize*6
end if