
;TV's server version 3.00
;Copyright (c) 2014, Tycho Veltmeijer.
;All rights reserved.

;callPoll events
POLLIN equ 0x1
POLLRDNORM equ POLLIN
POLLPRI equ 0x2
POLLOUT equ 0x4
POLLWRNORM equ POLLOUT
POLLERR equ 0x8
POLLHUP equ 0x10
POLLNVAL equ 0x20
POLLRDBAND equ 0x80
POLLWRBAND equ 0x200
POLLMSG equ 0x400
POLLRDHUP equ 0x2000

;------------------------------------------------------------------------------
;A bunch on functions called on the Windows that are not portabl to Linux
;param:
;  1 param
;return: -1 or a event descriptor
;-----------------------------------------------------------------------------
if used CreateSocketEvent | used CreateSocketEventClose | used CloseSocketEvent
CreateSocketEvent:
CreateSocketEventClose:
CloseSocketEvent:

mov peax,pdword [pesp+psize] ;return socket instead of event
ret psize
end if

;--ResetSocketEvent: seems to be unused in the project

;
;------------------------------------------------------------------------------
;This function will check if WSA event was signaled
;Since this functions does not exist in Linux and this project it's only used 
;to detect if connection is still alive, it does exectly that!
;param:
;  event: the event to be
;return: True or false
;-----------------------------------------------------------------------------
IsWSAEventSignaled:
call denter
;events bits, 0 on timeout and -1 on failure

push 0                        ;timeout
push POLLIN or POLLOUT or POLLRDHUP       ;event
push pdword [pebp+psize*2]    ;socket
call callPoll

xor ecx,ecx
cmp eax,-1
cmove eax,ecx
je endIsWSAEventSignaled

test ax,POLLRDHUP or POLLHUP
cmovz eax,ecx
jz endIsWSAEventSignaled

xor eax,eax
inc eax

endIsWSAEventSignaled:
call dleave
ret psize

;------------------------------------------------------------------------------
;This function will initilize the SOCKET stack
;return: 1 if NO error, else zero
;-----------------------------------------------------------------------------
if used initilize_socket_api
initilize_socket_api:
xor eax,eax
inc eax
ret
end if

;------------------------------------------------------------------------------
;This function will initilize a listen SOCKET
;default: IP v4 on TCP
;param:
;  flags: 0x1 Use IP v6
;         0x2 Use UDP
;  port:  The port to bind the socket to
;-----------------------------------------------------------------------------
if used CreateSocket
CreateSocket:
call denter

;set the address family
mov ebx,2  ;AF_INET
mov ecx,10 ;AF_INET6

mov edx,6   ;IPPROTO_TCP   (for tcp v4)
xor edi,edi ;IPPROTO_TCP   (for tcp v6)

test dword [pebp+psize*2] ,0x1 ;test the flag IPv6
cmovnz ebx,ecx ;set EBX on AF_INET6
cmovnz edx,edi

;set the type of socket and protocol
mov ecx,1   ;SOCK_STREAM
mov edi,2   ;SOCK_DGRAM

mov esi,17; IPPROTO_UDP
;-mov edx,6 ;IPPROTO_TCP

test dword [pebp+psize*2] ,0x2 ;test the flag UDP
cmovnz ecx,edi ;set edi on SOCK_DGRAM
cmovnz edx,esi ;set esi on IPPROTO_UDP

;Create the socket
if used is32
  push pedx ;IPPROTO_TCP or IPPROTO_UDP
  push pecx ;SOCK_STREAM or SOCK_DGRAM
  push pebx ;AF_INET or AF_INET6
  argpushsys 2,esp ;arg
  argpushsys 1,1 ;SYS_SOCKET
else
  argpushsys 3,0;pedx ;IPPROTO_TCP or IPPROTO_UDP
  argpushsys 2,pecx ;SOCK_STREAM or SOCK_DGRAM
  argpushsys 1,pebx ;AF_INET or AF_INET6
end if
linuxCall 41,102  ;sys_socket or 32:sys_socketcall
if used is32
  pop pebx
  add pesp,0x8
end if

test peax,peax
jnl noErrorSocket
;ppush noErrorSocket
;call setErrNo

plogmod LOG_ERROR,0,0,0,0,"Error creating socket, code: ",0x3,peax

xor eax,eax
jmp endCreateSocket

noErrorSocket:
mov pedx,peax

;size parameter
mov pesi,0x10

;make room on stack for sockaddr_in structure
sub pesp,0x10

test dword [pebp+psize*2] ,0x1 ;test the flag IPv6
jz createIPv4Structure

;add extra size on stack to create sockaddr_in6 structure
sub pesp,0x10
add pesi,0x10

createIPv4Structure:
mov pedi,pesp ;pointer to sockaddr_in(6)

;Zero the structure
xor eax,eax
mov ecx,esi ;size of the structure
rep stosb


;pointer to sockaddr_in(6)
mov pedi,pesp ;pointer to sockaddr_in(6)

;fill in member family
mov word [pedi],bx ;set family address


;Fill in member port
mov eax,dword [pebp+psize*3] ;get Port from stack
xchg al, ah ;convert to big-endian
mov word [pedi+2],ax ;fill in the port number

mov ebx,edx ;save socket on EBX

if used is32
  push pesi ;Size of sockaddr_in structure
  push pedi ;Pointer to sockaddr_in structure
  push pebx ;socket
  argpushsys 2,esp ;arg
  argpushsys 1,2 ;sys_bind
else
  argpushsys 3,pesi ;Size of sockaddr_in structure
  argpushsys 2,pedi ;Pointer to sockaddr_in structure
  argpushsys 1,pebx ;socket
end if
linuxCall 49,102  ;sys_bind or 32:sys_socketcall
if used is32
  pop pebx
  add pesp,0x8
end if

test peax,peax
jnl noErrorBind
;ppush noErrorBind
;call setErrNo

;jmp errCreateSocket

neg peax
plogmod LOG_ERROR,0,0,0,0,"Error binding socket, code: ",0x3,peax
jmp errCreateSocket

noErrorBind:

test dword [pebp+psize*2] ,0x2 ;test the flag UDP
jnz dontListen

if used is32
  push 0x7FFFFfff ;backlog
  push pebx        ;socket
  argpushsys 2,esp ;arg
  argpushsys 1,4           ;sys_listen
else
  argpushsys 2,0x7FFFFfff ;backlog
  argpushsys 1,pebx        ;socket
end if
linuxCall 50,102  ;sys_listen or 32:sys_socketcall
if used is32
  pop pebx
  add pesp,0x4
end if

test peax,peax
jnl noErrorListen; errCreateSocket
;ppush noErrorListen
;call setErrNo

;jmp errCreateSocket

neg peax
plogmod LOG_ERROR,0,0,0,0,"Error listen socket, code: ",0x3,peax
jmp errCreateSocket

noErrorListen:

dontListen:
mov peax,pebx

endCreateSocket:
call dleave
ret psize*2

;-
errCreateSocket:
argpushsys 1,pebx
linuxCall 3,6  ;sys_close

xor eax,eax
jmp endCreateSocket
end if

;------------------------------------------------------------------------------
;This function will wait for a client to connect
;param:
;  socket: the socket to listen to
;  sockaddr_in: pointer to sockaddr_in strucutre for remote address
;  addrlen: size of sockaddr_in
;return: -1 or a socket descriptor
;-----------------------------------------------------------------------------
if used Accept
Accept:
call denter

push pdword [pebp+psize*4]
mov pedx,pesp

recallAccept:
if used is32
  push pedx                  ;addrlen
  push pdword [pebp+psize*3] ;sockaddr_in
  push pdword [pebp+psize*2] ;Socket
  argpushsys 2,esp ;arg
  argpushsys 1,5 ;sys_accept
else
  argpushsys 3,pedx                  ;addrlen
  argpushsys 2,pdword [pebp+psize*3] ;sockaddr_in
  argpushsys 1,pdword [pebp+psize*2] ;Socket
end if
linuxCall 43,102  ;sys_accept or 32:sys_socketcall
if used is32
  add pesp,0xC
end if

;ppush endAccept
;call setErrNo
test peax,peax
jnl endAccept

neg peax

mov ebx,eax

push 1
call sleep

cmp ebx,4 ;EINTR
je recallAccept

cmp ebx,24 ;EMFILE
je recallAccept

cmp ebx,23 ;ENFILE
je recallAccept

cmp ebx,119 ;ENOBUFS
je recallAccept

cmp ebx,12 ;ENOMEM
je recallAccept

xor eax,eax
dec eax

endAccept:
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will return the local address on sockaddr_in structure
;param:
;  socket: the socket to listen to
;  sockaddr_in: pointer to sockaddr_in strucutre for local address
;  addrlen: size of sockaddr_in
;return: socket
;-----------------------------------------------------------------------------
if used GetLocalAddress
GetLocalAddress:
call denter

push pdword [pebp+psize*4]
mov pecx,pesp

mov pebx,pdword [pebp+psize*2] ;socket

if used is32
  push pecx
  push pdword [pebp+psize*3]
  push pebx ;sock
  argpushsys 2,esp ;arg
  argpushsys 1,6 ;SYS_GETSOCKNAME
else
  argpushsys 3,pecx
  argpushsys 2,pdword [pebp+psize*3]
  argpushsys 1,pebx
end if
linuxCall 51,102  ;sys_getsockname or 32:sys_socketcall
if used is32
  pop pebx
  add pesp,0x8
end if

mov peax,pebx ;return socket
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will close a socket, no shutdown or annything
;param:
;  Socket
;-----------------------------------------------------------------------------
if used CloseSocket | used CloseGeneralSocket
CloseSocket:
call denter

if used is32
  push 0;(=SHUT_RD)
  push pdword [pebp+psize*2]         ;SOCKET
  argpushsys 2,pesp                  ;arg
  argpushsys 1,13                    ;sys_shutdown
else
  argpushsys 2,0;(=SHUT_RD)          ;SHUT_WR
  argpushsys 1,pdword [pebp+psize*2] ;SOCKET
end if
linuxCall 48,102  ;shutdown or 32:sys_socketcall
if used is32
  add pesp,0x8
end if

mov pedi,pdword [pebp+psize*2]
jmp _closeSocketUnsafe
end if

;------------------------------------------------------------------------------
;This function will send and receive remaining data and finaly close the socket
;param:
;  Socket
;-----------------------------------------------------------------------------
if used SaflyCloseTCPSocket | used CloseSocket | used CloseGeneralSocket
SaflyCloseTCPSocket:
call denter

mov pedi,pdword [pebp+psize*2] ;SOCKET

if used is32
  push 1;(=SHUT_WR)          ;SHUT_WR
  push pedi                ;SOCKET
  argpushsys 2,esp ;arg
  argpushsys 1,13          ;sys_shutdown
else
  argpushsys 2,1;(=SHUT_WR)          ;SHUT_WR
  argpushsys 1,pedi                ;SOCKET
end if
linuxCall 48,102  ;shutdown or 32:sys_socketcall
if used is32
  add pesp,0x8
end if

sub pesp,80
if used is64
   mov r12,pesp
end if

recvRemainingData:
  if used is32
     mov pecx,pesp
  end if

  push pebp

if used is32
  push 0
  push 0
  push 0                     ;flags
  push 80                    ;length of buffer
  push ecx
  push pedi               ;SOCKET
  argpushsys 2,esp ;arg
  argpushsys 1,12                 ;sys_recvfrom
else
  argpushsys 6,0
  argpushsys 5,0
  argpushsys 4,0                     ;flags
  argpushsys 3,80                    ;length of buffer
  argpushsys 2,r12                   ;pointer to the buffer
  argpushsys 1,pedi               ;SOCKET
end if
linuxCall 45,102        ;recvfrom or 32:sys_socketcall
if used is32
  add pesp,0x18
end if

  pop pebp

test peax,peax
jz saflyCloseSocket

test peax,peax
jnl recvRemainingData
;ppush recvRemainingData
;call setErrNo

saflyCloseSocket:
_closeSocketUnsafe:
argpushsys 1,pedi
linuxCall 3,6  ;sys_close

call dleave
ret psize
end if

;------------------------------------------------------------------------------
;This function will send data
;param:
;  1:Socket
;  2:flags
;  3:timeout
;  4:buf
;  5:&len
;
;flags: RSF_RETURN_ON_ERROR (0x1) Return on error, do not try to resolve the error
;
;Return: CE_NOERR (0x0)
;        CE_NO_BUFFER_AVAILABLE (0x1)
;        CE_CONNECTION_CLOSED (0x2)
;        CE_UNKNOWN_ERROR (0x3)
;Return on &len: The data that is send (on success)
;-----------------------------------------------------------------------------
if used SendSocketData
SendSocketData:
call denter

;(R/E)BX = SOCKET
mov pebx,pdword [pebp+psize*2]
;(R/E)SI = buf
mov pesi,pdword [pebp+psize*5]
;(R/E)DI = &len
mov pedi,pdword [pebp+psize*6]

xor eax,eax
cmp dword [pedi],0
je successSendSocketData

retrySendSocketData:
push 0                     ;timeout
push POLLOUT or POLLRDHUP  ;event
push pebx                  ;socket
call callPoll

mov ecx,CE_UNKNOWN_ERROR
cmp eax,-1
cmove eax,ecx
je endSendSocketData

test eax,eax
je timeBlockedConnection

test ax,POLLRDHUP or POLLHUP
cmovnz eax,ecx
jnz endSendSocketData

test ax,POLLOUT
jz timeBlockedConnection

push pebp
push pesi
push pedi

if used is32
  push 0
  push 0
  push 0x40        ;flags = ;MSG_DONTWAIT
  push dword [pedi];len
  push pesi        ;Buff
  push pebx        ;SOCKET
  argpushsys 2,esp ;arg
  argpushsys 1,11  ;sendto
else
  argpushsys 6,0
  argpushsys 5,0
  argpushsys 4,0x40        ;flags = ;MSG_DONTWAIT
  mov edx,dword [pedi]     ;len;argpushsys 3,dword [pedi]  ;len
  argpushsys 2,pesi        ;Buff
  argpushsys 1,pebx        ;SOCKET
end if
linuxCall 44,102 ;sendto or 32:sys_socketcall
if used is32
  pop pebx
  add pesp,0x14
end if

pop pedi
pop pesi
pop pebp

test peax,peax
jz sendRetZero

;Get the last Socket error
test peax,peax
jnl successSendSocketData
;ppush successSendSocketData
;call setErrNo

neg peax

mov ecx,CE_UNKNOWN_ERROR

;if RSF_RETURN_ON_ERROR is set we do not resolve any error
test pdword [pebp+psize*3],RSF_RETURN_ON_ERROR
cmovnz peax,pecx ;set error code and return if the flag is set
jnz endSendSocketData

;Try to resolve the error
sendRetZero:
cmp eax,140 ;EWOULDBLOCK
je timeBlockedConnection

cmp eax,119 ;ENOBUFS
je timeBlockedConnection

cmp eax,12 ;ENOMEM
je timeBlockedConnection

mov ecx,CE_NO_BUFFER_AVAILABLE
cmp eax,115 ;EMSGSIZE
cmovz eax,ecx
jz endSendSocketData

;cmp eax,14 ;EFAULT
;je retrySendSocketData

mov eax,CE_CONNECTION_CLOSED
jmp endSendSocketData

timeBlockedConnection:
;wait 30 ms for the connectio to unblock
push 30
call sleep

;test if RSF_USE_TIMEOUT is set
test dword [pebp+psize*3],RSF_USE_TIMEOUT
jz retrySendSocketData

;substract 30 ms from the maximum allowed timeout
sub dword [pebp+psize*4],30
jnc retrySendSocketData

mov eax,CE_TIMEOUT
jmp endSendSocketData

successSendSocketData:
mov dword [pedi],eax

xor eax,eax ;no error

endSendSocketData:
call dleave
ret 5*psize
end if

;------------------------------------------------------------------------------
;This function will wait until the socket can be read OR when a timeout happend
;param:
;  1:Socket
;  2:events
;  3:timeout [word] wait amount of miliseconds
;   -1 for infinitive wait
;    0 for imidialty return


;Return on AX:
; success: events bits, 0 on timeout and -1 on failure

;-----------------------------------------------------------------------------
if used callPoll
callPoll:
call denter

lea pedi, pdword [pebp+psize*2] ;the polfd structure
if used is64
  xor eax,eax
  mov ax,word [pebp+psize*3]
  mov word [rdi+4],ax
end if

argpushsys 3,pdword [pebp+psize*4] ;timeout
argpushsys 2,1                     ;nfds
argpushsys 1,pedi                   ;pollfd
linuxCall 7,168 ;poll

;Get the last Socket error
test peax,peax
jnl pollSuccess
;ppush pollSuccess
;call setErrNo

mov eax,-1
jmp endCallPoll

pollSuccess:
test eax,eax
jz endCallPoll

xor eax,eax
mov ax,word [(pebp+psize*2)+6] ;return event

endCallPoll:
call dleave
ret psize*3
end if

;------------------------------------------------------------------------------
;This function will receive data
;param:
;  1:Socket
;  2:flags
;  3:timeout
;  4:buf
;  5:&len
;
;flags: RSF_RETURN_ON_ERROR (0x1) Return on error, do not try to resolve the error
;       RSF_PEEK_DATA (0x2) Leave the data on the internal buffer
;       RSF_RETURN_ON_FIRST_SUCCESSFUL_ATTEMPT (0x4) return afther one successful try receiving data from the client, do not check if 'size' is received
;
;Return: CE_NOERR (0x0)
;        CE_NO_BUFFER_AVAILABLE (0x1)
;        CE_CONNECTION_CLOSED (0x2)
;        CE_UNKNOWN_ERROR (0x3)

;Return on &len: The data that is received (on success)
;-----------------------------------------------------------------------------
if used RecvSocketData
RecvSocketData:
call denter

;(R/E)BX = SOCKET
mov pebx,pdword [pebp+psize*2]
;(R/E)DI = &len
mov pedi,pdword [pebp+psize*6]

; we cannot use MSG_WAITALL (0x100) since we need to keep track of the timeout value

;get flags
mov esi,dword [pebp+psize*3]

call getTickCount

;r12 or ECX = start time of this function
;r13 or EDX = received len
if used is32
   mov ecx,eax
   xor edx,edx
else
   mov r12,rax
   xor r13d,r13d
end if

;loopReceiveDataWithSaveTickCount:
if used is32
   push ecx      ;save tickcount
end if

xor eax,eax

mov ecx,esi
and ecx,RSF_PEEK_DATA    ;since RSF_PEEK_DATA == MSG_PEEK we can use it directly for recv

if used is32
   cmovnz edx,eax       ;zero len if RSF_PEEK_DATA is set
   push edx             ;save len value
else
   cmovnz r13d,eax      ;zero received buffer if RSF_PEEK_DATA is set
end if

;skip the sleep command the first time
jmp rcvSkipSleep

loopReceiveData:
;Set thread in sleep state (so we do not eat CPU)
push 1
call sleep

rcvSkipSleep:

mov ecx,esi
and ecx,RSF_PEEK_DATA    ;since RSF_PEEK_DATA == MSG_PEEK

mov eax,dword [pedi]    ;EAX = len

if used is32
   pop edx
   jnz peekFlagIsSet
     sub eax,edx          ;len - received buffer
   peekFlagIsSet:
   push edx
else
   jnz peekFlagIsSet
     sub eax,r13d         ;len - received buffer
   peekFlagIsSet:
end if

mov edx,eax

push pesi

mov pesi,pdword [pebp+psize*5]

or pecx,0x40;MSG_DONTWAIT

call denter

if used is32
  push 0
  push 0
  push pecx               ;flags
  push pedx               ;length of buffer
  push pesi               ;buff
  push pebx               ;SOCKET
  argpushsys 2,esp ;arg
  argpushsys 1,12         ;recvfrom
else
  argpushsys 6,0          ;
  argpushsys 5,0          ;
  argpushsys 4,pecx       ;flags
  argpushsys 3,pedx       ;len
  argpushsys 2,pesi       ;buff
  argpushsys 1,pebx       ;SOCKET
end if
linuxCall 45,102          ;recvfrom or 32:sys_socketcall
if used is32
  add pesp,0x18
end if

call dleave
pop pesi

;test peax,peax
;jz noErrorRecv

;Get the last Socket error
test peax,peax
jnl noErrorRecv

neg peax
;ppush noErrorRecv
;call setErrNo

;An error accord, return if RSF_RETURN_ON_ERROR is set
mov ecx,CE_UNKNOWN_ERROR

test esi,RSF_RETURN_ON_ERROR
cmovnz eax,ecx
jnz endRecvSocketData

recvRetZero:
push peax

push 1
call sleep

pop peax

cmp eax,140;EWOULDBLOCK
je checkOnTimeout

cmp eax,11;EAGAIN
je checkOnTimeout

cmp eax,12;ENOMEM
je checkOnTimeout

mov ecx,CE_CONNECTION_CLOSED
cmp eax,126 ;ENOTCONN
cmovz eax,ecx
jz endRecvSocketData

mov eax,CE_UNKNOWN_ERROR
jmp endRecvSocketData

noErrorRecv:
test esi,RSF_RETURN_ON_FIRST_SUCCESSFUL_ATTEMPT
jz dontReturnRecvSocketData

test eax,eax ;if no data received
jz checkOnTimeout

mov dword [pedi],eax  ;set &len
xor eax,eax ;CE_NO_ERROR
jmp endRecvSocketData

dontReturnRecvSocketData:
xor ecx,ecx  ;CE_NOERROR

;if PEEK flag is set, don't do ADD, do a move operation
test esi,RSF_PEEK_DATA

if used is32
   pop edx         ;pop received len
   cmovnz edx,eax  ;move data if PEEK flag is set (the next time it will give the same data, again and again and again)_
   jnz dontAddSizeSinceItIsPeek
   add edx,eax     ;add the received len from this try
   dontAddSizeSinceItIsPeek:
   push edx             ;save len value
   ;compair if the buffer is full
   cmp edx,dword [pedi]
else
   cmovnz r13d,eax  ;move data if PEEK flag is set (the next time it will give the same data, again and again and again)_
   jnz dontAddSizeSinceItIsPeek
   add r13d,eax    ;add the received len from this try
   dontAddSizeSinceItIsPeek:
   ;compair if the buffer is full
   cmp r13d,dword [pedi]
end if

cmovz eax,ecx  ;Set CE_NOERROR    ;cmovnc
jz endRecvSocketData              ;jae

test esi,RSF_PEEK_DATA    ;test on RSF_PEEK_DATA
jnz checkOnTimeout         ;jump if flag is set

;move the buffer +received bytes
add pdword [pebp+psize*5],peax

checkOnTimeout:
;check if we need to check on timeout
mov edx,dword [pebp+psize*4]  ;get timeout value
test edx,edx
jz loopReceiveData ;no Timout

call getTickCount

if used is32
   sub eax,dword [pesp+0x4] ;restore start time of this function and substract it from current time
else
   sub eax,r12d
end if

cmp eax,dword [pebp+psize*4] ;check with TimeOut value
jb loopReceiveData ;not yet timed out

_TIMEOUT:
mov eax,CE_TIMEOUT ;err code

setLenValueEndRecvSocketData:
if used is32
   pop edx               ;pop received len
   mov dword [pedi],edx  ;set &len value
else
   mov dword [pedi],r13d ;set &len value
end if

endRecvSocketData:
call dleave
ret psize*5
end if


;------------------------------------------------------------------------------
;This function will return the size of the send buffer for this socket
;param:
;  1:Socket
;
;Return: The size of the send buffer for this socket or zero

;-----------------------------------------------------------------------------
if used GetSendBufferSize
GetSendBufferSize:
push 7                     ;SO_SNDBUF
push pdword [pesp+psize*2] ;socket
call CallGetSockOpt
ret psize
end if


;------------------------------------------------------------------------------
;This function will return the size of the receive buffer for this socket
;param:
;  1:Socket
;
;Return: The size of the send buffer for this socket or zero

;-----------------------------------------------------------------------------
if used GetRecvBufferSize
GetRecvBufferSize:
push 0x8                   ;SO_RCVBUF
push pdword [pesp+psize*2] ;socket
call CallGetSockOpt
ret psize
end if

;------------------------------------------------------------------------------
;This function will call getsockopt with intheger as result
;param:
;  1:Socket
;  2:optname
;
;Return: The soptval or zero on error

;-----------------------------------------------------------------------------
if used CallGetSockOpt
CallGetSockOpt:
call denter

push 0
mov peax,pesp ;*optval
push 4      ;size of an integer

if used is32 
  
  push esp                     ;*optlen
  push eax                     ;*optval
  push pdword [pebp+psize*3]   ;optname
  push 1 ;SOL_SOCKET           ;level
  push pdword [pebp+psize*2]   ;socket
  argpushsys 2,esp ;arg
  argpushsys 1,15              ;getsockopt
else
  argpushsys 5,rsp                     ;*optlen
  argpushsys 4,rax                     ;*optval
  argpushsys 3,pdword [pebp+psize*3]   ;optname
  argpushsys 2,1 ;SOL_SOCKET           ;level
  argpushsys 1,pdword [pebp+psize*2]   ;socket
end if
linuxCall 55,102          ;getsockopt or 32:sys_socketcall
if used is32
  add pesp,0x14
end if

pop pecx ;pop size
pop pecx ;pop optval

;Get the last Socket error
test peax,peax
jnl endCallGetSockOpt
;ppush endCallGetSockOpt
;call setErrNo

xor ecx,ecx ;error

endCallGetSockOpt:
mov peax,pecx
call dleave
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will set the size of the send buffer for this socket
;param:
;  1:Socket
;  2:New send buffer size
;
;Return: Zero on succes

;-----------------------------------------------------------------------------
if used SetSendBufferSize
SetSendBufferSize:
push pdword [pesp+psize*2] ;New send buffer size
push 7 ;SO_SNDBUF
push pdword [pesp+psize*3] ;socket
call CallSetSockOpt
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will set the size of the receive buffer for this socket
;param:
;  1:Socket
;  2:New send buffer size
;
;Return: Zero on succes

;-----------------------------------------------------------------------------
if used SetRecvBufferSize
SetRecvBufferSize:
push pdword [pesp+psize*2] ;New receive buffer size
push 0x8 ;SO_RCVBUF
push pdword [pesp+psize*3] ;socket
call CallSetSockOpt
ret psize*2
end if

;------------------------------------------------------------------------------
;This function will call setsockopt for intheger options
;param:
;  1:Socket
;  2:optname
;  3:new value
;
;Return: Zero on succes

;-----------------------------------------------------------------------------
if used CallSetSockOpt
CallSetSockOpt:
pop peax
push 1 ;SOL_SOCKET
push peax
end if

;------------------------------------------------------------------------------
;This function will call setsockopt for intheger options
;param:
;  1:level
;  2:Socket
;  3:optname
;  4:new value
;
;Return: Zero on succes

;-----------------------------------------------------------------------------
if used CallSetSockOptEx | used CallSetSockOpt
CallSetSockOptEx:
call denter

lea peax,pdword [pebp+psize*5]   ;optval (new value)

if used is32 
  
  push 4                       ;optlen
  push eax                     ;*optval
  push pdword [pebp+psize*4]   ;optname
  push pdword [pebp+psize*2]   ;level
  push pdword [pebp+psize*3]   ;socket
  argpushsys 2,esp ;arg
  argpushsys 1,14              ;setsockopt
else
  argpushsys 5,4                       ;optlen
  argpushsys 4,rax                     ;*optval
  argpushsys 3,pdword [pebp+psize*4]   ;optname
  argpushsys 2,pdword [pebp+psize*2]   ;level
  argpushsys 1,pdword [pebp+psize*3]   ;socket
end if
linuxCall 54,102          ;setsockopt or 32:sys_socketcall
if used is32
  add pesp,0x14
end if

call dleave
ret psize*4
end if

;------------------------------------------------------------------------------
;This function will set keep-alve options
;param:
;  1:socket
;  2:onoff
;  3:keepalivetime
;  4:keepaliveinterval
;
;Return: boolean

;-----------------------------------------------------------------------------
SetKeepAlive:
call denter

push pdword [pebp+psize*3]   ;on or off
push 9                       ;SO_KEEPALIVE
push pdword [pebp+psize*2]   ;socket
call CallSetSockOpt

push pdword [pebp+psize*4]   ;keepalivetime
push 4                       ;TCP_KEEPIDLE
push pdword [pebp+psize*2]   ;socket
push 6                       ;SOL_TCP
call CallSetSockOptEx

push pdword [pebp+psize*5]   ;keepalivetime
push 5                       ;TCP_KEEPINTVL
push pdword [pebp+psize*2]   ;socket
push 6                       ;SOL_TCP
call CallSetSockOptEx

call dleave
ret psize*4