
;TV's server version 3.20
;Copyright (c) 2014 - 2022, Tycho Veltmeijer.
;All rights reserved.

;----------------------------------------------------
;                      LaunchCGI
;Launches CGI executable, send POST-command to CGI-application and send the header and result to the client
;
;input:
;    1:ProtocolSession
;    2:ClientSession
;    3:Pointer to a string with the CGI_path
;    4:Pointer to a string with the Enviroment
;    5:buffer size
;    6:Maximum timeout CGI-application
;    7:Pointer to a HTTPSession structure
;    8:Pointer to a ResponseHeader structure
;    9:flags:
;      0x1 use UTF-8 in environment instead of TF1-6

;
;output: (R/E)AX: True/False
;
;----------------------------------------------------

LaunchCGI:
call denter
;Create space on stack:
;--psize*3: SECURITY_ATTRIBUTES
;--psize*2: Return value for CreatePipe "CGI return Pipe"
;--psize*2: Return value for CreatePipe "Pipe For Post Command"
;--(...)  : StartupInfo
;--(psize*2+4*2): PROCESS_INFORMATION
sub pesp,psize*3 + psize*2 + psize*2 + (psize + psize*3 + 8*4 + psize + psize*4) + (psize*2+4*2)
mov pebx,pesp ;save stack pointer
MakeStack

if used is32
  cmp [is9x],1
  je dontConvertString__LaunchCGI
end if

test pdword [pebp+psize*10],0x1
jnz useUTF8Environment_1

  ;ENVIORNMENT to UTF-16
  push 0
  push pdword [pebp+psize*5]
  push 0x12
  call CreateOneString
  mov pdword [pebp+psize*5],peax

useUTF8Environment_1:
;CGI_path to UTF-16
push 0
push pdword [pebp+psize*4]
push 0x11
call CreateOneString
mov pdword [pebp+psize*4],peax

dontConvertString__LaunchCGI:
;--Fill in SECURITY_ATTRIBUTES--
mov dword [pebx],psize*3   ;set nLength
mov pdword [pebx+psize],0  ;set lpSecurityDescriptor
mov dword [pebx+psize*2],1 ;set bInheritHandle

;--Create CGI return Pipe--
argpush 4,0          ;nSize
argpush 3,pebx       ;lpPipeAttributes
lea pedx,pdword [pebx+psize*3]
argpush 2,pedx       ;hWritePipe
lea pecx,pdword [pebx+psize*4]
argpush 1,pecx       ;hReadPipe
call [CreatePipe]
test eax,eax
jz endLaunchCGI

;--Creating Pipe For Post Command--
argpush 4,0          ;nSize
argpush 3,pebx       ;lpPipeAttributes
lea pedx,pdword [pebx+psize*5]
argpush 2,pedx       ;hWritePipe
lea pecx,pdword [pebx+psize*6]
argpush 1,pecx       ;hReadPipe
call [CreatePipe]
test eax,eax
jz endLaunchCGIDeleteCGIPipe

;--Fill in StartupInfo--
lea pedi,pdword [pebx+psize*7] ;(R/E)DI => pointer to StartupInfo
mov ecx,(psize + psize*3 + 8*4 + psize + psize*4)
xor al,al
rep stosb ;zero StartupInfo
lea pedi,pdword [pebx+psize*7] ;(R/E)DI => pointer to StartupInfo

mov pdword [pedi],(psize + psize*3 + 8*4 + psize + psize*4) ;set cb

mov pecx,pdword [pebx+psize*6]  ;hReadPipe
mov pdword [pedi+psize + psize*3 + 8*4 + psize + psize],pecx ;set hStdInput

mov pecx,pdword [pebx+psize*3] ;hWritePipe
mov pdword [pedi+psize + psize*3 + 8*4 + psize + psize*2],pecx ;set hStdOutput
;mov pdword [pedi+psize + psize*3 + 8*4 + psize + psize*3],pecx ;set hStdError

mov dword [pedi+psize + psize*3 + 7*4],STARTF_USESHOWWINDOW  or STARTF_USESTDHANDLES ;set dwFlags  ;

lea pesi,pdword [pebx+16*psize+8*4] ;(R/E)SI => pointer to PROCESS_INFORMATION
mov pdword [pesi],0
mov pdword [pesi+psize],0
mov pdword [pesi+psize*2],0
if used is32
  mov dword [pesi+psize*3],0
end if

;Start CGI-application
argpush 10,pesi                      ;Process Info
argpush 9,pedi                       ;Startup Info
argpush 8,0                          ;Working directory
if used is32
  argpush 7,pdword [pebp+psize*5]    ;Environment
else
  mov pecx,pdword [pebp+psize*5]
  argpush 7,pecx                     ;Environment
end if

if used is32
  cmp [is9x],0
  je useUnicodeEnvironment

  argpush 6,NORMAL_PRIORITY_CLASS
  jmp usedANSIEnvironment
end if

useUnicodeEnvironment:
test pdword [pebp+psize*10],0x1
jz useUTF16Environment_1

  argpush 6,NORMAL_PRIORITY_CLASS     ;Flags
  jmp useUTF8Environment_2

useUTF16Environment_1:
  argpush 6,NORMAL_PRIORITY_CLASS or 0x00000400 ;CREATE_UNICODE_ENVIRONMENT     ;Flags

useUTF8Environment_2:
usedANSIEnvironment:
argpush 5,1                          ;Use Handle(s)
argpush 4,pebx                       ;SECURITY_ATTRIBUTES
argpush 3,pebx                       ;SECURITY_ATTRIBUTES
argpush 2,0                          ;COMMANDLINE
argpush 1,pdword [pebp+psize*4]      ;CGI-APP
if used is32
  cmp [is9x],0
  je useWCreateProcess

  call [CreateProcessA]
  jmp usedACreateProcess
end if

useWCreateProcess:
call [CreateProcessW]
usedACreateProcess:
test eax,eax
jz endLaunchCGIDeletePostPipe



mov pedx,pdword [pebp+psize*9] ;(R/E)DX => ResponeHeader
if used EnableChunking
  or dword [pedx],FRH_DONT_END_HEADER or FRH_DONT_SEND_CONTENT_TYPE or FRH_TRANSFER_ENCODING_CHUNKED ;add flags
else
  or dword [pedx],FRH_DONT_END_HEADER or FRH_DONT_SEND_CONTENT_TYPE  ;add flags
end if

if ~ used EnableChunking
;-;delete FRH_CONNECTION_KEEP_ALIVE flags
test dword [pedx],FRH_CONNECTION_KEEP_ALIVE
jz dontKeepConnectionAlive

  xor dword [pedx],FRH_CONNECTION_KEEP_ALIVE

dontKeepConnectionAlive:
end if
;check if a part of the POST command is in the buffer existing in HTTPSession
mov pedx,pdword [pebp+psize*8] ;(R/E)DX => HTTPSession
mov ecx,dword [pedx+psize*7+4] ;ECX => headerRestSize

jecxz noPostCommandInBuffer

  mov peax,pdword [pedx+psize*4] ;buffer
  if used is32
    add eax,dword [pedx+psize*7]  ;add headerSize to get pointer to headerRest
  else
    mov edx,dword [pedx+psize*7]  ;EDX => headerSize
    add rax,rdx                   ;add headerSize to get pointer to headerRest
  end if

  ;Write this POST command
  lea pedx,pdword [pebx+psize*7] ;(R/E)DI => Reuse the pointer to StartupInfo for lpNumberOfBytesWritten

  argpush 5,0                     ;lpOverlapped
  argpush 4,pedx                  ;lpNumberOfBytesWritten
  argpush 3,pecx                  ;nNumberOfBytesToWrite
  argpush 2,peax                  ;lpBuffer
  argpush 1,pdword [pebx+psize*5] ;hWritePipe
  call [WriteFile]


noPostCommandInBuffer:
;--;receive the size of one page
;--call GetSmallestPortionPageAllocation
mov eax,dword [pebp+psize*6] ;buffersize

if used is32
  mov ecx,eax ;ECX = size of buffer
else
  mov r12,rax ;r12 => size of buffer
end if

;allocate buffer
push peax
call GetMemLocation
mov pedi,peax ;(R/E)DI => pointer to buffer

if used is32
  sub ecx,4            ;substract the four bytes used by size of buffer integer
  mov dword [pedi],ecx ;dword [edi] = size of buffer
  add pedi,4           ;dword [edi-4] = size of buffer
end if

call [GetTickCount]
if used is32
  mov pdword [pebx+psize*9],eax ;(Reuse third member StartupInfo for the TickCount
else
  mov r13d,eax                  ;r13 => tickCount
end if

waitForInput:
  argpush 1,10
  call [Sleep] ;Wait for 10 miliseconds

  mov edx,dword [pebp+psize*7] ;EDX => maximum timeout
  test edx,edx
  jz unlimitedExecutionTime

    call [GetTickCount]
    mov edx,dword [pebp+psize*7] ;EDX => maximum timeout
    if used is32
      sub eax,pdword [pebx+psize*9] ;Tickcount - privious timeout => time CGI-application is running
    else
      sub eax,r13d                  ;Tickcount - privious timeout => time CGI-application is running
    end if
    cmp eax,edx
    ja timeoutReached

  unlimitedExecutionTime:
  ;check if connection is still alive
  push 1   ;buf (set to 'true' for if plugin doesn't support flag CSI_CONNECTION_ALIVE)
  mov peax,pesp

  push 4   ;len

  push pesp                                   ;*len
  push peax                                   ;*buf
  push CSI_CONNECTION_ALIVE                   ;type
  push pdword [pebp+psize*3]                  ;ClientSession
  push pdword [pebp+psize*2]                  ;ProtocolSession
  call [GetClientSessionInfo]
  test eax,eax
  jz connectionStillAlive

  pop peax ;len
  pop peax ;buf
  test peax,peax
  jz timeoutReached ;connection is not alive anymore

  connectionStillAlive:
  mov pedx,pdword [pebp+psize*8] ;(R/E)DX => HTTPSession
  mov pedx,pdword [pedx]         ;(R/E)DX => defaultVariables
  mov pedx,pdword [pedx+psize]   ;EDX => method
  cmp dword [pedx],'POST'        ;check if method is POST
  jne noPostCommand
    if used is32
      push dword [pedi-4]       ;len
    else
      push r12                 ;len
    end if

    push pesp                                   ;&len
    push pedi                                   ;buf
    push 100                                    ;timeout
    push RSF_RETURN_ON_FIRST_SUCCESSFUL_ATTEMPT ;flags
    push pdword [pebp+psize*3]                  ;ClientSession
    push pdword [pebp+psize*2]                  ;ProtocolSession
    call [RecvData]
    test eax,eax
    jz noRecvError

    cmp eax,3
    jbe timeoutReached

  noRecvError:
    pop peax ;Received data (len)

    test peax,peax
    jz noPostCommand
    ;jmp noPostCommand

    lea pecx,pdword [pebx+psize*7] ;(R/E)DI => Reuse the pointer to StartupInfo for lpNumberOfBytesWritten

    argpush 5,0                     ;lpOverlapped
    argpush 4,pecx                  ;lpNumberOfBytesWritten
    argpush 3,peax                  ;nNumberOfBytesToWrite
    argpush 2,pedi                  ;lpBuffer
    argpush 1,pdword [pebx+psize*5] ;hWritePipe
    call [WriteFile]

  noPostCommand:
  argpush 2,0                     ;HighSize
  argpush 1,pdword [pebx+psize*4] ;hReadPipe
  call [GetFileSize]
  test eax,eax
  jnz ReadBuffer

  lea pecx,pdword [pebx+psize*7] ;(R/E)CX => Reuse the pointer to StartupInfo for GetExitProcess

  argpush 2,pecx          ;lpExitCode
  argpush 1,pdword [pesi]  ;hProcess
  call [GetExitCodeProcess]

  mov ecx,dword [pebx+psize*7] ;ExitCode

  cmp ecx,103h ;STILL_ACTIVE
je waitForInput
jmp endLaunchCGIDeleteBufferEndProcessHandles

ReadBuffer:
  ;if(lenBuf<FileSize) FileSize=lenBuf;
  if used is32
    cmp dword [pedi-4],eax
    jae dontChangeReadSize

      mov eax,dword [pedi-4]

    dontChangeReadSize:
  else
    cmp r12,rax
    cmovb rax,r12
  end if

  lea pecx,pdword [pebx+psize*7] ;(R/E)DI => Reuse the pointer to StartupInfo for lpNumberOfBytesRead

  argpush 5,0                     ;lpOverlapped
  argpush 4,pecx                  ;lpNumberOfBytesRead
  argpush 3,peax                  ;nNumberOfBytesToRead
  argpush 2,pedi                  ;lpBuffer
  argpush 1,pdword [pebx+psize*4] ;hReadPipe
  call [ReadFile]

  cmp pdword [pebp+psize*9],0 ;ResponseHeader is set to zero is the header is send
  je HeaderIsAlreadySend

  cmp pdword [pebp+psize*9],1 ;ResponseHeader is set to one is the header is send
  je HeaderIsAlreadySend
    ;mov eax,0x0A0D0A0D ;two new lines
    mov pdword [pebx+psize*8],0 ;(Reuse second member StartupInfo for the pointer that is set to zero for the Status
    mov pecx,pdword [pebx+psize*7]   ;lpNumberOfBytesRead
    mov pedx,pedi ;(R/E)DX => pointer to buffer
    mov peax,pedi ;(R/E)AX => pointer to buffer

    loopBuffer:
      cmp pecx,4 ;check if size of bufer > 4
      jb endHeaderReached

      cmp dword [pedx],0x0A0D0A0D ;check on new lines
      je endHeaderReached

      cmp byte [pedx],':'
      jne nextLoopBuffer

        mov byte [pedx],0x00 ;set zero byte on ':' location

        ;generate hash from the name
        push peax ;push for function GenerateHashCaseInsensitiveSafe

        ;find next next new line
        mov peax,pedx ;(R/E)AX => location of ':'

        findNewLine:
          cmp ecx,2
          jb endFindNewLine

          cmp word [pedx],0x0A0D
          je FoundNewLine

          inc pedx
        loop findNewLine

      endFindNewLine:
        add pesp,psize ;remove the push for GenerateHashCaseInsensitiveSafe
        mov byte [peax],':' ;set ':' back on it's location
        jmp endHeaderReached

      FoundNewLine:
        ;swap pointer to ':' and the pointer for function GenerateHashCaseInsensitiveSafe
        xchg peax,pdword [pesp]
        push peax    ;repush for function GenerateHashCaseInsensitiveSafe
        call GenerateHashCaseInsensitiveSafe
        cmp eax,0x45AF5DAD   ;hash of 'status'

        pop peax
        mov byte [peax],':' ;set ':' back on it's location

        jne findNextNewLine

          inc peax ;skip ':'
          skipSpaces:
            cmp byte [peax],' '
            jne endSkipSpaces

            inc peax
          endSkipSpaces:

          mov byte [pedx],0
          mov pdword [pebx+psize*8],pedx  ;(Second member of StartupInfo) => pointer to replaced 0x0D byte

          mov pedx,pdword [pebp+psize*9] ;(R/E)DX => response header
          mov pdword [pedx+8],peax       ;set status code
          jmp endHeaderReached


        findNextNewLine:
        mov peax,pedx
        add peax,2 ;skip new line and send pointer to fname of CGI-variable

    nextLoopBuffer:
      inc pedx
    loop loopBuffer

  endHeaderReached:
    push DEFAULT_SENDTIMEOUT   ;timeout
    push pdword [pebp+psize*9] ;ResponseHeader
    push pdword [pebp+psize*3] ;ClientSession
    push pdword [pebp+psize*2] ;ProtocolSession
    call [SendHTTPHeader]


    mov peax,pdword [pebx+psize*8] ;(R/E)AX => pointer to replaced 0x0D byte
    test peax,peax
    jz statusWasNotSet

      mov byte [peax],0x0D           ;place the byte back
    statusWasNotSet:

    if used EnableChunking
      mov pdword [pebp+psize*9],1    ;set responseHeader to 1 as a signle to find the new line
    else
      mov pdword [pebp+psize*9],0    ;set responseHeader to 0
    end if

  HeaderIsAlreadySend:
  if used EnableChunking
  cmp pdword [pebp+psize*9],0
  je sendDataWithChunk

  mov pecx,pdword [pebx+psize*7]   ;lpNumberOfBytesRead
  mov pedx,pedi ;(R/E)DX => pointer to buffer
  HeaderCheck:
    cmp pecx,4 ;check if size of bufer > 4
    jb sendDataWithoutChunk

    cmp dword [pedx],0x0A0D0A0D ;check on new lines
    jne nextHeaderCheck

      mov pdword [pebp+psize*9],0
      sub pedx,pedi ;get size of header
      add pedx,2    ;add one 0A0D (the second will be send by the first chunk)
      mov pecx,pdword [pebx+psize*7] ;size
      sub pecx,pedx ;size of the content + 0A0D
      sub pecx,2    ;size of the content

      push pedx                   ;len

      push pesp                   ;&len
      push pedi                   ;buffer
      push DEFAULT_SENDTIMEOUT    ;timeout
      push RSF_RETURN_ON_COMPLETE ;flags
      push pdword [pebp+psize*3]  ;ClientSession
      push pdword [pebp+psize*2]  ;ProtocolSession
      call [SendData]

      pop peax ;remove len

      test pecx,pecx
      jz waitForInput ;continue reading the stream if there is no content
      call SendSizeChunk

      push pecx

      push pesp                   ;&len
      mov peax,pedi               ;(R/E)AX => buffer
      add peax,pedx               ;(R/E)AX => content - 2 (OAOD)
      add peax,2                  ;(R/E)AX => content
      push peax                   ;buffer
      push DEFAULT_SENDTIMEOUT    ;timeout
      push RSF_RETURN_ON_COMPLETE ;flags
      push pdword [pebp+psize*3]  ;ClientSession
      push pdword [pebp+psize*2]  ;ProtocolSession
      call [SendData]

      pop pecx  ;remove len
      test eax,eax
      jnz timeoutReached
      jmp waitForInput

  nextHeaderCheck:
    inc pedx
  loop HeaderCheck

  sendDataWithChunk:
  mov pecx,pdword [pebx+psize*7] ;size
  call SendSizeChunk

  sendDataWithoutChunk:
  end if
  lea peax,pdword [pebx+psize*7]
  push peax                   ;&len
  push pedi                   ;buffer
  push DEFAULT_SENDTIMEOUT    ;timeout
  push RSF_RETURN_ON_COMPLETE ;flags
  push pdword [pebp+psize*3]  ;ClientSession
  push pdword [pebp+psize*2]  ;ProtocolSession
  call [SendData]
  test eax,eax
  jnz timeoutReached

jmp waitForInput

timeoutReached:
  ;terminate the CGI-application
  argpush 2,0              ;return code
  argpush 1,pdword [pesi]  ;hProcess
  call [TerminateProcess]

endLaunchCGIDeleteBufferEndProcessHandles:
if used EnableChunking
call sendLastChunk
end if

if used is32
  cmp [is9x],1
  je dontFreeStrings__LaunchCGI
end if

test pdword [pebp+psize*10],0x1
jnz useUTF8Environment_3

  ;free UTF-16 environment
  push pdword [pebp+psize*5]
  call FreeMemLocation

useUTF8Environment_3:

;free UTF-16 CGI-Path
push pdword [pebp+psize*4]
call FreeMemLocation

dontFreeStrings__LaunchCGI:

;Free buffer
if used is32
  lea ecx,dword [pedi-4]
  push ecx
else
  push rdi
end if
call FreeMemLocation

;close handle hThread
argpush 1,pdword [pesi+psize] ;hThread
call [CloseHandle]

;close handle hProcess
argpush 1,pdword [pesi] ;hProcess
call [CloseHandle]

endLaunchCGIDeletePostPipe:
argpush 1,pdword [pebx+psize*6] ;hReadPipe (CGI return Pipe)
call [CloseHandle]

argpush 1,pdword [pebx+psize*5] ;hWritePipe (CGI return Pipe)
call [CloseHandle]

endLaunchCGIDeleteCGIPipe:
argpush 1,pdword [pebx+psize*4] ;hReadPipe (CGI return Pipe)
call [CloseHandle]

argpush 1,pdword [pebx+psize*3] ;hWritePipe (CGI return Pipe)
call [CloseHandle]

endLaunchCGI:
call dleave
ret psize*9

